Async {
	classvar <threads, <currentThread;

	*initClass {
		threads = List[];
		// XXX: replace this with a stack??
		currentThread = nil;
	}

	*run {
		arg function, args, completionFunction, abortFunction, mode = 'normal';
		var thread;

		thread = AsyncThread(function, args, completionFunction, abortFunction);

		if (#['normal', 'callback'].includes(mode).not, {
			error("Unknown async run mode"+ mode);
			this.verboseHalt
		});

		if ((mode === 'callback') and: { currentThread.notNil }, {
			error("Cannot add a callback async-thread while running an async-thread");
			this.verboseHalt
		});

		if (mode === 'callback', {
			threads.add(thread);
		}, {
			threads.insert(0, thread);
		});

		this.prKickStartIfNeeded
	}

	weAre {
		^currentThread.notNil
	}

	// Host application interface
	*deliverRequestResult {
		arg value;

		if (currentThread.notNil, {
			warn(thisMethod.asString + ": can't call me inside an async thread");
			this.verboseHalt;
		});

		if (threads.isEmpty, {
			warn(thisMethod.asString ++ ": no async threads");
		}, {
			this.prKick(value);
		});
	}

	// Failure
	*cleanup {
		if (currentThread.notNil, {
			currentThread.abort;
			threads.remove(currentThread);
			currentThread = nil;
		});
	}

	// Internals
	*prKickStartIfNeeded {
		var result, thread;

		while({ threads.notEmpty }, {
			thread = threads.last;

			if (thread.started, {
				^this
			});

			currentThread = thread;
			result = thread.resume;
			currentThread = nil;

			//postln("Got " ++ result);

			if (result === 'wait', {
				^this
			});

			if (result !== 'exit', {
				this.badKickResult;
				thread.abort;
				threads.pop;
				^this
			});

			threads.pop;
		});
	}

	*prKick {
		arg value;
		var thread, result;

		if (threads.isEmpty, { ^this });

		thread = threads.last;

		if (thread.notNil, {
			//debugln("Kicking an AsyncThread");

			currentThread = thread;
			result = thread.resume(value);
			currentThread = nil;

			//postln("Got " ++ result);

			if (#['exit', 'wait'].includes(result).not, {
				this.badKickResult;
				thread.abort;
				threads.pop;
				^this	
			});

			if (result === 'exit', {
				threads.pop;
				//debugln("AsyncThread terminated");
			});
			//"debugln("End of kick");
		}, {
			debugln("Found a 'nil' in threads (removing)");
			threads.pop;
		});
	}

	*badKickResult {
		arg result;
		error("Aborting thread because of bad kick result:");
		result.dump;
		this.dumpBackTrace;
	}

	// Super-Yield primitives

	*wait {
		^this.prWait;
	}

	*exit {
		^this.prExit;
	}

	*prWait {
		_AsyncThread_YieldForWait
		this.primitiveFailed
	}

	*prExit {
		_AsyncThread_YieldForExit
		this.primitiveFailed
	}
}
