ServerSync {
	classvar queue;

	*initClass {
		queue = List[];
	}

	*wait {
		arg server, action, successReaction, failureReaction, successReply = "/done", failureReply = "/fail", timeout = 0.2;
		var thread, succeded = false, routine;

		server = server ? Server.default;
		thread = thisThread;

		routine = Routine({
			timeout.wait;
			failureReaction.value;
			thread.next(\failed);
		});

		OSCresponder.pair(server.addr, successReply, {
			arg ...args;
			routine.stop;
			successReaction.valueArray(args);
			thread.next(\succeded);
		}, "/fail", {
			routine.stop;
			failureReaction.value;
			thread.next(\failed);
		});

		action.value;
		routine.play(AppClock);

		^(0.yield)
	}

 	*try {
		arg server, action, successReaction, failureReaction,
			successReply = "/done", failureReply = "/fail", timeout = 0.2, n = 3;
		var done = false;

		while( { n > 0 and: { done.not }}, {
			this.wait(server, action, {
				arg ...args;			var reactionResult;
				reactionResult = successReaction.valueArray(args);
				if (reactionResult !== \retry, {
					done = true;
				}, {
					done = false;
					n = n - 1;
				});	
			}, {
				arg ...args;
				n = n - 1;
			}, successReply, failureReply, timeout, n);
		});

		if (done.not, {
			failureReaction.value;
			^\failed
		}, {
			^\succeded
		});
	}

	*do {
		arg function;			var routine;
		routine = Routine({
			var nextRoutine;
			function.value;
			queue.removeAt(0);
			if (queue.size > 0, { nextRoutine = queue.first; nextRoutine.next });
		});
		queue.add(routine);
		if (queue.size == 1, { routine.next });
	}
}
