GtkComplexSurface : GtkDrawableArea {
	var proximity = false;

	var proximityGC, captionGC, cursorGC, captionLayout;

	var <modes, <mode, <modeName;
	var trigger, setter, axesNames, axesSpecs;

	var <showCaptions = true, <showProximityMarker = true;
	var <showBorder = true, <showHCursor = true, <showVCursor = true;
	var <>attackBeforeControls = true, <hideCursorsWhilePlaying = true;

	var <coords, pastCoords, pastTime, <speed, <xspeed, <yspeed;
	var <attacked = false;

	*new {
		arg ...modes;
		^super.new.initGtkComplexSurface(modes)
	}
	initGtkComplexSurface {
		arg someModes;

		modes = IdentityDictionary[];
		someModes.do({ |item| modes.add(item) }); 

		coords = Array[nil, nil, nil];
		pastCoords = Array[nil, nil, nil];
		this.prResetStats;

		captionLayout = PangoLayout(this.getPangoContext);

		buttonPressAction = {
			arg x, y, z, mods, buttons;

			if (z.notNil and: { buttons.includes(1) }, {
				this.updateCoords(x, y, z);
				this.sendAttack(x, y, z);
			});
		};
		motionAction = {
			arg x, y, z, mods, buttons;

			if (z.notNil and: { buttons.includes(1) }, {
				this.updateCoords(x, y, z);
				if (attacked.not, {
					this.sendAttack(x, y, z);
				}, {
					this.sendControls(x, y, z);
				});
			}, {
				if (attacked, {
					this.prResetStats;
					this.sendRelease(x, y, z);
				});
			});
		};
		buttonReleaseAction = {
			arg x, y, z, mods, buttons;

			if (buttons.includes(1), {
				this.updateCoords(x, y, z);
				this.sendRelease(x, y, z);
			});
		};

		proximityInAction = {
			proximity = true;
			this.invalidate;
		};
		proximityOutAction = {
			proximity = false;
			this.invalidate;
		};
		enterAction = {
			proximity = true;
			this.invalidate;
		};
		leaveAction = {
			proximity = false;
			this.invalidate;
		};

		this.updateCaptions;

		drawCallQueue.addArgCollector({ [proximityGC, captionGC, cursorGC, captionLayout] });
		drawCallQueue.add(\drawCaptions, 200);
		drawCallQueue.add(\drawCursor, 300);

		this.switch(someModes[0].key);

		//this.driven = true;
	}

	switch {
		arg newName;	var newMode;

		newMode = modes[newName];
		if (newMode.notNil, {
			modeName = newName;
			mode = newMode;
			showVCursor = (axesSpecs.size == 0) or: { axesSpecs[0].notNil };
			showHCursor = (axesSpecs.size < 2) or: { axesSpecs[1].notNil };
		});
	}

	prResetStats {
		speed = 0.0;
		xspeed = 0.0;
		yspeed = 0.0;
		pastTime = nil;
	}

	updateCoords {
		arg x, y, z;		var time, dt;

		time = thisThread.seconds;

		pastCoords[0] = coords[0];
		pastCoords[1] = coords[1];

		if (pastTime.notNil, {
			dt = time - pastTime;
			speed = hypot(x - pastCoords[0], y - pastCoords[1]) / dt;
			xspeed = (x - pastCoords[0]) / dt;
			yspeed = (y - pastCoords[1]) / dt;
		});

		coords[0] = x;
		coords[1] = y;
		coords[2] = z;

		pastTime = time;

		if (showVCursor or: showHCursor, {
			this.invalidate;
		});
	}

	sendAttack {
		arg x, y, z;

		if (attackBeforeControls, {
			if (attacked.not, {
				trigger.value(1.0, x, y, z);
				attacked = true;
			});
			this.sendControls(x, y, z);
		}, {
			this.sendControls(x, y, z);
			if (attacked.not, {
				trigger.value(1.0, x, y, z);
				attacked = true;
			});
		});
	}

	sendRelease {
		arg x, y, z;
		this.sendControls(x, y, z);
		trigger.value(0.0, x, y, z);
		attacked = false;
	}

	sendControls {
		arg ...args;
		setter.valueArray(axesSpecs.collectNotNil({
			arg spec, i;
			if (spec.notNil, {
				spec.map(args[i])
			}, {
				nil
			});
		}));
	}
	
	updateCaptions {
		var text;

		text = "";

		axesNames.do({
			arg name, i;
			if (name.notNil, {
				text = text ++ this.caption(i, name, axesSpecs[i]);
			});
		});

		captionLayout.setMarkup("<small>" ++ text ++ "</small>");
		this.invalidate;
	}

	caption {
		arg index, name, spec;
		var axis, text;

		axis = #["X", "Y", "Z"][index];

		if (spec.notNil, {
			text = 	axis ++ " axis:\n" ++
				"\tVariable: <b>" ++ name.asString ++ "</b>\n" ++
				"\tRange:    [<tt>" ++
				spec.minval.asString ++ "</tt>, <tt>" ++
				spec.maxval.asString ++ "</tt>] " ++
				spec.warp.asSpecifier.asString ++ " warp\n";
  		}, {
			text = axis ++ " axis is inactive\n";
		});

		^text
	}

	allocateResources {
		if (hasAllocatedResources || window.isNil, { ^this });
		super.allocateResources;

		proximityGC = GdkGC.new(window);
		proximityGC.setRgbFgColor("grey");

		captionGC = GdkGC.new(window);
		captionGC.setRgbFgColor("dark green");

		cursorGC = GdkGC.new(window);
		cursorGC.setRgbFgColor("dark grey");
	}

	drawCaptions {
		var w, h;

		if (showCaptions, {
			#w, h = captionLayout.getPixelSize;
			window.drawLayout(captionGC, 5, height - h - 5, captionLayout);
		});
	}

	showCaptions_ { arg value; showCaptions = value; this.invalidate }
	showProximityMarker_ { arg value; showProximityMarker = value; this.invalidate }
	showBorder_ { arg value; showBorder = value; this.invalidate }
	showHCursor_ { arg value; showHCursor = value; this.invalidate }
	showVCursor_ { arg value; showVCursor = value; this.invalidate }
	hideCursorsWhilePlaying_ { arg value; hideCursorsWhilePlaying = value; this.invalidate }

	drawBackground {
		super.drawBackground;
		if (showBorder, {
			window.drawRectangle(proximityGC, false, 0, 0, width, height);
		});
		if (proximity and: showProximityMarker, {
			window.drawRectangle(proximityGC, false, 2, 2, width - 5, height - 5);
		});
	}

	drawCursor {
		if (hideCursorsWhilePlaying.not or: attacked.not, {
			if (showVCursor and: coords[0].notNil, {
				window.drawLine(cursorGC, coords[0] * width, 0, coords[0] * width, height)
			});
			if (showHCursor and: coords[1].notNil, {
				window.drawLine(cursorGC, 0, coords[1] * height, width, coords[1] * height)
			});
		});
	}

	defaultSizeAsToplevel {	^#[320, 200] }

	proximityColor_ { arg name; proximityGC.setRgbFgColor(name); }
	captionColor_ { arg name; captionGC.setRgbFgColor(name); }
	cursorColor_ { arg name; cursorGC.setRgbFgColor(name); }

	finalize {
		this.unrefAll([
			proximityGC, captionGC, cursorGC, captionLayout
		]);
		super.finalize;
	}
}
