GtkPlus1dSlider : GtkPlus2dSlider {
	var action, triggerAction, index, <valueLabel, <captionLabel, <labelOnTop, <>inverted = false;
	var miniCaptionGC, miniCaptionLayout;

	*initClass {
		ObjectSignals.register(this, #['value-changed', 'value-set', 'attack', 'release'])
	}

	*new {
		arg spec, action, width = 200, height = 20, mode, signalOnInit = false, withLabels = false,
			triggerAction, labelOnTop = false, inverted = false;
		var object;

		if (mode == \horizontal, { width = 200; height = 20; });
		if (mode == \vertical, { width = 20; height = 200; });

		spec = if (spec.isKindOf(Association), { spec }, { "unnamed" -> spec });

		object = super.new(nil, nil, spec, spec, nil)
			.initGtkPlus1dSlider(action, width, height, signalOnInit, withLabels, triggerAction, labelOnTop, inverted)

		^object
	}

	initGtkPlus1dSlider {
		arg anAction, width, height, signalOnInit, withLabels, aTriggerAction, inLabelOnTop, inInverted;

		miniCaptionLayout = PangoLayout(this.getPangoContext);

		inverted = inInverted;
		this.setSizeRequest(width, height);
		action = anAction;
		triggerAction = aTriggerAction;
		trigger = { |value| triggerAction.value(value) };
		this.showCaptions_(false);
		this.showProximityMarker_(false);
		this.reset(signalOnInit);
		if (withLabels, {
			valueLabel = this.newValueLabel;
			captionLabel = this.newCaptionLabel;
		});
		labelOnTop = inLabelOnTop;
		drawCallQueue.add(\drawMiniCaption, 210);
	}

	state { ^this.value }
	state_ { |x| this.value = x }

	value { ^this.prMap }
	value_ { |value| this.set(value); action.value(value); this.emit('value-changed', value); }
	set {
		arg value;
		coords[index] = this.prUnmap(value);
		this.emit('value-set', value);
		this.updateMiniCaption;
		this.invalidate;
	}
	position { ^coords[index] }
	position_ {
		arg x;		var value;
		coords[index] = x;
		value = this.prMap(x);
		action.value(value);
		this.emit('value-changed', value);
		this.invalidate;
	} 
	setPosition {
		arg x;
		coords[index] = x;
		this.emit('value-set', this.prMap(x));
		this.invalidate;
	}

	prMap {
		arg x;
		x = x ? coords[index];
		if (inverted, {
			^specs[index].value.map(1.0 - x)
		}, {
			^specs[index].value.map(x)
		});
	}
	prUnmap {
		arg value;
		^specs[index].value.unmap(value)
	}

	reset {
		arg emit = false;
		if (emit, {
			this.value = specs[index].value.default;
		}, {
			this.set(specs[index].value.default);
		});
	}

	setSizeRequest {
		arg width, height;
		if (width > height, {
			this.showVCursor = true;
			this.showHCursor = false;
			index = 0;
			setter = { arg x, y; this.value = x };
		}, {
			this.showVCursor = false;
			this.showHCursor = true;
			index = 1;
			setter = { arg x, y; this.value = y };
		});
		super.setSizeRequest(width, height);
	}

	newCaptionLabel {
		^GtkLabel.new(specs[index].key.asString)
	}
	newValueLabel {
		arg width = 80, prec = 5;
		var label;
		label = GtkLabel.new;
		label.setSizeRequest(width);
		label.setText(this.value.asStringPrec(prec));
		this.watch('value-set', this, {
			arg value;
			label.setText(value.asStringPrec(prec));
		});
		^label
	}

	labelOnTop_ { |value| labelOnTop = value; this.invalidate }
	drawMiniCaption {
		var w, h, x, y, offset = 5;
		if (labelOnTop and: { index == 0 }, {
			#w, h = miniCaptionLayout.getPixelSize;
			x = (width - w) / 2;
			y = (height - h) / 2;
			window.drawLayout(miniCaptionGC, offset, y, miniCaptionLayout);
		});
	}
	updateMiniCaption {
		var text, prec = 5;
		text = specs[index].key.asString ++ ": " ++ this.value.asStringPrec(prec);
		miniCaptionLayout.setMarkup("<small>" ++ text ++ "</small>");
		this.invalidate;
	}
	allocateResources {
		super.allocateResources;
		miniCaptionGC = GdkGC.new(window);
		miniCaptionGC.setRgbFgColor("dark green");
	}
	finalize {
		this.unrefAll([
			miniCaptionGC, miniCaptionLayout
		]);
		super.finalize;
	}
}
