GtkPlusLazySequencer : GtkDrawableArea {
	var objectGC, focusedObjectGC, layout;
	var <objects, <>deliverAction, <sequence;
	var cursor, focused, <>mode;

	*new {
		arg deliverAction, mode = \automatic;
		^super.new.initGtkPlusLazySequencer(deliverAction, mode)
	}

	initGtkPlusLazySequencer {
		arg aDeliverAction, aMode;

		mode = aMode;
		deliverAction = aDeliverAction;
		objects = IdentityDictionary[];
		layout = PangoLayout(this.getPangoContext);
		drawCallQueue.add(\drawObjects, 200);

		sequence = List[];

		motionAction = {
			arg x, y, z, mods, buttons;

			cursor = Point(x * width, y * height);

			this.updateFocusedObject;
		};

		proximityInAction = {
			this.reset;
			this.invalidate;
		};
		proximityOutAction = {
			this.prUnfocusAndDeliver;
			this.invalidate;
		};
		enterAction = {
			this.reset;
			this.invalidate;
		};
		leaveAction = {
			this.prUnfocusAndDeliver;
			this.invalidate;
		};

		this.invalidate
	}

	reset {
		sequence = List[];
	}
	prUnfocusAndDeliver {
		if (focused.notNil, {
			sequence.add(focused);
			focused = nil;
		});
		if (mode === \automatic, {
			deliverAction.value(sequence);
		});
	}

	deliver {
		deliverAction.value(sequence);
	}

	addObject {
		arg id, name, x, y;
		objects[id] = [name, [x, y], nil];
		this.invalidate;
	}

	addObjectsInCircle {
		arg data, singleton, r = 0.3;	var n, theta, x, y;
		n = data.size;
		data.do({
			arg item, i;
			theta = i * (2*pi / n);
			#x, y = ([cos(theta), sin(theta)] * r) + [0.5, 0.5];
			this.addObject(item[0], item[1], x, y);
		});
		if (singleton.notNil, {
			this.addObject(singleton[0], singleton[1], 0.5, 0.5);
		});
	}

	drawObjects {
		var coords, x, y, w, h, gc;
		var name, rect;

		objects.keysValuesDo({
			arg key, data, i;
			#name, coords = data;
			#x, y = coords * [width, height];
			layout.setMarkup("<small>" ++ name ++ "</small>");
			data[2] = layout.getPixelSize;
			#w, h = data[2];

			rect = Rect(x - (w/2) - 5, y - (h/2) - 5, w + 10, h + 10);
			gc = if (key === focused, { focusedObjectGC }, { objectGC });

			window.drawLayout(gc, x - (w/2), y - (h/2), layout);
			window.drawRectangle(gc, false, rect.left, rect.top, rect.width, rect.height);
		});
	}

	updateFocusedObject {
		var name, x, y, w, h, coords, sizes, rect;

		objects.keysValuesDo({
			arg key, data, i;

			#name, coords, sizes = data;
			#x, y = coords * [width, height];
			if (sizes.notNil, {
				#w, h = sizes;

				rect = Rect(x - (w/2) - 5, y - (h/2) - 5, w + 10, h + 10);
				if (rect.containsPoint(cursor), {
					if (focused !== key, {
						focused = key;
						this.invalidate;
					});
					^true
				});
			});
		});
		if (focused !== nil, {
			sequence.add(focused);
			focused = nil;
			this.invalidate;
		});
		^false
	}

	allocateResources {
		if (hasAllocatedResources || window.isNil, { ^this });
		super.allocateResources;

		objectGC = GdkGC.new(window);
		objectGC.setRgbFgColor("dark grey");

		focusedObjectGC = GdkGC.new(window);
		focusedObjectGC.setRgbFgColor("dark red");
	}

	finalize {
		this.unrefAll([objectGC, focusedObjectGC, layout]);
		super.finalize;
	}
}
