GtkPlusMultiSlider : GtkSensitiveArea {
	var <n, mode;
	var hasAllocatedResources;
	var <drawCallQueue, proximity = false;

	var backgroundGC, borderGC, cursorGC;
	var <>setter, <specs, <showBorder = true, <showSeparators = true;

	var <values, <coords, <>locked = false, <>style = \lines;

	*new {
		arg n = 2, setter, specs, mode = \vertical;
		^super.new.initGtkPlusMultiSlider(n, setter, specs, mode)
	}
	initGtkPlusMultiSlider {
		arg aN, aSetter, someSpecs, aMode;

		n = aN;
		mode = aMode;

		if (specs.isArray, {
			specs = someSpecs ++ \unipolar.asSpec.dup(max(0, n - someSpecs.size));
		}, {
			specs = someSpecs ? \unipolar.asSpec;
		});
		setter = aSetter;

		values = Array.fill(n, { arg i; this.prDefaultValue(i) });
		coords = Array.fill(n, { arg i; this.prDefaultCoord(i) });

		hasAllocatedResources = false;

		this.addEvents('exposure-mask');
		this.connect('expose-event', {
			this.allocateResources;
			drawCallQueue.perform;
			false;
		});

		buttonPressAction = {
			arg x, y, z, mods, buttons;

			if (not(locked), {
				if (z.notNil and: { buttons.includes(1) }, {
					this.updateValues(x, y, z);
				});
			});
		};
		motionAction = {
			arg x, y, z, mods, buttons;

			if (not(locked), {
				if (z.notNil and: { buttons.includes(1) }, {
					this.updateValues(x, y, z);
				});
			});
		};
		buttonReleaseAction = {
			arg x, y, z, mods, buttons;

			if (not(locked), {
				if (buttons.includes(1), {
					this.updateValues(x, y, z);
				});
			});
		};

		drawCallQueue = MethodCallQueue.new(this);
		drawCallQueue.add(\drawBackground, 100);
		drawCallQueue.add(\drawCursors, 200);
	}

	updateValues {
		arg x, y, z;
		var i, value;

		if (mode == \horizontal, {
			y = 1.0 - y.clip(0.0, 1.0);

			if (x.notNil and: y.notNil, {
				i = (x * n).asInteger.clip(0, n - 1);
				value = this.prMap(i, y);
				values[i] = value;
				coords[i] = y;
				setter.value(i, value, values);
				this.invalidate;
			});
		}, {
			x = x.clip(0.0, 1.0);

			if (x.notNil and: y.notNil, {
				i = (y * n).asInteger.clip(0, n - 1);
				value = this.prMap(i, x);
				values[i] = value;
				coords[i] = x;
				setter.value(i, value, values);
				this.invalidate;
			});
		});
	}

	prMap {
		arg i, y;
		if (specs.isArray, {
			^specs[i].map(y);
		}, {
			^specs.map(y);
		});
	}
	prUnmap {
		arg i, value;
		if (specs.isArray, {
			^specs[i].unmap(value);
		}, {
			^specs.unmap(value);
		});
	}
	prDefaultCoord {
		arg i;
		if (specs.isArray, {
			^specs[i].unmap(specs[i].default);
		}, {
			^specs.unmap(specs.default);
		});
	}
	prDefaultValue {
		arg i;
		if (specs.isArray, {
			^specs[i].default;
		}, {
			^specs.default;
		});
	}

	values_ {
		arg someValues;
		if (someValues.isNumber, {
			someValues = someValues.dup(n);
		});
		values = someValues;
		coords = someValues.collect({ arg value, i; this.prUnmap(i, value) });
		setter.value(nil, nil, values);
		this.invalidate;
	}
	value { ^this.values }
	value_ { arg someValues; this.values(someValues) }

	set {
		arg someValues;
		if (someValues.isNumber, {
			someValues = someValues.dup(n);
		});
		values = someValues;
		coords = someValues.collect({ arg value, i; this.prUnmap(i, value) });
		this.invalidate;
	}

	interp { arg x; ^this.values.interp(x) }
	resample { arg i, n; ^this.values.resample(i, n) }
	
	allocateResources {
		if (hasAllocatedResources || window.isNil, { ^this; });

		backgroundGC = GdkGC.new(window);
		backgroundGC.setRgbFgColor("white");

		borderGC = GdkGC.new(window);
		borderGC.setRgbFgColor("grey");

		cursorGC = GdkGC.new(window);
		cursorGC.setRgbFgColor("dark grey");

		hasAllocatedResources = true;
	}

	showBorder_ { arg value; showBorder = value; this.invalidate }
	showSeparators_ { arg value; showSeparators = value; this.invalidate }

	drawBackground {
		var w;
		window.drawRectangle(backgroundGC, true, 0, 0, width, height);
		if (showBorder, {
			window.drawRectangle(borderGC, false, 0, 0, width, height);
		});
		if (showSeparators, {
			if (mode == \horizontal, {
				w = width / n;
				n.do({
					arg i;
					window.drawLine(borderGC, i * w, 0, i * w, height)			
				});
			}, {
				w = height / n;
				n.do({
					arg i;
					window.drawLine(borderGC, 0, i * w, width, i * w)			
				});
			});
		});
	}

	drawCursors {
		var w, y;

		if (mode == \horizontal, {
			w = width / n;
			if (style == \lines, {
				n.do({
					arg i;
					y = (1.0 - coords[i]) * height;
					window.drawLine(cursorGC, i * w, y, (i + 1) * w, y)
				});
			}, {
				n.do({
					arg i;
					y = (1.0 - coords[i]) * height;
					window.drawRectangle(cursorGC, true, i * w, y, w, height - y);
				});
			});
		}, {
			w = height / n;
			if (style == \lines, {
				n.do({
					arg i;
					y = coords[i] * width;
					window.drawLine(cursorGC, y, i * w, y, (i + 1) * w)
				});
			}, {
				n.do({
					arg i;
					y = coords[i] * width;
					window.drawLine(cursorGC, y, i * w, y, (i + 1) * w)
				});
			});
		});
	}

	defaultSizeAsToplevel {	^#[320, 200] }

	backgroundColor_ { arg name; backgroundGC.setRgbFgColor(name); this.invalidate }
	cursorColor_ { arg name; cursorGC.setRgbFgColor(name); this.invalidate }
	borderColor_ { arg name; borderGC.setRgbFgColor(name); this.invalidate }

	finalize {
		this.unrefAll([
			backgroundGC, borderGC, cursorGC
		]);
	}
}
