// Access to data member could be optimized by using
// different read-only classes (like GdkEventButton, GdkEventKey etc)
// built by specific methods (asGdkEventButton, asGdkEventKey etc)

GdkEvent : GBoxed {
	*initClass {
		^this.registerWrapperClass("GdkEvent")
	}

	type {
		_prGdkEventGetType;
		^this.primitiveFailed
	}
	
	coords {
		_prGdkEventGetCoords;
		^this.primitiveFailed
	}
	state {
		_prGdkEventGetState;
		^this.primitiveFailed
	}
	rootCoords {
		_prGdkEventGetRootCoords;
		^this.primitiveFailed
	}

	keyval {
		_prGdkEventGetKeyval;
		^this.primitiveFailed
	}

	button {
		_prGdkEventGetButton;
		^this.primitiveFailed
	}

	time {
		_prGdkEventGetTime;
		^this.primitiveFailed
	}

	isHint {
		_prGdkEventIsHint;
		^this.primitiveFailed
	}

	axis {
		arg n;
		_prGdkEventGetAxis;
		^this.primitiveFailed		
	}

	// Extra
	keyname {
		_prGdkEventGetKeyname;
		^this.primitiveFailed
	}

	// Extra
	stateAsString {
		var currentState, name;

		name = "";
		currentState = this.state;
		if (currentState.notNil, {
			if ((currentState & Gdk.shift_mask) != 0, {
				name = name ++ "Shift+";
			});
			if ((currentState & Gdk.control_mask) != 0, {
				name = name ++ "Control+";
			});
			if ((currentState & Gdk.mod1_mask) != 0, {
				name = name ++ "Alt+";
			});
			if ((currentState & Gdk.mod3_mask) != 0, {
				name = name ++ "Mod3+";
			});
			if ((currentState & Gdk.mod4_mask) != 0, {
				name = name ++ "Mod4+";
			});
			if ((currentState & Gdk.mod5_mask) != 0, {
				name = name ++ "Mod5+";
			});
		});
		
		^name
	}

	stateAsArray {
		var currentState, array;

		array = Array[];
		currentState = this.state;
		if (currentState.notNil, {
			if ((currentState & Gdk.shift_mask) != 0, {
				array = array.add("shift");
			});
			if ((currentState & Gdk.control_mask) != 0, {
				array = array.add("control");
			});
			if ((currentState & Gdk.mod1_mask) != 0, {
				array = array.add("alt");
			});
			if ((currentState & Gdk.mod3_mask) != 0, {
				array = array.add("mod3");
			});
			if ((currentState & Gdk.mod4_mask) != 0, {
				array = array.add("mod4");
			});
			if ((currentState & Gdk.mod5_mask) != 0, {
				array = array.add("mod5");
			});
		});
		
		^array
	}

	keyCombinationString {
		arg ignoreCase = true;
		var key, keyname, state, name;

		keyname = this.keyname; 

		if (keyname.notNil, {
			if ((keyname.size == 1) && ignoreCase, {
    				key = String.newFrom([keyname.at(0).toUpper]);
  			}, {
				key = keyname;
			});

			^[this.stateAsString, key];
		}, {
			^[nil, nil];
		});
	}

	modsAndKeyString {
		arg ignoreCase = true;
		var key, keyname;

		keyname = this.keyname; 

		if (keyname.notNil, {
			if ((keyname.size == 1) && ignoreCase, {
   				key = String.newFrom([keyname.at(0).toUpper]);
  			}, {
				key = keyname;
			});

			^this.stateAsArray.add(key);
		}, {
			^nil;
		});
	}

	getDeviceState {
		_prGdkEventGetDeviceState;
		^this.primitiveFailed
	}
}
