GtkPatchView : GtkVBox {
	var patch;
	var innerBox, buttonBox;

	*new {
		arg aPatch, depth = 1;
		^super.new(false, 2).initGtkPatchView(aPatch, depth)
	}

	initGtkPatchView {
		arg aPatch, depth;
		var frame;

		patch = aPatch;

		frame = GtkFrame.new;
		this.packStart(frame, false, false);
		innerBox = GtkVBox.new(false, 2);
		innerBox.setBorderWidth(2);
		frame.add(innerBox, false, false);

		buttonBox = [
			"Play" -> { patch.play },
			"Stop" -> { patch.synth.free },
			"Free and Destroy" -> { patch.synth.free; this.destroy },
			"Save" -> {
				GtkFileSelection.askAndDo("Save state", {
					arg filename;
					patch.state.writeArchive(filename)
				});
			},
			"Load" -> {
				GtkFileSelection.askAndDo("Load state", {
					arg filename;
					patch.state = Object.readArchive(filename)
				});
			},
			"Push" -> {
				~valuestack.push(patch.state)
			},
			"Pop" -> {
				if (~valuestack.notEmpty, {
					patch.state = ~valuestack.pop
				});
			},
			"Peek" -> {
				if (~valuestack.notEmpty, {
					patch.state = ~valuestack.peek
				});
			},
		].asButtons.asGtkHBox(true, 2);

		this.createAllPatchControls(patch, innerBox, buttonBox, depth);

		this.packEnd(buttonBox);

		this.showAllChildren
	}

	createAllPatchControls {
		arg patch, sliderBox, buttonBox, depth;
		var gui, frame, box;

		patch.argsForSynth.do({
			arg a, i;

			if (a.isKindOf(Patch), {
				if (a.argsForSynth.size > 0 and: { (depth === inf) or: { depth > 1 } }, {
					frame = GtkFrame(patch.argNamesForSynth[i].asString);
					box = GtkVBox.new(false, 0).setBorderWidth(2);
					frame.add(box);
					this.packStart(frame, false, false);
					this.reorderChild(frame, 0);
					this.createAllPatchControls(a, box, buttonBox, if (depth === inf, { inf }, { depth - 1}));
				});
			}, {
				gui = a.gui(patch.argNamesForSynth[i]);
				if (gui.notNil, {
					if (gui.isKindOf(GtkButton), {
						buttonBox.packStart(gui, true, true);
					}, {
						sliderBox.packStart(gui, false, false);
					});
				});
			});
		});

		
	}

	add {
		arg name, editor;
		var view;

		view = GtkNumberEditorView(name, editor);
		innerBox.packStart(view, false, false);
		view.show
	}
	addAll {
		arg items;
		items.do({ arg item; this.add(item.key, item.value) });
	}
}
