GtkTreeModel : GObject {
	getColumnType {
		arg n;
		this.getColumnType_request(n);
		^Async.wait
	}
	getColumnType_request {
		arg n;
		_prGtkTreeModelGetColumnType;
		^this.primitiveFailed
	}

	set {
		arg iter ... pairs;
		var type;

		pairs.do({
			arg pair;
			type = this.getColumnType(pair.key);
			this.setValue(iter, pair.key, type, pair.value);
		});
	}

	getPath {
		arg iter;
		this.getPath_request(iter);
		^Async.wait
	}
	getPath_request {
		arg iter;
		_prGtkTreeModelGetPath;
		^this.primitiveFailed
	}

	getIter {
		arg path;

		if (path.notNil, {
			this.getIter_request(path);
			^Async.wait
		}, {
			^this.getIterFirst
		});
	}
	getIter_request {
		arg path;
		_prGtkTreeModelGetIter;
		^this.primitiveFailed
	}

	getIterFirst {
		this.getIterFirst_request;
		^Async.wait
	}
	getIterFirst_request {
		arg path;
		_prGtkTreeModelGetIterFirst;
		^this.primitiveFailed
	}

	iterParent {
		arg iter;
		this.iterParent_request(iter);
		^Async.wait
	}
	iterParent_request {
		arg iter;
		_prGtkTreeModelIterParent;
		^this.primitiveFailed
	}

	iterChildren {
		arg iter;
		if (iter.notNil, {
			this.iterChildren_request(iter);
			^Async.wait
		}, {
			^this.getIterFirst
		});
	}
	iterChildren_request {
		arg iter;
		_prGtkTreeModelIterChildren;
		^this.primitiveFailed
	}

	iterHasChild {
		arg iter;
		this.iterHasChild_request(iter);
		^Async.wait
	}
	iterHasChild_request {
		arg iter;
		_prGtkTreeModelIterHasChild;
		^this.primitiveFailed

	}

	iterNChildren {
		arg iter;
		this.iterNChildren_request(iter);
		^Async.wait
	}
	iterNChildren_request {
		arg iter;
		_prGtkTreeModelIterNChildren;
		^this.primitiveFailed

	}

	iterNthChild {
		arg iter, n;
		this.iterNthChild_request(iter, n);
		^Async.wait
	}
	iterNthChild_request {
		arg iter, n;
		_prGtkTreeModelIterNthChild;
		^this.primitiveFailed
	}

	iterNext {
		arg iter, next;
		this.iterNext_request(iter, next);
		^Async.wait
	}
	iterNext_request {
		arg iter, next;
		_prGtkTreeModelIterNext;
		^this.primitiveFailed
	}

	getValue {
		arg iter, col;
		this.getValue_request(iter, col);
		^Async.wait
	}
	getValue_request {
		arg iter, col;
		_prGtkTreeModelGetValue;
		^this.primitiveFailed
	}

	// EXTRA
	iterDetectBetweenChildren {
		arg iter, predicate;
		var current;

		if (iter.isNil, {
			current = this.getIterFirst;
		}, {
			current = this.iterChildren(iter);
		});
	
		while({	current.notNil }, {
			if (predicate.value(current), {
				^current;
			});
			current = this.iterNext(current);
		});

		^nil
	}
	iterDetectInSubtree {
		arg iter, predicateGenerator;
		var predicate;
		
		predicate = predicateGenerator.next;
		while({ predicate.notNil }, {
			iter = this.iterDetectBetweenChildren(iter, predicate);
			if (iter.isNil, { ^nil });
			predicate = predicateGenerator.next;
		});
		^iter
	}
	iterDetectInSubtreeFail {
		arg iter, predicateGenerator, failFunction;
		var next, predicate, i;
		
		i = 0;
		predicate = predicateGenerator.next;
		while({ predicate.notNil }, {
			next = this.iterDetectBetweenChildren(iter, predicate);
			if (next.isNil, {
				next = failFunction.value(i, iter);
			});
			iter = next;
			predicate = predicateGenerator.next;
			i = i + 1;
		});
		^iter
	}

	iterDetectInSubtreeFailOnce {
		arg iter, predicateGenerator, failFunction;
		var next, predicate, i;
		
		i = 0;
		predicate = predicateGenerator.next;
		while({ predicate.notNil }, {
			next = this.iterDetectBetweenChildren(iter, predicate);
			if (next.isNil, {
				^failFunction.value(i, iter);
			});
			iter = next;
			predicate = predicateGenerator.next;
			i = i + 1;
		});
		^iter
	}

	iterReduceToRoot {
		arg iter, function, result;
		var iter1, iter2;

		if (iter.notNil, {
			iter1 = iter.copy;
		}, {
			iter1 = nil;
		});

		while ( { iter1.notNil }, {
			result = function.value(iter1, result);

			iter2 = this.iterParent(iter1);
			iter1 = iter2;
		});

		^result
	}

	getPathString {
		arg iter;
		this.getPathString_request(iter);
		^Async.wait
	}
	getPathString_request {
		arg iter;
		_prGtkTreeModelGetPathString;
		^this.primitiveFailed
	}

	getIndices {
		arg iter;
		var path, indices;
		path = this.getPath(iter);
		indices = path.getIndices;
		^indices
	}

	getIterFromString {
		arg string;
		var path, iter;

		if (string.isNil, {
			iter = this.getIter;
		}, {
			path = GtkTreePath.newFromString(this, string);
			iter = this.getIter(path);
		});
		^iter
	}

	getIterFromIndices {
		arg indices;
		^this.getIterFromString(indices.joinAsString(":", ""));
	}

	treeDo {
		arg iter, branchFunction, leafFunction, argument;
		var firstchild, forward;

		iter = iter ? this.getIterFirst;
		while( { iter.notNil}, {
			firstchild = this.iterChildren(iter);
			if (firstchild.isNil, {
				leafFunction.value(argument, iter);
			}, {
				forward = branchFunction.value(argument, iter);
				this.treeDo(firstchild, branchFunction, leafFunction, forward);
			});
			iter = this.iterNext(iter);
		});		
	}
}
