GtkWindow : GtkBin {
	classvar <>allwindows, current;

	*initClass {
		allwindows = Array.new;
		current = nil;
	}

	*new {
		arg...args;
		^super.new.performList(\initWithNewObjectAndRemember, args)
	}
	initWithNewObjectAndRemember {
		arg...args;
		this.performList(\initWithNewObject, args);
		GtkWindow.remember(this);
	}
	initWithNewObject {
		arg type = \toplevel;
		this.newGtkWindow_request(type);
		this.wrap(Async.wait);
	}
	newGtkWindow_request {
		arg type = \toplevel;
		_prGtkWindowNew;
		^this.primitiveFailed		
	}

	setTitle {
		arg title = "Supercollider Window";
		this.setTitle_request(title);
		Async.wait
	}
	setTitle_request {
		arg title;
		_prGtkWindowSetTitle;
		^this.primitiveFailed
	}

	getTitle {
		this.getTitle_request;
		^Async.wait
	}
	getTitle_request {
		_prGtkWindowGetTitle;
		^this.primitiveFailed
	}

	setDefaultSize {
		arg w, h;
		this.setDefaultSize_request(w, h);
		^Async.wait
	}
	setDefaultSize_request {
		arg w, h;
		_prGtkWindowSetDefaultSize;
		^this.primitiveFailed
	}

	getDefaultSize {
		this.getDefaultSize_request;
		^Async.wait
	}
	getDefaultSize_request {
		_prGtkWindowGetDefaultSize;
		^this.primitiveFailed
	}

	move {
		arg x, y;
		this.move_request(x, y);
		^Async.wait
	}
	move_request {
		arg x, y;
		_prGtkWindowMove;
		^this.primitiveFailed
	}


	setPosition {
		arg position = \none;
		this.setPosition_request(position);
		^Async.wait
	}
	setPosition_request {
		arg position = \none;
		_prGtkWindowSetPosition;
		^this.primitiveFailed
	}

	setTransientFor {
		arg parent;
		this.setTransientFor_request(parent);
		^Async.wait
	}
	setTransientFor_request {
		arg parent;
		_prGtkWindowSetTransientFor;
		^this.primitiveFailed
	}

	setFocus {
		arg widget;
		this.setFocus_request(widget);
		^Async.wait
	}
	setFocus_request {
		arg widget;
		_prGtkWindowSetFocus;
		^this.primitiveFailed
	}
	getFocus {
		this.getFocus_request;
		^GObject.wrap(Async.wait)
	}
	getFocus_request {
		_prGtkWindowGetFocus;
		^this.primitiveFailed
	}

	setDestroyWithParent {
		arg value = true;
		this.setDestroyWithParent_request(value);
		^Async.wait
	}
	setDestroyWithParent_request {
		arg value = true;
		_prGtkWindowSetDestroyWithParent;
		^this.primitiveFailed
	}

	getSize {
		this.getSize_request;
		^Async.wait
	}
	getSize_request {
		_prGtkWindowGetSize;
		^this.primitiveFailed
	}

	resize {
		arg w, h;
		this.resize_request(w, h);
		^Async.wait
	}
	resize_request {
		arg w, h;
		_prGtkWindowResize;
		^this.primitiveFailed
	}

	// EXTRA

	*remember {
		arg window;
		allwindows = allwindows.add(window);
		current = this;
	}
	*forget {
		arg window;
		allwindows.remove(window);
		if (allwindows.size > 0, {
			current = allwindows.last;
		}, {
			current = nil;
		});
	}
	*destroyAll {
		var list;

		list = allwindows.copy;
		list.do({
			arg window;
			window.destroy;
		});
		allwindows = [];
	}

	center {
		arg x, y, w, h;
		var reqw, reqh, defw, defh, cx, cy;
		
		#reqw, reqh = this.getSizeRequest;
		#defw, defh = this.getDefaultSize;

		if (reqw < 0, { reqw = defw });
		if (reqh < 0, { reqh = defh });

		w = w ? x;
		h = h ? y;

		cx = x + (w - reqw)/2;
		cy = y + (h - reqh)/2;

		this.setPosition(\none);
		this.move(cx, cy);
	}

	centerOnScreen {
		this.center(0, 0, Gdk.getScreenWidth, Gdk.getScreenHeight);
	}

	setHideOnDelete {
		this.setHideOnDelete_request;
		^Async.wait
	}
	setHideOnDelete_request {
		_prGtkWindowSetHideOnDelete;
		^this.primitiveFailed
	}

	// GObject overrides
	prDestroySignalHandler {
		super.prDestroySignalHandler;
		GtkWindow.forget(this);
	}

	enableHideOnEscape {
		this.connect("key-release-event", {
			arg widget, event;
			var stateName, keyName;
			if (event.notNil, {
				#stateName, keyName = event.keyCombinationString;
				if (keyName == "Escape", {
					this.hide;
				});
			});
		});
	}
}
