Recorder {
	var <filename, <group, <buffersize, <filefmt, <samplefmt, <firstbus, <channels;
	var <synth, <defname, <buffer, <ready;

	*initClass {
		ObjectSignals.register(this, #['got-ready'])
	}
	*new {
		arg filename, group, buffersize = 65536, filefmt, samplefmt, firstbus = 0, channels = 2;

		filefmt = filefmt ? "wav";
		samplefmt = samplefmt ? "int16";
		group = group ?? { Server.default.asGroup };

		^super.newCopyArgs(filename, group, buffersize, filefmt, samplefmt, firstbus, channels).initRecorder
	}
	initRecorder {
		var server;

		server = group.server;
		ready = false;

		defname = ("Recorder-synth-" ++ channels).asSymbol;

		Task({
			SynthDef(defname, {
				arg bufnum;
				DiskOut.ar(bufnum, In.ar(firstbus, channels));
			}).send(server);
			0.2.wait;
			buffer = Buffer.alloc(server, buffersize, channels);
			0.2.wait;
			buffer.write(filename, filefmt, samplefmt, 0, 0, true);
			0.2.wait;
			synth = Synth.newPaused(defname, [\bufnum, buffer.bufnum], group, \addToTail);
			NodeWatcher.register(synth);
			0.2.wait;
			"Done".postln;
			ready = true;
			this.emit('got-ready');
		}).play(SystemClock);
	}

	start {
		synth.run(true);
	}
	stop {
		synth.run(false);
	}

	toggle {
		synth.run(synth.isRunning.not);
	}

	isRunning { ^(synth.notNil and: { synth.isRunning }) }

	free {
		synth.free;
		buffer.close;
		buffer.free;
		NodeWatcher.unregister(synth);
	}
}

