ObjectSignals {
	classvar <allReactions, <allWatchers;
	classvar <allEmittedClasses;
	classvar <>checkEnabled, <>strictCheckEnabled, <>warningEnabled;

	*initClass {
		allReactions = IdentityDictionary.new;
		allWatchers = IdentityDictionary.new;
		checkEnabled = true;
		strictCheckEnabled = true;
		warningEnabled = true;
	}

	*register {
		arg class, classes;

		allEmittedClasses = allEmittedClasses ? IdentityDictionary.new;

		if (allEmittedClasses.at(class).notNil, {
			error("class " ++ class.asString ++ " has already registered its signals");
			this.verboseHalt;
		}, {
			allEmittedClasses.put(class, classes)
		});
	}

	*classEmitsClass {
		arg emitterClass, emittedClass;
		var emitterClasses, emittedClasses, registered = false;
		
		emittedClasses = ObjectSignals.allEmittedClasses.at(emitterClass);
		if (emittedClasses.notNil, {
			registered = true;
			if (emittedClasses.includes(emittedClass), {
				^true
			});
		});

		emitterClass.superclasses.do({
			arg class;
			emittedClasses = ObjectSignals.allEmittedClasses.at(class);
			if (emittedClasses.notNil, {
				registered = true;
				if (emittedClasses.includes(emittedClass), {
					^true
				});
			});
		});

		if (registered.not && warningEnabled, {
			Post << "*** class '" << emitterClass.asString << "' and its superclasses have no emitted "
				<< "classes registered\n";
			Post << "*** but you asked an object to watch for '" << emittedClass.asString << "' coming from there\n";
		});

		if (emittedClass === Anything, {
			^true
		}, {
			if (registered.not && strictCheckEnabled, {
				^false;
			}, {
				^true
			});
		});
	}
}
