GtkArrayedControl : GtkSensitiveArea {
	var hasAllocatedResources, whiteGC, blackGC, gc;
	var drawCallQueue;

	var <array, maxValue, minValue;

	*newWithArray {
		arg anArray, aMaxValue, aMinValue;
		^super.new.initGtkArrayedControl(anArray, aMaxValue, aMinValue)
	}
	*new {
		arg size, aMaxValue, aMinValue, value;

		value = value ?	if (abs(aMaxValue - aMinValue) > max(abs(aMaxValue), abs(aMinValue)), {
			value = 0.0;
		}, {
			value = aMinValue;
		});
		^super.newWithArray(Array.fill(size, value), aMaxValue, aMinValue)
	}

	initGtkArrayedControl {
		arg anArray, aMaxValue, aMinValue;

		array = anArray;
		maxValue = aMaxValue ? { array.max };
		minValue = aMinValue ? { array.min };

		hasAllocatedResources = false;

		XInput.enabled = true;

		this.addEvents('exposure-mask');
		this.connect('expose-event', {
			this.allocateResources;
			drawCallQueue.perform;
			false;
		});

		buttonPressAction = {
			arg x, y, z, mods, buttons;
			if (buttons.includes(1), {
				this.setValueFromPosition(x, y);
			});
		};
		motionAction = {
			arg x, y, z, mods, buttons;
			if (buttons.includes(1), {
				this.setValueFromPosition(x, y);
			});
		};
		buttonReleaseAction = {
			arg x, y, z, mods, buttons;
			if (buttons.includes(1), {
				this.setValueFromPosition(x, y);
			});
		};

		drawCallQueue = MethodCallQueue.new(this);
		drawCallQueue.add(\drawBackground, 100);
		drawCallQueue.add(\drawHistogram, 200);
	}

	allocateResources {
		if (hasAllocatedResources || window.isNil, { ^this; });

		whiteGC = this.getStyle.getWhiteGC;
		blackGC = this.getStyle.getBlackGC;
		gc = GdkGC.new(window);
		gc.setRgbFgColor("light blue");

		hasAllocatedResources = true;
	}

	drawBackground {
		window.drawRectangle(whiteGC, true, 0, 0, width, height);
	}

	drawHistogram {
		var dx, dy, y, zy;

		dx = width / array.size;

		zy = (maxValue.asFloat / (maxValue - minValue)) * height;

		array.do({
			arg value, i;

			y = ((value.asFloat / (maxValue - minValue)) * height) + zy;

			dy = max(y, zy) - min(y, zy);

			window.drawRectangle(gc, true, dx * i, min(y, zy), dx, dy);
			window.drawRectangle(blackGC, false, dx * i, min(y, zy), dx, dy);
		});

		window.drawLine(blackGC, 0, zy, width, zy);

	}

	setValueFromPosition {
		arg x, y;
		var dx, zy, index, value;
		
		if (x.inclusivelyBetween(0.0, 1.0) and: { y.inclusivelyBetween(0.0, 1.0) }, {
			zy = (maxValue.asFloat / (maxValue - minValue)) * height;

			index = floor(x * array.size).asInteger;
			value = maxValue - ((1.0 - y) * (maxValue - minValue));

			this.set(index, value);

			this.invalidate;
		});
	}

	set {
		arg index, value;
	}

	finalize {
		this.unrefAll([
			gc
		]);
	}

	defaultSizeAsToplevel {	^#[400, 300] }
}

GtkNumericArrayView : GtkArrayedControl {
	*new {
		arg anArray, aMaxValue, aMinValue;
		^super.newWithArray(anArray, aMaxValue, aMinValue)
	}

	set {
		arg index, value;
		array[index] = value;
		array.changed(this);
	}
}
