GtkClassTreeView : GtkVBox {
	var scrolled, <model, view, nameRenderer, nameColumn;
	var filterEntry, filter, rootEntry, root;
	var <selection;

	*new {
		^super.new(false, 0).initGtkClassTreeView
	}
	initGtkClassTreeView {
		var box, label, group;

		filter = WildExpr.new("");
		root = "Object";

		scrolled = GtkScrolledWindow.new;
		scrolled.setPolicy(\automatic, \automatic);
		this.packStart(scrolled, true, true, 0);

		model = GtkTreeStore.new(\gchararray);

		view = GtkTreeView.newWithModel(model);
		selection = view.getSelection;
		selection.setMode('single');
		scrolled.add(view);

		nameRenderer = GtkCellRendererText.new;
		nameColumn = GtkTreeViewColumn.newWithAttributes("Classes", nameRenderer, "text", 0);
		view.appendColumn(nameColumn);

		box = GtkHBox.new(false, 0);
		this.packEnd(box, false, false, 0);

		label = GtkLabel.new("Filter:");
		box.packStart(label, false, false, 0);		

		filterEntry = GtkEntry.new;
		filterEntry.connect('activate', {
			filter = WildExpr.new(filterEntry.getText);
			this.rebuildModel;
		});
		box.packEnd(filterEntry, true, true, 0);		

		box = GtkHBox.new(false, 0);
		this.packEnd(box, false, false, 0);

		label = GtkLabel.new("Root:");
		box.packStart(label, false, false, 0);		

		rootEntry = GtkEntry.new;
		rootEntry.connect('activate', {
			var name;
			name = rootEntry.getText;
			if (name.asSymbol.asClass.notNil, {
				root = name;
				this.rebuildModel;
			}, {
				rootEntry.setText(root);
			});
		});
		box.packEnd(rootEntry, true, true, 0);

		this.rebuildModel;

		this.showAllChildren
	}
	rebuildModel {
		var classObject, iter;

		model.clear;

		classObject = root.asSymbol.asClass;
		if (classObject.notNil, {
			iter = this.addClassToModel(nil, classObject);
			if (iter.notNil, {
				selection.selectIter(iter);
			});
		});
		
	}
	addClassToModel {
		arg parent, classObject;
		var name, iter, names;

		name = classObject.name.asString;

		if (filter.isNil or: { filter.match(name) or: { name == root } }, {
			iter = model.append(parent);
			model.set(iter, 0 -> name);

			names = classObject.subclasses.collect({ arg class; class.name.asString });
			if (names.notNil, {
				names.sort;

				names.do({
					arg name;
					this.addClassToModel(iter, name.asSymbol.asClass)
				});
			});
		});
	
		^iter
	}
}
