GtkEnvelopeArea : GtkSensitiveArea {
	var envelope, <>signalOnButtonRelease, envelopeIsChanged, <>maxPoints, <>minPoints, <style, <yspec;
	var <>lockExtremes;
	var <>locked = false;

	var drawCallQueue, whiteGC, blackGC, greyGC, captionGC, captionLayout;
	var hasAllocatedResources, borderWidth = 5;
	var selectedPoint, selXOff, selYOff;

	*initClass {
		ObjectSignals.register(this, #['envelope-changed', 'size-changed'])
	}

	*new {
		arg maxPoints, envelope, minPoints, lockExtremes = \xAtEndPoints, style = \linear, yspec;
		^super.new.initGtkEnvelopeArea(maxPoints, envelope, minPoints, lockExtremes, style, yspec)
	}

	initGtkEnvelopeArea {
		arg inMaxPoints, inEnvelope, inMinPoints, inLockExtremes, inStyle, inYSpec;

		if (inMaxPoints.notNil, {
			if (inEnvelope.isNil or: { inMaxPoints != inEnvelope.size }, {
				inEnvelope = Array.series(inMaxPoints, 0.0, 1.0 / (inMaxPoints - 1)).collect({|x| [x, 0.0]});
			});
		});

		envelope = if (inEnvelope.notNil, { Envelope.from(inEnvelope) }, { Envelope[] });
		maxPoints = inMaxPoints;
		minPoints = inMinPoints;
		style = inStyle;
		yspec = inYSpec ? \unipolar;

		captionLayout = PangoLayout.new(this.getPangoContext);

		if (inLockExtremes == \xAtEndPoints, {
			if (envelope.size > 0, {
				envelope[0] = [0.0, envelope[0][1]];
				envelope[envelope.size - 1] = [1.0, envelope[envelope.size - 1][1]];
			});
		});
		lockExtremes = inLockExtremes;

		selectedPoint = nil;
		signalOnButtonRelease = true;
		envelopeIsChanged = false;

		drawCallQueue = MethodCallQueue.new(this);
		drawCallQueue.add(\drawBackground, 100);
		drawCallQueue.add(\drawCaptions, 150);
		drawCallQueue.add(\drawPoints, 200);

		buttonPressAction = {
			arg x, y, z, marray, barray;
			var newSelectedPoint, cWindowX, cWindowY, pWindowX, pWindowY;

			if (not(locked), {
				envelope.do({
					arg point, i;

					cWindowX = x * width;
					cWindowY = y * height;
					pWindowX = this.toWindowX(point[0]);
					pWindowY = this.toWindowY(point[1]);

					if (hypot(cWindowX - pWindowX, cWindowY - pWindowY) < 3, {
						newSelectedPoint = i;
						selXOff = pWindowX - cWindowX;
						selYOff = pWindowY - cWindowY;
					});
				});

				if (newSelectedPoint != selectedPoint, { this.invalidate });
				selectedPoint = newSelectedPoint;

				if (selectedPoint.isNil, {
					selectedPoint = this.addPoint(x, 1.0 - y);
					selXOff = 0;
					selYOff = 0;
					if (newSelectedPoint.notNil, { this.invalidate });
				});
			});
		};
		buttonReleaseAction = {
			arg x, y, z, marray, barray;

			if (not(locked), {
				if (selectedPoint.notNil, {
					if (marray.includes('Shift'), {
						this.removePoint(selectedPoint);
					});
					selectedPoint = nil;
				});
				if (signalOnButtonRelease, {
					if (envelopeIsChanged, {
						envelopeIsChanged = false;
						this.emit('envelope-changed', envelope);
					});
				});
			});
		};
		motionAction = {
			arg x, y, z, marray, barray;
			var cWindowX, cWindowY, pWindowX, pWindowY, nx, ny;

			if (not(locked), {
				if (selectedPoint.notNil, {
					this.movePoint(selectedPoint,	this.toEnvelopeX((x * width) + selXOff),
													this.toEnvelopeY((y * height) + selYOff));
				});
			});
		};

		hasAllocatedResources = false;

		this.addEvents('exposure-mask');
		this.connect('expose-event', {
			this.allocateResources;
			drawCallQueue.perform;
			false;
		});
	}

	allocateResources {
		if (hasAllocatedResources || window.isNil, { ^this; });

		whiteGC = this.getStyle.getWhiteGC;
		blackGC = this.getStyle.getBlackGC;
 		greyGC = GdkGC.new(window);
		greyGC.setRgbFgColor("grey");
 		captionGC = GdkGC.new(window);
		captionGC.setRgbFgColor("light blue");

		hasAllocatedResources = true;

		this.updateCaptions;
	}

	addPoint {
		arg x, y;
		var index, point;

		if (maxPoints == envelope.size, { ^nil });

		index = envelope.size;
		if ((index == 0) or: (index == (envelope.size - 1)), {
			if (lockExtremes == \xAtEndPoints, {
				if (index == 0, { x = 0.0 }, { x = 1.0 });
			});
		});

		point = [x, y];

		envelope.add(point);
		envelope.sort({ arg i1, i2; i1[0] <= i2[0] });

		this.emit('size-changed', envelope.size);
		this.emit('envelope-changed', envelope);
		this.invalidate;
		^envelope.indexOf(point)
	}
	removePoint {
		arg index;

		if (minPoints == envelope.size, { ^nil });
		if (lockExtremes.notNil and: ((index == 0) or: (index == (envelope.size - 1))), { ^nil });

		envelope.removeAt(selectedPoint);
		this.emit('size-changed', envelope.size);
		this.emit('envelope-changed', envelope);
		this.invalidate
	}
	movePoint {
		arg index, x, y;
		var minx = 0.0, maxx = 1.0, miny = 0.0, maxy = 1.0;

		if ((index == 0) or: (index == (envelope.size - 1)), {
			if (lockExtremes == \full, { ^nil });
			if (lockExtremes == \x, { x = nil });
			if (lockExtremes == \xAtEndPoints, {
				if (index == 0, { x = 0.0 }, { x = 1.0 });
			});
		});

		if (index > 0, { minx = envelope[index - 1][0] });
		if (index < (envelope.size - 1), { maxx = envelope[index + 1][0] });
		if (x.notNil, {	envelope[selectedPoint][0] = x.clip(minx, maxx) });
		if (y.notNil, { envelope[selectedPoint][1] = y.clip(miny, maxy) });
		if (signalOnButtonRelease, {
			envelopeIsChanged = true;
		}, {
			this.emit('envelope-changed', envelope);
		});
		this.invalidate;
	}

	drawBackground {
		window.drawRectangle(whiteGC, true, 0, 0, width, height);
	}

	drawCaptions {
		window.drawLayout(captionGC, 5, 5, captionLayout);
	}

	updateCaptions {
		captionLayout.setText("Interpolation:\t" + style.asString ++ "\n" ++ "Y scale:\t\t" + yspec.asString ++ "\n");
		this.invalidate;
	}

	drawPoints {
		var x, y, gc;
		envelope.do({
			arg point, i;
			x = this.toWindowX(point[0]);
			y = this.toWindowY(point[1]);
			gc = if (selectedPoint.notNil and: { selectedPoint == i }, { blackGC }, {greyGC });
			window.drawArc(gc, false, x - 3, y - 3, 6, 6, 0, 360 * 64);
		});
		envelope.doAdjacentPairs({
			arg point1, point2;
			window.drawLine(greyGC, this.toWindowX(point1[0]), this.toWindowY(point1[1]),
					this.toWindowX(point2[0]), this.toWindowY(point2[1]));
		});
	}

	toWindowX { arg x; ^borderWidth + (x.asFloat * (width - (borderWidth * 2))) }
	toWindowY { arg y; ^borderWidth + ((1.0 - y.asFloat) * (height - (borderWidth * 2))) }
	toEnvelopeX { arg x; ^((x - borderWidth).asFloat / (width - (borderWidth * 2))) }
	toEnvelopeY { arg y; ^(1.0 - ((y - borderWidth).asFloat / (height - (borderWidth * 2)))) }

	style_ {
		arg value;
		style = value;
		this.updateCaptions;
	}
	yspec_ {
		arg value;
		yspec = value;
		this.updateCaptions;
	}

	points { ^envelope.deepCopy; }
	points_ {
		arg somePoints;
		envelope = Envelope.from(somePoints);
		this.emit('envelope-changed', this.points);
		this.invalidate;		
	}

	state { ^[this.points, style, yspec] }
	state_ { |x| this.points = x[0]; this.style = x[1]; this.yspec = x[2]; }

	asEnvelopeArray {
		var amparray, timearray, spec;

		spec = yspec.asSpec;
		amparray = envelope.collect({ |point| spec.map(point[1]) });
		timearray = Array.fill(envelope.size - 1, 0.0);
		envelope.doAdjacentPairs({ arg point1, point2, i; timearray[i] = point2[0] - point1[0] });

		^Env(amparray, timearray, style).asArray
	}

	size {
		arg...args; var w, h;

		if ((args.size == 1), {
			case([
				{ args[0] == 'free' } -> { w = -1; h = -1; },
				{ args[0] == 'mini' } -> { w = 320; h = 200; }
			]);
		}, {
			#w, h = args;
		});
		this.setSizeRequest(w, h);
		this.queueResize;
	}

	defaultSizeAsToplevel {	^#[320, 200] }

	finalize {
		this.unrefAll([greyGC, captionGC, captionLayout]);
		^super.finalize
	}
}
