GtkMethodsListView : GtkVBox {
	var <classObject;
	var scrolled, <model, view;
	var <selection;
	
	*new {
		arg classObject;
		^super.new(false, 0).initGtkMethodsListView(classObject);
	}
	initGtkMethodsListView {
		arg aClassObject;
		var renderer, column;

		classObject = aClassObject;

		scrolled = GtkScrolledWindow.new;
		scrolled.setPolicy(\automatic, \automatic);
		this.packStart(scrolled, true, true, 0);

		model = GtkTreeStore.new(\gchararray, \gint, \gboolean, \gchararray, \gchararray);

		view = GtkTreeView.newWithModel(model);
		selection = view.getSelection;
		selection.setMode('single');
		scrolled.add(view);

		renderer = GtkCellRendererText.new;
		column = GtkTreeViewColumn.newWithAttributes("Method", renderer, "text", 0, "weight", 1);
		view.appendColumn(column);

	/*	renderer = GtkCellRendererText.new;
		column = GtkTreeViewColumn.newWithAttributes("File", renderer, "text", 3);
		view.appendColumn(column);

		renderer = GtkCellRendererText.new;
		column = GtkTreeViewColumn.newWithAttributes("Line", renderer, "text", 4);
		view.appendColumn(column);
	*/
		this.rebuildModel;

		this.showAllChildren
	}

	rebuildModel {
		var classes, addedMethods, classIter;
		var name;

		model.clear;

		if (classObject.notNil, {
			addedMethods = List[];

			classes = [classObject] ++ classObject.superclasses;

			classes.do({
				arg aClass;

				if (aClass !== classObject, {
					classIter = model.append(nil);
					model.set(classIter, 0 -> aClass.asString, 1 -> 700, 2 -> false);
				}, {
					classIter = nil;
				});

				this.addMethodsFrom(aClass, "Instance methods", classIter, addedMethods);

				name = classObject.name.asString;

				if (name.beginsWith("Meta_").not, {
					aClass = ("Meta_" ++ name).asSymbol.asClass;
					if (aClass.notNil, {
						this.addMethodsFrom(aClass, "Class methods", classIter, addedMethods);
					});
				});
			});
		});
	}

	addMethodsFrom {
		arg aClass, title, parent, addedMethods;
		var addedHere, methodListIter, iter;

		if (aClass.methods.notNil, {
			methodListIter = model.append(parent);
			model.set(methodListIter, 0 -> title, 1 -> 700, 2 -> false);

			addedHere = List[];
			aClass.methods.do({
				arg method;
				if (addedMethods.includes(method).not, {
					addedMethods.add(method);
					addedHere.add(method);
				});
			});
			addedHere.sort({ arg a, b; a.name < b.name });
			addedHere.do({
				arg method;
				iter = model.append(methodListIter);
				model.set(iter, 0 -> method.name.asString,
						1 -> 400, 2 -> true,
						3 -> method.filenameSymbol.asString,
						4 -> method.charPos.asString);
			});
		});
	}

	classObject_ {
		arg aClassObject;

		classObject = aClassObject;
		this.rebuildModel;
	}

	location {
		var iter, name, filename, charPos, methodObject, classes;

		iter = selection.getSelected;
		if (iter.notNil and: { model.getValue(iter, 2) }, {
			name = model.getValue(iter, 0);
			filename = model.getValue(iter, 3);
			charPos = model.getValue(iter, 4).asInteger;
			^[filename, charPos]
		});
		^#[nil, nil]
	}
}
