GtkMultiPackager : GtkFrame {
	var <vchildren, <packagers, realchild, showFrame = true;
	var <label, currentPackagerName, updateLabelAction, <>defaultPackagerName = \default;

	*new {
		^super.new.initGtkMultiPackager
	}
	initGtkMultiPackager {
		updateLabelAction = {
			if (showFrame.not, {
				this.setLabel;
			}, {
				this.setLabel(this.prMakeLabel);
			});
		};
	
		vchildren = List[];
		packagers = IdentityDictionary[];
	}
	add { arg ...widgets; this.addAll(widgets) }
	addAll {
		arg ...args;
		args.do({
			arg widgets;
			widgets = widgets.collectNotNil({|w| w});
			vchildren.addAll(widgets);
			this.prRefAll(widgets)
		});
	}
	prRefAll { arg widgets; widgets.do({ |widget| widget.ref } ) }
	prUnrefVChildren { vchildren.do({ |widget| if (widget.ptr.notNil, { widget.unref }) }) }
	addPackager { arg id, packager; packagers[id] = packager }
	replacePackager { arg id, packager; this.addPackager(id, packager) }
	showAllChildren { arg depth; if (realchild.notNil, { realchild.showAllChildren(depth) }) }
	switch {
		arg name;
		if (packagers.includesKey(name), {
			if (realchild.notNil, {
				this.prRemoveAllFromRealParents;
				this.remove(realchild);
			});
			realchild = packagers[name].value;
			super.add(realchild);
			currentPackagerName = name;
			this.prUpdateLabel;
			realchild.show;
		});
	}
	cycle {
		arg dir = \forward;
		var names, index, step;
		step = if (dir === \forward, { 1 }, { -1 });
		names = packagers.keys.asArray.sort;
		index = names.indexOf(currentPackagerName);
		if (index.notNil, {
			this.switch(names@@(index+step));
		});
	}
	prRemoveAllFromRealParents { vchildren.do({ |widget| widget.removeFromParent }) }
	clear {
		this.prUnrefVChildren;
		if (realchild.notNil, {
			this.remove(realchild);
		});
		vchildren.clear;
	}
	showFrame_ {
		arg value;
		if (value, {
			this.setShadowType('etched-in');
		}, {
			this.setShadowType('none');
		});
		this.prUpdateLabel;
		showFrame = value;
	}	
	label_ { arg aLabel; label = aLabel; this.prUpdateLabel; }
	prUpdateLabel { updateLabelAction.value }

	prMakeLabel {
		if (label.isNil, {
			^currentPackagerName.asString;
		}, {
			^(label + "(" ++ currentPackagerName.asString ++ ")")
		});
	}

	show {
		if (currentPackagerName.isNil, {
			this.switch(this.defaultPackagerName);
		});
		super.show;
	}

	finalize {
		this.prUnrefVChildren;
		vchildren.clear;
		super.finalize
	}

	asToplevel {
		arg w, h;
		var window;
		window = super.asToplevel(w, h);
		this.showFrame = false;
		this.setLabel;
		updateLabelAction = {
			window.setTitle(this.prMakeLabel);
		};
		this.prUpdateLabel;
		^window
	}
}
