GtkMultiSourceEditor : GtkFrame {
	var <editor, <buffers, <>path;
	*new {
		^super.new.initGtkMultiSourceEditor
	}
	initGtkMultiSourceEditor {

		editor = GtkSourceEditor.new;
		this.add(editor.show);		

		buffers = Dictionary[];
		this.setLabel(this.prAddBuffer(nil, editor.buffer));

		editor.view.connect('populate-popup', {
			arg view, menu;		var items;
			menu.appendItems([
				\separator,
				"Buffer" -> [
					"New" -> {
						this.prAddBuffer;
						this.prRenameCurrentBufferFromUI
					},
					"New scratch" -> { this.prAddBuffer },
					"Rename current" -> { this.prRenameCurrentBufferFromUI },
					"Remove current" -> { this.prRemoveCurrentBufferFromUI }
				]
			]);

			items = [];
			buffers.keys.asArray.sort.do({
				arg name, i;
				items = items.add((i+1).asString ++ ":" + name -> {
					this.switch(name);
				});
			});

			if (buffers.size < 8, {
				menu.appendItems([\separator] ++ items);
			}, {
				menu.appendItems(["Buffers" -> items]);
			});
		});

	}
	prAddBuffer {
		arg name, buffer, source, switch = true;
		var altname, index = 0;
		
		name = name ?? { "Unnamed" + Date.getDate.stamp	};
		buffer = buffer ?? { GtkSourceBuffer.new };

		while({
			altname = name ++ if (index == 0, { "" }, { "_" ++ index.asStringPad(2) });
			buffers.keys.includes(altname)
		}, {
			index = index + 1;
		});

		name = altname;

		buffer.ref;
		buffers.put(name, buffer);

		if (source.notNil, {
			buffer.setText(source.asString);
		});

		if (switch, { this.switch(name) });

		^name
	}
	switch {
		arg name;		var buffer;
		buffer = buffers[name];
		if (buffer.notNil, {
			editor.buffer = buffer;
			this.setLabel(name);
		});
	}

	prRenameCurrentBufferFromUI {
		"New name: ".asGtkDialog({
			arg newName;		var oldName;
			if (buffers.includesKey(newName), {
				this.prRenameCurrentBufferFromUI;
			}, {
				this.setLabel(newName);
				oldName = buffers.findKeyForValue(editor.buffer);
				buffers[newName] = editor.buffer;
				buffers.removeAt(oldName);
			});
		})
	}

	prRemoveCurrentBufferFromUI {
		this.removeBuffer(buffers.findKeyForValue(editor.buffer))
	}

	removeBuffer {
		arg name;
		buffers.removeAt(name);
		editor.buffer.unref;
		if (buffers.size > 0, {
			name = buffers.keys.asArray.sort.first;
			this.switch(name);
		}, {
			this.prAddBuffer;
		});
	}

	sources { ^buffers.collect({ arg buffer; buffer.getText }) }
	sources_ {
		arg sources;	var toBeRemoved;

		this.clear;
		toBeRemoved = buffers.keys;

		sources.keysValuesDo({
			arg name, source;
			this.prAddBuffer(name, nil, source);
		});
		toBeRemoved.do({ |name| this.removeBuffer(name) });

	}

	clear {
		this.unrefAll(buffers);
		buffers = Dictionary[];
		this.prAddBuffer;
	}

	finalize {
		this.unrefAll(buffers);
		super.finalize;
	}
}
