GtkPlusKbd {
	classvar <bindings;
	classvar <>echo = false, <>locked = false;
	classvar lastCombination, forgetNextRelease;

	*initClass {
		bindings = IdentityDictionary[\global -> KbdBindingSet.new];
		lastCombination = nil;
		forgetNextRelease = false;
	}

	*process {
		arg source, combination, direction;

		if (#["Shift_L", "Shift_R", "Control_L", "Control_R", "Alt_L", "Mode_switch",
			"Caps_Lock", "Num_Lock", "Scroll_Lock"].occurencesOf(combination.last) > 0, {
			^this
		});

		if (direction == \press, {
			if (lastCombination.notNil, {
				if (combination != lastCombination[1], {
					if (echo, { Post << "Forcing release because of switch" << nl });
					this.process(lastCombination[0], lastCombination[1], \release);
					forgetNextRelease = true;
					this.process(source, combination, \press);
				}, {
				});
				^nil
			});
			forgetNextRelease = false;
			lastCombination = [source, combination];
		}, {
			if (forgetNextRelease, {
				forgetNextRelease = false;
				^nil
			});
 			lastCombination = nil;
		});

		if (echo, {
			Post << "GtkPlusKbd-process: " << combination << if (direction == \press, {
				" pressed"
			}, {
				" released"
			}) << nl;
		});

		this.performReaction(source, combination, direction);
	}

	*performReaction {
		arg source, combination, direction;
		var action;
		
		bindings[source].do({
			arg set;
			action = set.find(direction, combination);
			if (action.notNil, {
				action.value(combination, direction)
				^this
			});
		});
		bindings[\global].find(direction, combination).value(combination, direction)
	}

	*enable {
		arg window;

		window.addEvents(['key-press-mask', 'key-release-mask']);

		window.connect('key-press-event', {
			arg window, event;
			if (GtkPlusKbd.locked, {
				false
			}, {
				if (event.notNil, {
					GtkPlusKbd.process(window, event.modsAndKeyString, \press);
				});
				true
			});
		});

		window.connect('key-release-event', {
			arg window, event;
			if (GtkPlusKbd.locked, {
				false
			}, {
				if (event.notNil, {
					GtkPlusKbd.process(window, event.modsAndKeyString, \release);
				});
				true
			});
		});
	}

	*bind {
		arg direction ...combination;				var action;
		action = combination.pop;
		this.bindCombination(direction, combination, action);
	}
	*bindCombination {
		arg direction, combination, action;
		if (combination.notNil, {
			bindings[\global].addCombination(direction, combination, action);
		});
	}

	*addBindingSet {
		arg source... sets;

		sets.do({
			arg set;

			if (bindings[source].isNil, {
				bindings[source] = List[];
			});
			if (bindings[source].includes(set).not, {
				bindings[source].add(set);
			});
		});
	}
	*removeBindingSet {
		arg source ...sets;
		sets.do({
			arg set;
			if (bindings[source].notNil, {
				bindings[source].remove(set);
			});
		});
	}
	*setBindingSets {
		arg source ...sets;
		bindings[source] = sets.asList;
	}

	*unbind {
		arg direction ...combination;
		bindings[\global][direction].removeAt(combination.join("-"));
	}

	*clear {
		bindings[\global][\press].clear;
		bindings[\global][\release].clear;
	}
}
