GtkPresetView : GtkMultiPackager {
	var <presets, takeAction, restoreAction;
	var <nSpinner, <nameEntry;
	var <storeButton, <newButton, <recallButton, <loadButton, <saveButton;

	*new {
		arg presets, takeAction, restoreAction, label, buttonStyle = \textual;
		^super.new.initGtkPresetView(presets, takeAction, restoreAction, label, buttonStyle)
	}
	of {
		arg ...items;
		takeAction = { items.collectMsg(\state) };
		restoreAction = { |data| items.do({ |item, i| item.state = data[i] }) };
	}
	initGtkPresetView {
		arg somePresets, aTakeAction, aRestoreAction, aLabel, buttonStyle;
		var box, table;

		presets = somePresets;
		this.watch('changed', presets, { this.onPresetChanged });
		takeAction = aTakeAction;
		restoreAction = aRestoreAction;

		this.label = aLabel;
		this.showFrame = false;

		nSpinner = GtkIntNumberBox(0, max(presets.size - 1, 0), {
			this.restoreCurrentPreset
		});

		nameEntry = GtkTextField.new;
		if (presets.size > 0, { nameEntry.setText(presets[0][0]) });
		nameEntry.connect('activate', {
			var n;
			n = nSpinner.value;
			if (n < presets.size, {
				presets[n][0] = nameEntry.getText;
			});
		});

		if (buttonStyle == \iconic, {
			storeButton = GtkButton.newWithStockImage('gtk-color-picker', callback: { this.store });
			newButton = GtkButton.newWithStockImage('gtk-new', callback: { this.new });
			recallButton = GtkButton.newWithStockImage('gtk-jump-to', callback: { this.recall });
		
			loadButton = GtkButton.newWithStockImage('gtk-open', callback: { this.load });
			saveButton = GtkButton.newWithStockImage('gtk-save-as', callback: { this.save });
		}, {
			storeButton = GtkButton.newWithLabelAndCallback("take", { this.store });
			newButton = GtkButton.newWithLabelAndCallback("new", { this.new });
			recallButton = GtkButton.newWithLabelAndCallback("restore", { this.recall });
		
			loadButton = GtkButton.newWithLabelAndCallback("load", { this.load });
			saveButton = GtkButton.newWithLabelAndCallback("save", { this.save });
		});

		this.add(nSpinner, nameEntry, storeButton, newButton, recallButton, loadButton, saveButton);

		this.addPackager(\normal, {
			var mainbox, table;
	
			mainbox = GtkVBox.new;

			table = GtkTable(2, 2);
			mainbox.packStart(table, false, false);

			table.attach(GtkLabel("Number:"), 0, 1, 0, 1, 0, 0);		
			table.attach(nSpinner, 1, 2, 0, 1, [\expand, \fill], 0);

			table.attach(GtkLabel("Name:"), 0, 1, 1, 2, 0, 0);
			nameEntry.locked = false;
			table.attach(nameEntry, 1, 2, 1, 2, [\expand, \fill], 0);

			box = GtkHBox.new;
			mainbox.packStart(box);		

			box.packMany([storeButton, newButton, recallButton, loadButton, saveButton]);
			mainbox.showAllChildren;

			mainbox
		});
		defaultPackagerName = \normal;

		this.addPackager(\mini, {
			var mainbox, table;
	
			mainbox = GtkHBox.new;

			mainbox.packStart(nSpinner, false, false);
			nameEntry.locked = true;
			mainbox.packStart(nameEntry, true, true);

			mainbox.showAllChildren;
		});

		this.restoreCurrentPreset;
	}

	store {
		arg n, data, name;

		n = n ? nSpinner.getValue.asInteger;
		if (n < presets.size, {
			data = data ? takeAction.value;
			presets.update(n, name, data);
		});
	}

	recall {
		arg preset;
		preset = preset ? this.currentPreset;
		if (preset.notNil, { restoreAction.value(preset[1]) });
	}

	new {
		arg data, name;
		nameEntry.setText(presets.add(takeAction.value));
		nSpinner.value = presets.size - 1;
	}

	load {
		GtkFileSelection.askAndDo("Load presets", {
			arg filename;
			presets.read(filename);
		});
	}

	save {
		GtkFileSelection.askAndDo("Save presets", {
			arg filename;
			presets.write(filename);
		});
	}

	showFileButtons {
		arg value = true;
		if (value, {
			loadButton.show;
			saveButton.show;
		}, {
			loadButton.hide;
			saveButton.hide;
		});
	}

	currentPreset {	^presets[nSpinner.value] }

	restoreCurrentPreset {
		var preset;
		preset = this.currentPreset;
		if (preset.notNil, {
			restoreAction.value(preset[1]);
			nameEntry.setText(preset[0]);
		});
	}

	onPresetChanged {
		arg somePresets;
		nSpinner.setRange(0, max(presets.size - 1, 1), 1);
		if (presets.size > 0, {
			nameEntry.setText(presets[nSpinner.value][0]);
		});
		this.restoreCurrentPreset;
	}
}