GtkSCCodeView : GtkVBox {
	var <scrolled, <textbuffer, <textview;
	var outputbuffer, filename;
	var filenameLabel, positionLabel;

	*new {
		^super.new(false, 0).initGtkSCCodeView
	}

	initGtkSCCodeView {
		var box;

		box = GtkHBox.new(false, 0);
		this.packStart(box, false, false, 0);

		filenameLabel = GtkLabel.new("");
		box.packStart(filenameLabel, true, true, 0);

		positionLabel = GtkLabel.new("");
		positionLabel.setSizeRequest(60, -1);
		box.packStart(positionLabel, false, false, 0);

		scrolled = GtkScrolledWindow.new;
		scrolled.setPolicy(\automatic, \automatic);
		this.packStart(scrolled, true, true, 0);

		outputbuffer = GtkConsoleOutputBuffer.get;

		textview = GtkTextView.new;
		textbuffer = textview.getBuffer;
		scrolled.add(textview);

		textview.connect('key-press-event', {
			arg widget, event;
			this.onKeyPressEvent(event)
		});

		this.showAllChildren
	}

	filename_ {
		arg aFilename;

		filename = aFilename;
		filenameLabel.setText(filename);
	}

	updatePositionLabel {
		arg line = 0, column = 0, char = 0;
		positionLabel.setText(line.asString ++ ":" ++ column ++ ":" ++ char);		
	}

	load {
		arg aFilename, charPos;
		var f;

		if (aFilename != filename, {
			this.filename = aFilename;

			if (filename.notNil, {
				f = File.new(filename, "r");
				textbuffer.setText(f.readAllString);
				f.close
			}, {
				textbuffer.setText("");
			});
		});
		if (charPos.notNil, {
			this.scrollToChar(charPos);
		});
	}

	*load {
		arg filename, charPos;
		^this.new.load(filename, charPos);
	}

	scrollToChar {
		arg charPos;
		var iter;
		if (charPos.notNil, {
			iter = textbuffer.getIterAtOffset(charPos);
			textview.scrollToIter(iter, 0.0, 0.0, 0.0);
			this.updatePositionLabel(0, 0, charPos);
		});
	}

	sourceOf {
		arg object;
		if (object.notNil, {
			this.load(object.filenameSymbol.asString.standardizePath)
		}, {
			textbuffer.setText("");
		});
	}

	*sourceOf {
		arg object;
		^this.new.sourceOf(object)
	}


	onKeyPressEvent {
		arg event;
		var text;

		if (event.isNil or: { event.keyval != Gdk.keys[\Return] }, {
			^false
		});

		case([
			{ event.state == Gdk.shift_mask } -> {
				text = textbuffer.getText;
				this.interpret(text);
				^true
			},
			{ event.state == Gdk.mod1_mask } -> {
				text = textbuffer.getText;
				this.interpret(text);
				^true
			},
			{ event.state == Gdk.control_mask } -> {
				text = textbuffer.getText;
				this.interpret(text);
				^true
			},

		]);
 
		false
	}

	interpret {
		arg text;
		if (text.last != Char.nl, { text = text.add(Char.nl) });
		outputbuffer.appendWithTag(text, 'command');
		text.interpretPrint;
	}
}
