GtkSCProjectView : GtkMultiPackager {
	var <consoleView, <editor, <path;
	var consoleScrolled, editorScrolled;
	var consoleSignalC;

	*new {
		^super.new.initGtkSCProjectView
	}
	initGtkSCProjectView {
		consoleView = GtkTextView.newWithBuffer(GtkConsoleOutputBuffer.get);
		editor = GtkMultiSourceEditor.new;		

		consoleView.setEditable(false);
		consoleSignalC = GtkConsoleOutputBuffer.get.connect('changed', {
			consoleView.scrollToEnd
		});
		consoleView.scrollToEnd;

		editor.editor.view.connect('populate-popup', {
			arg view, menu;		var items;
			menu.appendItems([
				\separator,
				"Clear console" -> { GtkConsoleOutputBuffer.get.setText("") },
				"Cycle arrangement" -> { this.cycle },
				\separator,
				"Project" -> [
					"Load" -> { this.prLoadFromUI(true) },
					"Revert" -> { this.prLoadFromUI },
					"Save" -> { this.prSaveFromUI },
					"Save as..." -> { this.prSaveFromUI(true) },
					"Clear" -> { this.clear; path = nil }
				]
			]);
		});

		consoleScrolled = GtkScrolledWindow.new.setPolicy(\automatic, \automatic);
		consoleScrolled.add(consoleView.show);
		editorScrolled = GtkScrolledWindow.new.setPolicy(\automatic, \automatic);
		editorScrolled.addWithViewport(editor.show);

		this.add(consoleScrolled, editorScrolled);

		this.addPackager(\vertical, {
			var box, paned;

			paned = GtkVPaned.new;
			paned.add1(consoleScrolled);
			paned.add2(editorScrolled);
			
			paned.setPosition(0);
			paned.showChildren;

			paned
		});
		this.addPackager(\horizontal, {
			var box, paned;

			paned = GtkHPaned.new;
			paned.add1(consoleScrolled);
			paned.add2(editorScrolled);

			paned.setPosition(0);
			paned.showChildren;
			
			paned
		});
		this.defaultPackagerName = \vertical;
	}

	clear {
		editor.clear;
	}

	load {
		arg filename;
		var project, sources;
		if (File.exists(filename), {
			project = Object.readArchive(filename);
			editor.sources = project[\sources];
			path = filename;
		});
	}
	save {
		arg filename;
		var project;

		if (File.exists(filename), {
			unixCmd("cp " ++ filename + " " + filename ++ ".backup-" ++ Date.getDate.asSortableString);
		});
		project = IdentityDictionary[];
		project[\sources] = editor.sources;
		project.writeArchive(filename);
		path = filename;
	}

	prSaveFromUI {
		arg askname = false;
		if (askname or: { path.isNil }, {
			GtkFileSelection.askAndDo("Write project", {
				arg filename;
				this.save(filename);
			});
		}, {
			this.save(path);
		});
	}
	prLoadFromUI {
		arg askname = false;
		if (askname or: { path.isNil }, {
			GtkFileSelection.askAndDo("Read project", {
				arg filename;
				this.load(filename);
			});
		}, {
			this.load(path);
		});
	}

	defaultSizeAsToplevel {	^#[800, 600] }

	finalize {
		GtkConsoleOutputBuffer.get.disconnect(consoleSignalC);
		super.finalize;
	}
}
