GtkSampleSetSelector : GtkMultiGroupAndElementSelector {
	classvar <>defaultSampleRate = 44100, <playbackSynthdef;
	var <sizeLabels, <playbackButtons;
	*initClass {
		playbackSynthdef = SynthDef(this.class.name.asString ++ "-play", {
			arg out = 0, bufnum = 0, duration = 1.0, attack = 0.001, decay = 0.001;
			var output;
			output = PlayBuf.ar(1, bufnum, BufRateScale.kr(bufnum), 1.0) *
				Linen.kr(Trig.kr(1.0, duration - decay), attack, 1.0, decay, doneAction: 2);
			Out.ar(out, [output, output]);
		});
	}
	*new {
		arg groups, n, label;
		^super.new(groups, n, label).initGtkSampleSetSelector
	}
	initGtkSampleSetSelector {
		var label, button;

		this.watch('selection-changed', this, {
			arg k, i ,j;
			sizeLabels[k].setText(this.prSizeLabelText(k), 0.0, 0.5);
		});

		this.replacePackager(\normal, {
			var table, n;

			n = selectors.size;
			table = GtkTable(if (n > 1, { n + 1 }, { 1 }), 5);
			table.packArrays(
				selectors, [0, 0],
				rndButtons, [0, 0],
				rndInGrpButtons, [0, 0],
				sizeLabels, [0, 0],
				playbackButtons, [0, 0]
			);
			if (n > 1, {
				table.attach(allRndButton.show, 1, 2, n, n + 1);
				table.attach(allRndInGrpButton.show, 2, 3, n, n + 1);
			});
			table
		});

		this.addPackager(\tworows, {
			var table, n;

			n = selectors.size;
			table = GtkTable(if (n > 1, { (2*n) + 1 }, { 2 }), 3);
			n.do({
				arg i;
				table.attach(selectors[i], 0, 1, (i*2), (i*2) + 1);
				table.attach(rndButtons[i], 1, 2, (i*2), (i*2) + 1);
				table.attach(rndInGrpButtons[i], 2, 3, (i*2), (i*2) + 1);
				table.attach(sizeLabels[i], 0, 2, (i*2) + 1, (i*2) + 2);
				table.attach(playbackButtons[i], 2, 3, (i*2) + 1, (i*2) + 2);
			});
			if (n > 1, {
				table.attach(allRndButton, 1, 2, 2*n, (2*n) + 1);
				table.attach(allRndInGrpButton, 2, 3, 2*n, (2*n) + 1);
			});
			table.showChildren;
			table
		});
	}

	prMakeWidgets {
		arg groups, n;

		super.prMakeWidgets(groups, n);

		sizeLabels = Array.fill(n, {
			arg k;
			GtkLabel(this.prSizeLabelText(k)).setSizeRequest(120, -1);
		});
		playbackButtons = Array.fill(n, {
			arg k;
			GtkTriggerButton("Play", {
				var synth, buffer, msg;

				buffer = selectors[k].element;
                synth = Synth.basicNew(this.class.playbackSynthdef, buffer.server, -1);
                msg = synth.newMsg(buffer.server.asTarget, \addToTail, [
					\bufnum, buffer.bufnum,
					\duration, buffer.numFrames.asFloat / GtkSampleSetSelector.defaultSampleRate
				]);

				this.class.playbackSynthdef.send(buffer.server, msg);
			});
		});

		this.addAll(sizeLabels, playbackButtons);
	}

	prSizeLabelText {
		arg k;			var buffer, frames, text;
		buffer = selectors[k].element;
		text = if (buffer.notNil, {
			frames = buffer.numFrames;
			frames.asString +"-" + (frames.asFloat / GtkSampleSetSelector.defaultSampleRate).asStringPrec(5);
		}, {
			""
		});
		^text
	}

	shortestSampleDuration {
		^selectors.collectNotNil({ |selector| selector.element.numFrames / GtkSampleSetSelector.defaultSampleRate }).minItem
	}

	finalize {
		this.forgetEverything;
		super.finalize;
	}
}
