GtkScribble : GtkWindow {
	var drawingarea, window, whiteGC, blackGC;
	var hasAllocatedResources;
	var points, nPoints, count;

	*new {
		^super.new.init;
	}
	init {
		this.setDefaultSize(640, 400);
		this.setTitle("Scribble");

		points = [];
		nPoints = 4;
		count = 0;
		hasAllocatedResources = false;

		drawingarea = GtkDrawingArea.new;
		drawingarea.addEvents([	'exposure-mask',
					'button-press-mask', 'button-release-mask',
					'pointer-motion-mask', 'pointer-motion-hint-mask'	]);

		this.add(drawingarea);

		drawingarea.connect("motion-notify-event", {
			arg widget, event;
			this.onMotionNotifyEvent(event)
		});

		drawingarea.connect("realize", {
			window = drawingarea.window;
			window.setCursor("crosshair");
		});

		drawingarea.connect("configure-event", {
			true;
		});

		drawingarea.connect("expose-event", {
			this.allocateResources;
			this.redraw;
			false;
		});

		this.showAllChildren(true);
	}

	allocateResources {
		if (hasAllocatedResources, { ^this; });

		whiteGC = drawingarea.getStyle.getWhiteGC;
		blackGC = drawingarea.getStyle.getBlackGC;

		hasAllocatedResources = true;
	}
	redraw {
		var w, h;

		if (window.notNil, {
			#w, h = window.getSize;
			window.drawRectangle(whiteGC, true, 0, 0, w, h);

/*			points.do({
				arg point;
				window.drawPoint(blackGC, w * point.x, h * point.y)
			});
*/
			points.doAdjacentPairs({
				arg point1, point2;
				window.drawLine(blackGC, w * point1.x, h * point1.y, w * point2.x, h * point2.y)
			});
		});
	}

	onMotionNotifyEvent {
		arg event;
		var x, y, z, state;
		var w, h;

		if (event.isNil, { ^true });

		#w, h = window.getSize;

		if (event.isHint, {
			state = event.getDeviceState;
			x = event.axis(1) / w;
			y = event.axis(2) / h;
			z = event.axis(3);
		}, {
			#x, y = event.coords;

			x = x.asFloat / w;
			y = y.asFloat / h;
			z = event.axis(3);
			state = event.state;
		});

		//[x, y, z].postln;

		if ((state & Gdk.button1_mask) != 0, {
			points = points.add(Point(x, y));
			if (points.size > nPoints, { points.removeAt(0) });

			count = count + 1;
			if (count > 30, {
				count = 0;
				nPoints = nPoints + 1;
			});

			this.invalidate;
		});
	}

	invalidate {
		drawingarea.queueDraw;
	}
}
