GtkSensitiveArea : GtkDrawingArea {
	classvar <driven, refreshTimeout, <>refreshDelay = 40;
	var window, realized = false, needRefresh = true, width, height, <driven = false;

	var >motionAction, >buttonPressAction, >buttonReleaseAction;
	var >proximityInAction, >proximityOutAction, >keyAction;
	var >leaveAction, >enterAction;

	*initClass {
		driven = List[];
	}
	*new {
		^super.new.initGtkSensitiveArea
	}
	initGtkSensitiveArea {
		XInput.enabled = true;

		this.setExtensionEvents(\all);
		this.addEvents([
				'leave-notify-mask', 'enter-notify-mask',
				'button-press-mask', 'button-release-mask',
				'pointer-motion-mask', 'pointer-motion-hint-mask',
				'proximity-in-mask', 'proximity-out-mask'
		]);

		this.connect('realize', {
			//"REALIZE EVENT".postln;
			window = this.window;
			window.setCursor('crosshair');
			#width, height = window.getSize;
			realized = true;
		});

		this.connect('configure-event', {
			//"CONFIGURE EVENT".postln;
			window = this.window;
			if (window.notNil, {
				#width, height = window.getSize;
			});
			true;
		});

		this.connect('motion-notify-event', {
			arg widget, event;
			this.onMotionNotifyEvent(event);
			true
		});
		this.connect('button-press-event', {
			arg widget, event;
			this.onButtonPressEvent(event);
			true
		});
		this.connect('button-release-event', {
			arg widget, event;
			this.onButtonReleaseEvent(event);
			true
		});
		this.connect('proximity-in-event', {
			arg widget, event;
			//"PROXIMITY IN".postln;
			proximityInAction.value(widget, event);
			true
		});
		this.connect('proximity-out-event', {
			arg widget, event;
			//"PROXIMITY OUT".postln;
			proximityOutAction.value(widget, event);
			true
		});
		this.connect('leave-notify-event', {
			arg widget, event;
			leaveAction.value(widget, event);
			true
		});
		this.connect('enter-notify-event', {
			arg widget, event;
			enterAction.value(widget, event);
			true
		});
		this.connect('key-press-event', {
			arg widget, event;
			\x.postln;
			this.onKeyEvent(event, \press);
		});
		this.connect('key-release-event', {
			arg widget, event;
			\x.postln;
			this.onKeyEvent(event, \release);
		});

		keyAction = { false };
	}

	onMotionNotifyEvent {
		arg event;
		var x, y, z, w, h, state, marray, barray;

		if (event.isNil, { ^true });

		if (event.isHint, {
			state = event.getDeviceState;
			x = event.axis(1) / width;
			y = event.axis(2) / height;
			z = event.axis(3);
		}, {
			#x, y = event.coords;

			x = x.asFloat / width;
			y = y.asFloat / height;
			z = event.axis(3);
			state = event.state;
		});

		#marray, barray = this.translateInfo(event.state);
		motionAction.value(x, y, z, marray, barray);
	}

	onButtonPressEvent {
		arg event;
		var x, y, z, state, marray, barray;

		if (event.isNil, { ^true });

		state = event.getDeviceState;
		#x, y = event.coords;
		x = x.asFloat / width;
		y = y.asFloat / height;
		z = event.axis(3);

		#marray, barray = this.translateInfo(event.state, event.button);
		buttonPressAction.value(x, y, z, marray, barray);
	}

	onButtonReleaseEvent {
		arg event;
		var x, y, z, state, marray, barray;

		if (event.isNil, { ^true });

		state = event.getDeviceState;
		#x, y = event.coords;
		x = x.asFloat / width;
		y = y.asFloat / height;
		z = 0.0;

		#marray, barray = this.translateInfo(event.state, event.button);
		buttonReleaseAction.value(x, y, z, marray, barray);
	}

	translateInfo {
		arg state, inButtons;
		var buttons, marray, barray;

		if (inButtons.notNil, {
			buttons = inButtons;
		}, {
			buttons = (state >> 8) & 31;
		});

		barray = Array.new(3);

		if ((state & Gdk.button1_mask) != 0, { barray.add(1) });
		if ((state & Gdk.button2_mask) != 0, { barray.add(2) });
		if ((state & Gdk.button3_mask) != 0, { barray.add(3) });

		marray = [];

		if ((state & Gdk.shift_mask) != 0, {
			marray = marray.add('Shift');
		});
		if ((state & Gdk.control_mask) != 0, {
			marray = marray.add('Control');
		});
		if ((state & Gdk.mod1_mask) != 0, {
			marray = marray.add('Alt');
		});
		if ((state & Gdk.mod3_mask) != 0, {
			marray = marray.add('Mod3');
		});
		if ((state & Gdk.mod4_mask) != 0, {
			marray = marray.add('Mod4');
		});
		if ((state & Gdk.mod5_mask) != 0, {
			marray = marray.add('Mod5');
		});

		^[marray, barray]
	}

	invalidate {
		if (driven, {
			needRefresh = true
		}, {
			this.queueDraw;
		});
	}

	queueRedrawIfNeeded {
		if (realized, {
			if (needRefresh, {
				this.queueDraw;
				needRefresh = false;
			});
		});
	}

	onKeyEvent {
		arg event, direction;
		var stateName, keyName;
		if (event.notNil, {
			#stateName, keyName = event.keyCombinationString; 
			^keyAction.value(stateName, keyName, direction);
		});
		^false
	}

	driven_ {
		arg value;
		driven = value;
		if (driven, {
			GtkSensitiveArea.addDriven(this);
		}, {
		});
	} 

	finalize {
		if (driven, { GtkSensitiveArea.removeDriven(this) });
		super.finalize;
	}

	*addDriven {
		arg area;
		driven.add(area);
		if (driven.size == 1, {
			refreshTimeout = Gtk.timeoutAdd(refreshDelay, {
				driven.doMsg(\queueRedrawIfNeeded);
				refreshDelay
			});
		});
	}
	*removeDriven {
		arg area;
		driven.remove(area);
		if (driven.size == 0, {
			Gtk.timeoutRemove(refreshTimeout);
		});
	}
	*forgetAllDriven {
		driven.clear;
		Gtk.timeoutRemove(refreshTimeout);
	}
}
