GtkServerStatusDisplay : GtkVBox {
	var server;
	var scrolled, model, treeview, entries;
	var <active, timeout;

	*new {
		arg aServer, inActive = true;
		^super.new(false, 0).initGtkServerStatusDisplay(aServer, inActive)
	}

	initGtkServerStatusDisplay {
		arg aServer, inActive;
		var renderer, column, iter;

		server = aServer ? Server.default;

		scrolled = GtkScrolledWindow.new;
		scrolled.setShadowType('etched-in');
		scrolled.setPolicy(\never, \automatic);
		this.packStart(scrolled, true, true, 0);

		model = GtkTreeStore.new(\gchararray, \gchararray, \gint);

		treeview = GtkTreeView.newWithModel(model);
		treeview.getSelection.setMode('none');
		scrolled.add(treeview);

		renderer = GtkCellRendererText.new;
		column = GtkTreeViewColumn.newWithAttributes("Name", renderer, "text", 0, "weight", 2);
		treeview.appendColumn(column);

		renderer = GtkCellRendererText.new;
		column = GtkTreeViewColumn.newWithAttributes("Value", renderer, "text", 1);
		treeview.appendColumn(column);

		entries = #[\UGens, \Synths, \Groups, \SynthDefs, \AvgCPU, \PeakCPU];

		entries.do({
			arg name;
			model.set(model.append, 0 -> (name ++ ":"), 1 -> "n/a", 2 -> 700);
		});

		if (inActive, {
			this.active = true;
		}, {
			server.addDependant(this);
			server.addStatusWatcher;
		});

		this.showAllChildren
	}

	active_ {
		arg value;
		if (value == active, { ^this });

		if (value == true, {
			server.removeDependant(this);

			timeout = Gtk.timeoutAdd(1000, {
				this.timeoutHandler;
				true;
			});
		}, {
			Gtk.timeoutRemove(timeout);
			timeout = nil;

			server.addDependant(this);
			server.addStatusWatcher;
		});
		
		active = value;
	}

	timeoutHandler {
		this.updateEntries
	}

	set {
		arg key, value;
		var index, iter;

		index = entries.indexOf(key);
		if (index.isNil, { ^this });
	
		iter = model.getIterFromString(index.asString);
		if (iter.isNil, { ^this });

		if (iter.isKindOf(GtkTreeIter).not, {
			Post << "Eh... WTF... I got a " << iter << " instead of a GtkTreeIter" << nl;
			^this
		});

		model.set(iter, 1 -> value.asString);
	}

	updateEntries {
		this.set(\UGens, server.numUGens);
		this.set(\Synths, server.numSynths);
		this.set(\Groups, server.numGroups);
		this.set(\SynthDefs, server.numSynthDefs);
		this.set(\AvgCPU, server.avgCPU);
		this.set(\PeakCPU, server.peakCPU);
	}

	update {
		arg changed;

		if (changed == server, {
			{ this.updateEntries }.defer;
		});
	}

	finalize {
		if (active, {
			Gtk.timeoutRemove(timeout);
		}, {
			server.removeDependant(this);
		});
	}

	defaultSizeAsToplevel {	^#[320, 400] }
}
