GtkServerStatusLine : GtkHBox {
	var server;
	var entries, labels;
	var <active, timeout;

	*new {
		arg aServer, inActive = true;
		^super.new(false, 0).initGtkServerStatusDisplay(aServer, inActive)
	}

	initGtkServerStatusDisplay {
		arg aServer, inActive;
		var label;

		server = aServer ? Server.default;

		entries = #[\UGens, \Synths, \Groups, \SynthDefs, \AvgCPU, \PeakCPU];

		labels = Array.fill(entries.size, {
			arg i;
			this.packStart(GtkLabel(entries[i].asString ++ ": "), false, false);
			label = GtkLabel.new;
			label.setSizeRequest(#[40, 40, 40, 40, 100, 100][i], -1);
			this.packStart(label, false, false);
			label
		});

		if (inActive, {
			this.active = true;
		}, {
			server.addDependant(this);
			server.addStatusWatcher;
		});

		this.showChildren
	}

	active_ {
		arg value;
		if (value == active, { ^this });

		if (value == true, {
			server.removeDependant(this);

			timeout = Gtk.timeoutAdd(1000, {
				this.timeoutHandler;
				true;
			});
		}, {
			Gtk.timeoutRemove(timeout);
			timeout = nil;

			server.addDependant(this);
			server.addStatusWatcher;
		});
		
		active = value;
	}

	timeoutHandler {
		this.updateEntries
	}

	updateEntries {
		labels[0].setText(server.numUGens.asString);
		labels[1].setText(server.numSynths.asString);
		labels[2].setText(server.numGroups.asString);
		labels[3].setText(server.numSynthDefs.asString);
		labels[4].setText(server.avgCPU.asStringPrec(3));
		labels[5].setText(server.peakCPU.asStringPrec(3));
	}

	update {
		arg changed;

		if (changed == server, {
			{ this.updateEntries }.defer;
		});
	}

	finalize {
		if (active, {
			Gtk.timeoutRemove(timeout);
		}, {
			server.removeDependant(this);
		});
	}

	defaultSizeAsToplevel {	^#[400, 32] }
}
