// It's derived from GtkPatchView

GtkSoundSystemView : GtkVBox {
	var system;
	var innerBox, buttonBox;

	*new {
		arg aSystem, depth = inf;
		^super.new(false, 2).initGtkSoundSystemView(aSystem, depth)
	}

	initGtkSoundSystemView {
		arg aSystem, depth;
		var frame, includedPatches;

		system = aSystem;

		frame = GtkFrame.new;
		this.packStart(frame, false, false);
		innerBox = GtkVBox.new(false, 2);
		innerBox.setBorderWidth(2);
		frame.add(innerBox, false, false);

		buttonBox = [
			"Play" -> { system.play },
			"Stop" -> { system.patch.synth.free },
			"Free and Destroy" -> { system.patch.synth.free; this.destroy },
			"Save" -> {
				GtkFileSelection.askAndDo("Save state", {
					arg filename;
					system.state.writeArchive(filename)
				});
			},
			"Load" -> {
				GtkFileSelection.askAndDo("Load state", {
					arg filename;
					system.state = Object.readArchive(filename)
				});
			},
			"Push" -> {
				~valuestack.push(system.state)
			},
			"Pop" -> {
				if (~valuestack.notEmpty, {
					system.state = ~valuestack.pop
				});
			},
			"Peek" -> {
				if (~valuestack.notEmpty, {
					system.state = ~valuestack.peek
				});
			},
		].asButtons.asGtkHBox(true, 2);

		includedPatches = this.createPatchControlsRecursively(system.patch, innerBox, buttonBox, depth);

		system.patches.keysValuesDo({
			arg name, patch;
			if (includedPatches.includes(patch).not, {
				includedPatches.addAll(this.createPatchControlsRecursively(patch, innerBox, buttonBox, inf));
			});
		});

		this.packEnd(buttonBox);

		this.showAllChildren
	}

	createPatchControlsRecursively {
		arg patch, sliderBox, buttonBox, depth;
		var gui, frame, box, includedPatches;

		depth = if (depth === inf, { inf }, { depth - 1 });

		includedPatches = List[patch];

		patch.argsForSynth.do({
			arg a, i;

			if (a.isKindOf(Patch), {
				if (a.argsForSynth.size > 0 and: { (depth === inf) or: { depth > 0 } }, {
					frame = GtkFrame(patch.argNamesForSynth[i].asString);
					box = GtkVBox.new(false, 0).setBorderWidth(2);
					frame.add(box);
					this.packStart(frame, false, false);
					this.reorderChild(frame, 0);
					includedPatches.addAll(this.createPatchControlsRecursively(a, box, buttonBox, depth));
				});
			}, {
				gui = a.gui(patch.argNamesForSynth[i]);
				if (gui.notNil, {
					if (gui.isKindOf(GtkButton), {
						buttonBox.packStart(gui, true, true);
					}, {
						sliderBox.packStart(gui, false, false);
					});
				});
			});
		});

		^includedPatches		
	}

	add {
		arg name, editor;
		var view;

		view = GtkNumberEditorView(name, editor);
		innerBox.packStart(view, false, false);
		view.show
	}
	addAll {
		arg items;
		items.do({ arg item; this.add(item.key, item.value) });
	}
}
