GtkSourceEditor : GtkVBox {
	var <view, <scrolled, <buffer, <bottomBox, <lineLabel, <columnLabel;
	var <path;

	*new {
		arg filename;
		^super.new.initGtkSourceEditor(filename)
	}
	initGtkSourceEditor {
		arg filename;

		buffer = GtkSourceBuffer.new;
		view = GtkSourceView.newWithBuffer(buffer);
		view.setProperty('tabs-width', 4);
		view.setProperty('show-line-numbers', true);

		scrolled = GtkScrolledWindow.new;
		scrolled.setPolicy(\automatic, \automatic);
		scrolled.add(view.show);
		this.packStart(scrolled, true, true);

		bottomBox = GtkHBox.new;
		this.packStart(bottomBox, false, false);

		lineLabel = GtkDisplayLabel(whiteBackground: false);
		columnLabel = GtkDisplayLabel(whiteBackground: false);
		bottomBox.packStart(lineLabel, false, false);
		bottomBox.packStart(columnLabel, false, false);	
		bottomBox.showChildren;

		view.connect('move-cursor', {
			arg ...args;
			this.prUpdatePositionDisplay;
		});

		view.connect('populate-popup', {
			arg view, menu;
			menu.appendItems([
				\separator,
				"Evaluate selection" -> {
					var code;
					code = buffer.getSelectedText;
					code.interpretPrint;
				},
				"Evaluate line" -> {
					this.evaluateLine
				},
				"Evaluate buffer" -> {
					var code;
					code = buffer.getText;
					code.interpretPrint;
				},
			]);
		});

		view.connect('button-press-event', {
			arg widget, event;	var button2, shift;

			button2 = event.button == 2;
			shift = (event.state & Gdk.shift_mask) != 0;

			if (button2 and: shift, {
				this.evaluateLine;
				true;
			}, {
				false;
			});
		});

		this.read(filename);

		this.prUpdatePositionDisplay;
		this.showChildren;
	}

	evaluateLine {
		var code, iter, line, i1, i2;
		iter = buffer.getIterAtMark(buffer.getInsert);
		line = iter.getLine;
				
		i1 = buffer.getIterAtLine(line);
		i2 = buffer.getIterAtMark(buffer.getInsert).forwardToLineEnd;

		code = buffer.getText(i1, i2);
		code.interpretPrint;
	}

	read {
		arg filename;		var file;
		if (filename.notNil, {
			if (File.exists(filename), {
				file = File(filename, "r");
				buffer.setText(file.readAllString);
				file.close;
				path = filename;
			});
		});
	}
	needsScrolledWindow { ^true }

	prUpdatePositionDisplay {
		var iter;
		iter = buffer.getIterAtMark(buffer.getInsert);
		lineLabel.value = iter.getLine;
		columnLabel.value = iter.getLineOffset;
	}

	buffer_ {
		arg aBuffer;
		buffer = aBuffer;
		view.setBuffer(buffer);
	}
}