SampleSet {
	var groups;
	*initClass {
		ObjectSignals.register(this, #['size-changed', 'sample-loaded', 'group-added'])
	}
	*new { ^super.new.initSampleSet }
	*load { arg server, groupSizes, filenamemaker; ^this.new.load(server, groupSizes, filenamemaker)}
	initSampleSet {
		groups = List[];
	}
	at { |i| ^groups[i] }
	choose { ^groups.choose.choose }
	chooseGroup { ^groups.choose }
	addGroups {
		arg i;
		i.do({
			groups.add(List[]);
			this.emit('sample-loaded', groups.size);
		});
		this.emit('size-changed');
	}
	addSample {
		arg i, sample;
		groups[i].add(sample);
		this.emit('sample-loaded', i, groups[i].size);
		this.emit('size-changed');
	}

	load {
		arg server, groupSizes, filenamemaker, assumeNonEmpty = true, preCreateAction, postCreateAction, completionAction;

		preCreateAction = preCreateAction ? {
			arg filename;
			Post << "Creating buffer for " << filename << nl;
		};
		postCreateAction = postCreateAction ? {
			arg size;
			Post << "Buffer size: " << size << nl;
		};
		completionAction = completionAction ? {
			Post << "Samples loaded" << nl;
		};

		ServerSync.do({
			this.addGroups(groupSizes.size);

			groupSizes.do({
				arg nsamples, i;
				var clickgroup, filename, groupname;

				nsamples.do({
					arg j;
					var buffer, result;
	
					filename = filenamemaker.value(i, j);
					result = ServerSync.try(server, {
						preCreateAction.value(filename);
						buffer = Buffer.read(server, filename);
					}, nil, {
						"Could not create buffer".postln;
					});

					if (result === \succeded, {
						ServerSync.try(server, {
							//"Querying info about buffer ".post; buffer.bufnum.postln;
							server.sendMsg("/b_query", buffer.bufnum);
						}, {
							arg t, r, m;

							if (assumeNonEmpty and: (m[2] == 0), {
								\retry
							}, {
								buffer.numFrames = m[2];
								buffer.numChannels = m[3];

								postCreateAction.value(buffer.numFrames);

								this.addSample(i, buffer);

								\done
							});

						}, {
							"Could not get buffer info".postln;
						}, "/b_info");
					});
			
				});
			});
			completionAction.value;
		});
	}

	updateInfo {
		arg assumeNonEmpty = true, server;

		groups.do({
			arg group;
			group.do({
				arg buffer;

				ServerSync.try(server, {
					//"Querying info about buffer ".post; buffer.bufnum.postln;
					server.sendMsg("/b_query", buffer.bufnum);
				}, {
					arg t, r, m;

					if (assumeNonEmpty and: (m[2] == 0), {
						\retry
					}, {
						buffer.numFrames = m[2];
						buffer.numChannels = m[3];
						\done
					});
				}, {
					"Could not get buffer info".postln;
				}, "/b_info");
			});
		});
	}

	nGroups { ^groups.size }
}