Folder : MultiLevelIdentityDictionary {
	var <context, refs;
	var <>currentSubfolderPath;

	nodeType { ^Subfolder }

	*initClass {
		ObjectSignals.register(this, #['item-added', 'item-removed', 'item-renamed', 'item-moved', 'cleared'])
	}
	*new {
		arg aContext;
		^super.new.initFolder(aContext)
	}
	initFolder {
		arg aContext;

		context = aContext;
		currentSubfolderPath = [];
	}

	addItemsWithOrderModel {
		arg items, model;
		var subfolderpath, key, modelKeys;
		var subfolder;

		items.do({
			arg path, item;

			#subfolderpath, key = path.splitBeforeLast;

			subfolder = this.createSubfolderAtPath(subfolderpath);
			modelKeys = model.keysInSubfolder(subfolderpath);

			if (modelKeys.includes(key), {
				subfolder.insertWithOrderModel(key, item, modelKeys);
			});
		});
	}

	clear {
		dictionary = this.nodeType.new;
		this.checkFolderConsistency;
		this.emit('cleared');
	}

	merge {
		arg that, clear = true;

		if (clear, { this.clear } );

		if ((context.notNil || that.context.notNil) and: { context !== that.context }, {
			error("merging folders in different contexts");
			this.verboseHalt
		});

		that.leafDo({
			arg path, object;
			this.putAtPath(path, object);
		});
	}
	replace {
		arg that;
		this.merge(that, true);
	}

	newClone {
		^this.class.new(context).merge(this)
	}

	pathIsUsed {
		arg path;
		^this.atPath(path).notNil
	}

	pathToIndices {
		arg path;
		var indices, index, item;

		path = path.asArray;

		indices = Array.new(path.size);
		item = dictionary;
		path.do({
			arg name, i;
			if (item.isKindOf(this.nodeType).not && (i != (path.size - 1)), {
				^nil
			});

 			index = item.indexOf(name);
			if (index.isNil, {
				^nil
			});
	
			indices = indices.add(index);
			item = item.atIndex(index);
		});

		^indices
	}
	pathAsIndicesString {
		arg path, fallback = nil;
		var indices;

		indices = this.pathToIndices(path); 
		if (indices.notNil, {
			^indices.joinAsString(":", "");
		}, {
			^fallback
		});
	}
	indicesToPath {
		arg indices;
		var item, path, key;

		item = dictionary;
		path = [];
		indices.do({
			arg index;
			key = item.keyByIndex(index);
			if (key.isNil, {
				^nil
			}, {
				path = path.add(key);
				item = item.at(key);
			});
		});
		^path
	}
	indicesStringToPath {
		arg indices;
		^this.indicesToPath(indices.asString.split($:).collect({ arg i; i.asInteger }))
	}

	prIsMovingInTheSameFolder {
		arg oldfolderindices, newfolderindices, position, targetitem;

		if (oldfolderindices != newfolderindices, { ^false });
		if (['into-or-before', 'into-or-after'].includes(position) && targetitem.isKindOf(this.nodeType), {
			^false
		});
		
		^true
	}

	moveByIndices {
		arg oldindices, newindices, position = 'before';
		var oldfolderindices, newfolderindices;
		var oldsubfolder, newsubfolder;
		var key, item, targetitem, oldIndex, newIndex;
		var newkey, inTheSameFolder, oldpath, newpath;

		if (oldindices == newindices, { ^nil });

		oldpath = this.indicesToPath(oldindices);

		oldfolderindices = oldindices.copyFromStart(-1);
		newfolderindices = newindices.copyFromStart(-1);

		oldIndex = oldindices.last;
		newIndex = newindices.last;

		targetitem = this.atIndices(newindices);

		if (targetitem.isKindOf(FolderAggregateItem), { ^this });

		if (this.prIsMovingInTheSameFolder(oldfolderindices, newfolderindices, position, targetitem), {
			oldsubfolder = this.atIndices(oldfolderindices);
			item = oldsubfolder[oldpath.last];
			case([
				{ ['before', 'into-or-before'].includes(position) } -> {
					newkey = oldsubfolder.moveByIndex(oldIndex, newIndex);
					newpath = oldpath.copyButLast ++ [newkey];
					this.emit('item-moved', oldindices, position, newindices, newkey, oldpath, newpath, item);
					^this
				},
				{ ['after', 'into-or-after'].includes(position) } -> {
					newkey = oldsubfolder.moveByIndex(oldIndex, newIndex + 1);
					newpath = oldpath.copyButLast ++ [newkey];
					this.emit('item-moved', oldindices, position, newindices, newkey, oldpath, newpath, item);
					^this
				}
			], {
				error("unknown position symbol");
				^this.verboseHalt;
			});		
		}, {
			oldsubfolder = this.atIndices(oldfolderindices);
			newsubfolder = this.atIndices(newfolderindices);

			if (['into-or-before', 'into-or-after'].includes(position) && (oldfolderindices == newindices), { 
				^nil
			});

			#key, item = oldsubfolder.removeAtIndexWithKey(oldIndex);

			case([
				{position == 'before'} -> {
					newkey = newsubfolder.safeInsertAt(key, newIndex, item);
					newpath = this.indicesToPath(newfolderindices) ++ [newkey];
					this.emit('item-moved', oldindices, position, newindices, newkey, oldpath, newpath, item);
					^this
				}, 
				{position == 'after'} -> {
					newkey = newsubfolder.safeInsertAt(key, newIndex + 1, item);
					newpath = this.indicesToPath(newfolderindices) ++ [newkey];
					this.emit('item-moved', oldindices, position, newindices, newkey, newpath, oldpath, item);
					^this
				},
				// Here we know that the target is a folder
				{position == 'into-or-before'} -> {
					newkey = targetitem.safePut(key, item);
					newpath = this.indicesToPath(newfolderindices) ++ [newkey];
					this.emit('item-moved', oldindices, position, newindices, newkey, newpath, oldpath, item);
					^this
				}, 
				{position == 'into-or-after'} -> {
					newkey = targetitem.safePut(key, item);
					newpath = this.indicesToPath(newfolderindices) ++ [newkey];
					this.emit('item-moved', oldindices, position, newindices, newkey, newpath, oldpath, item);
					^this
				}

			], {
				error("unknown position symbol");
				this.verboseHalt;
			});
		});
	}

	pathIsItem {
		arg path;
		var item;
		item = this.atPath(path);
		^(item.notNil and: {item.isKindOf(this.nodeType).not })
	}
	pathIsSubfolder {
		arg path;
		^this.atPath(path).isKindOf(this.nodeType)
	}
	pathIsAggregate {
		arg path;
		^this.atPath(path).isKindOf(FolderAggregateItem)
	}
	pathCanHaveChildren {
		arg path;		var item;
		item = this.atPath(path);
		^(item.isKindOf(FolderAggregateItem).not && item.isKindOf(this.nodeType).not)
	}

	subfolderOrParent {
		arg path;
		if (path.notNil, {
			if (this.atPath(path).isKindOf(this.nodeType).not, {
				^path.copyButLast;
			});
		}, {
			^[];
		});
	}
	indicesAreSubfolder {
		arg indices;
		^this.atIndices(indices).isKindOf(this.nodeType)
	}

	keysInSubfolder {
		arg path;
		var subfolder;
		subfolder = this.atPath(path);
		if (subfolder.isNil or: {subfolder.isKindOf(this.nodeType).not}, {
			^nil
		});
		^subfolder.keys
	}

	atIndices {
		arg indices;
		var item;

		item = dictionary;
		indices.do({
			arg index; 
			item = item.atIndex(index);
			if (item.isNil, {
				^nil
			});
		});
		^item
	}
	atIndicesWithPath {
		arg indices;
		var path, item, key, result;

		item = dictionary;
		path = [];
		indices.do({
			arg index; 
			result = item.atIndexWithKey(index);
			if (result.isNil, {
				^nil
			}, {
				#key, item = result;
				path = path.add(key);
			});
		});
		^[path, item]
	}

	createNewItemOfTypeByName {
		arg path, typename, name, initializer;
		var class;

		class = this.itemTypes.at(typename.asString);
		if (class.isNil, {
			^nil
		}, {
			^this.createNewItemOfType(path, class, name ? typename, initializer);
		});
		^nil		
	}
	itemTypeNameByType {
		arg class;
		this.itemTypes.keysValuesDo({
			arg typeName, itsClass;
			if (itsClass === class, {
				^typeName
			});
		});
	}
	createNewItemOfType {
		arg path, type, name, initializer;
		var item;

		path = path.asArray;

		if (type.needsContext, {
			item = type.new(context);
		}, {
			item = type.new;
		});

		initializer.value(item)

		^this.addItem(item, name, path);
	}

	addItem {
		arg item, name, path;
		var typeName, fullpath;

		path = path ? currentSubfolderPath;
		if (this.pathIsAggregate(path), { ^nil });

		name = name ? (this.itemTypes.findKeyForValue(item.class) ? item.class.name);

		name = this.makeUniqueName(name.asSymbol, path);
		fullpath = path ++ [name];
		this.prPutAtPath(fullpath, item);
		^fullpath
	}

	realAtPath {
		arg path;
		var item;
		item = super.atPath(path);
		if (item.isKindOf(FolderAggregateItem), {
			^item.object
		}, {
			^item
		});
	}

	add {
		arg item;
		^this.addItem(item, item.class.asString)
	}

	putAtPath {
		arg fullpath, item;
		if (this.pathIsAggregate(fullpath.copyButLast), { ^nil });
		if (this.complexItemTypes.includes(item.class), {
			item = item.asFolderAggregate
		});
		this.prPutAtPath(fullpath, item);
	}
	putAllAtPath {
		arg folderpath, basename, items;
		items.do({
			arg item, n;
			this.putAtPath(folderpath + [basename.asString + "n." ++ (n + 1).asString], item);
		});
	}

	prPutAtPath {
		arg fullpath, item;
		var olditem, indices;

		olditem = this.atPath(fullpath);

		super.putAtPath(fullpath, item);

		if (olditem.notNil, {
			this.emit('item-removed', fullpath, olditem, this.pathToIndices(fullpath));
		});
		this.emit('item-added', fullpath, item);
	}
	removeAtPath {
		arg fullpath;
		var item, indices;

		indices = this.pathToIndices(fullpath);
		item = super.removeAtPath(fullpath);
		if (item.notNil, {
			this.emit('item-removed', fullpath, item, indices)
		});
		^item
	}
	removeAtPaths {
		arg list;
		list.do({ arg path; this.removeAtPath(path) })
	}
	removeAllLeavesSuchThat {
		arg predicate;
		var toBeRemoved;
		toBeRemoved = [];
		this.leafDo({
			arg path, object;
			if (predicate.value(path, object), {
				toBeRemoved = toBeRemoved.add(path);
			});
		});
		toBeRemoved.do({
			arg path;
			this.removeAtPath(path);
		});
	}

	moveAtPath {
		arg oldpath, newpath;
		var item;
		
		item = super.moveAtPath(oldpath, newpath);
		// WE SHOULD EMIT A SIGNAL HERE! MORON!
	}
	renameAtPath {
		arg fullpath, newname;
		var subfolderpath, name, subfolder, item;
		var newfullpath;

		#subfolderpath, name = fullpath.splitBeforeLast;

		if (name == newname, {
			^nil
		});

		subfolder = this.atPath(subfolderpath);
		if (subfolder.isKindOf(this.nodeType), {
			item = subfolder.rename(name, newname);
			if (item.notNil, {
				newfullpath = subfolderpath ++ [newname];
				this.emit('item-renamed', this.pathToIndices(newfullpath), newfullpath, item, fullpath);
				^item
			}, {
				^nil
			});
		});
	}

	createSubfolderAtPath {
		arg path;
		var path2, item;

		path2 = [];
		item = dictionary;
		path.do({
			arg name;
			path2 = path2.add(name); 
			item = item.atFail(name, {
				var newitem; 
				newitem = this.newInternalNode;
				item.put(name, newitem);
				this.emit('item-added', path2, item);
				newitem
			});
		});
		^item
	} 

	createNewSubfolder {
		arg path, name;
		var fullpath;

		path = path.asArray;

		if (this.pathIsAggregate(path), { ^nil });

		name = name ? this.makeUniqueName(this.subfolderLogicName.asSymbol, path);
		fullpath = path ++ [name];
		this.putAtPath(fullpath, this.nodeType.new);
		^fullpath
	}
	createNewSubfolderIfMissing {
		arg path, name;
		var fullpath;
		fullpath = path ++ [name];
		if (this.pathIsUsed(fullpath).not, {
			^this.createNewSubfolder(path, name);
		});
		^fullpath
	}

	createItemCopy {
		arg path;
		var folderpath, newpath, name, newname, item, newitem;

		path = path.asArray;

		if (this.pathIsAggregate(path.copyButLast), { ^nil });

		item = this.atPath(path);
		if (item.isNil, { ^nil });

		folderpath = path.copy;
		name = folderpath.pop;

		newname = this.makeUniqueName(name.asString + "copy", folderpath);
		newpath = folderpath ++ [newname];
		newitem = item.deepCopy;

		this.prPutAtPath(newpath, newitem);
		^newpath
	}

	mirrorFromSelection {
		arg paths;
		var folder;

		folder = this.class.new(context);

		paths.do({
			arg path;
			var item;

			item = this.atPath(path);
			if (item.notNil, {
				if (item.isKindOf(this.nodeType), {
					this.leafDoFrom(path, {
						arg path, item;
						folder.putAtPath(path, item);
					});
				}, {
					folder.putAtPath(path, item);
				});
			});
		});

		^folder
	}

	makeUniqueName {
		arg basename, path;
		var subfolder;

		path = path ? [];
		path = path.asArray;

		subfolder = this.atPath(path);
		if (subfolder.isNil, {
			^basename.asSymbol
		}, {
			^subfolder.makeUniqueName(basename)
		});
	}

	createGtkView {
		^GtkFolderView.new(this)
	}

	itemTypes {
		^Dictionary[]
	}
	complexItemTypes {
		^Dictionary[]
	}

	subfolderLogicName {
		^"Subfolder"
	}

	checkFolderConsistency {
	}

	// XML support
	fillXmlNode {
		arg node;

		this.orderedTreeDo({
			arg path, subfolder, parent, index;
			subfolder.createXmlNode(parent, path.last, "index" -> index.asString);
		}, {
			arg path, object, parent, index;
			object.createXmlNode(parent, path.last, "index" -> index.asString);
		}, node);
	}
	parseXmlNode {
		arg node, context;
		super.parseXmlNode(node, context);
		this.checkFolderConsistency;
	}
	handleXmlChild {
		arg child;

		if (child.name == dictionary.class.xmlNodeName, {
			this.clear;
			this.prParseXmlSubfolderNode([], child)
		}, {
			super.handleXmlChild
		});
	}
	prParseXmlSubfolderNode {
		arg path, node;
		var child, name, index, object, pair, items;
		items = [];
		child = node.children;
		while({child.notNil}, {
			if (child.name == this.nodeType.xmlNodeName , {
				name = child.getProperty(this.nodeType.namePropertyName);

				index = child.getProperty("index");
				if (index.notNil, {
					index = index.asInteger
				}, {
					if (this.ordered, {
						Post << "NO INDICES IN XML REPRESENTATION\n";
						^this
					});
				});
				
				if (name.notNil, {
					items = items.add([index, 'subfolder', path ++ [name.asSymbol], child]);
				});
			}, {
				index = child.getProperty("index");
				if (index.notNil, {
					index = index.asInteger
				}, {
					if (this.ordered, {
						Post << "NO INDICES IN XML REPRESENTATION\n";
						^this
					});
				});

				pair = child.buildItem(context: context);
				if (pair.notNil, {
					if (pair.key.notNil, {
						items = items.add([index, 'item', pair.key.asSymbol, pair.value]);
					}, {
						Post << "IGNORING NODE: NO ITEM NAME" << "\n";
					});
				}, {
					Post << "IGNORING NODE: UNKNOWN CLASS" << "\n";
				});
			});
			child = child.next;
		});

		items.sort({ arg item1, item2; item1.at(0) < item2.at(0) });
		items.do({
			arg item;
			if (item.at(1) === 'subfolder', {
				this.prParseXmlSubfolderNode(item.at(2), item.at(3))
			}, {
				this.putAtPath(path ++ item.at(2), item.at(3));
			});
		});
	}

	// Experimental

	leafPaths {
		var array;

		array = [];
		this.leafDo({
			arg path, object;
			array = array.add(path);
		});

		^array
	}

	defaultPath {
		^this.leafPaths.first;
	}

	ordered {
		^(this.nodeType.tryPerform(\ordered) ? false )
	}

	*isGoodItemName {
		arg name;

		name = name.asString;

		^(name.contains("/").not and: {name.contains("*").not and: {name != "..."}})
	}

	orderedTreeDo {
		arg branchFunc, leafFunc, argument0, postBranchFunc;
		var result;

		if (this.ordered, {
			result = this.doOrderedTreeDo([], this.dictionary, branchFunc, leafFunc, argument0, postBranchFunc, 0);
			^result;
		}, {
			^this.treeDo(branchFunc, leafFunc, argument0, postBranchFunc);
		});
	}
	orderedTreeDoFrom {
		arg folderpath, branchFunc, leafFunc, argument0, postBranchFunc;
		var folder, result;

		if (this.ordered, {
			folderpath = folderpath.asArray;

			folder = this.atPath(folderpath);
			if (folder.isKindOf(this.nodeType), {
				// XXX: This is wrong: we should get the index of this subfolder in its parent
				result = this.doOrderedTreeDo(folderpath, folder, branchFunc, leafFunc, argument0, postBranchFunc, 0);
			}, {
				result = nil;
			});

			^result;
		}, {
			^this.treeDoFrom(folderpath, branchFunc, leafFunc, argument0, postBranchFunc);
		});
	}
	doOrderedTreeDo {
		arg path, object, branchFunc, leafFunc, argument, postBranchFunc, index;
		var result;
		if (object.isKindOf(this.nodeType), {
			if (branchFunc.notNil, {
				result = branchFunc.value(path, object, argument, index);
			}, {
				result = argument;
			});
			object.keysValuesDo({
				arg name, subobject, subindex;
				this.doOrderedTreeDo(path ++ [name], subobject, branchFunc, leafFunc, result, postBranchFunc, subindex)
			});
			if (postBranchFunc.notNil, {
				result = postBranchFunc.value(path, object, result, index);
			});
			^result
		}, {
			leafFunc.value(path, object, argument, index);
		})
	}
}
