OrderedDictionary : Dictionary {
	var keys;

	*new {
		arg n = 3;
		^super.new(n).initSortedDictionary
	}
	initSortedDictionary {
		keys = Array.new;
	}

	atIndex {
		arg index;
		if (index >= keys.size, {
			^nil
		}, {
			^this.at(keys.at(index))
		});
	}
	atIndexWithKey {
		arg index;
		var key;

		if (index >= keys.size, {
			^nil
		}, {
			key = keys.at(index);
			^[key, this.at(key)]
		});
	}

	put {
		arg key, value;
		var index;

		super.put(key, value);
		index = keys.indexOfEqual(key);
		if (index.isNil, {
			keys = keys.add(key);
		});
	}

	insert {
		arg key, otherkey, value;
		var index, otherindex;

		otherindex = keys.indexOfEqual(otherkey);

		if (otherindex.isNil, {
			^put(key, otherkey)
		}, {
			super.put(key, value);
			index = keys.indexOfEqual(key);
			if (index.notNil, {
				keys.remove(key);
			});
			keys = keys.insert(otherindex, key);
		});
	}
	insertAt {
		arg key, index, value;
		var oldindex;

		super.put(key, value);
		oldindex = keys.indexOfEqual(key);
		if (oldindex.notNil, {
			keys.remove(key);
		});
		if (index.isNil, {
			keys = keys.add(key);
		}, {
			keys = keys.insert(index, key);
		});
	}
	safeInsertAt {
		arg key, index, value;
		key = this.makeUniqueName(key);
		this.insertAt(key, index, value);
		^key
	}
	safePut {
		arg key, value;
		key = this.makeUniqueName(key);
		this.put(key, value);
		^key
	}

	move {
		arg key, otherkey;
		var index, otherindex;
	
		index = keys.indexOfEqual(key); 
		if (index.isNil, { ^this });

		keys.removeAt(index);
		otherindex = keys.indexOfEqual(otherkey);
		if (otherindex.notNil, {
			keys = keys.insert(otherindex, key);
		}, {
			keys = keys.add(key);
		});
	}

	rename {
		arg key, newkey;
		var item, index, otherindex;
	
		index = keys.indexOfEqual(key); 
		if (index.isNil, { ^nil });
		if (this.at(newkey).notNil, { ^nil });

		item = super.removeAt(key);
		super.put(newkey, item);
		keys.put(index, newkey);
		^item
	}

	moveByIndex {
		arg index, otherindex;
		var key;

		key = keys.at(index);
		if (otherindex.notNil, {
			keys = keys.insert(otherindex, key);
		}, {
			keys = keys.add(key);
		});
		if (otherindex < index, {
			keys.removeAt(index + 1);
		}, {
			keys.removeAt(index);
		});
		^key
	}

	removeAtIndex {
		arg index;
		if (index >= keys.size, {
			^nil
		}, {
			^this.removeAt(keys.at(index))
		});
	}
	removeAtIndexWithKey {
		arg index;
		var key;
		if (index >= keys.size, {
			^nil
		}, {
			key = keys.at(index);
			^[key, this.removeAt(key)]
		});
	}

	removeAt {
		arg key;
		var val, index;

		index = keys.indexOfEqual(key);
		if (index.notNil, {
			val = super.removeAt(key);
			keys.removeAt(index);
			^val
		}, {
			^nil
		});
	}
	removeAtFail {
		arg key, function;
		var val, index;
		val = super.removeAtFail(key, function);
		index = keys.indexOfEqual(key);
		if (key.notNil, {
			keys.removeAt(index);
		});
		^val
	}

	keysValuesDo {
		arg function;

		keys.do({
			arg key, index;
			function.value(key, this.at(key), index);
		});
	}
	internalKeysValuesDo {
		arg function;
		^super.keysValuesDo(function);
	}

	keys {
		^keys.copy
	}

	sort {
		arg function;
		keys.sort(function);
	}

	fixKeyList {
		var index;
		this.internalKeysValuesDo({
			arg k, v, i;
			index = keys.indexOfEqual(k);
			
			if (index.isNil, {
				keys = keys.add(k);
			});
		});
		keys = keys.select({ arg key; key.notNil });
	}

	indexOf {
		arg key;
		^keys.indexOfEqual(key)
	}
	keyByIndex {
		arg index;
		if (index < keys.size, {
			^keys.at(index);
		}, {
			^nil
		});
	}

	makeUniqueName {
		arg basename;
		var n, name;

		basename = basename.asSymbol;

		if (keys.includes(basename).not, { ^basename });

		n = 2;
		name = (basename.asString ++ " n." ++ n.asString).asSymbol; 
		while({ this.at(name).notNil }, {
			n = n + 1;
			name = (basename.asString ++ " n." ++ n.asString).asSymbol;
		});

		^name
	}

	insertWithOrderModel {
		arg key, item, modelKeys;
		var index;

		modelKeys.removeAllSuchThat({ arg aKey; aKey !== key and: { keys.includes(aKey).not } });
		index = modelKeys.detectIndex({ arg aKey; aKey === key })
		^this.insertAt(key, index, item);
	}
}
