PathLookupTable : IdentityDictionary {
	var folder;
	*new {
		arg theFolder;
		^super.new.init(theFolder)
	}
	init {
		arg theFolder;
		
		folder = theFolder;

		this.watch('item-added', folder, {
			arg path, item;
			this.addItem(path, item);
		});
		this.watch('item-removed', folder, {
			arg path, item;
			this.removeItem(item);
		});
		this.watch('item-renamed', folder, {
			arg dummy, newpath, item, oldpath;
			this.moveItem(oldpath, newpath, item);
		});
		this.watch('item-moved', folder, {
			arg dummy1, dummy2, dummy3, oldpath, newpath, item;
			this.moveItem(oldpath, newpath, item);
		});
		this.watch('cleared', folder, {
			arg dummy1, dummy2, dummy3, oldpath, newpath, item;
			this.clear;
		});
	}

	// ----------

	addItem {
		arg path, item;
		if (item.isKindOf(folder.nodeType), {
			this.addSubfolder(path, item);
		}, {
			if (item.isKindOf(FolderAggregateItem), {
				this.addAggregate(path, item);
			}, {
				this.addSimpleItem(path, item);
			});
		});
	}
	addAggregate {
		arg path, item;
		var mnames, mobjects;

		mnames = item.membersNames;
		mobjects = item.members;
		mobjects.do({
			arg item, index;
			this.put(item, path ++ [mnames[index]]);
		});
	}
	addSimpleItem {
		arg path, item;
		this.put(item, path);
	}
	addSubfolder {
		arg path, item;
		item.do({
			arg key, subitem;
			this.put(subitem, path ++ [key]);
		});		
	}

	// ----------

	moveItem {
		arg oldpath, newpath, item;
		if (item.isKindOf(folder.nodeType), {
			this.moveSubfolder(oldpath, newpath, item);
		}, {
			if (item.isKindOf(FolderAggregateItem), {
				this.moveAggregate(oldpath, newpath, item);
			}, {
				this.moveSimpleItem(oldpath, newpath, item);
			});
		});
	}
	moveSimpleItem {
		arg oldpath, newpath, item;
		this.removeAt(item);
		this.put(item, newpath);
	}
	moveAggregate {
		arg oldpath, newpath, item;
		var names;

		names = item.membersNames;

		item.members.do({
			arg subitem, index;
			this.removeAt(subitem);
			this.put(subitem, newpath ++ [names[index]]);
		});		
	}
	moveSubfolder {
		arg oldpath, newpath, item;
		item.do({
			arg key, subitem;
			this.removeItem(subitem);
			this.put(subitem, newpath ++ [key]);
		});		
	}

	// ----------

	removeItem {
		arg item;
		if (item.isKindOf(folder.nodeType), {
			this.removeSubfolder(item);
		}, {
			if (item.isKindOf(FolderAggregateItem), {
				this.removeAggregate(item);
			}, {
				this.removeSimpleItem(item);
			});
		});
	}
	removeSimpleItem {
		arg item;
		this.removeAt(item);
	}
	removeAggregate {
		arg item;
		item.members.do({
			arg subitem;
			this.removeAt(subitem);
		});		
	}
	removeSubfolder {
		arg item;
		item.do({
			arg key, subitem;
			this.removeItem(subitem)
		});		
	}
}
