XmlNode {
	var <ptr, <docPtr;
	classvar <classNodeNames;

	*registerClass {
		arg classobject, name;
		classNodeNames = classNodeNames ? Dictionary.new;
		name = name ? classobject.xmlNodeName;
		classNodeNames.put(name, classobject);
	}

	buildItem {
		arg classobject, context;
		var name, itemname, object;

		name = this.name;

		classobject = classobject ? XmlNode.classNodeNames.at(name);
		if (classobject.isNil, {
			Post << "XmlNode-buildItem: no class for node of type " << name << "\n"; 
			^nil
		});

		itemname = this.getProperty(classobject.namePropertyName);		

		object = this.buildObject(classobject, context);
		
		if (object.isNil, {
			^nil
		}, {
			^(itemname -> object)
		});
	}

	buildObject {
		arg classobject, context;
		var name, object;

		name = this.name;

		if (classobject.isNil, {
			classobject = XmlNode.classNodeNames.at(name);
		});

		if (classobject.isNil, {
			object = nil;
		}, {
			if (classobject.needsContext, {
				object = classobject.new(context);
				object.parseXmlNode(this, context);
			}, {
				object = classobject.new;
				object.parseXmlNode(this);
			});
		});

		 ^object
	}

	*new {
		arg inDoc;
		^super.new.init(inDoc);
	}
	init {
		arg inDocPtr;
		ptr = 0;
		docPtr = inDocPtr;
	}

	*wrap {
		arg docPtr, inPtr;
		if (docPtr.notNil && inPtr.notNil, {
			^super.new.wrap(docPtr, inPtr)
		}, {
			^nil
		});
	}
	
	wrap {
		arg inDocPtr, inPtr;
		ptr = inPtr;
		docPtr = inDocPtr;
	}

	valid {
		^((ptr != 0) && (docPtr != 0))
	}

	name {
		_prXmlNodeGetName;
		^this.primitiveFailed
	}
	children {
		^XmlNode.wrap(docPtr, this.prChildren)
	}
	prChildren {
		_prXmlNodeGetChildren;
		^this.primitiveFailed
	}

	next {
		^XmlNode.wrap(docPtr, this.prNext)
	}
	prNext {
		_prXmlNodeGetNext;
		^this.primitiveFailed
	}

	addChild {
		arg name, content;
		^XmlNode.wrap(docPtr, this.prNewChild(name, content))
	}
	prNewChild {
		arg name, content;
		_prXmlNodeNewChild;
		^this.primitiveFailed
	}

	addTextChild {
		arg tag, body;
		^XmlNode.wrap(docPtr, this.prNewTextChild(tag, body))
	}
	prNewTextChild {
		arg tag, body;
		_prXmlNodeNewTextChild;
		^this.primitiveFailed
	}

 
	setProperty {
		arg name, value;
		_prXmlNodeSetProperty;
		^this.primitiveFailed
	}
	getProperty {
		arg name;
		_prXmlNodeGetProperty;
		^this.primitiveFailed
	}

	getContent {
		arg expand = true;
		_prXmlNodeGetContent;
		^this.primitiveFailed
	}
}
