#include <iostream>

#include <GC.h>
#include <VMGlobals.h>
#include <PyrObject.h>
#include <PyrKernel.h>
#include <PyrPrimitive.h>
#include <PyrMessage.h>

#undef DEBUG_VERBOSE_SUPERYIELD

static PyrSymbol *sym_async;
static PyrSymbol *sym_asyncthread;

static PyrSymbol *sym_currentThread;
static PyrSymbol *sym_wait;
static PyrSymbol *sym_exit;
static PyrSymbol *sym_realThread;
static PyrSymbol *sym_realParentThread;

static PyrClass *class_async;
static PyrClass *class_asyncthread;

extern PyrSymbol *s_prstart;
extern bool classFindClassVar(PyrClass** classobj, PyrSymbol *name, int *index);
extern bool classFindInstVar(PyrClass* classobj, PyrSymbol *name, int *index);
extern void switchToThread(struct VMGlobals *g, struct PyrThread *newthread, int oldstate, int *numArgsPushed);

static PyrSlot *classVarSlot(struct VMGlobals *g, PyrClass* classobj, PyrSymbol *name) {
  PyrClass *classobj0 = classobj;
  int vindex;

  if (classFindClassVar(&classobj0, name, &vindex))
    return g->classvars[classobj0->classIndex.ui].uo->slots + vindex;
  else
    return 0;
}

static PyrSlot *objectVarSlot(PyrObject *object, PyrSymbol *name) {
  int vindex;

  if (classFindInstVar(object->classptr, name, &vindex))
    return object->slots + vindex;
  else
    return 0;
}

static PyrObject *get_current_async_thread(struct VMGlobals *g) {
  PyrSlot *currentThreadSlot;

  class_async = sym_async->u.classobj;
  if (class_async == 0) {
    error("class Async not found");
    return 0;
  }

  class_asyncthread = sym_asyncthread->u.classobj;
  if (class_asyncthread == 0) {
    error("class AsyncThread not found");
    return 0;
  }

  currentThreadSlot = classVarSlot(g, class_async, sym_currentThread);
  if (currentThreadSlot == 0) {
    error("could not find currentThread variable of class Async.\n");
    return 0;
  }

  if (!isKindOfSlot(currentThreadSlot, class_asyncthread)) {
    error("misplaced super-yield. currentThread is currently set to:\n");
    dumpObjectSlot(currentThreadSlot);
    return 0;
  }

  return currentThreadSlot->uo;
}

static int prAsyncThread_SuperYield(struct VMGlobals *g, int numArgsPushed, PyrSymbol *result) {
  PyrSlot value;
  PyrSlot *realThreadSlot, *realParentThreadSlot;
  PyrObject *currentAsyncThread;
 
  currentAsyncThread = get_current_async_thread(g);
  if (currentAsyncThread == 0) return errFailed;

  realThreadSlot = objectVarSlot(currentAsyncThread, sym_realThread);
  if (realThreadSlot == 0) {
    error("could not find realThread variable of current async thread.\n");
    dumpObject(currentAsyncThread);
    return errFailed;
  }
  realParentThreadSlot = objectVarSlot(currentAsyncThread, sym_realParentThread);
  if (realParentThreadSlot == 0) {
    error("could not find realParentThread variable of current async thread.\n");
    dumpObject(currentAsyncThread);
    return errFailed;
  }

  if (NotNil(realThreadSlot)) {
    error("cannot perform a super-yield twice before a resume. Previous super-yield was from:\n");
    dumpObjectSlot(realThreadSlot);
    return errFailed;
  }
  SetObject(realThreadSlot, (PyrObject *)g->thread);
  SetObject(realParentThreadSlot, g->thread->parent.uo);

  value.ucopy = g->sp->ucopy;

  // We switch directly to the parent of the current async thread
  // but we store the current thread because we will start again from there

  PyrThread *parent = ((PyrThread *)currentAsyncThread)->parent.uot;

#ifdef  DEBUG_VERBOSE_SUPERYIELD
  post("Super-yielding to:\n");
  dumpObject((PyrObject *)parent);
#endif

  switchToThread(g, parent, tYieldToParent, &numArgsPushed);

  SetSymbol(g->sp - numArgsPushed + 1, result);

  return errNone;
}

static int prAsyncThread_YieldForWait(struct VMGlobals *g, int numArgsPushed) {
  return prAsyncThread_SuperYield(g, numArgsPushed, sym_wait);
}

static int prAsyncThread_YieldForExit(struct VMGlobals *g, int numArgsPushed) {
  return prAsyncThread_SuperYield(g, numArgsPushed, sym_exit);
}

static int prAsyncThread_Resume(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *a, *b, threadSlot, value, *realThreadSlot, *realParentThreadSlot;
  PyrObject *threadObject;
  PyrThread *thread, *parent;
  int state;
  bool isLong;
  
  a = g->sp - 1;
  b = g->sp;
  
  threadObject = a->uo;
  
  realThreadSlot = objectVarSlot(threadObject, sym_realThread);
  if (realThreadSlot == 0) {
    error("could not find realThread variable of currentAsyncThread.\n");
    return 0;
  }
  realParentThreadSlot = objectVarSlot(threadObject, sym_realParentThread);
  if (realParentThreadSlot == 0) {
    error("could not find realParentThread variable of currentAsyncThread.\n");
    return 0;
  }

  // If realThread is not nil, we previously made a super-yield, possibly
  // from a thread *inside* the AsyncThread. We resume the inner (real) thread.
  // We can call this a "long resume".
  // In this case the "parent" of the real thread is not the current thread but it's
  // store in the realParentThread variable.
  if (NotNil(realThreadSlot)) {
    thread = realThreadSlot->uot;
	parent = realParentThreadSlot->uot;
    SetNil(realThreadSlot);
	isLong = true; 
  } else {
    thread = a->uot;
	parent = g->thread;
	isLong = false;
  }
  
  state = thread->state.ui;
  
  if (state == tInit) {
    threadSlot.ucopy = a->ucopy;
    value.ucopy = b->ucopy;
    
	SetObject(&thread->parent, parent);
	g->gc->GCWrite(thread, parent);
  
    thread->beats.uf = parent->beats.uf;
    thread->seconds.uf = parent->seconds.uf;
    thread->clock.ucopy = parent->clock.ucopy;
    g->gc->GCWrite(thread, &parent->clock);

    switchToThread(g, thread, tYieldToChild, &numArgsPushed);
    
    g->sp = g->gc->Stack()->slots - 1;
    g->receiver.ucopy = (++g->sp)->uf = threadSlot.ucopy;
    (++g->sp)->ucopy = value.ucopy;
    
    sendMessage(g, s_prstart, 2);
  } else if (state == tYieldToParent) {
    SetObject(&thread->parent, parent);
    g->gc->GCWrite(thread, parent);
    
    thread->beats.uf = parent->beats.uf;
    thread->seconds.uf = parent->seconds.uf;
    thread->clock.ucopy = parent->clock.ucopy;
    g->gc->GCWrite(thread, &parent->clock);
    
    value.ucopy = b->ucopy;

    switchToThread(g, thread, tYieldToChild, &numArgsPushed);

    (g->sp - numArgsPushed + 1)->ucopy = value.ucopy;
  } else if (state == tDone) {
    a->ucopy = thread->terminalValue.ucopy;
  } else if (state == tRunning) {
    error("Tried to resume the running thread.\n");
    return errFailed;
  } else {
    error("Thread in strange state: %d\n", state);
    return errFailed;
  }
  return errNone;
}

void initialize_async_primitives(void) {
  int base = nextPrimitiveIndex();
  int index = 0;

  sym_async = getsym("Async");
  sym_asyncthread = getsym("AsyncThread");

  sym_currentThread = getsym("currentThread");
  sym_wait = getsym("wait");
  sym_exit = getsym("exit");
  sym_realThread = getsym("realThread");
  sym_realParentThread = getsym("realParentThread");

  definePrimitive(base, index++, "_AsyncThread_YieldForWait", prAsyncThread_YieldForWait, 1, 0);
  definePrimitive(base, index++, "_AsyncThread_YieldForExit", prAsyncThread_YieldForExit, 1, 0);
  definePrimitive(base, index++, "_AsyncThread_Resume", prAsyncThread_Resume, 2, 0);
}

// --------

