// MetaControl
// Copyright (C) 2003 Maurizio Umberto Puxeddu

#include <iostream>

#include <glib/gthread.h>
#include <gtk/gtkmain.h>
#include <gtk/gtkwindow.h>
#include <gtk/gtkscrolledwindow.h>
#include <gtk/gtktextview.h>

#include "sclangInterpreter.h"

#include "gtkrequest.h"

extern long compiledOK;

static gboolean scroll_to_end(gpointer widget, gpointer data) {
  GtkTextBuffer *buffer = GTK_TEXT_BUFFER(widget);
  GtkTextView *view = GTK_TEXT_VIEW(data);

  static GtkTextMark *mark = 0;

  if (mark == 0) {
    GtkTextIter iter;
    gtk_text_buffer_get_end_iter(buffer, &iter);
    mark = gtk_text_buffer_create_mark(buffer, NULL, &iter, FALSE);
  }

  gtk_text_view_scroll_to_mark(view, mark, 0.0, TRUE, 1.0, 0.0);
}

static GtkWidget *make_console_window(GtkTextBuffer *buffer) {
  GtkWidget *scrolled, *toplevel, *view;
  
  view = gtk_text_view_new_with_buffer(buffer);
  gtk_text_view_set_editable(GTK_TEXT_VIEW(view), FALSE);

  g_signal_connect(G_OBJECT(buffer), "changed", G_CALLBACK(scroll_to_end), (gpointer)view);

  scrolled = gtk_scrolled_window_new(0, 0);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_container_add(GTK_CONTAINER(scrolled), view);
  
  toplevel = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_default_size(GTK_WINDOW(toplevel), 800, 600);
  gtk_container_add(GTK_CONTAINER(toplevel), scrolled);
  
  g_signal_connect(GTK_OBJECT(toplevel), "destroy", G_CALLBACK(gtk_main_quit), NULL);

  gtk_widget_show(GTK_WIDGET(view));
  gtk_widget_show(GTK_WIDGET(scrolled));

  return toplevel;
}

int main(int argc, char *argv[]) {
  GtkWidget *consoleWindow;

  g_thread_init(0);
  gtk_init(&argc, &argv);
           
  SclangInterpreter sclangInterpreter;

  consoleWindow = make_console_window(sclangInterpreter.consoleBuffer());

  GtkRequestQueue::initialize();
  GtkRequestQueue::connectMainloopCallback();

  SclangInterpreter::initialize();

  gtk_widget_show(GTK_WIDGET(consoleWindow));
  gtk_main();

  GtkRequestQueue::finalize();
  SclangInterpreter::finalize();

  return 0;
}
