#ifndef _safeQueue_H_
#define _safeQueue_H_

#include <unistd.h>

#include <iostream>
#include <list>

#include <sigc++/sigc++.h>

#include <glib/gmain.h>
#include <glib/gthread.h>
#include <glib/giochannel.h>

template <class Type>
class SafeQueue : public SigC::Object {
public:
  SafeQueue(void);
  ~SafeQueue();

  void post(const Type &type);
  SigC::Signal1<void, const Type &> message_arrived;

private:
  GMutex *m_mutex;
  std::list<Type> m_queue;
  int m_fds[2];

  GIOChannel *m_channel;

  bool on_bang(GIOCondition c);
  static gboolean dispatch_io_channel_signal(GIOChannel *source, GIOCondition condition, gpointer data);
};

template <class Type>
gboolean SafeQueue<Type>::dispatch_io_channel_signal(GIOChannel *source, GIOCondition condition, gpointer data) {
  SafeQueue<Type> *self = (SafeQueue<Type> *)data;
  self->on_bang(condition);
  return TRUE;
}


template <class Type>
SafeQueue<Type>::SafeQueue(void) {
  pipe(m_fds);

  m_mutex = g_mutex_new();
  m_channel = g_io_channel_unix_new(m_fds[0]);
  g_io_add_watch(m_channel, G_IO_IN, SafeQueue<Type>::dispatch_io_channel_signal, this);
}

template <class Type>
SafeQueue<Type>::~SafeQueue() {
  close(m_fds[0]);
  close(m_fds[1]);
  g_mutex_free(m_mutex);
  g_io_channel_unref(m_channel);
}

template <class Type>
void SafeQueue<Type>::post(const Type &object) {
  char c = 't';

  g_mutex_lock(m_mutex);
  m_queue.push_back(object);
  g_mutex_unlock(m_mutex);

  write(m_fds[1], &c, sizeof(c));
}

template <class Type>
bool SafeQueue<Type>::on_bang(GIOCondition c) {
  Type body;
  int result;
  char b;

  if ((result = read(m_fds[0], &b, sizeof(b))) == 1) {
    g_mutex_lock(m_mutex);
    body = m_queue.front();
    m_queue.pop_front();
    g_mutex_unlock(m_mutex);
    message_arrived(body);
  } else if (result < 0) {
    std::cout << "SafeQueue: input error" << std::endl;
  }

  return true;
}

//

template <class Type>
class SafeQueue2 : public SigC::Object {
public:
  SafeQueue2(void);
  ~SafeQueue2();

  void post(Type *type);
  SigC::Signal1<void, Type *> message_arrived;

private:
  GMutex *m_mutex;
  std::list<Type *> m_queue;
  int m_fds[2];

  GIOChannel *m_channel;
  bool on_bang(GIOCondition c);
  static gboolean dispatch_io_channel_signal(GIOChannel *source, GIOCondition condition, gpointer data);
};

template <class Type>
gboolean SafeQueue2<Type>::dispatch_io_channel_signal(GIOChannel *source, GIOCondition condition, gpointer data) {
  SafeQueue2<Type> *self = (SafeQueue2<Type> *)data;
  self->on_bang(condition);
  return TRUE;
}

template <class Type>
SafeQueue2<Type>::SafeQueue2(void) {
  m_mutex = g_mutex_new();
  pipe(m_fds);

  m_channel = g_io_channel_unix_new(m_fds[0]);
  g_io_add_watch(m_channel, G_IO_IN, SafeQueue2<Type>::dispatch_io_channel_signal, this);
}

template <class Type>
SafeQueue2<Type>::~SafeQueue2() {
  close(m_fds[0]);
  close(m_fds[1]);
  g_mutex_free(m_mutex);
  g_io_channel_unref(m_channel);
}

template <class Type>
void SafeQueue2<Type>::post(Type *object) {
  char c = 't';

  g_mutex_lock(m_mutex);
  m_queue.push_back(object);
  g_mutex_unlock(m_mutex);

  write(m_fds[1], &c, sizeof(c));
}

template <class Type>
bool SafeQueue2<Type>::on_bang(GIOCondition c) {
  Type *body;
  int result;
  char b;

  if ((result = read(m_fds[0], &b, sizeof(b))) == 1) {
    g_mutex_lock(m_mutex);
    body = m_queue.front();
    m_queue.pop_front();
    g_mutex_unlock(m_mutex);
    message_arrived(body);
  } else if (result < 0) {
    std::cout << "SafeQueue2: input error" << std::endl;
  }

  return true;
}

#endif // _safeQueue_H_
