#ifndef _sclangInterpreter_H_
#define _sclangInterpreter_H_

#include <stdarg.h>

#include <list>
#include <string>

#include <sigc++/sigc++.h>

#include <gtk/gtktextbuffer.h>

#include "callableQueue.h"
#include "safeQueue.h"

class SclangInterpreter : public SigC::Object {
public:
  SclangInterpreter(void);
  ~SclangInterpreter();

  static int eval_string_sync(const std::string &s);
  static int eval_string_async(const std::string &s);

  static void test(void);

  static int initialize(void);
  static void reset(void);
  static void finalize(void);
  
  static int make_tmp_file(void);
  static const char *get_last_tmp_file(void);

  static void post(const std::string &text);
  static void initialize_primitives(void);
  
  static SclangInterpreter *get_instance(void);

  static int vm_tick(void);
  GtkTextBuffer *consoleBuffer(void) { return m_consoleBuffer; }


private:
  static SclangInterpreter *m_self;

  static gboolean call_clock_timeout(gpointer data);
  bool on_clock_timeout(void);
  
  std::list<std::string> m_tempFiles;

  static void vm_start(void);
  static void vm_stop(void);

  static int prMetacNullPrimitive(struct VMGlobals *g, int numArgsPushed);

#if 0
  static int prNewTextWindow(struct VMGlobals *g, int numArgsPushed);
  static int prOpenTextFile(struct VMGlobals *g, int numArgsPushed);
#endif

  static int prSCWindow_New(struct VMGlobals *g, int numArgsPushed);

  CallableQueue m_callableQueue;
  GtkTextBuffer *m_consoleBuffer;
};

#endif // _sclangInterpreter_H_
