#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include <sys/param.h>
#include <unistd.h>

#include "simpleConfigFile.h"

#define STANDARD_CONFIG_FILE_PATH1 "~/.Supercollider3/library.config"
#define STANDARD_CONFIG_FILE_PATH2 "/etc/Supercollider3/library.config"

SimpleConfigFile::SimpleConfigFile(void) {
}

SimpleConfigFile::~SimpleConfigFile() {
}

SimpleConfigFile *SimpleConfigFile::load(const char *filename) {
  SimpleConfigFile *self = new SimpleConfigFile();
  if (self->readFile(filename)) {
    delete self;
    self = 0;
  }
  return self;
}

SimpleConfigFile *SimpleConfigFile::loadFromStandardPaths(void) {
  SimpleConfigFile *self = 0;
  const char *config_file_path = getenv("SC_CONFIG_FILE");

  if (config_file_path != 0)
    self = SimpleConfigFile::load(config_file_path);
  if (self == 0) {
    char filename[MAXPATHLEN];
    getcwd(filename, MAXPATHLEN - 16);
    strcat(filename, "library.config");
    self = SimpleConfigFile::load(filename);
  }
  if (self == 0)
    self = SimpleConfigFile::load(STANDARD_CONFIG_FILE_PATH1);
  if (self == 0)
    self = SimpleConfigFile::load(STANDARD_CONFIG_FILE_PATH2);
  return self;
}

bool SimpleConfigFile::readFile(const char *filename) {
  FILE *file;
  char *realpath;

  realpath = (char *)malloc(MAXPATHLEN + 1);
  realpath = unixStandardizePath(filename, realpath);

  file = fopen(realpath, "r");
  if (file == 0) {
    free(realpath);
    return true;
  }

  parseFile(file);

  fclose(file);

  free(realpath);
  return false;
}

void SimpleConfigFile::parseFile(FILE *file) {
  char buffer[MAXPATHLEN + 1];

  while (fgets(buffer, MAXPATHLEN + 1, file) == buffer) {
    int len = strlen(buffer);
    if (len == 0 || (len == 1 && buffer[0] == '\n')) continue;
    if (buffer[len - 1] == '\n') buffer[len - 1] = '\0';

    char *realpath = (char *)malloc(MAXPATHLEN + 1);

    if (unixStandardizePath(buffer + 1, realpath) != realpath) {
      free(realpath);
    } else {
      if (buffer[0] == '+') {
	addIncludedDirectory(realpath);      
      } else if (buffer[0] == '-') {
	addExcludedDirectory(realpath);
      } else {
	free(realpath);
      }
    }
  }
}
