#ifndef _gtkasyncrequests_H_
#define _gtkasyncrequests_H_

#include <pango/pango-layout.h>
#include <pango/pango-context.h>
#include <gtk/gtk.h>
#include <gtk/gtkeditable.h>
#include <gdk/gdkwindow.h>
#include <gdk/gdkcursor.h>
#include <gdk/gdkdrawable.h>
#include <gtk/gtkadjustment.h>
#include <gtk/gtkscrolledwindow.h>
#include <gtk/gtkstyle.h>
#include <gtk/gtkmisc.h>
#include <gtk/gtkfixed.h>
#include <gtk/gtkhseparator.h>
#include <gtk/gtkvseparator.h>
#include <gtk/gtkwidget.h>
#include <gtk/gtkframe.h>
#include <gtk/gtkpaned.h>
#include <gtk/gtkhpaned.h>
#include <gtk/gtkvpaned.h>
#include <gtk/gtkeventbox.h>
#include <gtk/gtkcontainer.h>
#include <gtk/gtklabel.h>
#include <gtk/gtkbox.h>
#include <gtk/gtknotebook.h>
#include <gtk/gtkhbox.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtkbin.h>
#include <gtk/gtkwindow.h>
#include <gtk/gtkbutton.h>
#include <gtk/gtkradiobutton.h>
#include <gtk/gtkimage.h>
#include <gtk/gtkmenu.h>
#include <gtk/gtkmenushell.h>
#include <gtk/gtkmenuitem.h>
#include <gtk/gtkseparatormenuitem.h>
#include <gtk/gtktextview.h>
#include <gtk/gtktextbuffer.h>
#include <gtk/gtktextmark.h>
#include <gtk/gtkentry.h>
#include <gtk/gtktable.h>
#include <gtk/gtkrange.h>
#include <gtk/gtkscale.h>
#include <gtk/gtkvscale.h>
#include <gtk/gtkhscale.h>
#include <gtk/gtkdrawingarea.h>
#include <gtk/gtktreeview.h>
#include <gtk/gtktreeselection.h>
#include <gtk/gtktreeviewcolumn.h>
#include <gtk/gtktreestore.h>
#include <gtk/gtkliststore.h>
#include <gtk/gtkcellrenderertext.h>
#include <gtk/gtkcellrenderertoggle.h>
#include <gtk/gtktreeselection.h>

#include <glade/glade.h>

#include <libgnomecanvas/libgnomecanvas.h>

#include <gtksourceview/gtksourceview.h>
#include <gtksourceview/gtksourcebuffer.h>

#include <SCBase.h>
#include <PyrSlot.h>
#include <PyrObject.h>
#include <PyrKernel.h>
#include <PyrPrimitive.h>
#include <PyrMessage.h>
#include <GC.h>
#include <VMGlobals.h>

#include "gtkrequest.h"

extern bool GValue_to_slot(const GValue *m_value, PyrSlot *slot);
extern void SetEnumSlot(PyrSlot *slot, GType enum_type, int value);
extern bool createGBoxedWrapper(VMGlobals *g, PyrSlot *slot, const char *classname, void *ptr, GType type, bool borrowed = false);
extern bool createGBoxedWrapper(VMGlobals *g, PyrSlot *slot, PyrClass *classObject, void *ptr, GType type, bool borrowed);

#if 0
#  define GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE { std::cout << "Perform method at " << __LINE__ << std::endl; }
#else
#  define GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE { }
#endif

class GtkRequest_ReturnBoxed : public GtkRequest {
public:
  GtkRequest_ReturnBoxed(const char *wrapperClass) :
    m_wrapperClass(wrapperClass), m_ptr(0), m_type(G_TYPE_INVALID)
  {
  }
  
  void setBoxedResult(GType type, void *boxed) {
    m_type = type;
    m_ptr = boxed;
  }

  void copyBoxedResult(GType type, void *boxed) {
    m_type = type;
    m_ptr = g_boxed_copy(type, boxed);
  }

  void nilBoxedResult(void) {
    m_ptr = 0;
    m_type = G_TYPE_INVALID;
  }
  
  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    if ((m_ptr == 0) || createGBoxedWrapper(g, resultSlot, m_wrapperClass, m_ptr, m_type, false)) {
      SetNil(g->sp);
    }
    return 1;
  };

private:
  const char *m_wrapperClass;
  void *m_ptr;
  GType m_type;
};

// ------ GBoxed

class GBoxedFreeRequest : public GtkRequest {
public:
  GBoxedFreeRequest(GType type, void *ptr) :
    m_type(type), m_ptr(ptr)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    g_boxed_free(m_type, m_ptr);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GType m_type;
  void *m_ptr;
};

class GBoxedCopyRequest : public GtkRequest {
public:
  GBoxedCopyRequest(PyrClass *wrapperClass, GType type, void *original) :
    m_wrapperClass(wrapperClass), m_type(type), m_original(original), m_copy(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_copy = g_boxed_copy(m_type, m_original);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    if ((m_copy == 0) || createGBoxedWrapper(g, resultSlot, m_wrapperClass, m_copy, m_type, false)) {
      SetNil(resultSlot);
    }
    return 1;
  };
  
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  PyrClass *m_wrapperClass;
  GType m_type;
  void *m_original;
  void *m_copy;
};

// ------ GObject

class GObjectRefRequest : public GtkRequest {
public:
  GObjectRefRequest(GObject *object) :
    m_object(object)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    g_object_ref(m_object);
  }
  
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GObject *m_object;
};

class GObjectUnrefRequest : public GtkRequest {
public:
  GObjectUnrefRequest(GObject *object) :
    m_object(object)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    g_object_unref(m_object);
  }
  
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GObject *m_object;
};

class GtkObjectSinkRequest : public GtkRequest {
public:
  GtkObjectSinkRequest(GtkObject *object) :
    m_object(object)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_object_sink(m_object);
  }
  
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkObject *m_object;
};

class GObjectGetPropertyRequest : public GtkRequest {
public:
  GObjectGetPropertyRequest(GObject *object, const char *name, int size) :
    m_object(object)
  {
	safe_string_copy(m_name, name, size, MEDIUM_STRING_SIZE);
	m_destroy_value = false;
  }

  ~GObjectGetPropertyRequest() {
    if (m_destroy_value) g_value_unset(&m_value);
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    memset(&m_value, 0, sizeof(m_value));
	GParamSpec *propspec;
	propspec = g_object_class_find_property(G_OBJECT_GET_CLASS(m_object), m_name);
	if (propspec != 0) {
	  g_value_init(&m_value, propspec->value_type); 
	  m_destroy_value = true;
	  g_object_get_property(m_object, m_name, &m_value);
	}
  }
  
  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    PyrSlot slot;
    if (!GValue_to_slot(&m_value, &slot)) {
      resultSlot->ucopy = slot.ucopy;
    } else {
      SetNil(resultSlot);
    } 
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GObject *m_object;
  char m_name[MEDIUM_STRING_SIZE];
  GValue m_value;
  bool m_destroy_value;
};

class GObjectSetPropertyRequest : public GtkRequest {
public:
  GObjectSetPropertyRequest(GObject *object, const char *name, int size, GValue *value) :
    m_object(object)
  {
	safe_string_copy(m_name, name, size, MEDIUM_STRING_SIZE);
    memset(&m_value, 0, sizeof(m_value));
    g_value_init(&m_value, G_VALUE_TYPE(value));
    g_value_copy(value, &m_value);
  }

  ~GObjectSetPropertyRequest() {
    g_value_unset(&m_value);
  }
  
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    g_object_set_property(m_object, m_name, &m_value);
  }
    
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GObject *m_object;
  char m_name[MEDIUM_STRING_SIZE];
  GValue m_value;
};

// ------ GdkGC

class GdkGCNewRequest : public GtkRequest {
public:
  GdkGCNewRequest(GdkDrawable *drawable) :
    m_drawable(drawable),
    m_gc(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_gc = gdk_gc_new(m_drawable);
  }



  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_gc));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GdkDrawable *m_drawable;
  GdkGC *m_gc;
};

class GdkGCSetRgbBgColorByNameRequest : public GtkRequest {
public:
  GdkGCSetRgbBgColorByNameRequest(GdkGC *gc, const char *colorname, int size) :
    m_gc(gc)
  {
    safe_string_copy(m_colorname, colorname, size, SMALL_STRING_SIZE);  
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    GdkColor color;
    gdk_color_parse(m_colorname, &color);
    gdk_gc_set_rgb_bg_color(m_gc, &color);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GdkGC *m_gc;
  char m_colorname[SMALL_STRING_SIZE];
};

class GdkGCSetRgbFgColorByNameRequest : public GtkRequest {
public:
  GdkGCSetRgbFgColorByNameRequest(GdkGC *gc, const char *colorname, int size) :
    m_gc(gc)
  {
    safe_string_copy(m_colorname, colorname, size, SMALL_STRING_SIZE);  
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    GdkColor color;
    gdk_color_parse(m_colorname, &color);
    gdk_gc_set_rgb_fg_color(m_gc, &color);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GdkGC *m_gc;
  char m_colorname[SMALL_STRING_SIZE];
};

class GdkGCSetRgbBgColorRequest : public GtkRequest {
public:
  GdkGCSetRgbBgColorRequest(GdkGC *gc, GdkColor color) :
    m_gc(gc), m_color(color)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gdk_gc_set_rgb_bg_color(m_gc, &m_color);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GdkGC *m_gc;
  GdkColor m_color;
};

class GdkGCSetRgbFgColorRequest : public GtkRequest {
public:
  GdkGCSetRgbFgColorRequest(GdkGC *gc, GdkColor color) :
    m_gc(gc), m_color(color)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gdk_gc_set_rgb_fg_color(m_gc, &m_color);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GdkGC *m_gc;
  GdkColor m_color;
};

class GdkGCSetLineAttributesRequest : public GtkRequest {
public:
  GdkGCSetLineAttributesRequest(GdkGC *gc, int width, GdkLineStyle style, GdkCapStyle cap, GdkJoinStyle join) :
    m_gc(gc), m_width(width), m_style(style), m_cap(cap), m_join(join)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gdk_gc_set_line_attributes(m_gc, m_width, m_style, m_cap, m_join);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GdkGC *m_gc;
  int m_width;
  GdkLineStyle m_style;
  GdkCapStyle m_cap;
  GdkJoinStyle m_join;
};

// ------ GdkWindow

class GdkWindowSetCursorRequest : public GtkRequest {
public:
  GdkWindowSetCursorRequest(GdkWindow *window, GdkCursorType type) :
    m_window(window), m_type(type)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gdk_window_set_cursor(m_window, gdk_cursor_new(m_type));
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GdkWindow *m_window;
  GdkCursorType m_type;
};

// ------ GdkDrawable

class GdkDrawableDrawRectangleRequest : public GtkRequest {
public:
  GdkDrawableDrawRectangleRequest(GdkDrawable *drawable, GdkGC *gc, bool filled, int x, int y, int w, int h) :
    m_drawable(drawable), m_gc(gc), m_filled(filled), m_x(x), m_y(y), m_w(w), m_h(h)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gdk_draw_rectangle(m_drawable, m_gc, m_filled, m_x, m_y, m_w, m_h);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GdkDrawable *m_drawable;
  GdkGC *m_gc;
  bool m_filled;
  int m_x, m_y, m_w, m_h;
};

class GdkDrawableDrawArcRequest : public GtkRequest {
public:
  GdkDrawableDrawArcRequest(GdkDrawable *drawable, GdkGC *gc, bool filled, int x, int y, int w, int h,
			    int angle1, int angle2) :
    m_drawable(drawable), m_gc(gc), m_filled(filled), m_x(x), m_y(y), m_w(w), m_h(h),
    m_angle1(angle1), m_angle2(angle2)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gdk_draw_arc(m_drawable, m_gc, m_filled, m_x, m_y, m_w, m_h, m_angle1, m_angle2);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GdkDrawable *m_drawable;
  GdkGC *m_gc;
  bool m_filled;
  int m_x, m_y, m_w, m_h;
  int m_angle1, m_angle2;
};

class GdkDrawableDrawLineRequest : public GtkRequest {
public:
  GdkDrawableDrawLineRequest(GdkDrawable *drawable, GdkGC *gc, int x1, int y1, int x2, int y2) :
    m_drawable(drawable), m_gc(gc), m_x1(x1), m_y1(y1), m_x2(x2), m_y2(y2)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gdk_draw_line(m_drawable, m_gc, m_x1, m_y1, m_x2, m_y2);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GdkDrawable *m_drawable;
  GdkGC *m_gc;
  int m_x1, m_y1, m_x2, m_y2;
};

class GdkDrawableDrawPointRequest : public GtkRequest {
public:
  GdkDrawableDrawPointRequest(GdkDrawable *drawable, GdkGC *gc, int x, int y) :
    m_drawable(drawable), m_gc(gc), m_x(x), m_y(y)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gdk_draw_point(m_drawable, m_gc, m_x, m_y);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GdkDrawable *m_drawable;
  GdkGC *m_gc;
  int m_x, m_y;
};

class GdkDrawableDrawLayoutRequest : public GtkRequest {
public:
  GdkDrawableDrawLayoutRequest(GdkDrawable *drawable, GdkGC *gc, int x, int y, PangoLayout *layout) :
    m_drawable(drawable), m_gc(gc), m_x(x), m_y(y), m_layout(layout)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gdk_draw_layout(m_drawable, m_gc, m_x, m_y, m_layout);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GdkDrawable *m_drawable;
  GdkGC *m_gc;
  int m_x, m_y;
  PangoLayout *m_layout;
};

class GdkDrawableGetSizeRequest : public GtkRequest {
public:
  GdkDrawableGetSizeRequest(GdkDrawable *drawable) :
    m_drawable(drawable),
    m_w(-1), m_h(-1)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gdk_drawable_get_size(m_drawable, &m_w, &m_h);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    PyrObject *array = newPyrArray(g->gc, 2, 0, false);
    PyrSlot *slots = array->slots;
    
    SetInt(slots + 0, m_w);
    SetInt(slots + 1, m_h);
    
    array->size = 2;

    SetObject(resultSlot, array);

    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GdkDrawable *m_drawable;
  int m_w, m_h;
};

// ------ Gtk

class GtkTimeoutAddRequest : public GtkRequest {
public:
  GtkTimeoutAddRequest(int interval, int id1, gboolean (*callback)(void *)) :
    m_interval(interval), m_id1(id1), m_callback(callback),
    m_id2(-1)
  {
  }

  void perform(void) {
    m_id2 = gtk_timeout_add(m_interval, m_callback, (void *)m_id1);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, m_id2);
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  gboolean (*m_callback)(void *);
  int m_interval;
  int m_id1, m_id2;
};

class GtkTimeoutRemoveRequest : public GtkRequest {
public:
  GtkTimeoutRemoveRequest(int id2) :
    m_id2(id2)
  {
  }

  void perform(void) {
    gtk_timeout_remove(m_id2);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  int m_id2;
};

// ------ GtkWidget

class GtkWidgetSetSizeRequestRequest : public GtkRequest {
public:
  GtkWidgetSetSizeRequestRequest(GtkWidget *widget, int w, int h) :
    m_widget(widget),
    m_w(w), m_h(h)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_widget_set_size_request(m_widget, m_w, m_h);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
  int m_w, m_h;
};

class GtkWidgetGetSizeRequestRequest : public GtkRequest {
public:
  GtkWidgetGetSizeRequestRequest(GtkWidget *widget) :
    m_widget(widget),
    m_w(-1), m_h(-1)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_widget_get_size_request(m_widget, &m_w, &m_h);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    PyrObject *array = newPyrArray(g->gc, 2, 0, false);
    PyrSlot *slots = array->slots;
    
    SetInt(slots + 0, m_w);
    SetInt(slots + 1, m_h);
    
    array->size = 2;

    SetObject(resultSlot, array);

    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
  int m_w, m_h;
};

class GtkWidgetShowRequest : public GtkRequest {
public:
  GtkWidgetShowRequest(GtkWidget *widget) :
    m_widget(widget)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_widget_show(m_widget);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
};

class GtkWidgetShowAllRequest : public GtkRequest {
public:
  GtkWidgetShowAllRequest(GtkWidget *widget) :
    m_widget(widget)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_widget_show_all(m_widget);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
};

class GtkWidgetHideRequest : public GtkRequest {
public:
  GtkWidgetHideRequest(GtkWidget *widget) :
    m_widget(widget)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_widget_hide(m_widget);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
};

class GtkWidgetHideAllRequest : public GtkRequest {
public:
  GtkWidgetHideAllRequest(GtkWidget *widget) :
    m_widget(widget)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_widget_hide_all(m_widget);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
};

class GtkWidgetDestroyRequest : public GtkRequest {
public:
  GtkWidgetDestroyRequest(GtkWidget *widget) :
    m_widget(widget)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_widget_destroy(m_widget);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
};

class GtkWidgetQueueDrawRequest : public GtkRequest {
public:
  GtkWidgetQueueDrawRequest(GtkWidget *widget) :
    m_widget(widget)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_widget_queue_draw(m_widget);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
};

class GtkWidgetQueueResizeRequest : public GtkRequest {
public:
  GtkWidgetQueueResizeRequest(GtkWidget *widget) :
    m_widget(widget)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_widget_queue_resize(m_widget);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
};

class GtkWidgetSetEventsRequest : public GtkRequest {
public:
  GtkWidgetSetEventsRequest(GtkWidget *widget, GdkEventMask mask) :
    m_widget(widget),
    m_mask(mask)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_widget_set_events(m_widget, m_mask);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
  GdkEventMask m_mask;
};

class GtkWidgetAddEventsRequest : public GtkRequest {
public:
  GtkWidgetAddEventsRequest(GtkWidget *widget, GdkEventMask mask) :
    m_widget(widget),
    m_mask(mask)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_widget_add_events(m_widget, m_mask);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
  GdkEventMask m_mask;
};

class GtkWidgetSetExtensionEventsRequest : public GtkRequest {
public:
  GtkWidgetSetExtensionEventsRequest(GtkWidget *widget, GdkExtensionMode mode) :
    m_widget(widget),
    m_mode(mode)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_widget_set_extension_events(m_widget, m_mode);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
  GdkExtensionMode m_mode;
};

class GtkWidgetGetWindowRequest : public GtkRequest {
public:
  GtkWidgetGetWindowRequest(GtkWidget *widget) :
    m_widget(widget),
    m_window(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_window = m_widget->window;
  }
  
  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    if (m_window == 0) {
      SetNil(resultSlot);
    } else {
      SetInt(resultSlot, int(m_window));
    }

    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
  GdkWindow *m_window;
};

class GtkWidgetSetStateRequest : public GtkRequest {
public:
  GtkWidgetSetStateRequest(GtkWidget *widget, GtkStateType state) :
    m_widget(widget),
    m_state(state)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_widget_set_state(m_widget, m_state);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
  GtkStateType m_state;
};

class GtkWidgetGetStyleRequest : public GtkRequest {
public:
  GtkWidgetGetStyleRequest(GtkWidget *widget) :
    m_widget(widget),
    m_style(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_style = gtk_widget_get_style(m_widget);
  }
  
  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    if (m_style == 0) {
      SetNil(resultSlot);
    } else {
      SetInt(resultSlot, int(m_style));
    }

    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
  GtkStyle *m_style;
};

class GtkWidgetGetPangoContextRequest : public GtkRequest {
public:
  GtkWidgetGetPangoContextRequest(GtkWidget *widget) :
    m_widget(widget),
    m_context(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_context = gtk_widget_get_pango_context(m_widget);
  }
  
  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    if (m_context == 0) {
      SetNil(resultSlot);
    } else {
      SetInt(resultSlot, int(m_context));
    }

    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
  PangoContext *m_context;
};

class GtkWidgetIsRealizedRequest : public GtkRequest_ReturnBoolean {
public:
  GtkWidgetIsRealizedRequest(GtkWidget *widget) :
    m_widget(widget)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = GTK_WIDGET_REALIZED(m_widget);
  }
  
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
};

class GtkWidgetIsToplevelRequest : public GtkRequest_ReturnBoolean {
public:
  GtkWidgetIsToplevelRequest(GtkWidget *widget) :
    m_widget(widget)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = GTK_WIDGET_TOPLEVEL(m_widget);
  }
  
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
};

class GtkWidgetIsMappedRequest : public GtkRequest_ReturnBoolean {
public:
  GtkWidgetIsMappedRequest(GtkWidget *widget) :
    m_widget(widget)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = GTK_WIDGET_MAPPED(m_widget);
  }
  
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
};

class GtkWidgetIsVisibleRequest : public GtkRequest_ReturnBoolean {
public:
  GtkWidgetIsVisibleRequest(GtkWidget *widget) :
    m_widget(widget)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = GTK_WIDGET_VISIBLE(m_widget);
  }
  
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
};

class GtkWidgetCanFocusRequest : public GtkRequest_ReturnBoolean {
public:
  GtkWidgetCanFocusRequest(GtkWidget *widget) :
    m_widget(widget)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = GTK_WIDGET_CAN_FOCUS(m_widget);
  }
  
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
};

class GtkWidgetGetParentWindowRequest : public GtkRequest {
public:
  GtkWidgetGetParentWindowRequest(GtkWidget *widget) :
    m_widget(widget),
    m_window(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_window = gtk_widget_get_parent_window(m_widget);
  }
  
  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    if (m_window == 0) {
      SetNil(resultSlot);
    } else {
      SetInt(resultSlot, int(m_window));
    }

    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
  GdkWindow *m_window;
};

class GtkWidgetSetParentRequest : public GtkRequest {
public:
  GtkWidgetSetParentRequest(GtkWidget *widget, GtkWidget *parent) :
    m_widget(widget),
	m_parent(parent)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_widget_set_parent(m_widget, m_parent);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
  GtkWidget *m_parent;
};

class GtkWidgetUnparentRequest : public GtkRequest {
public:
  GtkWidgetUnparentRequest(GtkWidget *widget) :
    m_widget(widget)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_widget_unparent(m_widget);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
};

class GtkWidgetGetParentRequest : public GtkRequest {
public:
  GtkWidgetGetParentRequest(GtkWidget *widget) :
    m_widget(widget),
    m_parent(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_parent = gtk_widget_get_parent(m_widget);
  }
  
  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    if (m_parent == 0) {
      SetNil(resultSlot);
    } else {
      SetInt(resultSlot, int(m_parent));
    }

    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
  GtkWidget *m_parent;
};

class GtkWidgetGetToplevelRequest : public GtkRequest {
public:
  GtkWidgetGetToplevelRequest(GtkWidget *widget) :
    m_widget(widget),
    m_toplevel(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_toplevel = gtk_widget_get_toplevel(m_widget);
  }
  
  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    if (m_toplevel == 0) {
      SetNil(resultSlot);
    } else {
      SetInt(resultSlot, int(m_toplevel));
    }

    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
  GtkWidget *m_toplevel;
};

class GtkWidgetGetDoubleBufferedRequest : public GtkRequest_ReturnBoolean {
public:
  GtkWidgetGetDoubleBufferedRequest(GtkWidget *widget) :
    m_widget(widget)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = GTK_WIDGET_DOUBLE_BUFFERED(m_widget);
  }
  
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
};

class GtkWidgetSetDoubleBufferedRequest : public GtkRequest {
public:
  GtkWidgetSetDoubleBufferedRequest(GtkWidget *widget, bool value) :
    m_widget(widget), m_value(value)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_widget_set_double_buffered(m_widget, m_value);
  }
  
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
  bool m_value;
};

class GtkWidgetGrabFocusRequest : public GtkRequest {
public:
  GtkWidgetGrabFocusRequest(GtkWidget *widget) :
    m_widget(widget)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_widget_grab_focus(m_widget);
  }
  
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
};

class GtkWidgetGetSensitiveRequest : public GtkRequest_ReturnBoolean {
public:
  GtkWidgetGetSensitiveRequest(GtkWidget *widget) :
    m_widget(widget)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = GTK_WIDGET_SENSITIVE(m_widget);
  }
  
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
};

class GtkWidgetSetSensitiveRequest : public GtkRequest {
public:
  GtkWidgetSetSensitiveRequest(GtkWidget *widget, bool value) :
    m_widget(widget), m_value(value)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_widget_set_sensitive(m_widget, m_value);
  }
  
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
  bool m_value;
};

class GtkWidgetModifyBgRequest : public GtkRequest {
public:
  GtkWidgetModifyBgRequest(GtkWidget *widget, GtkStateType state, GdkColor color) :
    m_widget(widget), m_state(state), m_color(color)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_widget_modify_bg(m_widget, m_state, &m_color);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
  GtkStateType m_state;
  GdkColor m_color;
};

class GtkWidgetReparentRequest : public GtkRequest {
public:
  GtkWidgetReparentRequest(GtkWidget *widget, GtkWidget *newParent) :
    m_widget(widget), m_newParent(newParent)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_widget_reparent(m_widget, m_newParent);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_widget;
  GtkWidget *m_newParent;
};

// ------ GtkWindow

class GtkWindowNewRequest : public GtkRequest {
public:
  GtkWindowNewRequest(GtkWindowType type) :
    m_type(type),
    m_window(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_window = (GtkWindow *)gtk_window_new(m_type);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_window));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWindowType m_type;
  GtkWindow *m_window;
};

class GtkWindowSetTitleRequest : public GtkRequest {
public:
  GtkWindowSetTitleRequest(GtkWindow *window, const char *s, int size) :
    m_window(window)
  {
    safe_string_copy(m_s, s, size, MEDIUM_STRING_SIZE);
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_window_set_title(m_window, m_s);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWindow *m_window;
  char m_s[MEDIUM_STRING_SIZE];
};

class GtkWindowGetTitleRequest : public GtkRequest {
public:
  GtkWindowGetTitleRequest(GtkWindow *window) :
    m_window(window)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    const char *text = gtk_window_get_title(m_window);
    m_size = safe_string_copy(m_s, text, -1, MEDIUM_STRING_SIZE);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    PyrString *strobj = newPyrStringN(g->gc, m_size, 0, true);
    memcpy(strobj->s, m_s, m_size);
   
    SetObject(resultSlot, strobj);

    return 1;
  };

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWindow *m_window;
  char m_s[MEDIUM_STRING_SIZE];
  int m_size;
};

class GtkWindowSetDefaultSizeRequest : public GtkRequest {
public:
  GtkWindowSetDefaultSizeRequest(GtkWindow *window, int w, int h) :
    m_window(window),
    m_w(w), m_h(h)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_window_set_default_size(m_window, m_w, m_h);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWindow *m_window;
  int m_w, m_h;
};

class GtkWindowResizeRequest : public GtkRequest {
public:
  GtkWindowResizeRequest(GtkWindow *window, int w, int h) :
    m_window(window),
    m_w(w), m_h(h)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_window_resize(m_window, m_w, m_h);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWindow *m_window;
  int m_w, m_h;
};

class GtkWindowGetDefaultSizeRequest : public GtkRequest {
public:
  GtkWindowGetDefaultSizeRequest(GtkWindow *window) :
    m_window(window),
    m_w(-1), m_h(-1)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_window_get_default_size(m_window, &m_w, &m_h);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    PyrObject *array = newPyrArray(g->gc, 2, 0, false);
    PyrSlot *slots = array->slots;
    
    SetInt(slots + 0, m_w);
    SetInt(slots + 1, m_h);
    
    array->size = 2;

    SetObject(resultSlot, array);

    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWindow *m_window;
  int m_w, m_h;
};

class GtkWindowGetSizeRequest : public GtkRequest {
public:
  GtkWindowGetSizeRequest(GtkWindow *window) :
    m_window(window),
    m_w(-1), m_h(-1)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_window_get_size(m_window, &m_w, &m_h);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    PyrObject *array = newPyrArray(g->gc, 2, 0, false);
    PyrSlot *slots = array->slots;
    
    SetInt(slots + 0, m_w);
    SetInt(slots + 1, m_h);
    
    array->size = 2;

    SetObject(resultSlot, array);

    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWindow *m_window;
  int m_w, m_h;
};

class GtkWindowMoveRequest : public GtkRequest {
public:
  GtkWindowMoveRequest(GtkWindow *window, int x, int y) :
    m_window(window),
    m_x(x), m_y(y)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_window_move(m_window, m_x, m_y);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWindow *m_window;
  int m_x, m_y;
};

class GtkWindowSetPositionRequest : public GtkRequest {
public:
  GtkWindowSetPositionRequest(GtkWindow *window, GtkWindowPosition position) :
    m_window(window),
    m_position(position)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_window_set_position(m_window, m_position);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWindow *m_window;
  GtkWindowPosition m_position;
};

class GtkWindowSetFocusRequest : public GtkRequest {
public:
  GtkWindowSetFocusRequest(GtkWindow *window, GtkWidget *widget) :
    m_window(window), m_widget(widget)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_window_set_focus(m_window, m_widget);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWindow *m_window;
  GtkWidget *m_widget;
};

class GtkWindowGetFocusRequest : public GtkRequest {
public:
  GtkWindowGetFocusRequest(GtkWindow *window) :
    m_window(window),
    m_widget(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_widget = gtk_window_get_focus(m_window);
  }
  
  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    if (m_widget == 0) {
      SetNil(resultSlot);
    } else {
      SetInt(resultSlot, int(m_widget));
    }

    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWindow *m_window;
  GtkWidget *m_widget;
};


class GtkWindowSetTransientForRequest : public GtkRequest {
public:
  GtkWindowSetTransientForRequest(GtkWindow *window, GtkWindow *parent) :
    m_window(window),
    m_parent(parent)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_window_set_transient_for(m_window, m_parent);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWindow *m_window;
  GtkWindow *m_parent;
};

class GtkWindowSetDestroyWithParentRequest : public GtkRequest {
public:
  GtkWindowSetDestroyWithParentRequest(GtkWindow *window, bool value) :
    m_window(window),
    m_value(value)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_window_set_destroy_with_parent(m_window, m_value);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWindow *m_window;
  bool m_value;
};

class GtkWindowSetHideOnDeleteRequest : public GtkRequest {
public:
  GtkWindowSetHideOnDeleteRequest(GtkWindow *window) :
    m_window(window)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    g_signal_connect((GObject *)m_window, "delete_event", G_CALLBACK(gtk_widget_hide_on_delete), 0);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWindow *m_window;
};

// ------ GtkFileSelection

class GtkFileSelectionNewRequest : public GtkRequest {
public:
  GtkFileSelectionNewRequest(const char *s, int size) :
    m_fileselection(0)
  {
    safe_string_copy(m_s, s, size, MEDIUM_STRING_SIZE);
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_fileselection = (GtkFileSelection *)gtk_file_selection_new(m_s);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_fileselection));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkFileSelection *m_fileselection;
  char m_s[FILENAME_STRING_SIZE];
};

class GtkFileSelectionSetFilenameRequest : public GtkRequest {
public:
  GtkFileSelectionSetFilenameRequest(GtkFileSelection *fileselection, const char *s, int size) :
    m_fileselection(fileselection)
  {
    safe_string_copy(m_s, s, size, FILENAME_STRING_SIZE);
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_file_selection_set_filename(m_fileselection, m_s);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkFileSelection *m_fileselection;
  char m_s[FILENAME_STRING_SIZE];
};

class GtkFileSelectionGetFilenameRequest : public GtkRequest {
public:
  GtkFileSelectionGetFilenameRequest(GtkFileSelection *fileselection) :
    m_fileselection(fileselection), m_size(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    const char *filename = gtk_file_selection_get_filename(m_fileselection);
    m_size = safe_string_copy(m_s, filename, -1, FILENAME_STRING_SIZE);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    PyrString *strobj = newPyrStringN(g->gc, m_size, 0, true);
    memcpy(strobj->s, m_s, m_size);
    
    SetObject(resultSlot, strobj);
    return 1;
  }
  
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkFileSelection *m_fileselection;
  char m_s[FILENAME_STRING_SIZE];
  int m_size;
};

class GtkFileSelectionShowFileopButtonsRequest : public GtkRequest {
public:
  GtkFileSelectionShowFileopButtonsRequest(GtkFileSelection *fileselection) :
    m_fileselection(fileselection)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_file_selection_show_fileop_buttons(m_fileselection);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkFileSelection *m_fileselection;
};

class GtkFileSelectionHideFileopButtonsRequest : public GtkRequest {
public:
  GtkFileSelectionHideFileopButtonsRequest(GtkFileSelection *fileselection) :
    m_fileselection(fileselection)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_file_selection_hide_fileop_buttons(m_fileselection);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkFileSelection *m_fileselection;
};

class GtkFileSelectionGetSelectionsRequest : public GtkRequest {
public:
  GtkFileSelectionGetSelectionsRequest(GtkFileSelection *fileselection) :
    m_fileselection(fileselection), m_n(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gchar **filenames = gtk_file_selection_get_selections(m_fileselection);

    int i;

    for (i = 0; i < 16; ++i) {
      if (filenames[i] == 0) break;
      m_sizes[i] = safe_string_copy(m_filenames[i], filenames[i], -1, FILENAME_STRING_SIZE);
    }
    m_n = i - 1;
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    PyrObject *array = newPyrArray(g->gc, m_n, 0, false);
    PyrSlot *slots = array->slots;
    
    for (int i = 0; i < m_n; ++i) {
      PyrString *strobj = newPyrStringN(g->gc, m_sizes[i], 0, true);
      memcpy(strobj->s, m_filenames[i], m_sizes[i]);
      SetObject(slots + i, strobj);
    }
    array->size = m_n;
    SetObject(resultSlot, array);
    return 1;
  }
  
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkFileSelection *m_fileselection;
  char m_filenames[16][FILENAME_STRING_SIZE];
  int m_sizes[16];
  int m_n;
};

class GtkFileSelectionSetSelectMultipleRequest : public GtkRequest {
public:
  GtkFileSelectionSetSelectMultipleRequest(GtkFileSelection *fileselection, bool value) :
    m_fileselection(fileselection), m_value(value)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_file_selection_set_select_multiple(m_fileselection, m_value);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkFileSelection *m_fileselection;
  bool m_value;
};

class GtkFileSelectionGetSelectMultipleRequest : public GtkRequest_ReturnBoolean {
public:
  GtkFileSelectionGetSelectMultipleRequest(GtkFileSelection *fileselection) :
    m_fileselection(fileselection)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = gtk_file_selection_get_select_multiple(m_fileselection);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkFileSelection *m_fileselection;
};

class GtkFileSelectionCompleteRequest : public GtkRequest {
public:
  GtkFileSelectionCompleteRequest(GtkFileSelection *fileselection, const char *s, int size) :
    m_fileselection(fileselection)
  {
    safe_string_copy(m_s, s, size, FILENAME_STRING_SIZE);
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_file_selection_complete(m_fileselection, m_s);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkFileSelection *m_fileselection;
  char m_s[FILENAME_STRING_SIZE];
};

// ------ GtkButton

class GtkButtonNewRequest : public GtkRequest {
public:
  GtkButtonNewRequest(void) :
    m_button(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_button = (GtkButton *)gtk_button_new();
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_button));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkButton *m_button;
};

class GtkButtonNewWithLabelRequest : public GtkRequest {
public:
  GtkButtonNewWithLabelRequest(const char *s, int size) :
    m_button(0)
  {
    safe_string_copy(m_s, s, size, MEDIUM_STRING_SIZE);
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_button = (GtkButton *)gtk_button_new_with_label(m_s);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_button));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkButton *m_button;
  char m_s[MEDIUM_STRING_SIZE];
};

class GtkButtonNewFromStockRequest : public GtkRequest {
public:
  GtkButtonNewFromStockRequest(const char *s, int size) :
    m_button(0)
  {
    safe_string_copy(m_s, s, size, SMALL_STRING_SIZE);
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_button = (GtkButton *)gtk_button_new_from_stock(m_s);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_button));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkButton *m_button;
  char m_s[SMALL_STRING_SIZE];
};

class GtkButtonSetLabelRequest : public GtkRequest {
public:
  GtkButtonSetLabelRequest(GtkButton *button, const char *s, int size) :
    m_button(button)
  {
    safe_string_copy(m_s, s, size, MEDIUM_STRING_SIZE);
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_button_set_label(m_button, m_s);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkButton *m_button;
  char m_s[MEDIUM_STRING_SIZE];
};

class GtkButtonGetLabelRequest : public GtkRequest {
public:
  GtkButtonGetLabelRequest(GtkButton *button) :
    m_button(button)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    const char *text = gtk_button_get_label(m_button);
    m_size = safe_string_copy(m_s, text, -1, MEDIUM_STRING_SIZE);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    PyrString *strobj = newPyrStringN(g->gc, m_size, 0, true);
    memcpy(strobj->s, m_s, m_size);
   
    SetObject(resultSlot, strobj);
    return 1;
  };

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkButton *m_button;
  char m_s[MEDIUM_STRING_SIZE];
  int m_size;
};

// ------ GtkToggleButton

class GtkToggleButtonNewRequest : public GtkRequest {
public:
  GtkToggleButtonNewRequest(void) :
    m_button(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_button = (GtkToggleButton *)gtk_toggle_button_new();
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_button));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkToggleButton *m_button;
};

class GtkToggleButtonNewWithLabelRequest : public GtkRequest {
public:
  GtkToggleButtonNewWithLabelRequest(const char *s, int size) :
    m_button(0)
  {
    safe_string_copy(m_s, s, size, MEDIUM_STRING_SIZE);
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_button = (GtkToggleButton *)gtk_toggle_button_new_with_label(m_s);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_button));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkToggleButton *m_button;
  char m_s[MEDIUM_STRING_SIZE];
};

class GtkToggleButtonSetModeRequest : public GtkRequest {
public:
  GtkToggleButtonSetModeRequest(GtkToggleButton *button, bool value) :
    m_button(button), m_value(value)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_toggle_button_set_mode(m_button, m_value);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkToggleButton *m_button;
  bool m_value;
};

class GtkToggleButtonGetModeRequest : public GtkRequest_ReturnBoolean {
public:
  GtkToggleButtonGetModeRequest(GtkToggleButton *button) :
    m_button(button)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = gtk_toggle_button_get_mode(m_button);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkToggleButton *m_button;
};

class GtkToggleButtonSetActiveRequest : public GtkRequest {
public:
  GtkToggleButtonSetActiveRequest(GtkToggleButton *button, bool value) :
    m_button(button), m_value(value)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_toggle_button_set_active(m_button, m_value);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkToggleButton *m_button;
  bool m_value;
};

class GtkToggleButtonGetActiveRequest : public GtkRequest_ReturnBoolean {
public:
  GtkToggleButtonGetActiveRequest(GtkToggleButton *button) :
    m_button(button)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = gtk_toggle_button_get_active(m_button);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkToggleButton *m_button;
};

class GtkToggleButtonSetInconsistentRequest : public GtkRequest {
public:
  GtkToggleButtonSetInconsistentRequest(GtkToggleButton *button, bool value) :
    m_button(button), m_value(value)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_toggle_button_set_inconsistent(m_button, m_value);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkToggleButton *m_button;
  bool m_value;
};

class GtkToggleButtonGetInconsistentRequest : public GtkRequest_ReturnBoolean {
public:
  GtkToggleButtonGetInconsistentRequest(GtkToggleButton *button) :
    m_button(button)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = gtk_toggle_button_get_inconsistent(m_button);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkToggleButton *m_button;
};

// ------ GtkToggleButton

class GtkCheckButtonNewRequest : public GtkRequest {
public:
  GtkCheckButtonNewRequest(void) :
    m_button(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_button = (GtkCheckButton *)gtk_check_button_new();
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_button));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkCheckButton *m_button;
};

class GtkCheckButtonNewWithLabelRequest : public GtkRequest {
public:
  GtkCheckButtonNewWithLabelRequest(const char *s, int size) :
    m_button(0)
  {
    safe_string_copy(m_s, s, size, MEDIUM_STRING_SIZE);
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_button = (GtkCheckButton *)gtk_check_button_new_with_label(m_s);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_button));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkCheckButton *m_button;
  char m_s[MEDIUM_STRING_SIZE];
};

// ------ GtkRadioButton

class GtkRadioButtonNewRequest : public GtkRequest {
public:
  GtkRadioButtonNewRequest(void *ptr) :
    m_button(0), m_ptr(ptr)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_button = (GtkRadioButton *)gtk_radio_button_new((GSList *)m_ptr);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_button));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkRadioButton *m_button;
  void *m_ptr;
};

class GtkRadioButtonNewWithLabelRequest : public GtkRequest {
public:
  GtkRadioButtonNewWithLabelRequest(void *ptr, const char *s, int size) :
    m_button(0), m_ptr(ptr)
  {
    safe_string_copy(m_s, s, size, MEDIUM_STRING_SIZE);
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_button = (GtkRadioButton *)gtk_radio_button_new_with_label((GSList *)m_ptr, m_s);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_button));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkRadioButton *m_button;
  char m_s[MEDIUM_STRING_SIZE];
  void *m_ptr;
};

class GtkRadioButtonNewFromWidgetRequest : public GtkRequest {
public:
  GtkRadioButtonNewFromWidgetRequest(GtkRadioButton *other) :
    m_button(0), m_other(other)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_button = (GtkRadioButton *)gtk_radio_button_new_from_widget(m_other);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_button));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkRadioButton *m_button;
  GtkRadioButton *m_other;
};

class GtkRadioButtonNewWithLabelFromWidgetRequest : public GtkRequest {
public:
  GtkRadioButtonNewWithLabelFromWidgetRequest(GtkRadioButton *other, const char *s, int size) :
    m_button(0), m_other(other)
  {
    safe_string_copy(m_s, s, size, MEDIUM_STRING_SIZE);
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_button = (GtkRadioButton *)gtk_radio_button_new_with_label_from_widget(m_other, m_s);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_button));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkRadioButton *m_button;
  char m_s[MEDIUM_STRING_SIZE];
  GtkRadioButton *m_other;
};

class GtkRadioButtonGetGroupRequest : public GtkRequest {
public:
  GtkRadioButtonGetGroupRequest(GtkRadioButton *button) :
    m_button(button), m_ptr(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_ptr = gtk_radio_button_get_group(m_button);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetPtr(resultSlot, m_ptr);
    return 1;
  };

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkRadioButton *m_button;
  void *m_ptr;
};

class GtkRadioButtonSetGroupRequest : public GtkRequest {
public:
  GtkRadioButtonSetGroupRequest(GtkRadioButton *button, void *ptr) :
    m_button(button), m_ptr(ptr)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_radio_button_set_group(m_button, (GSList *)m_ptr);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkRadioButton *m_button;
  void *m_ptr;
};

// ------ GtkBin

class GtkBinGetChildRequest : public GtkRequest {
public:
  GtkBinGetChildRequest(GtkBin *bin) :
    m_bin(bin), m_child(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_child = gtk_bin_get_child(m_bin);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    if (m_child == 0)
      SetNil(resultSlot);
    else
      SetInt(resultSlot, int(m_child));
    return 1;
  };

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkBin *m_bin;
  GtkWidget *m_child;
};

class GtkContainerAddRequest : public GtkRequest {
public:
  GtkContainerAddRequest(GtkContainer *container, GtkWidget *child) :
    m_container(container),
    m_child(child)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_container_add(m_container, m_child);
  }
  
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkContainer *m_container;
  GtkWidget *m_child;
};

class GtkContainerRemoveRequest : public GtkRequest {
public:
  GtkContainerRemoveRequest(GtkContainer *container, GtkWidget *child) :
    m_container(container),
    m_child(child)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_container_remove(m_container, m_child);
  }
  
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkContainer *m_container;
  GtkWidget *m_child;
};

class GtkContainerChildrenRequest : public GtkRequest {
public:
  GtkContainerChildrenRequest(GtkContainer *container) :
    m_container(container), m_children(0), m_n(0)
  {
  }

  ~GtkContainerChildrenRequest() {
    g_list_free(m_children);
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_children = g_list_copy(gtk_container_children(m_container));
    m_n = g_list_length(m_children);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    GList *item;
    
    int i;
    PyrObject *array = newPyrArray(g->gc, m_n, 0, false);
    PyrSlot *slots = array->slots;
    
    for (i = 0, item = m_children; i < m_n; ++i, item = g_list_next(item)) {
      GtkWidget *child = (GtkWidget *)item->data;
      if (child == 0) {
	SetNil(slots + i);
      } else {
	SetInt(slots + i, int(child));
      }
    }
    array->size = m_n;
    SetObject(resultSlot, array);
    return 1;
  };
  
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkContainer *m_container;
  GList *m_children;
  int m_n;
};

class GtkContainerNChildrenRequest : public GtkRequest {
public:
  GtkContainerNChildrenRequest(GtkContainer *container) :
    m_container(container), m_n(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    GList *children;
    children = g_list_copy(gtk_container_children(m_container));
    m_n = g_list_length(children);
    g_list_free(children);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, m_n);
    return 1;
  };
  
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkContainer *m_container;
  int m_n;
};

class GtkContainerSetBorderWidthRequest : public GtkRequest {
public:
  GtkContainerSetBorderWidthRequest(GtkContainer *container, int n) :
    m_container(container),
    m_n(n)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_container_set_border_width(m_container, m_n);
  }
  
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkContainer *m_container;
  int m_n;
};

// ----------- GtkImage

class GtkImageNewFromStockRequest : public GtkRequest {
public:
  GtkImageNewFromStockRequest(const char *s, int size, GtkIconSize iconsize) :
    m_image(0), m_iconsize(iconsize)
  {
    safe_string_copy(m_s, s, size, SMALL_STRING_SIZE);
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_image = (GtkImage *)gtk_image_new_from_stock(m_s, m_iconsize);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_image));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkImage *m_image;
  char m_s[SMALL_STRING_SIZE];
  GtkIconSize m_iconsize;
};

class GtkImageNewFromFileRequest : public GtkRequest {
public:
  GtkImageNewFromFileRequest(const char *s, int size) :
    m_image(0)
  {
    safe_string_copy(m_s, s, size, FILENAME_STRING_SIZE);
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_image = (GtkImage *)gtk_image_new_from_file(m_s);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_image));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkImage *m_image;
  char m_s[FILENAME_STRING_SIZE];
};

// ---------

class GtkFixedNewRequest : public GtkRequest {
public:
  GtkFixedNewRequest(void) :
    m_fixed(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_fixed = (GtkFixed *)gtk_fixed_new();
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_fixed));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkFixed *m_fixed;
};

class GtkFixedPutRequest : public GtkRequest {
public:
  GtkFixedPutRequest(GtkFixed *fixed, GtkWidget *widget, int x, int y) :
    m_fixed(fixed), m_widget(widget), m_x(x), m_y(y)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_fixed_put(m_fixed, m_widget, m_x, m_y);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkFixed *m_fixed;
  GtkWidget *m_widget;
  int m_x, m_y;
};

class GtkFixedMoveRequest : public GtkRequest {
public:
  GtkFixedMoveRequest(GtkFixed *fixed, GtkWidget *widget, int x, int y) :
    m_fixed(fixed), m_widget(widget), m_x(x), m_y(y)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_fixed_move(m_fixed, m_widget, m_x, m_y);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkFixed *m_fixed;
  GtkWidget *m_widget;
  int m_x, m_y;
};

class GtkFixedSetHasWindowRequest : public GtkRequest {
public:
  GtkFixedSetHasWindowRequest(GtkFixed *fixed, bool value) :
    m_fixed(fixed), m_value(value)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_fixed_set_has_window(m_fixed, m_value);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkFixed *m_fixed;
  bool m_value;
};

class GtkFixedGetHasWindowRequest : public GtkRequest {
public:
  GtkFixedGetHasWindowRequest(GtkFixed *fixed) :
    m_fixed(fixed)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = gtk_fixed_get_has_window(m_fixed);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    if (m_value)
      SetTrue(resultSlot);
    else
      SetFalse(resultSlot);
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkFixed *m_fixed;
  bool m_value;
};


// ---------

class GtkHBoxNewRequest : public GtkRequest {
public:
  GtkHBoxNewRequest(bool homo, int spacing) :
    m_homo(homo), m_spacing(spacing),
    m_hbox(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_hbox = (GtkHBox *)gtk_hbox_new(m_homo, m_spacing);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_hbox));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  bool m_homo;
  int m_spacing;
  GtkHBox *m_hbox;
};

class GtkVBoxNewRequest : public GtkRequest {
public:
  GtkVBoxNewRequest(bool homo, int spacing) :
    m_homo(homo), m_spacing(spacing),
    m_vbox(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_vbox = (GtkVBox *)gtk_vbox_new(m_homo, m_spacing);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_vbox));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  bool m_homo;
  int m_spacing;
  GtkVBox *m_vbox;
};

class GtkBoxPackStartRequest : public GtkRequest {
public:
  GtkBoxPackStartRequest(GtkBox *box, GtkWidget *child, bool expand, bool fill, int padding) :
    m_box(box), m_child(child),
    m_expand(expand), m_fill(fill), m_padding(padding)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_box_pack_start(m_box, m_child, m_expand, m_fill, m_padding);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkBox *m_box;
  GtkWidget *m_child;
  bool m_expand, m_fill;
  int m_padding;
};

class GtkBoxPackEndRequest : public GtkRequest {
public:
  GtkBoxPackEndRequest(GtkBox *box, GtkWidget *child, bool expand, bool fill, int padding) :
    m_box(box), m_child(child),
    m_expand(expand), m_fill(fill), m_padding(padding)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_box_pack_end(m_box, m_child, m_expand, m_fill, m_padding);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkBox *m_box;
  GtkWidget *m_child;
  bool m_expand, m_fill;
  int m_padding;
};

class GtkBoxReorderChildRequest : public GtkRequest {
public:
  GtkBoxReorderChildRequest(GtkBox *box, GtkWidget *child, int n) :
    m_box(box), m_child(child), m_n(n)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_box_reorder_child(m_box, m_child, m_n);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkBox *m_box;
  GtkWidget *m_child;
  int m_n;
};

// ----------------

class GtkEventBoxNewRequest : public GtkRequest {
public:
  GtkEventBoxNewRequest(void) :
    m_eventbox(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_eventbox = (GtkEventBox *)gtk_event_box_new();
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_eventbox));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkEventBox *m_eventbox;
};

// ----------------

class GtkNotebookNewRequest : public GtkRequest {
public:
  GtkNotebookNewRequest(void) :
    m_notebook(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_notebook = (GtkNotebook *)gtk_notebook_new();
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_notebook));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkNotebook *m_notebook;
};

class GtkNotebookAppendPageRequest : public GtkRequest {
public:
  GtkNotebookAppendPageRequest(GtkNotebook *notebook, GtkWidget *page, GtkWidget *label) :
    m_notebook(notebook), m_page(page), m_label(label)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_notebook_append_page(m_notebook, m_page, m_label);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkNotebook *m_notebook;
  GtkWidget *m_page;
  GtkWidget *m_label;
};

class GtkNotebookPrependPageRequest : public GtkRequest {
public:
  GtkNotebookPrependPageRequest(GtkNotebook *notebook, GtkWidget *page, GtkWidget *label) :
    m_notebook(notebook), m_page(page), m_label(label)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_notebook_prepend_page(m_notebook, m_page, m_label);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkNotebook *m_notebook;
  GtkWidget *m_page;
  GtkWidget *m_label;
};

class GtkNotebookInsertPageRequest : public GtkRequest {
public:
  GtkNotebookInsertPageRequest(GtkNotebook *notebook, GtkWidget *page, GtkWidget *label, int n) :
    m_notebook(notebook), m_page(page), m_label(label), m_n(n)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_notebook_insert_page(m_notebook, m_page, m_label, m_n);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkNotebook *m_notebook;
  GtkWidget *m_page;
  GtkWidget *m_label;
  int m_n;
};

class GtkNotebookRemovePageRequest : public GtkRequest {
public:
  GtkNotebookRemovePageRequest(GtkNotebook *notebook, int n) :
    m_notebook(notebook), m_n(n)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_notebook_remove_page(m_notebook, m_n);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkNotebook *m_notebook;
  int m_n;
};

class GtkNotebookGetCurrentPageRequest : public GtkRequest {
public:
  GtkNotebookGetCurrentPageRequest(GtkNotebook *notebook) :
    m_notebook(notebook), m_n(-1)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_n = gtk_notebook_get_current_page(m_notebook);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, m_n);
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkNotebook *m_notebook;
  int m_n;
};

class GtkNotebookGetNthPageRequest : public GtkRequest {
public:
  GtkNotebookGetNthPageRequest(GtkNotebook *notebook, int n) :
    m_notebook(notebook), m_n(n), m_page(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_page = gtk_notebook_get_nth_page(m_notebook, m_n);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    if (m_page == 0)
      SetNil(resultSlot);
    else
      SetInt(resultSlot, int(m_page));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkNotebook *m_notebook;
  int m_n;
  GtkWidget *m_page;
};

class GtkNotebookGetNPagesRequest : public GtkRequest {
public:
  GtkNotebookGetNPagesRequest(GtkNotebook *notebook) :
    m_notebook(notebook), m_n(-1)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_n = gtk_notebook_get_n_pages(m_notebook);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, m_n);
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkNotebook *m_notebook;
  int m_n;
};

class GtkNotebookPageNumRequest : public GtkRequest {
public:
  GtkNotebookPageNumRequest(GtkNotebook *notebook, GtkWidget *page) :
    m_notebook(notebook), m_page(page), m_n(-1)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_n = gtk_notebook_page_num(m_notebook, m_page);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    if (m_n >= 0)
      SetInt(resultSlot, m_n);
    else
      SetNil(resultSlot);
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkNotebook *m_notebook;
  GtkWidget *m_page;
  int m_n;
};

class GtkNotebookSetCurrentPageRequest : public GtkRequest {
public:
  GtkNotebookSetCurrentPageRequest(GtkNotebook *notebook, int n) :
    m_notebook(notebook), m_n(n)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_notebook_set_current_page(m_notebook, m_n);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkNotebook *m_notebook;
  int m_n;
};

class GtkNotebookPrevPageRequest : public GtkRequest {
public:
  GtkNotebookPrevPageRequest(GtkNotebook *notebook) :
    m_notebook(notebook)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_notebook_prev_page(m_notebook);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkNotebook *m_notebook;
};

class GtkNotebookNextPageRequest : public GtkRequest {
public:
  GtkNotebookNextPageRequest(GtkNotebook *notebook) :
    m_notebook(notebook)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_notebook_next_page(m_notebook);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkNotebook *m_notebook;
};

class GtkNotebookSetScrollableRequest : public GtkRequest {
public:
  GtkNotebookSetScrollableRequest(GtkNotebook *notebook, bool value) :
    m_notebook(notebook), m_value(value)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_notebook_set_scrollable(m_notebook, m_value);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkNotebook *m_notebook;
  bool m_value;
};

class GtkNotebookGetScrollableRequest : public GtkRequest {
public:
  GtkNotebookGetScrollableRequest(GtkNotebook *notebook) :
    m_notebook(notebook), m_value(false)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = gtk_notebook_get_scrollable(m_notebook);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    if (m_value)
      SetTrue(resultSlot);
    else
      SetFalse(resultSlot);
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkNotebook *m_notebook;
  bool m_value;
};

class GtkNotebookPopupDisableRequest : public GtkRequest {
public:
  GtkNotebookPopupDisableRequest(GtkNotebook *notebook) :
    m_notebook(notebook)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_notebook_popup_disable(m_notebook);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkNotebook *m_notebook;
};

class GtkNotebookPopupEnableRequest : public GtkRequest {
public:
  GtkNotebookPopupEnableRequest(GtkNotebook *notebook) :
    m_notebook(notebook)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_notebook_popup_enable(m_notebook );
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkNotebook *m_notebook;
};

// --------- GtkLabel

class GtkLabelNewRequest : public GtkRequest {
public:
  GtkLabelNewRequest(const char *s, int size) :
    m_label(0), m_text(0)
  {
    if (s != 0) {
      safe_string_copy(m_s, s, size, MEDIUM_STRING_SIZE);
      m_text = m_s;
    } else {
      m_text = 0;
    }
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_label = (GtkLabel *)gtk_label_new(m_text);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_label));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkLabel *m_label;
  char m_s[MEDIUM_STRING_SIZE];
  char *m_text;
};

class GtkLabelSetTextRequest : public GtkRequest {
public:
  GtkLabelSetTextRequest(GtkLabel *label, const char *s, int size) :
    m_label(label), m_text(0)
  {
    if (s != 0) {
      safe_string_copy(m_s, s, size, MEDIUM_STRING_SIZE);
      m_text = m_s;
    } else {
      m_text = 0;
    }
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_label_set_text(m_label, m_text);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkLabel *m_label;
  char m_s[MEDIUM_STRING_SIZE];
  char *m_text;
};

class GtkLabelSetMarkupRequest : public GtkRequest {
public:
  GtkLabelSetMarkupRequest(GtkLabel *label, const char *s, int size) :
    m_label(label), m_text(0)
  {
    if (s != 0) {
      safe_string_copy(m_s, s, size, MEDIUM_STRING_SIZE);
      m_text = m_s;
    } else {
      m_text = 0;
    }
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_label_set_markup(m_label, m_text);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkLabel *m_label;
  char m_s[MEDIUM_STRING_SIZE];
  char *m_text;
};

class GtkLabelGetTextRequest : public GtkRequest {
public:
  GtkLabelGetTextRequest(GtkLabel *label) :
    m_label(label)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    const char *text = gtk_label_get_text(m_label);
    m_size = safe_string_copy(m_s, text, -1, MEDIUM_STRING_SIZE);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;

    PyrString *strobj = newPyrStringN(g->gc, m_size, 0, true);
    memcpy(strobj->s, m_s, m_size);
    
    SetObject(resultSlot, strobj);
    return 1;
  };

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkLabel *m_label;
  char m_s[MEDIUM_STRING_SIZE];
  int m_size;
};

class GtkLabelSetJustifyRequest : public GtkRequest {
public:
  GtkLabelSetJustifyRequest(GtkLabel *label, GtkJustification mode) :
    m_label(label), m_mode(mode)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_label_set_justify(m_label, m_mode);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkLabel *m_label;
  GtkJustification m_mode;
};

// --------- GtkFrame

class GtkFrameNewRequest : public GtkRequest {
public:
  GtkFrameNewRequest(const char *s, int size) :
    m_frame(0), m_text(0)
  {
    if (s != 0) {
      safe_string_copy(m_s, s, size, MEDIUM_STRING_SIZE);
      m_text = m_s;
    } else {
      m_text = 0;
    }
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_frame = (GtkFrame *)gtk_frame_new(m_text);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_frame));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkFrame *m_frame;
  char m_s[MEDIUM_STRING_SIZE];
  char *m_text;
};

class GtkFrameSetLabelRequest : public GtkRequest {
public:
  GtkFrameSetLabelRequest(GtkFrame *frame, const char *s, int size) :
    m_frame(frame), m_text(0)
  {
    if (s != 0) {
      safe_string_copy(m_s, s, size, MEDIUM_STRING_SIZE);
      m_text = m_s;
    } else {
      m_text = 0;
    }
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_frame_set_label(m_frame, m_text);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkFrame *m_frame;
  char m_s[MEDIUM_STRING_SIZE];
  char *m_text;
};

class GtkFrameSetShadowTypeRequest : public GtkRequest {
public:
  GtkFrameSetShadowTypeRequest(GtkFrame *frame, GtkShadowType type) :
    m_frame(frame), m_type(type)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_frame_set_shadow_type(m_frame, m_type);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkFrame *m_frame;
  GtkShadowType m_type;
};

// ------------- GtkTextView

class GtkTextViewNewRequest : public GtkRequest {
public:
  GtkTextViewNewRequest(void) :
    m_textview(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_textview = (GtkTextView *)gtk_text_view_new();
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_textview));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextView *m_textview;
};

class GtkTextViewNewWithBufferRequest : public GtkRequest {
public:
  GtkTextViewNewWithBufferRequest(GtkTextBuffer *buffer) :
    m_textview(0), m_buffer(buffer)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_textview = (GtkTextView *)gtk_text_view_new_with_buffer(m_buffer);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_textview));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextView *m_textview;
  GtkTextBuffer *m_buffer;
};

class GtkTextViewGetBufferRequest : public GtkRequest {
public:
  GtkTextViewGetBufferRequest(GtkTextView *textview) :
    m_textview(textview), m_buffer(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_buffer = gtk_text_view_get_buffer(m_textview);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_buffer));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextView *m_textview;
  GtkTextBuffer *m_buffer;
};

class GtkTextViewSetBufferRequest : public GtkRequest {
public:
  GtkTextViewSetBufferRequest(GtkTextView *textview, GtkTextBuffer *buffer) :
    m_textview(textview), m_buffer(buffer)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_text_view_set_buffer(m_textview, m_buffer);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextView *m_textview;
  GtkTextBuffer *m_buffer;
};

class GtkTextViewSetEditableRequest : public GtkRequest {
public:
  GtkTextViewSetEditableRequest(GtkTextView *textview, bool value) :
    m_textview(textview), m_value(value)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_text_view_set_editable(m_textview, m_value);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextView *m_textview;
  bool m_value;
};

class GtkTextViewGetEditableRequest : public GtkRequest {
public:
  GtkTextViewGetEditableRequest(GtkTextView *textview) :
    m_textview(textview), m_value(false)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = gtk_text_view_get_editable(m_textview);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    if (m_value)
      SetTrue(resultSlot);
    else
      SetFalse(resultSlot);
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextView *m_textview;
  bool m_value;
};

class GtkTextViewSetCursorVisibleRequest : public GtkRequest {
public:
  GtkTextViewSetCursorVisibleRequest(GtkTextView *textview, bool value) :
    m_textview(textview), m_value(value)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_text_view_set_cursor_visible(m_textview, m_value);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextView *m_textview;
  bool m_value;
};

class GtkTextViewScrollToMarkRequest : public GtkRequest {
public:
  GtkTextViewScrollToMarkRequest(GtkTextView *textview, GtkTextMark *mark, double within_margin,
				 bool use_align, double xalign, double yalign) :
    m_textview(textview), m_mark(mark),
    m_within_margin(within_margin),
    m_use_align(use_align), m_xalign(xalign), m_yalign(yalign)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_text_view_scroll_to_mark(m_textview, m_mark, m_within_margin, m_use_align, m_xalign, m_yalign);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextView *m_textview;
  GtkTextMark *m_mark;
  double m_within_margin;
  bool m_use_align;
  float m_xalign;
  float m_yalign;
};

class GtkTextViewScrollToIterRequest : public GtkRequest {
public:
  GtkTextViewScrollToIterRequest(GtkTextView *textview, GtkTextIter *iter, double within_margin,
				 bool use_align, double xalign, double yalign) :
    m_textview(textview), m_iter(iter),
    m_within_margin(within_margin),
    m_use_align(use_align), m_xalign(xalign), m_yalign(yalign)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_text_view_scroll_to_iter(m_textview, m_iter, m_within_margin, m_use_align, m_xalign, m_yalign);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextView *m_textview;
  GtkTextIter *m_iter;
  double m_within_margin;
  bool m_use_align;
  float m_xalign;
  float m_yalign;
};

// ------------- GtkTextBuffer

class GtkTextBufferNewRequest : public GtkRequest {
public:
  GtkTextBufferNewRequest(void) :
    m_buffer(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_buffer = (GtkTextBuffer *)gtk_text_buffer_new(0);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_buffer));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextBuffer *m_buffer;
};

class GtkTextBufferSetTextRequest : public GtkRequest {
public:
  GtkTextBufferSetTextRequest(GtkTextBuffer *buffer, const char *s, int size) :
    m_buffer(buffer)
  {
    m_size = safe_string_copy(m_s, s, size, LARGE_STRING_SIZE);
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_text_buffer_set_text(m_buffer, m_s, m_size);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextBuffer *m_buffer;
  char m_s[LARGE_STRING_SIZE];
  int m_size;
};

class GtkTextBufferGetTextRequest : public GtkRequest {
public:
  GtkTextBufferGetTextRequest(GtkTextBuffer *buffer, GtkTextIter *beginIter, GtkTextIter *endIter, bool hidden) :
    m_buffer(buffer),
    m_beginIter(beginIter), m_endIter(endIter), m_hidden(hidden),
    m_text(0), m_size(0)
  {
  }

  ~GtkTextBufferGetTextRequest() {
    g_free(m_text);
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    GtkTextIter beginOfBufferIter, endOfBufferIter;

    if (m_beginIter == 0) {
      gtk_text_buffer_get_start_iter(m_buffer, &beginOfBufferIter);
      m_beginIter = &beginOfBufferIter;
    }
    
    if (m_endIter == 0) {
      gtk_text_buffer_get_end_iter(m_buffer, &endOfBufferIter);
      m_endIter = &endOfBufferIter;
    }

    m_text = gtk_text_buffer_get_text(m_buffer, m_beginIter, m_endIter, m_hidden);
    m_size = strlen(m_text);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    PyrString *strobj = newPyrStringN(g->gc, m_size, 0, true);
    memcpy(strobj->s, m_text, m_size);
    
    SetObject(resultSlot, strobj);
    return 1;
  };

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextBuffer *m_buffer;
  GtkTextIter *m_beginIter;
  GtkTextIter *m_endIter;
  bool m_hidden;
  char *m_text;
  int m_size;
};

class GtkTextIterIsStartRequest : public GtkRequest_ReturnBoolean {
public:
  GtkTextIterIsStartRequest(GtkTextIter *iter) :
    m_iter(iter)
  {
  }
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = gtk_text_iter_is_start(m_iter);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextIter *m_iter;
};

class GtkTextIterIsEndRequest : public GtkRequest_ReturnBoolean {
public:
  GtkTextIterIsEndRequest(GtkTextIter *iter) :
    m_iter(iter)
  {
  }
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = gtk_text_iter_is_end(m_iter);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextIter *m_iter;
};

class GtkTextIterGetLineRequest : public GtkRequest_ReturnInteger {
public:
  GtkTextIterGetLineRequest(GtkTextIter *iter) :
    m_iter(iter)
  {
  }
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = gtk_text_iter_get_line(m_iter);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextIter *m_iter;
};

class GtkTextIterGetOffsetRequest : public GtkRequest_ReturnInteger {
public:
  GtkTextIterGetOffsetRequest(GtkTextIter *iter) :
    m_iter(iter)
  {
  }
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = gtk_text_iter_get_offset(m_iter);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextIter *m_iter;
};

class GtkTextIterGetLineOffsetRequest : public GtkRequest_ReturnInteger {
public:
  GtkTextIterGetLineOffsetRequest(GtkTextIter *iter) :
    m_iter(iter)
  {
  }
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = gtk_text_iter_get_line_offset(m_iter);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextIter *m_iter;
};

class GtkTextIterGetLineIndexRequest : public GtkRequest_ReturnInteger {
public:
  GtkTextIterGetLineIndexRequest(GtkTextIter *iter) :
    m_iter(iter)
  {
  }
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = gtk_text_iter_get_line_index(m_iter);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextIter *m_iter;
};

class GtkTextIterSetLineOffsetRequest : public GtkRequest {
public:
  GtkTextIterSetLineOffsetRequest(GtkTextIter *iter, int offset) :
    m_iter(iter), m_offset(offset)
  {
  }
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_text_iter_set_line_offset(m_iter, m_offset);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextIter *m_iter;
  int m_offset;
};

class GtkTextIterForwardToLineEndRequest : public GtkRequest {
public:
  GtkTextIterForwardToLineEndRequest(GtkTextIter *iter) :
    m_iter(iter)
  {
  }
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_text_iter_forward_to_line_end(m_iter);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextIter *m_iter;
};

class GtkTextIterForwardLineRequest : public GtkRequest {
public:
  GtkTextIterForwardLineRequest(GtkTextIter *iter) :
    m_iter(iter)
  {
  }
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_text_iter_forward_line(m_iter);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextIter *m_iter;
};

class GtkTextBufferGetIterAtMarkRequest : public GtkRequest_ReturnBoxed {
public:
  GtkTextBufferGetIterAtMarkRequest(GtkTextBuffer *buffer, GtkTextMark *mark) :
    GtkRequest_ReturnBoxed("GtkTextIter"),
    m_buffer(buffer), m_mark(mark)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    GtkTextIter iter;
    gtk_text_buffer_get_iter_at_mark(m_buffer, &iter, m_mark);
    copyBoxedResult(gtk_text_iter_get_type(), &iter);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextBuffer *m_buffer;
  GtkTextMark *m_mark;
};

class GtkTextBufferGetIterAtLineOffsetRequest : public GtkRequest_ReturnBoxed {
public:
  GtkTextBufferGetIterAtLineOffsetRequest(GtkTextBuffer *buffer, int line, int offset) :
    GtkRequest_ReturnBoxed("GtkTextIter"),
    m_buffer(buffer), m_line(line), m_offset(offset)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    GtkTextIter iter;
    gtk_text_buffer_get_iter_at_line_offset(m_buffer, &iter, m_line, m_offset);
    copyBoxedResult(gtk_text_iter_get_type(), &iter);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextBuffer *m_buffer;
  int m_line, m_offset;
};

class GtkTextBufferGetIterAtLineIndexRequest : public GtkRequest_ReturnBoxed {
public:
  GtkTextBufferGetIterAtLineIndexRequest(GtkTextBuffer *buffer, int line, int index) :
    GtkRequest_ReturnBoxed("GtkTextIter"),
    m_buffer(buffer), m_line(line), m_index(index)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    GtkTextIter iter;
    gtk_text_buffer_get_iter_at_line_index(m_buffer, &iter, m_line, m_index);
    copyBoxedResult(gtk_text_iter_get_type(), &iter);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextBuffer *m_buffer;
  int m_line, m_index;
};

class GtkTextBufferGetIterAtLineRequest : public GtkRequest_ReturnBoxed {
public:
  GtkTextBufferGetIterAtLineRequest(GtkTextBuffer *buffer, int line) :
    GtkRequest_ReturnBoxed("GtkTextIter"),
    m_buffer(buffer), m_line(line)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    GtkTextIter iter;
    gtk_text_buffer_get_iter_at_line(m_buffer, &iter, m_line);
    copyBoxedResult(gtk_text_iter_get_type(), &iter);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextBuffer *m_buffer;
  int m_line;
};

class GtkTextBufferGetIterAtOffsetRequest : public GtkRequest_ReturnBoxed {
public:
  GtkTextBufferGetIterAtOffsetRequest(GtkTextBuffer *buffer, int offset) :
    GtkRequest_ReturnBoxed("GtkTextIter"),
    m_buffer(buffer), m_offset(offset)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    GtkTextIter iter;
    gtk_text_buffer_get_iter_at_offset(m_buffer, &iter, m_offset);
    copyBoxedResult(gtk_text_iter_get_type(), &iter);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextBuffer *m_buffer;
  int m_offset;
};

class GtkTextBufferGetStartIterRequest : public GtkRequest_ReturnBoxed {
public:
  GtkTextBufferGetStartIterRequest(GtkTextBuffer *buffer) :
    GtkRequest_ReturnBoxed("GtkTextIter"),
    m_buffer(buffer)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    GtkTextIter iter;
    gtk_text_buffer_get_start_iter(m_buffer, &iter);
    copyBoxedResult(gtk_text_iter_get_type(), &iter);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextBuffer *m_buffer;
};

class GtkTextBufferGetEndIterRequest : public GtkRequest_ReturnBoxed {
public:
  GtkTextBufferGetEndIterRequest(GtkTextBuffer *buffer) :
    GtkRequest_ReturnBoxed("GtkTextIter"),
    m_buffer(buffer)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    GtkTextIter iter;
    gtk_text_buffer_get_end_iter(m_buffer, &iter);
    copyBoxedResult(gtk_text_iter_get_type(), &iter);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextBuffer *m_buffer;
};

class GtkTextBufferGetBoundsRequest : public GtkRequest {
public:
  GtkTextBufferGetBoundsRequest(GtkTextBuffer *buffer) :
    m_buffer(buffer), m_ptr1(0), m_ptr2(0)
  {
  }
  
  void perform(GType type, void *boxed) {
    GtkTextIter iter1, iter2;
    gtk_text_buffer_get_bounds(m_buffer, &iter1, &iter2);
    m_ptr1 = (GtkTextIter *)g_boxed_copy(gtk_text_iter_get_type(), &iter1);
    m_ptr2 = (GtkTextIter *)g_boxed_copy(gtk_text_iter_get_type(), &iter2);
  }
  
  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    PyrClass *gtktreeiterclassobj = getsym("GtkTextIter")->u.classobj;

    PyrObject *iter1Object = instantiateObject(g->gc, gtktreeiterclassobj, 0, true, true);
  
    SetInt(iter1Object->slots + 0, (int)m_ptr1);
    SetInt(iter1Object->slots + 1, gtk_text_iter_get_type());

    PyrObject *iter2Object = instantiateObject(g->gc, gtktreeiterclassobj, 0, true, true);
  
    SetInt(iter2Object->slots + 0, (int)m_ptr2);
    SetInt(iter2Object->slots + 1, gtk_text_iter_get_type());

    PyrObject *array = newPyrArray(g->gc, 2, 0, false);
    PyrSlot *slots = array->slots;

    SetObject(slots + 0, iter1Object);
    SetObject(slots + 1, iter2Object);

    array->size = 2;

    SetObject(resultSlot, array);
    return 1;
  };
  
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextBuffer *m_buffer;
  void *m_ptr1, *m_ptr2;
};

class GtkTextBufferInsertRequest : public GtkRequest {
public:
  GtkTextBufferInsertRequest(GtkTextBuffer *buffer, GtkTextIter *iter, const char *s, int size) :
    m_buffer(buffer), m_iter(iter)
  {
    m_size = safe_string_copy(m_s, s, size, LARGE_STRING_SIZE);
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_text_buffer_insert(m_buffer, m_iter, m_s, m_size);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextBuffer *m_buffer;
  GtkTextIter *m_iter;
  char m_s[LARGE_STRING_SIZE];
  int m_size;
};

class GtkTextBufferInsertAtCursorRequest : public GtkRequest {
public:
  GtkTextBufferInsertAtCursorRequest(GtkTextBuffer *buffer, const char *s, int size) :
    m_buffer(buffer)
  {
    m_size = safe_string_copy(m_s, s, size, LARGE_STRING_SIZE);
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_text_buffer_insert_at_cursor(m_buffer, m_s, m_size);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextBuffer *m_buffer;
  char m_s[LARGE_STRING_SIZE];
  int m_size;
};

class GtkTextBufferInsertWithTagRequest : public GtkRequest {
public:
  GtkTextBufferInsertWithTagRequest(GtkTextBuffer *buffer, GtkTextIter *iter, const char *s, int size,
				    const char *tagName) :
    m_buffer(buffer), m_iter(iter)
  {
    m_size = safe_string_copy(m_s, s, size, LARGE_STRING_SIZE);
    safe_string_copy(m_tagName, tagName, -1, SMALL_STRING_SIZE);
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_text_buffer_insert_with_tags_by_name(m_buffer, m_iter, m_s, m_size, m_tagName);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextBuffer *m_buffer;
  GtkTextIter *m_iter;
  GtkTextTag *m_tag;
  char m_s[LARGE_STRING_SIZE];
  int m_size;
  char m_tagName[SMALL_STRING_SIZE];
};

class GtkTextBufferGetMarkRequest : public GtkRequest {
public:
  GtkTextBufferGetMarkRequest(GtkTextBuffer *buffer, const char *name, int size) :
    m_buffer(buffer), m_mark(0)
  {
    safe_string_copy(m_name, name, size, MEDIUM_STRING_SIZE);
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_mark = gtk_text_buffer_get_mark(m_buffer, m_name);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    if (m_mark == 0)
      SetNil(resultSlot);
    else
      SetInt(resultSlot, int(m_mark));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextBuffer *m_buffer;
  char m_name[SMALL_STRING_SIZE];
  GtkTextMark *m_mark;
};

class GtkTextBufferGetInsertRequest : public GtkRequest {
public:
  GtkTextBufferGetInsertRequest(GtkTextBuffer *buffer) :
    m_buffer(buffer)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_mark = gtk_text_buffer_get_insert(m_buffer);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    if (m_mark == 0)
      SetNil(resultSlot);
    else
      SetInt(resultSlot, int(m_mark));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextBuffer *m_buffer;
  GtkTextMark *m_mark;
};

class GtkTextBufferGetSelectionBoundRequest : public GtkRequest {
public:
  GtkTextBufferGetSelectionBoundRequest(GtkTextBuffer *buffer) :
    m_buffer(buffer)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_mark = gtk_text_buffer_get_selection_bound(m_buffer);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    if (m_mark == 0)
      SetNil(resultSlot);
    else
      SetInt(resultSlot, int(m_mark));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextBuffer *m_buffer;
  GtkTextMark *m_mark;
};


class GtkTextBufferGetSelectionBoundsRequest : public GtkRequest {
public:
  GtkTextBufferGetSelectionBoundsRequest(GtkTextBuffer *buffer) :
    m_buffer(buffer), m_ptr1(0), m_ptr2(0)
  {
  }
  
  void perform(GType type, void *boxed) {
    GtkTextIter iter1, iter2;
    gtk_text_buffer_get_selection_bounds(m_buffer, &iter1, &iter2);
    m_ptr1 = (GtkTextIter *)g_boxed_copy(gtk_text_iter_get_type(), &iter1);
    m_ptr2 = (GtkTextIter *)g_boxed_copy(gtk_text_iter_get_type(), &iter2);
  }
  
  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    PyrClass *gtktreeiterclassobj = getsym("GtkTextIter")->u.classobj;

    PyrObject *iter1Object = instantiateObject(g->gc, gtktreeiterclassobj, 0, true, true);
  
    SetInt(iter1Object->slots + 0, (int)m_ptr1);
    SetInt(iter1Object->slots + 1, gtk_text_iter_get_type());

    PyrObject *iter2Object = instantiateObject(g->gc, gtktreeiterclassobj, 0, true, true);
  
    SetInt(iter2Object->slots + 0, (int)m_ptr2);
    SetInt(iter2Object->slots + 1, gtk_text_iter_get_type());

    PyrObject *array = newPyrArray(g->gc, 2, 0, false);
    PyrSlot *slots = array->slots;

    SetObject(slots + 0, iter1Object);
    SetObject(slots + 1, iter2Object);

    array->size = 2;

    SetObject(resultSlot, array);
    return 1;
  };
  
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextBuffer *m_buffer;
  void *m_ptr1, *m_ptr2;
};

class GtkTextBufferCreateMarkRequest : public GtkRequest {
public:
  GtkTextBufferCreateMarkRequest(GtkTextBuffer *buffer, const char *s, GtkTextIter *iter,
				 bool leftGravity) :
    m_buffer(buffer), m_iter(iter), m_leftGravity(leftGravity)
  {
    if (s == 0) {
      m_name = 0;
    } else {
      safe_string_copy(m_s, s, -1, SMALL_STRING_SIZE);
      m_name = m_s;
    }
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_mark = gtk_text_buffer_create_mark(m_buffer, m_name, m_iter, m_leftGravity);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    if (m_mark == 0)
      SetNil(resultSlot);
    else
      SetInt(resultSlot, int(m_mark));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextBuffer *m_buffer;
  GtkTextIter *m_iter;
  char *m_name;
  char m_s[SMALL_STRING_SIZE];
  bool m_leftGravity;
  GtkTextMark *m_mark;
};

class GtkTextBufferDeleteMarkRequest : public GtkRequest {
public:
  GtkTextBufferDeleteMarkRequest(GtkTextBuffer *buffer, GtkTextMark *mark) :
    m_buffer(buffer), m_mark(mark)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_text_buffer_delete_mark(m_buffer, m_mark);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTextBuffer *m_buffer;
  GtkTextMark *m_mark;
};

// ------ GtkMenuShell

class GtkMenuShellAppendRequest : public GtkRequest {
public:
  GtkMenuShellAppendRequest(GtkMenuShell *menushell, GtkWidget *child) :
    m_menushell(menushell), m_child(child)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_menu_shell_append(m_menushell, m_child);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkMenuShell *m_menushell;
  GtkWidget *m_child;
};

class GtkMenuShellPrependRequest : public GtkRequest {
public:
  GtkMenuShellPrependRequest(GtkMenuShell *menushell, GtkWidget *child) :
    m_menushell(menushell), m_child(child)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_menu_shell_prepend(m_menushell, m_child);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkMenuShell *m_menushell;
  GtkWidget *m_child;
};

class GtkMenuShellInsertRequest : public GtkRequest {
public:
  GtkMenuShellInsertRequest(GtkMenuShell *menushell, GtkWidget *child, int position) :
    m_menushell(menushell), m_child(child), m_position(position)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_menu_shell_insert(m_menushell, m_child, m_position);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkMenuShell *m_menushell;
  GtkWidget *m_child;
  int m_position;
};

// ------ GtkMenu

class GtkMenuNewRequest : public GtkRequest {
public:
  GtkMenuNewRequest(void) :
    m_menu(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_menu = (GtkMenu *)gtk_menu_new();
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_menu));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkMenu *m_menu;
};

class GtkMenuPopupRequest : public GtkRequest {
public:
  GtkMenuPopupRequest(GtkMenu *menu, int button, int time) :
    m_menu(menu), m_button(button), m_time(time)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_menu_popup(m_menu, 0, 0, 0, 0, m_button, m_time);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkMenu *m_menu;
  int m_button, m_time;
};

// ---------- MenuItem

class GtkMenuItemNewWithLabelRequest : public GtkRequest {
public:
  GtkMenuItemNewWithLabelRequest(const char *s, int size) :
    m_menuitem(0)
  {
    safe_string_copy(m_s, s, size, MEDIUM_STRING_SIZE);
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_menuitem = (GtkMenuItem *)gtk_menu_item_new_with_label(m_s);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_menuitem));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkMenuItem *m_menuitem;
  char m_s[MEDIUM_STRING_SIZE];
};

class GtkSeparatorMenuItemNewRequest : public GtkRequest {
public:
  GtkSeparatorMenuItemNewRequest(void) :
    m_menuitem(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_menuitem = (GtkSeparatorMenuItem *)gtk_separator_menu_item_new();
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_menuitem));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSeparatorMenuItem *m_menuitem;
};

class GtkMenuItemSetSubmenuRequest : public GtkRequest {
public:
  GtkMenuItemSetSubmenuRequest(GtkMenuItem *menuitem, GtkWidget *widget) :
    m_menuitem(menuitem), m_widget(widget)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_menu_item_set_submenu(m_menuitem, m_widget);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkMenuItem *m_menuitem;
  GtkWidget *m_widget;
};

// ------ GtkHPaned

class GtkHPanedNewRequest : public GtkRequest {
public:
  GtkHPanedNewRequest(void) :
    m_hpaned(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_hpaned = (GtkHPaned *)gtk_hpaned_new();
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_hpaned));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkHPaned *m_hpaned;
};

class GtkVPanedNewRequest : public GtkRequest {
public:
  GtkVPanedNewRequest(void) :
    m_vpaned(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_vpaned = (GtkVPaned *)gtk_vpaned_new();
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_vpaned));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkVPaned *m_vpaned;
};

// -------- GtkPaned

class GtkPanedAdd1Request : public GtkRequest {
public:
  GtkPanedAdd1Request(GtkPaned *paned, GtkWidget *widget) :
    m_paned(paned), m_widget(widget)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_paned_add1(m_paned, m_widget);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkPaned *m_paned;
  GtkWidget *m_widget;
};

class GtkPanedPack1Request : public GtkRequest {
public:
  GtkPanedPack1Request(GtkPaned *paned, GtkWidget *widget, bool resize, bool shrink) :
    m_paned(paned), m_widget(widget), m_resize(resize), m_shrink(shrink)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_paned_pack1(m_paned, m_widget, m_resize, m_shrink);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkPaned *m_paned;
  GtkWidget *m_widget;
  bool m_resize, m_shrink;
};

class GtkPanedAdd2Request : public GtkRequest {
public:
  GtkPanedAdd2Request(GtkPaned *paned, GtkWidget *widget) :
    m_paned(paned), m_widget(widget)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_paned_add2(m_paned, m_widget);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkPaned *m_paned;
  GtkWidget *m_widget;
};

class GtkPanedPack2Request : public GtkRequest {
public:
  GtkPanedPack2Request(GtkPaned *paned, GtkWidget *widget, bool resize, bool shrink) :
    m_paned(paned), m_widget(widget), m_resize(resize), m_shrink(shrink)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_paned_pack2(m_paned, m_widget, m_resize, m_shrink);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkPaned *m_paned;
  GtkWidget *m_widget;
  bool m_resize, m_shrink;
};

class GtkPanedSetPositionRequest : public GtkRequest {
public:
  GtkPanedSetPositionRequest(GtkPaned *paned, int position) :
    m_paned(paned), m_position(position)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_paned_set_position(m_paned, m_position);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkPaned *m_paned;
  int m_position;
};

class GtkPanedGetPositionRequest : public GtkRequest {
public:
  GtkPanedGetPositionRequest(GtkPaned *paned) :
    m_paned(paned), m_position(-1)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_position = gtk_paned_get_position(m_paned);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, m_position);
    return 1;
  }
 
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkPaned *m_paned;
  int m_position;
};

// ------ GtkEntry

class GtkEntryNewRequest : public GtkRequest {
public:
  GtkEntryNewRequest(void) :
    m_entry(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_entry = (GtkEntry *)gtk_entry_new();
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_entry));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkEntry *m_entry;
};

class GtkEntryGetTextRequest : public GtkRequest {
public:
  GtkEntryGetTextRequest(GtkEntry *entry) :
    m_entry(entry),
    m_text(0), m_size(0)
  {
  }

  ~GtkEntryGetTextRequest() {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    // NOTE: This will be cause of some mess in future. We should copy.
    m_text = gtk_entry_get_text(m_entry);
    m_size = strlen(m_text);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    PyrString *strobj = newPyrStringN(g->gc, m_size, 0, true);
    memcpy(strobj->s, m_text, m_size);
    
    SetObject(resultSlot, strobj);
    return 1;
  };

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkEntry *m_entry;
  const char *m_text;
  int m_size;
};

class GtkEntrySetTextRequest : public GtkRequest {
public:
  GtkEntrySetTextRequest(GtkEntry *entry, const char *s, int size) :
    m_entry(entry)
  {
    m_size = safe_string_copy(m_s, s, size, LARGE_STRING_SIZE);
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_entry_set_text(m_entry, m_s);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkEntry *m_entry;
  char m_s[LARGE_STRING_SIZE];
  int m_size;
};

class GtkEntrySetEditableRequest : public GtkRequest {
public:
  GtkEntrySetEditableRequest(GtkEntry *entry, bool value) :
    m_entry(entry),
    m_value(value)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_entry_set_editable(m_entry, m_value);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkEntry *m_entry;
  bool m_value;
};

// ------ GtkEntry

class GtkDrawingAreaNewRequest : public GtkRequest {
public:
  GtkDrawingAreaNewRequest(void) :
    m_drawingarea(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_drawingarea = (GtkDrawingArea *)gtk_drawing_area_new();
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_drawingarea));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkDrawingArea *m_drawingarea;
};

// ------ GtkEntry

class GtkTableNewRequest : public GtkRequest {
public:
  GtkTableNewRequest(int n, int m, bool homo) :
    m_table(0), m_n(n), m_m(m), m_homo(homo)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_table = (GtkTable *)gtk_table_new(m_n, m_m, m_homo);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_table));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTable *m_table;
  int m_n, m_m;
  bool m_homo;
};

class GtkTableResizeRequest : public GtkRequest {
public:
  GtkTableResizeRequest(GtkTable *table, int n, int m) :
    m_table(table), m_n(n), m_m(m)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_table_resize(m_table, m_n, m_m);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTable *m_table;
  int m_n, m_m;
};

class GtkTableAttachRequest : public GtkRequest {
public:
  GtkTableAttachRequest(GtkTable *table, GtkWidget *child,
			int left, int right, int top, int bottom,
			GtkAttachOptions xoptions, GtkAttachOptions yoptions, int xpadding, int ypadding) :
    m_table(table), m_child(child),
    m_left(left), m_right(right), m_top(top), m_bottom(bottom),
    m_xoptions(xoptions), m_yoptions(yoptions), m_xpadding(xpadding), m_ypadding(ypadding)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_table_attach(m_table, m_child, m_left, m_right, m_top, m_bottom, m_xoptions, m_yoptions, m_xpadding, m_ypadding);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTable *m_table;
  GtkWidget *m_child;
  int m_left, m_right, m_top, m_bottom;
  GtkAttachOptions m_xoptions, m_yoptions;
  int m_xpadding, m_ypadding;
};

class GtkTableSetColSpacingsRequest : public GtkRequest {
public:
  GtkTableSetColSpacingsRequest(GtkTable *table, int n) :
    m_table(table), m_n(n)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_table_set_col_spacings(m_table, m_n);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTable *m_table;
  int m_n;
};

class GtkTableSetRowSpacingsRequest : public GtkRequest {
public:
  GtkTableSetRowSpacingsRequest(GtkTable *table, int n) :
    m_table(table), m_n(n)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_table_set_col_spacings(m_table, m_n);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTable *m_table;
  int m_n;
};

// ----- GtkMisc

class GtkMiscSetAlignmentRequest : public GtkRequest {
public:
  GtkMiscSetAlignmentRequest(GtkMisc *misc, double x, double y) :
    m_misc(misc), m_x(x), m_y(y)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_misc_set_alignment(m_misc, m_x, m_y);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkMisc *m_misc;
  double m_x, m_y;
};

class GtkMiscSetPaddingRequest : public GtkRequest {
public:
  GtkMiscSetPaddingRequest(GtkMisc *misc, int x, int y) :
    m_misc(misc), m_x(x), m_y(y)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_misc_set_padding(m_misc, m_x, m_y);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkMisc *m_misc;
  int m_x, m_y;
};

// ------ GtkAdjustment

class GtkAdjustmentNewRequest : public GtkRequest {
public:
  GtkAdjustmentNewRequest(double value, double lower, double upper, double step_increment,
			  double page_increment, double page_size) :
    m_adjustment(0), m_value(value), m_lower(lower), m_upper(upper),
    m_step_increment(step_increment), m_page_increment(page_increment), m_page_size(page_size)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_adjustment = (GtkAdjustment *)gtk_adjustment_new(m_value, m_lower, m_upper, m_step_increment, m_page_increment, m_page_size);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_adjustment));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkAdjustment *m_adjustment;
  double m_value, m_lower, m_upper, m_step_increment, m_page_increment, m_page_size;
};

class GtkAdjustmentGetValueRequest : public GtkRequest {
public:
  GtkAdjustmentGetValueRequest(GtkAdjustment *adjustment) :
    m_adjustment(adjustment)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = gtk_adjustment_get_value(m_adjustment);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetFloat(resultSlot, m_value);
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkAdjustment *m_adjustment;
  double m_value;
};

class GtkAdjustmentSetValueRequest : public GtkRequest {
public:
  GtkAdjustmentSetValueRequest(GtkAdjustment *adjustment, double value) :
    m_adjustment(adjustment), m_value(value)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_adjustment_set_value(m_adjustment, m_value);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkAdjustment *m_adjustment;
  double m_value;
};

class GtkAdjustmentClampPageRequest : public GtkRequest {
public:
  GtkAdjustmentClampPageRequest(GtkAdjustment *adjustment, double lower, double upper) :
    m_adjustment(adjustment), m_lower(lower), m_upper(upper)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_adjustment_clamp_page(m_adjustment, m_lower, m_upper);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkAdjustment *m_adjustment;
  double m_lower, m_upper;
};

class GtkAdjustmentChangedRequest : public GtkRequest {
public:
  GtkAdjustmentChangedRequest(GtkAdjustment *adjustment) :
    m_adjustment(adjustment)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_adjustment_changed(m_adjustment);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkAdjustment *m_adjustment;
};

class GtkAdjustmentValueChangedRequest : public GtkRequest {
public:
  GtkAdjustmentValueChangedRequest(GtkAdjustment *adjustment) :
    m_adjustment(adjustment)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_adjustment_value_changed(m_adjustment);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkAdjustment *m_adjustment;
};

// ------ GtkSizeGroup

class GtkSizeGroupNewRequest : public GtkRequest {
public:
  GtkSizeGroupNewRequest(GtkSizeGroupMode mode) :
    m_group(0), m_mode(mode)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_group = (GtkSizeGroup *)gtk_size_group_new(m_mode);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_group));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSizeGroup *m_group;
  GtkSizeGroupMode m_mode;
};

class GtkSizeGroupAddWidgetRequest : public GtkRequest {
 public:
  GtkSizeGroupAddWidgetRequest(GtkSizeGroup *group, GtkWidget *widget) :
    m_group(group),
    m_widget(widget)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_size_group_add_widget(m_group, m_widget);
  }
  
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSizeGroup *m_group;
  GtkWidget *m_widget;
};

class GtkSizeGroupRemoveWidgetRequest : public GtkRequest {
public:
  GtkSizeGroupRemoveWidgetRequest(GtkSizeGroup *group, GtkWidget *widget) :
    m_group(group),
    m_widget(widget)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_size_group_remove_widget(m_group, m_widget);
  }
  
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSizeGroup *m_group;
  GtkWidget *m_widget;
};

class GtkSizeGroupSetModeRequest : public GtkRequest {
public:
  GtkSizeGroupSetModeRequest(GtkSizeGroup *group, GtkSizeGroupMode mode) :
    m_group(group), m_mode(mode)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_size_group_set_mode(m_group, m_mode);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSizeGroup *m_group;
  GtkSizeGroupMode m_mode;
};

// ------ GtkScrolledWindow

class GtkScrolledWindowNewRequest : public GtkRequest {
public:
  GtkScrolledWindowNewRequest() :
    m_window(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_window = (GtkScrolledWindow *)gtk_scrolled_window_new(0, 0);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_window));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkScrolledWindow *m_window;
};

class GtkScrolledWindowSetShadowTypeRequest : public GtkRequest {
public:
  GtkScrolledWindowSetShadowTypeRequest(GtkScrolledWindow *window, GtkShadowType type) :
    m_window(window), m_type(type)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_scrolled_window_set_shadow_type(m_window, m_type);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkScrolledWindow *m_window;
  GtkShadowType m_type;
};

class GtkScrolledWindowSetPolicyRequest : public GtkRequest {
public:
  GtkScrolledWindowSetPolicyRequest(GtkScrolledWindow *window, GtkPolicyType hpol, GtkPolicyType vpol) :
    m_window(window), m_hpol(hpol), m_vpol(vpol)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_scrolled_window_set_policy(m_window, m_hpol, m_vpol);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkScrolledWindow *m_window;
  GtkPolicyType m_hpol, m_vpol;
};

class GtkScrolledWindowAddWithViewportRequest : public GtkRequest {
public:
  GtkScrolledWindowAddWithViewportRequest(GtkScrolledWindow *scrolledWindow, GtkWidget *child) :
    m_scrolledWindow(scrolledWindow),
    m_child(child)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_scrolled_window_add_with_viewport(m_scrolledWindow, m_child);
  }
  
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkScrolledWindow *m_scrolledWindow;
  GtkWidget *m_child;
};

// ------- GtkStyle

class GtkStyleGetWhiteGCRequest : public GtkRequest {
public:
  GtkStyleGetWhiteGCRequest(GtkStyle *style) :
    m_style(style)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_gc = m_style->white_gc;
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_gc));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkStyle *m_style;
  GdkGC *m_gc;
};

class GtkStyleGetBlackGCRequest : public GtkRequest {
public:
  GtkStyleGetBlackGCRequest(GtkStyle *style) :
    m_style(style)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_gc = m_style->black_gc;
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_gc));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkStyle *m_style;
  GdkGC *m_gc;
};

// ------- GtkTreeModel

class GtkTreeModelGetColumnTypeRequest : public GtkRequest {
public:
  GtkTreeModelGetColumnTypeRequest(GtkTreeModel *model, int column) :
    m_model(model), m_column(column), m_type(G_TYPE_INVALID)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_type = gtk_tree_model_get_column_type(m_model, m_column);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_type));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeModel *m_model;
  int m_column;
  GType m_type;
};

class GtkTreeModelGetIterRequest : public GtkRequest_ReturnBoxed {
public:
  GtkTreeModelGetIterRequest(GtkTreeModel *model, GtkTreePath *path) :
    GtkRequest_ReturnBoxed("GtkTreeIter"),
    m_model(model), m_path(path)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    GtkTreeIter iter;
    gtk_tree_model_get_iter(m_model, &iter, m_path);
    copyBoxedResult(gtk_tree_iter_get_type(), &iter);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeModel *m_model;
  GtkTreePath *m_path;
};

class GtkTreeModelGetIterFirstRequest : public GtkRequest_ReturnBoxed {
public:
  GtkTreeModelGetIterFirstRequest(GtkTreeModel *model) :
    GtkRequest_ReturnBoxed("GtkTreeIter"),
    m_model(model)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    GtkTreeIter iter;
    if (gtk_tree_model_get_iter_first(m_model, &iter))
      copyBoxedResult(gtk_tree_iter_get_type(), &iter);
    else
      nilBoxedResult();
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeModel *m_model;
};

class GtkTreeModelGetPathRequest : public GtkRequest_ReturnBoxed {
public:
  GtkTreeModelGetPathRequest(GtkTreeModel *model, GtkTreeIter *iter) :
    GtkRequest_ReturnBoxed("GtkTreePath"),
    m_model(model), m_iter(iter)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    GtkTreeIter parent;
    GtkTreePath *path = gtk_tree_model_get_path(m_model, m_iter);
    if (path != 0)
      setBoxedResult(gtk_tree_path_get_type(), path);
    else
      nilBoxedResult();
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeModel *m_model;
  GtkTreeIter *m_iter;
};

class GtkTreeModelGetPathStringRequest : public GtkRequest {
public:
  GtkTreeModelGetPathStringRequest(GtkTreeModel *model, GtkTreeIter *iter) :
    m_model(model), m_iter(iter)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    GtkTreeIter parent;
    GtkTreePath *path = gtk_tree_model_get_path(m_model, m_iter);
    if (path != 0) {
      gchar *s = gtk_tree_path_to_string(path);
      m_size = safe_string_copy(m_pathstring0, s, -1, SMALL_STRING_SIZE);
      m_pathstring = m_pathstring0;
      g_free(s);
      gtk_tree_path_free(path);
    } else {
      m_pathstring = 0;
    }
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    if (m_pathstring) {
      PyrString *strobj = newPyrStringN(g->gc, m_size, 0, true);
      memcpy(strobj->s, m_pathstring, m_size);
      SetObject(resultSlot, strobj);
    } else {
      SetNil(resultSlot);
    }
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeModel *m_model;
  GtkTreeIter *m_iter;
  char m_pathstring0[SMALL_STRING_SIZE];
  int m_size;
  char *m_pathstring;
};

/*
  extern bool GValue_to_slot(const GValue *value, PyrSlot *slot);

  class GtkTreeModelGetValueRequest : public GtkRequest {
  public:
  GtkTreeModelGetValueRequest(GtkTreeModel *model, GtkTreeIter *iter, int column) :
  m_model(model), m_iter(iter), m_column(column)
  {
  }

  ~GtkTreeModelGetValueRequest() {
  g_value_unset(&m_value);
  }

  void perform(void) {
  GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
  gtk_tree_model_get_value(m_model, m_iter, m_column, &m_value);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    GValue_to_slot(&m_value, resultSlot);
    return 1;
  };

    
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

  private:
  GtkTreeModel *m_model;
  GtkTreeIter *m_iter;
  int m_column;
  GValue m_value;
  };
*/

class GtkTreeModelIterParentRequest : public GtkRequest_ReturnBoxed {
public:
  GtkTreeModelIterParentRequest(GtkTreeModel *model, GtkTreeIter *iter) :
    GtkRequest_ReturnBoxed("GtkTreeIter"),
    m_model(model), m_iter(iter)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    GtkTreeIter parent;
    if (gtk_tree_model_iter_parent(m_model, &parent, m_iter))
      copyBoxedResult(gtk_tree_iter_get_type(), &parent);
    else
      nilBoxedResult();
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeModel *m_model;
  GtkTreeIter *m_iter;
};


class GtkTreeModelIterNextRequest : public GtkRequest_ReturnBoxed {
public:
  GtkTreeModelIterNextRequest(GtkTreeModel *model, GtkTreeIter *iter) :
    GtkRequest_ReturnBoxed("GtkTreeIter"),
    m_model(model), m_iter(iter)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    GtkTreeIter *newiter = (GtkTreeIter *)g_boxed_copy(gtk_tree_iter_get_type(), m_iter);
    if (!gtk_tree_model_iter_next(m_model, newiter)) {
      g_boxed_free(gtk_tree_iter_get_type(), newiter);
      nilBoxedResult();
    } else {
      setBoxedResult(gtk_tree_iter_get_type(), newiter);
    }
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeModel *m_model;
  GtkTreeIter *m_iter;
};

class GtkTreeModelIterHasChildRequest : public GtkRequest_ReturnBoolean {
public:
  GtkTreeModelIterHasChildRequest(GtkTreeModel *model, GtkTreeIter *iter) :
    m_model(model), m_iter(iter)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = gtk_tree_model_iter_has_child(m_model, m_iter);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeModel *m_model;
  GtkTreeIter *m_iter;
};

class GtkTreeModelIterNChildrenRequest : public GtkRequest {
public:
  GtkTreeModelIterNChildrenRequest(GtkTreeModel *model, GtkTreeIter *iter) :
    m_model(model), m_iter(iter), m_n(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_n = gtk_tree_model_iter_n_children(m_model, m_iter);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, m_n);
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeModel *m_model;
  GtkTreeIter *m_iter;
  int m_n;
};

class GtkTreeModelIterChildrenRequest : public GtkRequest_ReturnBoxed {
public:
  GtkTreeModelIterChildrenRequest(GtkTreeModel *model, GtkTreeIter *iter) :
    GtkRequest_ReturnBoxed("GtkTreeIter"),
    m_model(model), m_iter(iter)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    GtkTreeIter children;
    if (gtk_tree_model_iter_children(m_model, &children, m_iter))
      copyBoxedResult(gtk_tree_iter_get_type(), &children);
    else
      nilBoxedResult();
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeModel *m_model;
  GtkTreeIter *m_iter;
};

class GtkTreeModelIterNthChildRequest : public GtkRequest_ReturnBoxed {
public:
  GtkTreeModelIterNthChildRequest(GtkTreeModel *model, GtkTreeIter *iter, int n) :
    GtkRequest_ReturnBoxed("GtkTreeIter"),
    m_model(model), m_iter(iter), m_n(n)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    GtkTreeIter children;
    if (gtk_tree_model_iter_nth_child(m_model, &children, m_iter, m_n))
      copyBoxedResult(gtk_tree_iter_get_type(), &children);
    else
      nilBoxedResult();
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeModel *m_model;
  GtkTreeIter *m_iter;
  int m_n;
};

// ------ GtkTreePath
class GtkTreePathNewFromStringRequest : public GtkRequest_ReturnBoxed {
public:
  GtkTreePathNewFromStringRequest(GtkTreeModel *model, const char *pathstring, int size) :
    GtkRequest_ReturnBoxed("GtkTreePath"),
    m_model(model)
  {
    safe_string_copy(m_pathstring, pathstring, size, SMALL_STRING_SIZE);
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    GtkTreePath *path = gtk_tree_path_new_from_string(m_pathstring);
    setBoxedResult(gtk_tree_path_get_type(), path);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeModel *m_model;
  char m_pathstring[SMALL_STRING_SIZE];
};

#define MAX_TREE_PATH_INDICES 128
typedef gint TreePathIndexArray[MAX_TREE_PATH_INDICES];

class GtkTreePathGetIndicesRequest : public GtkRequest {
public:
  GtkTreePathGetIndicesRequest(GtkTreePath *path) :
    m_path(path)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;

    m_n = gtk_tree_path_get_depth(m_path);
    gint *indices = gtk_tree_path_get_indices(m_path);
    
    for (int i = 0; i < m_n; ++i) {
      m_indices[i] = indices[i];
    }
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    PyrObject *array = newPyrArray(g->gc, m_n, 0, false);
    PyrSlot *slots = array->slots;
  
    for (int i = 0; i < m_n; ++i)
      SetInt(slots + i, m_indices[i]);
  
    array->size = m_n;
  
    SetObject(resultSlot, array);
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreePath *m_path;
  TreePathIndexArray m_indices;
  int m_n;
};

class GtkTreePathAsStringRequest : public GtkRequest {
public:
  GtkTreePathAsStringRequest(GtkTreePath *path) :
    m_path(path)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;

    gchar *s = gtk_tree_path_to_string(m_path);
    m_size = safe_string_copy(m_s, s, -1, SMALL_STRING_SIZE);
    g_free(s);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    PyrString *strobj = newPyrStringN(g->gc, m_size, 0, true);
    memcpy(strobj->s, m_s, m_size);
    SetObject(resultSlot, strobj);
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreePath *m_path;
  char m_s[SMALL_STRING_SIZE];
  int m_size;
};

class GtkTreeModelGetValueRequest : public GtkRequest {
public:
  GtkTreeModelGetValueRequest(GtkTreeModel *model, GtkTreeIter *iter, int column) :
    m_model(model), m_iter(iter), m_column(column)
  {
  }

  ~GtkTreeModelGetValueRequest() {
    g_value_unset(&m_value);
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    memset(&m_value, 0, sizeof(m_value));
    gtk_tree_model_get_value(m_model, m_iter, m_column, &m_value);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    PyrSlot slot;
    if (!GValue_to_slot(&m_value, &slot)) {
      resultSlot->ucopy = slot.ucopy;
    } else {
      SetNil(resultSlot);
    } 
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeModel *m_model;
  GtkTreeIter *m_iter;
  int m_column;
  GValue m_value;
};

// ------ GtkTreeRow

class GtkTreeRowReferenceNewRequest : public GtkRequest {
public:
  GtkTreeRowReferenceNewRequest(GtkTreeModel *model, GtkTreePath *path) :
    m_model(model), m_path(path)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_reference = gtk_tree_row_reference_new(m_model, m_path);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_reference));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeModel *m_model;
  GtkTreePath *m_path;
  GtkTreeRowReference *m_reference;
};

class GtkTreeRowReferenceGetPathRequest : public GtkRequest_ReturnBoxed {
public:
  GtkTreeRowReferenceGetPathRequest(GtkTreeRowReference *reference) :
    GtkRequest_ReturnBoxed("GtkTreePath"),
    m_reference(reference)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    GtkTreePath *path = gtk_tree_row_reference_get_path(m_reference);
    if (path != 0)
      setBoxedResult(gtk_tree_path_get_type(), path);
    else
      nilBoxedResult();
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeRowReference *m_reference;
};

class GtkTreeRowReferenceValidRequest : public GtkRequest_ReturnBoolean {
public:
  GtkTreeRowReferenceValidRequest(GtkTreeRowReference *reference) :
    m_reference(reference)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = gtk_tree_row_reference_valid(m_reference);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeRowReference *m_reference;
};

class GtkTreeRowReferenceFreeRequest : public GtkRequest {
public:
  GtkTreeRowReferenceFreeRequest(GtkTreeRowReference *reference) :
    m_reference(reference)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_tree_row_reference_free(m_reference);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeRowReference *m_reference;
};

class GtkTreeRowReferenceCopyRequest : public GtkRequest {
public:
  GtkTreeRowReferenceCopyRequest(GtkTreeRowReference *reference) :
    m_reference(reference)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_copy = gtk_tree_row_reference_copy(m_reference);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_copy));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeRowReference *m_reference;
  GtkTreeRowReference *m_copy;
};

// ------ GtkTreeView

class GtkTreeViewNewRequest : public GtkRequest {
public:
  GtkTreeViewNewRequest(void) :
    m_treeview(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_treeview = (GtkTreeView *)gtk_tree_view_new();
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_treeview));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeView *m_treeview;
};

class GtkTreeViewNewWithModelRequest : public GtkRequest {
public:
  GtkTreeViewNewWithModelRequest(GtkTreeModel *model) :
    m_treeview(0), m_model(model)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_treeview = (GtkTreeView *)gtk_tree_view_new_with_model(m_model);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_treeview));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeView *m_treeview;
  GtkTreeModel *m_model;
};

class GtkTreeViewSetRulesHintRequest : public GtkRequest {
public:
  GtkTreeViewSetRulesHintRequest(GtkTreeView *treeview, bool value) :
    m_treeview(treeview), m_value(value)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_tree_view_set_rules_hint(m_treeview, m_value);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeView *m_treeview;
  bool m_value;
};

class GtkTreeViewSetSearchColumnRequest : public GtkRequest {
public:
  GtkTreeViewSetSearchColumnRequest(GtkTreeView *treeview, int n) :
    m_treeview(treeview), m_n(n)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_tree_view_set_search_column(m_treeview, m_n);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeView *m_treeview;
  bool m_n;
};

class GtkTreeViewAppendColumnRequest : public GtkRequest {
public:
  GtkTreeViewAppendColumnRequest(GtkTreeView *treeview, GtkTreeViewColumn *column) :
    m_treeview(treeview), m_column(column)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_tree_view_append_column(m_treeview, m_column);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeView *m_treeview;
  GtkTreeViewColumn *m_column;
};

class GtkTreeViewExpandRowRequest : public GtkRequest {
public:
  GtkTreeViewExpandRowRequest(GtkTreeView *treeview, GtkTreePath *path, bool expandAll) :
    m_treeview(treeview), m_path(path), m_expandAll(expandAll)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_tree_view_expand_row(m_treeview, m_path, m_expandAll);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeView *m_treeview;
  GtkTreePath *m_path;
  bool m_expandAll;
};

class GtkTreeViewExpandAllRequest : public GtkRequest {
public:
  GtkTreeViewExpandAllRequest(GtkTreeView *treeview) :
    m_treeview(treeview)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_tree_view_expand_all(m_treeview);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeView *m_treeview;
};

class GtkTreeViewGetSelectionRequest : public GtkRequest {
public:
  GtkTreeViewGetSelectionRequest(GtkTreeView *treeview) :
    m_treeview(treeview)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_selection = gtk_tree_view_get_selection(m_treeview);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_selection));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeView *m_treeview;
  GtkTreeSelection *m_selection;
};

class GtkTreeViewGetModelRequest : public GtkRequest {
public:
  GtkTreeViewGetModelRequest(GtkTreeView *treeview) :
    m_treeview(treeview)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_model = gtk_tree_view_get_model(m_treeview);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_model));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeView *m_treeview;
  GtkTreeModel *m_model;
};

class GtkTreeViewScrollToCellRequest : public GtkRequest {
public:
  GtkTreeViewScrollToCellRequest(GtkTreeView *treeview, GtkTreePath *path, GtkTreeViewColumn *column,
				 bool useAlign, double rowAlign, double colAlign) :
    m_treeview(treeview), m_path(path), m_column(column), m_useAlign(useAlign),
    m_rowAlign(rowAlign), m_colAlign(colAlign)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_tree_view_scroll_to_cell(m_treeview, m_path, m_column, m_useAlign, m_rowAlign, m_colAlign);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeView *m_treeview;
  GtkTreePath *m_path;
  GtkTreeViewColumn *m_column;
  bool m_useAlign;
  double m_rowAlign;
  double m_colAlign;
};

class GtkTreeViewSetReorderableRequest : public GtkRequest {
public:
  GtkTreeViewSetReorderableRequest(GtkTreeView *treeview, bool value) :
    m_treeview(treeview), m_value(value)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_tree_view_set_reorderable(m_treeview, m_value);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeView *m_treeview;
  bool m_value;
};

class GtkTreeViewGetReorderableRequest : public GtkRequest_ReturnBoolean {
public:
  GtkTreeViewGetReorderableRequest(GtkTreeView *treeview) :
    m_treeview(treeview)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = gtk_tree_view_get_reorderable(m_treeview);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeView *m_treeview;
};

#define DRAG_TARGETS_MAX 8
typedef char DragTargetNameArray[DRAG_TARGETS_MAX][SMALL_STRING_SIZE];
typedef int DragTargetFlagArray[DRAG_TARGETS_MAX];
typedef int DragTargetInfoArray[DRAG_TARGETS_MAX];

class GtkTreeViewEnableModelDragSourceRequest : public GtkRequest_ReturnBoolean {
public:
  GtkTreeViewEnableModelDragSourceRequest(GtkTreeView *treeview, 
					  GdkModifierType modifier, int n,
					  DragTargetNameArray &nameArray,
					  DragTargetFlagArray &flagArray,
					  DragTargetInfoArray &infoArray,
					  GdkDragAction action) :
    m_treeview(treeview), m_modifier(modifier), m_n(n), m_action(action)
  {
    for (int i = 0; i < n; ++i) {
      safe_string_copy(m_nameArray[i], nameArray[i], -1, MEDIUM_STRING_SIZE);
      m_flagArray[i] = flagArray[i];
      m_infoArray[i] = infoArray[i];
    }
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;

    GtkTargetEntry *targets = new GtkTargetEntry[m_n];

    for (int i = 0; i < m_n; ++i) {
      targets[i].target = &m_nameArray[i][0];
      targets[i].flags = (guint)m_flagArray[i];
      targets[i].info = (guint)m_infoArray[i];
    }

    gtk_tree_view_enable_model_drag_source(m_treeview, m_modifier, targets, m_n, m_action);

    delete [] targets;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeView *m_treeview;
  GdkModifierType m_modifier;
  int m_n;
  DragTargetNameArray m_nameArray;
  DragTargetFlagArray m_flagArray;
  DragTargetInfoArray m_infoArray;
  GdkDragAction m_action;
};

class GtkTreeViewEnableModelDragDestRequest : public GtkRequest {
public:
  GtkTreeViewEnableModelDragDestRequest(GtkTreeView *treeview, 
					int n,
					DragTargetNameArray &nameArray,
					DragTargetFlagArray &flagArray,
					DragTargetInfoArray &infoArray,
					GdkDragAction action) :
    m_treeview(treeview), m_n(n), m_action(action)
  {
    for (int i = 0; i < n; ++i) {
      safe_string_copy(m_nameArray[i], nameArray[i], -1, MEDIUM_STRING_SIZE);
      m_flagArray[i] = (guint)flagArray[i];
      m_infoArray[i] = (guint)infoArray[i];
    }
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;

    GtkTargetEntry *targets = new GtkTargetEntry[m_n];

    for (int i = 0; i < m_n; ++i) {
      targets[i].target = &m_nameArray[i][0];
      targets[i].flags = m_flagArray[i];
      targets[i].info = m_infoArray[i];
    }

    gtk_tree_view_enable_model_drag_dest(m_treeview, targets, m_n, m_action);

    delete [] targets;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeView *m_treeview;
  GdkModifierType m_modifier;
  int m_n;
  DragTargetNameArray m_nameArray;
  DragTargetFlagArray m_flagArray;
  DragTargetInfoArray m_infoArray;
  GdkDragAction m_action;
};

class GtkTreeViewGetDestRowAtPosRequest : public GtkRequest {
public:
  GtkTreeViewGetDestRowAtPosRequest(GtkTreeView *treeview, gint x, gint y) :
    m_treeview(treeview), m_x(x), m_y(y)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_succeded = gtk_tree_view_get_dest_row_at_pos(m_treeview, m_x, m_y, &m_path, &m_dropPosition);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    if (m_succeded) {
      PyrObject *array = newPyrArray(g->gc, 2, 0, false);
      PyrSlot *slots = array->slots;
    
      if (m_path == 0) {
	SetNil(slots + 0);
      } else {
	PyrClass *pathclassobj = getsym("GtkTreePath")->u.classobj;
	PyrObject *pathObject = instantiateObject(g->gc, pathclassobj, 0, true, true);
	
	SetInt(pathObject->slots + 0, (int)m_path);
	SetInt(pathObject->slots + 1, gtk_tree_path_get_type());
	
	SetObject(slots + 0, pathObject);
      } 

      SetEnumSlot(slots + 1, GTK_TYPE_TREE_VIEW_DROP_POSITION, m_dropPosition);
      
      array->size = 2;
      
      SetObject(resultSlot, array);
    } else {
      SetNil(resultSlot);
    }
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeView *m_treeview;
  gint m_x, m_y;
  GtkTreePath *m_path;
  GtkTreeViewDropPosition m_dropPosition;
  bool m_succeded;
};

// -------- GtkTreeViewColumn

class GtkTreeViewColumnNewRequest : public GtkRequest {
public:
  GtkTreeViewColumnNewRequest(void) :
    m_column(0)
  {
  }
    
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_column = (GtkTreeViewColumn *)gtk_tree_view_column_new();
  }
    
  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_column));
    return 1;
  }
    
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeViewColumn *m_column;
};

class GtkTreeViewColumnSetTitleRequest : public GtkRequest {
public:
  GtkTreeViewColumnSetTitleRequest(GtkTreeViewColumn *column, const char *s, int size) :
    m_column(column)
  {
    safe_string_copy(m_s, s, size, MEDIUM_STRING_SIZE);
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_tree_view_column_set_title(m_column, m_s);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeViewColumn *m_column;
  char m_s[MEDIUM_STRING_SIZE];
};

typedef char AttributeNameArray[16][SMALL_STRING_SIZE];
typedef int AttributeColumnArray[16];

class GtkTreeViewColumnNewWithAttributesRequest : public GtkRequest {
public:
  GtkTreeViewColumnNewWithAttributesRequest(const char *title, int titleSize,
					    GtkCellRenderer *renderer,
					    int nAttributes,
					    AttributeNameArray &anames, AttributeColumnArray &acols) :
    m_column(0), m_renderer(renderer)
  {
    safe_string_copy(m_title, title, titleSize, MEDIUM_STRING_SIZE);

    m_nAttributes = std::min(16, nAttributes);

    for (int i = 0; i < m_nAttributes; ++i) {
      safe_string_copy(m_anames[i], anames[i], -1, SMALL_STRING_SIZE);
      m_acols[i] = acols[i];
    }
  }
    
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_column = gtk_tree_view_column_new();
    gtk_tree_view_column_set_title(m_column, m_title);

    gtk_tree_view_column_pack_start(m_column, m_renderer, true);

    for (int i = 0; i < m_nAttributes; ++i) {
      gtk_tree_view_column_add_attribute(m_column, m_renderer, m_anames[i], m_acols[i]);
    }
  }
    
  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_column));
    return 1;
  }
    
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeViewColumn *m_column;
  char m_title[MEDIUM_STRING_SIZE];
  GtkCellRenderer *m_renderer;
  int m_nAttributes;
  AttributeNameArray m_anames;
  AttributeColumnArray m_acols;
};

class GtkTreeViewColumnSetClickableRequest : public GtkRequest {
public:
  GtkTreeViewColumnSetClickableRequest(GtkTreeViewColumn *column, bool value) :
    m_column(column), m_value(value)
  {
  }
  
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_tree_view_column_set_clickable(m_column, m_value);
  }
    
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeViewColumn *m_column;
  bool m_value;
};

class GtkTreeViewColumnSetSizingRequest : public GtkRequest {
public:
  GtkTreeViewColumnSetSizingRequest(GtkTreeViewColumn *column, GtkTreeViewColumnSizing sizing) :
    m_column(column), m_sizing(sizing)
  {
  }
    
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_tree_view_column_set_sizing(m_column, m_sizing);
  }
    
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeViewColumn *m_column;
  GtkTreeViewColumnSizing m_sizing;
};

class GtkTreeViewColumnSetFixedWidthRequest : public GtkRequest {
public:
  GtkTreeViewColumnSetFixedWidthRequest(GtkTreeViewColumn *column, int value) :
    m_column(column), m_value(value)
  {
  }
    
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_tree_view_column_set_fixed_width(m_column, m_value);
  }
    
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeViewColumn *m_column;
  int m_value;
};

// ---------- GtkTreeStore

class GtkTreeStoreNewRequest : public GtkRequest {
public:
  GtkTreeStoreNewRequest(int nTypes, GType *types) :
    m_store(0)
  {
    m_nTypes = std::min(32, nTypes);
    for (int i = 0; i < m_nTypes; ++i)
      m_types[i] = types[i];
  }
    
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_store = (GtkTreeStore *)gtk_tree_store_newv(m_nTypes, m_types);
  }
    
  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_store));
    return 1;
  }
    
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeStore *m_store;
  int m_nTypes;
  GType m_types[32];
};

class GtkTreeStoreSetValueRequest : public GtkRequest {
public:
  GtkTreeStoreSetValueRequest(GtkTreeStore *store, GtkTreeIter *iter, int column, GValue *value) :
    m_store(store), m_iter(iter), m_column(column)
  {
    memset(&m_value, 0, sizeof(m_value));
    g_value_init(&m_value, G_VALUE_TYPE(value));
    g_value_copy(value, &m_value);
  }
    
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_tree_store_set_value(m_store, m_iter, m_column, &m_value);
    g_value_unset(&m_value);
  }
    
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeStore *m_store;
  GtkTreeIter *m_iter;
  int m_column;
  GValue m_value;
};

class GtkTreeStoreClearRequest : public GtkRequest {
public:
  GtkTreeStoreClearRequest(GtkTreeStore *store) :
    m_store(store)
  {
  }
    
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_tree_store_clear(m_store);
  }
    
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeStore *m_store;
};

class GtkTreeStoreRemoveRequest : public GtkRequest_ReturnBoolean {
public:
  GtkTreeStoreRemoveRequest(GtkTreeStore *store, GtkTreeIter *iter) :
    m_store(store), m_iter(iter)
  {
  }
    
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = gtk_tree_store_remove(m_store, m_iter);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeStore *m_store;
  GtkTreeIter *m_iter;
};

class GtkTreeStoreAppendRequest : public GtkRequest_ReturnBoxed {
public:
  GtkTreeStoreAppendRequest(GtkTreeStore *store, GtkTreeIter *parent) :
    GtkRequest_ReturnBoxed("GtkTreeIter"),
    m_store(store), m_parent(parent)
  {
  }
    
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    GtkTreeIter iter;
    gtk_tree_store_append(m_store, &iter, m_parent);
    copyBoxedResult(gtk_tree_iter_get_type(), &iter);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeStore *m_store;
  GtkTreeIter *m_parent;
};

class GtkTreeStoreInsertBeforeRequest : public GtkRequest_ReturnBoxed {
public:
  GtkTreeStoreInsertBeforeRequest(GtkTreeStore *store, GtkTreeIter *parent, GtkTreeIter *sibling) :
    GtkRequest_ReturnBoxed("GtkTreeIter"),
    m_store(store), m_parent(parent), m_sibling(sibling)
  {
  }
    
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    GtkTreeIter iter;
    gtk_tree_store_insert_before(m_store, &iter, m_parent, m_sibling);
    copyBoxedResult(gtk_tree_iter_get_type(), &iter);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeStore *m_store;
  GtkTreeIter *m_parent;
  GtkTreeIter *m_sibling;
};

class GtkTreeStoreInsertAfterRequest : public GtkRequest_ReturnBoxed {
public:
  GtkTreeStoreInsertAfterRequest(GtkTreeStore *store, GtkTreeIter *parent, GtkTreeIter *sibling) :
    GtkRequest_ReturnBoxed("GtkTreeIter"),
    m_store(store), m_parent(parent), m_sibling(sibling)
  {
  }
    
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    GtkTreeIter iter;
    gtk_tree_store_insert_after(m_store, &iter, m_parent, m_sibling);
    copyBoxedResult(gtk_tree_iter_get_type(), &iter);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeStore *m_store;
  GtkTreeIter *m_parent;
  GtkTreeIter *m_sibling;
};

// ---------- GtkListStore

class GtkListStoreNewRequest : public GtkRequest {
public:
  GtkListStoreNewRequest(int nTypes, GType *types) :
    m_store(0)
  {
    m_nTypes = std::min(32, nTypes);
    for (int i = 0; i < nTypes; ++i)
      m_types[i] = types[i];
  }
    
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_store = (GtkListStore *)gtk_list_store_newv(m_nTypes, m_types);
  }
    
  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_store));
    return 1;
  }
    
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkListStore *m_store;
  int m_nTypes;
  GType m_types[32];
};

class GtkListStoreSetValueRequest : public GtkRequest {
public:
  GtkListStoreSetValueRequest(GtkListStore *store, GtkTreeIter *iter, int column, GValue *value) :
    m_store(store), m_iter(iter), m_column(column)
  {
    memset(&m_value, 0, sizeof(m_value));
    g_value_init(&m_value, G_VALUE_TYPE(value));
    g_value_copy(value, &m_value);
  }
    
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_list_store_set_value(m_store, m_iter, m_column, &m_value);
    g_value_unset(&m_value);
  }
    
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkListStore *m_store;
  GtkTreeIter *m_iter;
  int m_column;
  GValue m_value;
};

class GtkListStoreClearRequest : public GtkRequest {
public:
  GtkListStoreClearRequest(GtkListStore *store) :
    m_store(store)
  {
  }
    
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_list_store_clear(m_store);
  }
    
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkListStore *m_store;
};

class GtkListStoreRemoveRequest : public GtkRequest_ReturnBoolean {
public:
  GtkListStoreRemoveRequest(GtkListStore *store, GtkTreeIter *iter) :
    m_store(store), m_iter(iter)
  {
  }
    
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = gtk_list_store_remove(m_store, m_iter);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkListStore *m_store;
  GtkTreeIter *m_iter;
};

class GtkListStoreAppendRequest : public GtkRequest_ReturnBoxed {
public:
  GtkListStoreAppendRequest(GtkListStore *store) :
    GtkRequest_ReturnBoxed("GtkTreeIter"),
    m_store(store)
  {
  }
    
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    GtkTreeIter iter;
    gtk_list_store_append(m_store, &iter);
    copyBoxedResult(gtk_tree_iter_get_type(), &iter);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkListStore *m_store;
};

// ------ GtkCellRendererToggle

class GtkCellRendererToggleNewRequest : public GtkRequest {
public:
  GtkCellRendererToggleNewRequest(void) :
    m_renderer(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_renderer = (GtkCellRendererToggleNewRequest *)gtk_cell_renderer_toggle_new();
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_renderer));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkCellRendererToggleNewRequest *m_renderer;
};

class GtkCellRendererTextNewRequest : public GtkRequest {
public:
  GtkCellRendererTextNewRequest(void) :
    m_renderer(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_renderer = (GtkCellRendererTextNewRequest *)gtk_cell_renderer_text_new();
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_renderer));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkCellRendererTextNewRequest *m_renderer;
};

// -------- GtkTreeSelection

class GtkTreeSelectionGetSelectedRequest : public GtkRequest_ReturnBoxed {
public:
  GtkTreeSelectionGetSelectedRequest(GtkTreeSelection *selection) :
    GtkRequest_ReturnBoxed("GtkTreeIter"),
    m_selection(selection)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    GtkTreeIter iter;
    if (gtk_tree_selection_get_selected(m_selection, 0, &iter) && iter.stamp != 0)
      copyBoxedResult(gtk_tree_iter_get_type(), &iter);
    else
      nilBoxedResult();
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeSelection *m_selection;
};

#define MAX_SELECTED_ITEMS 32
typedef GtkTreePath *TreePathArray[MAX_SELECTED_ITEMS];

class GtkTreeSelectionGetSelectedRowsRequest : public GtkRequest {
public:
  GtkTreeSelectionGetSelectedRowsRequest(GtkTreeSelection *selection) :
    m_selection(selection)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    GList *list, *item;
    int i;

    list = gtk_tree_selection_get_selected_rows(m_selection, 0);

    for (item = list, i = 0; item != 0 && i < MAX_SELECTED_ITEMS; item = g_list_next(item), ++i) {
      GtkTreePath *path = (GtkTreePath *)item->data;
      m_iterArray[i] = path;
    }
    m_n = i;
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    PyrObject *array = newPyrArray(g->gc, m_n, 0, false);
    PyrSlot *slots = array->slots;

    for (int i = 0; i < m_n; ++i) {
      GtkTreePath *path = m_iterArray[i];

      if (path == 0) {
	SetNil(slots + i);
      } else {
	PyrClass *wrapperclassobj = getsym("GtkTreePath")->u.classobj;
	PyrObject *wrapperObject = instantiateObject(g->gc, wrapperclassobj, 0, true, true);
	
	SetInt(wrapperObject->slots + 0, (int)path);
	SetInt(wrapperObject->slots + 1, gtk_tree_path_get_type());
    
	SetObject(slots + i, wrapperObject);
      }
    }

    array->size = m_n;

    SetObject(resultSlot, array);

    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeSelection *m_selection;
  TreePathArray m_iterArray;
  int m_n;
};


class GtkTreeSelectionSelectIterRequest : public GtkRequest {
public:
  GtkTreeSelectionSelectIterRequest(GtkTreeSelection *selection, GtkTreeIter *iter) :
    m_selection(selection), m_iter(iter)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_tree_selection_select_iter(m_selection, m_iter);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeSelection *m_selection;
  GtkTreeIter *m_iter;
};

class GtkTreeSelectionSelectAllRequest : public GtkRequest {
public:
  GtkTreeSelectionSelectAllRequest(GtkTreeSelection *selection) :
    m_selection(selection)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_tree_selection_select_all(m_selection);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeSelection *m_selection;
};

class GtkTreeSelectionUnselectAllRequest : public GtkRequest {
public:
  GtkTreeSelectionUnselectAllRequest(GtkTreeSelection *selection) :
    m_selection(selection)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_tree_selection_unselect_all(m_selection);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeSelection *m_selection;
};

class GtkTreeSelectionSetModeRequest : public GtkRequest {
public:
  GtkTreeSelectionSetModeRequest(GtkTreeSelection *selection, GtkSelectionMode mode) :
    m_selection(selection), m_mode(mode)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_tree_selection_set_mode(m_selection, m_mode);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkTreeSelection *m_selection;
  GtkSelectionMode m_mode;
};

// ------ GtkHScale

class GtkHScaleNewRequest : public GtkRequest {
public:
  GtkHScaleNewRequest(GtkAdjustment *adjustment) :
    m_adjustment(adjustment),
    m_scale(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_scale = (GtkHScale *)gtk_hscale_new(m_adjustment);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_scale));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkAdjustment *m_adjustment;
  GtkHScale *m_scale;
};

class GtkHScaleNewWithRangeRequest : public GtkRequest {
public:
  GtkHScaleNewWithRangeRequest(double min, double max, double step) :
    m_min(min), m_max(max), m_step(step),
    m_scale(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_scale = (GtkHScale *)gtk_hscale_new_with_range(m_min, m_max, m_step);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_scale));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  double m_min;
  double m_max;
  double m_step;
  GtkHScale *m_scale;
};

// ------ GtkVScale

class GtkVScaleNewRequest : public GtkRequest {
public:
  GtkVScaleNewRequest(GtkAdjustment *adjustment) :
    m_adjustment(adjustment),
    m_scale(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_scale = (GtkVScale *)gtk_vscale_new(m_adjustment);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_scale));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkAdjustment *m_adjustment;
  GtkVScale *m_scale;
};

class GtkVScaleNewWithRangeRequest : public GtkRequest {
public:
  GtkVScaleNewWithRangeRequest(double min, double max, double step) :
    m_min(min), m_max(max), m_step(step),
    m_scale(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_scale = (GtkVScale *)gtk_vscale_new_with_range(m_min, m_max, m_step);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_scale));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  double m_min;
  double m_max;
  double m_step;
  GtkVScale *m_scale;
};

// ------ GtkScale

class GtkScaleSetDigitsRequest : public GtkRequest {
public:
  GtkScaleSetDigitsRequest(GtkScale *scale, int n) :
    m_scale(scale), m_n(n)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_scale_set_digits(m_scale, m_n);
  }
  
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkScale *m_scale;
  int m_n;
};

class GtkScaleGetDigitsRequest : public GtkRequest {
public:
  GtkScaleGetDigitsRequest(GtkScale *scale) :
    m_scale(scale), m_n(-1)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_n = gtk_scale_get_digits(m_scale);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_n));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkScale *m_scale;
  int m_n;
};

class GtkScaleSetDrawValueRequest : public GtkRequest {
public:
  GtkScaleSetDrawValueRequest(GtkScale *scale, bool value) :
    m_scale(scale), m_value(value)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_scale_set_draw_value(m_scale, m_value);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkScale *m_scale;
  bool m_value;
};

class GtkScaleGetDrawValueRequest : public GtkRequest_ReturnBoolean {
public:
  GtkScaleGetDrawValueRequest(GtkScale *scale) :
    m_scale(scale)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = gtk_scale_get_draw_value(m_scale);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkScale *m_scale;
};

class GtkScaleSetValuePositionRequest : public GtkRequest {
public:
  GtkScaleSetValuePositionRequest(GtkScale *scale, GtkPositionType position) :
    m_scale(scale), m_position(position)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_scale_set_value_pos(m_scale, m_position);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkScale *m_scale;
  GtkPositionType m_position;
};

class GtkScaleGetValuePositionRequest : public GtkRequest {
public:
  GtkScaleGetValuePositionRequest(GtkScale *scale) :
    m_scale(scale)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_position = gtk_scale_get_value_pos(m_scale);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetEnumSlot(resultSlot, GTK_TYPE_POSITION_TYPE, m_position);
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkScale *m_scale;
  GtkPositionType m_position;
};

// ------ GtkRange

class GtkRangeSetUpdatePolicyRequest : public GtkRequest {
public:
  GtkRangeSetUpdatePolicyRequest(GtkRange *range, GtkUpdateType policy) :
    m_range(range), m_policy(policy)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_range_set_update_policy(m_range, m_policy);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkRange *m_range;
  GtkUpdateType m_policy;
};

class GtkRangeGetUpdatePolicyRequest : public GtkRequest {
public:
  GtkRangeGetUpdatePolicyRequest(GtkRange *range) :
    m_range(range)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_policy = gtk_range_get_update_policy(m_range);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetEnumSlot(resultSlot, GTK_TYPE_UPDATE_TYPE, m_policy);
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkRange *m_range;
  GtkUpdateType m_policy;
};

class GtkRangeSetValueRequest : public GtkRequest {
public:
  GtkRangeSetValueRequest(GtkRange *range, double value) :
    m_range(range), m_value(value)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_range_set_value(m_range, m_value);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkRange *m_range;
  double m_value;
};

class GtkRangeGetValueRequest : public GtkRequest {
public:
  GtkRangeGetValueRequest(GtkRange *range) :
    m_range(range)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = gtk_range_get_value(m_range);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetFloat(resultSlot, m_value);
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkRange *m_range;
  double m_value;
};

class GtkRangeSetAdjustmentRequest : public GtkRequest {
public:
  GtkRangeSetAdjustmentRequest(GtkRange *range, GtkAdjustment *adjustment) :
    m_range(range), m_adjustment(adjustment)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_range_set_adjustment(m_range, m_adjustment);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkRange *m_range;
  GtkAdjustment *m_adjustment;
};

class GtkRangeGetAdjustmentRequest : public GtkRequest {
public:
  GtkRangeGetAdjustmentRequest(GtkRange *range) :
    m_range(range), m_adjustment(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_adjustment = gtk_range_get_adjustment(m_range);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_adjustment));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkRange *m_range;
  GtkAdjustment *m_adjustment;
};

class GtkRangeSetInvertedRequest : public GtkRequest {
public:
  GtkRangeSetInvertedRequest(GtkRange *range, bool value) :
    m_range(range), m_value(value)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_range_set_inverted(m_range, m_value);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkRange *m_range;
  bool m_value;
};

class GtkRangeGetInvertedRequest : public GtkRequest_ReturnBoolean {
public:
  GtkRangeGetInvertedRequest(GtkRange *range) :
    m_range(range)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = gtk_range_get_inverted(m_range);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkRange *m_range;
};

class GtkRangeSetIncrementsRequest : public GtkRequest {
public:
  GtkRangeSetIncrementsRequest(GtkRange *range, double step, double page) :
    m_range(range), m_step(step), m_page(page)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_range_set_increments(m_range, m_step, m_page);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkRange *m_range;
  double m_step, m_page;
};

class GtkRangeSetRangeRequest : public GtkRequest {
public:
  GtkRangeSetRangeRequest(GtkRange *range, double min, double max) :
    m_range(range), m_min(min), m_max(max)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_range_set_range(m_range, m_min, m_max);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkRange *m_range;
  double m_min, m_max;
};

// ------ PangoContext

class PangoContextNewRequest : public GtkRequest {
public:
  PangoContextNewRequest(void) :
    m_context(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_context = 0; //pango_context_new();
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_context));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  PangoContext *m_context;
};

// ------ PangoLayout

class PangoLayoutNewRequest : public GtkRequest {
public:
  PangoLayoutNewRequest(PangoContext *context) :
    m_layout(0), m_context(context)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_layout = pango_layout_new(m_context);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_layout));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  PangoLayout *m_layout;
  PangoContext *m_context;
};

class PangoLayoutSetTextRequest : public GtkRequest {
public:
  PangoLayoutSetTextRequest(PangoLayout *layout, const char *s, int size) :
    m_layout(layout), m_size(size)
  {
    safe_string_copy(m_s, s, size, LARGE_STRING_SIZE);
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    pango_layout_set_text(m_layout, m_s, m_size);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  PangoLayout *m_layout;
  char m_s[LARGE_STRING_SIZE];
  int m_size;
};

class PangoLayoutSetMarkupRequest : public GtkRequest {
public:
  PangoLayoutSetMarkupRequest(PangoLayout *layout, const char *s, int size) :
    m_layout(layout), m_size(size)
  {
    safe_string_copy(m_s, s, size, LARGE_STRING_SIZE);
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    pango_layout_set_markup(m_layout, m_s, m_size);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  PangoLayout *m_layout;
  char m_s[LARGE_STRING_SIZE];
  int m_size;
};

class PangoLayoutGetPixelSizeRequest : public GtkRequest {
public:
  PangoLayoutGetPixelSizeRequest(PangoLayout *layout) :
    m_layout(layout), m_w(-1), m_h(-1)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    pango_layout_get_pixel_size(m_layout, &m_w, &m_h);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    PyrObject *array = newPyrArray(g->gc, 2, 0, false);
    PyrSlot *slots = array->slots;

    SetInt(slots + 0, m_w);
    SetInt(slots + 1, m_h);

    array->size = 2;

    SetObject(resultSlot, array);

    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  PangoLayout *m_layout;
  int m_w, m_h;
};

// ------ GtkEditable

class GtkEditableSetEditableRequest : public GtkRequest {
public:
  GtkEditableSetEditableRequest(GtkEditable *editable, bool value) :
    m_editable(editable), m_value(value)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_editable_set_editable(m_editable, m_value);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkEditable *m_editable;
  bool m_value;
};

class GtkEditableGetEditableRequest : public GtkRequest_ReturnBoolean {
public:
  GtkEditableGetEditableRequest(GtkEditable *editable) :
    m_editable(editable)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = gtk_editable_get_editable(m_editable);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkEditable *m_editable;
};

// ------ GtkSpinButton

class GtkSpinButtonNewRequest : public GtkRequest {
public:
  GtkSpinButtonNewRequest(GtkAdjustment *adjustment, double climbRate, int digits) :
    m_spinbutton(0), m_adjustment(adjustment), m_climbRate(climbRate), m_digits(digits)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_spinbutton = (GtkSpinButton *)gtk_spin_button_new(m_adjustment, m_climbRate, m_digits);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_spinbutton));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSpinButton *m_spinbutton;
  GtkAdjustment *m_adjustment;
  double m_climbRate;
  int m_digits;
};

class GtkSpinButtonNewWithRangeRequest : public GtkRequest {
public:
  GtkSpinButtonNewWithRangeRequest(double min, double max, double step) :
    m_spinbutton(0), m_min(min), m_max(max), m_step(step)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_spinbutton = (GtkSpinButton *)gtk_spin_button_new_with_range(m_min, m_max, m_step);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_spinbutton));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSpinButton *m_spinbutton;
  double m_min, m_max, m_step;
};

class GtkSpinButtonSetAdjustmentRequest : public GtkRequest {
public:
  GtkSpinButtonSetAdjustmentRequest(GtkSpinButton *spinbutton, GtkAdjustment *adjustment) :
    m_spinbutton(spinbutton), m_adjustment(adjustment)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_spin_button_set_adjustment(m_spinbutton, m_adjustment);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSpinButton *m_spinbutton;
  GtkAdjustment *m_adjustment;
};

class GtkSpinButtonGetAdjustmentRequest : public GtkRequest {
public:
  GtkSpinButtonGetAdjustmentRequest(GtkSpinButton *spinbutton) :
    m_spinbutton(spinbutton), m_adjustment(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_adjustment = gtk_spin_button_get_adjustment(m_spinbutton);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_adjustment));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSpinButton *m_spinbutton;
  GtkAdjustment *m_adjustment;
};

class GtkSpinButtonSetDigitsRequest : public GtkRequest {
public:
  GtkSpinButtonSetDigitsRequest(GtkSpinButton *spinbutton, int digits) :
    m_spinbutton(spinbutton), m_digits(digits)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_spin_button_set_digits(m_spinbutton, m_digits);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSpinButton *m_spinbutton;
  int m_digits;
};

class GtkSpinButtonGetDigitsRequest : public GtkRequest {
public:
  GtkSpinButtonGetDigitsRequest(GtkSpinButton *spinbutton) :
    m_spinbutton(spinbutton)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_digits = gtk_spin_button_get_digits(m_spinbutton);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, m_digits);
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSpinButton *m_spinbutton;
  int m_digits;
};

class GtkSpinButtonSetIncrementsRequest : public GtkRequest {
public:
  GtkSpinButtonSetIncrementsRequest(GtkSpinButton *spinbutton, double step, double page) :
    m_spinbutton(spinbutton), m_step(step), m_page(page)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_spin_button_set_increments(m_spinbutton, m_step, m_page);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSpinButton *m_spinbutton;
  double m_step, m_page;
};

class GtkSpinButtonGetIncrementsRequest : public GtkRequest {
public:
  GtkSpinButtonGetIncrementsRequest(GtkSpinButton *spinbutton) :
    m_spinbutton(spinbutton)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_spin_button_get_increments(m_spinbutton, &m_step, &m_page);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    PyrObject *array = newPyrArray(g->gc, 2, 0, false);
    PyrSlot *slots = array->slots;

    SetFloat(slots + 0, m_step);
    SetFloat(slots + 1, m_page);

    array->size = 2;

    SetObject(resultSlot, array);
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSpinButton *m_spinbutton;
  double m_step, m_page;
};

class GtkSpinButtonSetRangeRequest : public GtkRequest {
public:
  GtkSpinButtonSetRangeRequest(GtkSpinButton *spinbutton, double min, double max) :
    m_spinbutton(spinbutton), m_min(min), m_max(max)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_spin_button_set_range(m_spinbutton, m_min, m_max);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSpinButton *m_spinbutton;
  double m_min, m_max;
};

class GtkSpinButtonGetRangeRequest : public GtkRequest {
public:
  GtkSpinButtonGetRangeRequest(GtkSpinButton *spinbutton) :
    m_spinbutton(spinbutton)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_spin_button_get_range(m_spinbutton, &m_min, &m_max);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    PyrObject *array = newPyrArray(g->gc, 2, 0, false);
    PyrSlot *slots = array->slots;

    SetFloat(slots + 0, m_min);
    SetFloat(slots + 1, m_max);

    array->size = 2;

    SetObject(resultSlot, array);

    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSpinButton *m_spinbutton;
  double m_min, m_max;
};

class GtkSpinButtonSetValueRequest : public GtkRequest {
public:
  GtkSpinButtonSetValueRequest(GtkSpinButton *spinbutton, double value) :
    m_spinbutton(spinbutton), m_value(value)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_spin_button_set_value(m_spinbutton, m_value);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSpinButton *m_spinbutton;
  double m_value;
};

class GtkSpinButtonGetValueRequest : public GtkRequest {
public:
  GtkSpinButtonGetValueRequest(GtkSpinButton *spinbutton) :
    m_spinbutton(spinbutton)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = gtk_spin_button_get_value(m_spinbutton);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetFloat(resultSlot, m_value);
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSpinButton *m_spinbutton;
  double m_value;
};

class GtkSpinButtonGetValueAsIntegerRequest : public GtkRequest {
public:
  GtkSpinButtonGetValueAsIntegerRequest(GtkSpinButton *spinbutton) :
    m_spinbutton(spinbutton)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = gtk_spin_button_get_value_as_int(m_spinbutton);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, m_value);
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSpinButton *m_spinbutton;
  int m_value;
};

class GtkSpinButtonSetUpdatePolicyRequest : public GtkRequest {
public:
  GtkSpinButtonSetUpdatePolicyRequest(GtkSpinButton *spinbutton, GtkSpinButtonUpdatePolicy policy) :
    m_spinbutton(spinbutton), m_policy(policy)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_spin_button_set_update_policy(m_spinbutton, m_policy);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSpinButton *m_spinbutton;
  GtkSpinButtonUpdatePolicy m_policy;
};

class GtkSpinButtonGetUpdatePolicyRequest : public GtkRequest {
public:
  GtkSpinButtonGetUpdatePolicyRequest(GtkSpinButton *spinbutton) :
    m_spinbutton(spinbutton)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_policy = gtk_spin_button_get_update_policy(m_spinbutton);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetEnumSlot(resultSlot, GTK_TYPE_SPIN_BUTTON_UPDATE_POLICY, m_policy);
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSpinButton *m_spinbutton;
  GtkSpinButtonUpdatePolicy m_policy;
};

class GtkSpinButtonSetNumericRequest : public GtkRequest {
public:
  GtkSpinButtonSetNumericRequest(GtkSpinButton *spinbutton, bool value) :
    m_spinbutton(spinbutton), m_value(value)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_spin_button_set_numeric(m_spinbutton, m_value);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSpinButton *m_spinbutton;
  bool m_value;
};

class GtkSpinButtonGetNumericRequest : public GtkRequest_ReturnBoolean {
public:
  GtkSpinButtonGetNumericRequest(GtkSpinButton *spinbutton) :
    m_spinbutton(spinbutton)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = gtk_spin_button_get_numeric(m_spinbutton);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSpinButton *m_spinbutton;
};

class GtkSpinButtonSetWrapRequest : public GtkRequest {
public:
  GtkSpinButtonSetWrapRequest(GtkSpinButton *spinbutton, bool value) :
    m_spinbutton(spinbutton), m_value(value)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_spin_button_set_wrap(m_spinbutton, m_value);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSpinButton *m_spinbutton;
  bool m_value;
};

class GtkSpinButtonGetWrapRequest : public GtkRequest_ReturnBoolean {
public:
  GtkSpinButtonGetWrapRequest(GtkSpinButton *spinbutton) :
    m_spinbutton(spinbutton)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = gtk_spin_button_get_wrap(m_spinbutton);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSpinButton *m_spinbutton;
};

class GtkSpinButtonSetSnapToTicksRequest : public GtkRequest {
public:
  GtkSpinButtonSetSnapToTicksRequest(GtkSpinButton *spinbutton, bool value) :
    m_spinbutton(spinbutton), m_value(value)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_spin_button_set_snap_to_ticks(m_spinbutton, m_value);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSpinButton *m_spinbutton;
  bool m_value;
};

class GtkSpinButtonGetSnapToTicksRequest : public GtkRequest_ReturnBoolean {
public:
  GtkSpinButtonGetSnapToTicksRequest(GtkSpinButton *spinbutton) :
    m_spinbutton(spinbutton)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = gtk_spin_button_get_snap_to_ticks(m_spinbutton);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSpinButton *m_spinbutton;
};

class GtkSpinButtonUpdateRequest : public GtkRequest {
public:
  GtkSpinButtonUpdateRequest(GtkSpinButton *spinbutton) :
    m_spinbutton(spinbutton)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_spin_button_update(m_spinbutton);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSpinButton *m_spinbutton;
};

class GtkSpinButtonSpinRequest : public GtkRequest {
public:
  GtkSpinButtonSpinRequest(GtkSpinButton *spinbutton, GtkSpinType type, double amount) :
    m_spinbutton(spinbutton), m_type(type), m_amount(amount)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_spin_button_spin(m_spinbutton, m_type, m_amount);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSpinButton *m_spinbutton;
  GtkSpinType m_type;
  double m_amount;
};

// ------ GtkCombo

class GtkComboNewRequest : public GtkRequest {
public:
  GtkComboNewRequest(void) :
    m_combo(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_combo = (GtkCombo *)gtk_combo_new();
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_combo));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkCombo *m_combo;
};

#define MAX_COMBO_STRINGS 64
typedef char ComboStringArray[MAX_COMBO_STRINGS][MEDIUM_STRING_SIZE];

class GtkComboSetPopdownStringsRequest : public GtkRequest {
public:
  GtkComboSetPopdownStringsRequest(GtkCombo *combo, int n, ComboStringArray &strings) :
    m_combo(combo)
  {
    m_n = std::min(MAX_COMBO_STRINGS, n);

    for (int i = 0; i < m_n; ++i) {
      safe_string_copy(m_strings[i], strings[i], -1, MEDIUM_STRING_SIZE);
    }
  }
    
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    GList *stringList = 0;
    
    for (int i = 0; i < m_n; ++i)
      stringList = g_list_append(stringList, m_strings[i]);

    gtk_combo_set_popdown_strings(m_combo, stringList);

    g_list_free(stringList);
  }
    
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkCombo *m_combo;
  int m_n;
  ComboStringArray m_strings;
};

class GtkComboSetValueInListRequest : public GtkRequest {
public:
  GtkComboSetValueInListRequest(GtkCombo *combo, bool value, bool okIfEmpty) :
    m_combo(combo), m_value(value), m_okIfEmpty(okIfEmpty)
  {
  }
    
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_combo_set_value_in_list(m_combo, m_value, m_okIfEmpty);
  }
    
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkCombo *m_combo;
  bool m_value, m_okIfEmpty;
};

class GtkComboSetUseArrowsRequest : public GtkRequest {
public:
  GtkComboSetUseArrowsRequest(GtkCombo *combo, bool value) :
    m_combo(combo), m_value(value)
  {
  }
    
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_combo_set_use_arrows(m_combo, m_value);
  }
    
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkCombo *m_combo;
  bool m_value;
};

class GtkComboSetCaseSensitiveRequest : public GtkRequest {
public:
  GtkComboSetCaseSensitiveRequest(GtkCombo *combo, bool value) :
    m_combo(combo), m_value(value)
  {
  }
    
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_combo_set_case_sensitive(m_combo, m_value);
  }
    
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkCombo *m_combo;
  bool m_value;
};

class GtkComboSetItemStringRequest : public GtkRequest {
public:
  GtkComboSetItemStringRequest(GtkCombo *combo, GtkItem *item, const char *s, int size) :
    m_combo(combo), m_item(item)
  {
    safe_string_copy(m_s, s, size, MEDIUM_STRING_SIZE);
  }
    
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_combo_set_item_string(m_combo, m_item, m_s);
  }
    
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkCombo *m_combo;
  GtkItem *m_item;
  char m_s[MEDIUM_STRING_SIZE];
};

class GtkComboDisableActivateRequest : public GtkRequest {
public:
  GtkComboDisableActivateRequest(GtkCombo *combo) :
    m_combo(combo)
  {
  }
    
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_combo_disable_activate(m_combo);
  }
    
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkCombo *m_combo;
};

// ------ GladeXML

class GladeXMLNewRequest : public GtkRequest {
public:
  GladeXMLNewRequest(const char *filename, int filenamesize, const char *root, int rootsize,
		     const char *lamespace, int lamespacesize) :
    m_xml(0)
  {
    safe_string_copy(m_filename, filename, filenamesize, FILENAME_STRING_SIZE);
    if (root != 0) {
      safe_string_copy(m_root0, root, rootsize, SMALL_STRING_SIZE);
      m_root = m_root0;
    } else {
      m_root = 0;
    }
    if (lamespace != 0) {
      safe_string_copy(m_lamespace0, lamespace, lamespacesize, SMALL_STRING_SIZE);
      m_lamespace = m_lamespace0;
    } else {
      m_lamespace = 0;
    }
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_xml = glade_xml_new(m_filename, m_root, m_lamespace);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_xml));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GladeXML *m_xml;
  char m_filename[FILENAME_STRING_SIZE];
  char *m_root;
  char m_root0[SMALL_STRING_SIZE];
  char *m_lamespace;
  char m_lamespace0[SMALL_STRING_SIZE];
};


extern void sc_glade_xml_connect_function(const gchar *handler_name, GObject *object,
					  const gchar *signal_name, const gchar *signal_data,
					  GObject *connect_object, gboolean after, gpointer user_data);


class GladeXMLSignalAutoconnectRequest : public GtkRequest {
public:
  GladeXMLSignalAutoconnectRequest(GladeXML *xml, PyrObject *wrapper) :
    m_xml(xml), m_wrapper(wrapper)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    glade_xml_signal_autoconnect_full (m_xml, sc_glade_xml_connect_function, m_wrapper);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GladeXML *m_xml;
  PyrObject *m_wrapper;
};

// ------ GtkSelectionData

class GtkSelectionDataSetToPointerRequest : public GtkRequest {
public:
  GtkSelectionDataSetToPointerRequest(GtkSelectionData *data, int ptr) :
    m_data(data), m_ptr(ptr)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_selection_data_set(m_data, gdk_atom_intern ("ANONPOINTER", FALSE), 8, (const guchar *)&m_ptr, sizeof(m_ptr));
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSelectionData *m_data;
  int m_ptr;
};

class GtkSelectionDataGetAsPointerRequest : public GtkRequest {
public:
  GtkSelectionDataGetAsPointerRequest(GtkSelectionData *data) :
    m_data(data)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    if (m_data->data != 0)
      memcpy(&m_ptr, m_data->data, sizeof(m_ptr));
    else
      m_ptr = 0;
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    if (m_ptr != 0)
      SetInt(resultSlot, int(m_ptr));
    else
      SetNil(resultSlot);
    return 1;
  }
  
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSelectionData *m_data;
  int m_ptr;
};

// ------ GtkHSeparator

class GtkHSeparatorNewRequest : public GtkRequest {
public:
  GtkHSeparatorNewRequest(void) :
    m_separator(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_separator = (GtkHSeparator *)gtk_hseparator_new();
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_separator));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkHSeparator *m_separator;
};

// ------ GtkVSeparator

class GtkVSeparatorNewRequest : public GtkRequest {
public:
  GtkVSeparatorNewRequest(void) :
    m_separator(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_separator = (GtkVSeparator *)gtk_vseparator_new();
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_separator));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkVSeparator *m_separator;
};

class GtkColorSelectionNewRequest : public GtkRequest {
public:
  GtkColorSelectionNewRequest(void) :
    m_selection(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_selection = (GtkColorSelection *)gtk_color_selection_new();
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_selection));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkColorSelection *m_selection;
};

class GtkColorSelectionSetCurrentColorRequest : public GtkRequest {
public:
  GtkColorSelectionSetCurrentColorRequest(GtkColorSelection *selection, GdkColor color) :
    m_selection(selection), m_color(color)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_color_selection_set_current_color(m_selection, &m_color);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkColorSelection *m_selection;
  GdkColor m_color;
};

class GtkColorSelectionGetCurrentColorRequest : public GtkRequest {
public:
  GtkColorSelectionGetCurrentColorRequest(GtkColorSelection *selection) :
    m_selection(selection)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_color_selection_get_current_color(m_selection, &m_color);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    PyrObject *array = newPyrArray(g->gc, 3, 0, false);
    PyrSlot *slots = array->slots;

    SetInt(slots + 0, m_color.red);
    SetInt(slots + 1, m_color.green);
    SetInt(slots + 2, m_color.blue);

    array->size = 3;

    SetObject(resultSlot, array);

    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkColorSelection *m_selection;
  GdkColor m_color;
};

class GtkColorSelectionSetUpdatePolicyRequest : public GtkRequest {
public:
  GtkColorSelectionSetUpdatePolicyRequest(GtkColorSelection *selection, GtkUpdateType policy) :
    m_selection(selection), m_policy(policy)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_color_selection_set_update_policy(m_selection, m_policy);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkColorSelection *m_selection;
  GtkUpdateType m_policy;
};

class GtkColorSelectionSetHasOpacityControlRequest : public GtkRequest {
public:
  GtkColorSelectionSetHasOpacityControlRequest(GtkColorSelection *selection, bool value) :
    m_selection(selection), m_value(value)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gtk_color_selection_set_has_opacity_control(m_selection, m_value);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkColorSelection *m_selection;
  bool m_value;
};

class GtkColorSelectionGetHasOpacityControlRequest : public GtkRequest_ReturnBoolean {
public:
  GtkColorSelectionGetHasOpacityControlRequest(GtkColorSelection *selection) :
    m_selection(selection)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_value = gtk_color_selection_get_has_opacity_control(m_selection);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkColorSelection *m_selection;
};

// ------ GnomeCanvas

class GnomeCanvasNewRequest : public GtkRequest {
public:
  GnomeCanvasNewRequest(bool antialiased) :
    m_canvas(0), m_antialiased(antialiased)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    if (m_antialiased)
      m_canvas = gnome_canvas_new();
    else
      m_canvas = gnome_canvas_new_aa();      
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_canvas));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkWidget *m_canvas;
  bool m_antialiased;
};

class GnomeCanvasRootRequest : public GtkRequest {
public:
  GnomeCanvasRootRequest(GnomeCanvas *canvas) :
    m_canvas(canvas)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_group = gnome_canvas_root(m_canvas);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_group));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GnomeCanvas *m_canvas;
  GnomeCanvasGroup *m_group;
};

class GnomeCanvasSetScrollRegionRequest : public GtkRequest {
public:
  GnomeCanvasSetScrollRegionRequest(GnomeCanvas *canvas, double x, double y, double w, double h) :
    m_canvas(canvas), m_x(x), m_y(y), m_w(w), m_h(h)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    gnome_canvas_set_scroll_region(m_canvas, m_x, m_y, m_w, m_h);
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GnomeCanvas *m_canvas;
  double m_x, m_y, m_w, m_h;
};

class GnomeCanvasGroupAddRequest : public GtkRequest {
public:
  GnomeCanvasGroupAddRequest(GnomeCanvasGroup *group, GType type) :
    m_item(0), m_group(group), m_type(type)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_item = gnome_canvas_item_new(m_group, m_type, 0);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    if (m_item == 0)
      SetNil(resultSlot);
    else
      SetInt(resultSlot, int(m_item));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GnomeCanvasItem *m_item;
  GnomeCanvasGroup *m_group;
  GType m_type;
};

class GnomeCanvasItemSetRequest : public GtkRequest {
public:
  GnomeCanvasItemSetRequest(GnomeCanvasItem *item, const char *name, GValue *value) :
    m_item(item)
  {
    safe_string_copy(m_name, name, -1, SMALL_STRING_SIZE);  

    memset(&m_value, 0, sizeof(m_value));
    g_value_init(&m_value, G_VALUE_TYPE(value));
    g_value_copy(value, &m_value);
  }
    
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    g_object_freeze_notify (G_OBJECT(m_item));
    g_object_set_property(G_OBJECT(m_item), m_name, &m_value);
    gnome_canvas_item_request_update(m_item);
    g_object_thaw_notify (G_OBJECT(m_item));
    g_value_unset(&m_value);
  }
    
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GnomeCanvasItem *m_item;
  char m_name[SMALL_STRING_SIZE];
  GValue m_value;
};

#define MAX_GNOME_ITEMS_POINTS 128
typedef double GnomeItemPointsArray[MAX_GNOME_ITEMS_POINTS * 2];

class GnomeCanvasItemSetPointsRequest : public GtkRequest {
public:
  GnomeCanvasItemSetPointsRequest(GnomeCanvasItem *item, GnomeItemPointsArray &points, int n) :
    m_item(item)
  {
    m_n = (n < MAX_GNOME_ITEMS_POINTS) ? n : MAX_GNOME_ITEMS_POINTS;
    for (int i = 0; i < n * 2; i += 2) {
      m_points[i] = points[i];
      m_points[i + 1] = points[i + 1];
    }
  }
    
  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    GValue value = { 0 };
    g_value_init(&value, gnome_canvas_points_get_type());
    GnomeCanvasPoints *points = gnome_canvas_points_new(m_n);
    for (int i = 0; i < m_n * 2; i += 2) {
      points->coords[i] = m_points[i];
      points->coords[i + 1] = m_points[i + 1];
    }
    g_value_set_boxed(&value, points);
    g_object_freeze_notify (G_OBJECT(m_item));
    g_object_set_property(G_OBJECT(m_item), "points", &value);
    gnome_canvas_item_request_update(m_item);
    g_object_thaw_notify (G_OBJECT(m_item));

    gnome_canvas_points_unref(points);
    g_value_unset(&value);    
  }
    
  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GnomeCanvasItem *m_item;
  int m_n;
  GnomeItemPointsArray m_points;
};

// GtkSourceView

class GtkSourceBufferNewRequest : public GtkRequest {
public:
  GtkSourceBufferNewRequest(void) :
    m_buffer(0)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_buffer = (GtkSourceBuffer *)gtk_source_buffer_new(0);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_buffer));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSourceBuffer *m_buffer;
};

class GtkSourceViewNewWithBufferRequest : public GtkRequest {
public:
  GtkSourceViewNewWithBufferRequest(GtkSourceBuffer *buffer) :
    m_sourceview(0), m_buffer(buffer)
  {
  }

  void perform(void) {
    GTK_ASYNC_REQUEST_DUMP_PERFORM_LINE;
    m_sourceview = (GtkSourceView *)gtk_source_view_new_with_buffer(m_buffer);
  }

  int pushResult(VMGlobals *g) {
    PyrSlot *resultSlot = ++g->sp;
    SetInt(resultSlot, int(m_sourceview));
    return 1;
  }

  SIZE_METHOD_DEFINITION;
  CODE_LOCATION_METHOD_DEFINITION;

private:
  GtkSourceView *m_sourceview;
  GtkSourceBuffer *m_buffer;
};


#endif // _gtkasyncrequests_H_
