#include <iostream>

#include <time.h>
#include <sys/poll.h>

#include "sclangInterpreter.h"

#include <glib-object.h>
#include <gdk/gdkdrawable.h>
#include <gtk/gtkwidget.h>
#include <gtk/gtkcontainer.h>
#include <gtk/gtkbin.h>
#include <gtk/gtkenums.h>
#include <gtk/gtkwindow.h>
#include <gtk/gtkeventbox.h>
#include <gtk/gtknotebook.h>
#include <gtk/gtkdrawingarea.h>
#include <gtk/gtkentry.h>
#include <gtk/gtkmain.h>
#include <gtk/gtkframe.h>
#include <gtk/gtkpaned.h>
#include <gtk/gtkhpaned.h>
#include <gtk/gtkvpaned.h>
#include <gtk/gtkbox.h>
#include <gtk/gtkhbox.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtkbutton.h>
#include <gtk/gtkmenu.h>
#include <gtk/gtkmenuitem.h>
#include <gtk/gtklabel.h>
#include <gtk/gtktable.h>
#include <gtk/gtkliststore.h>
#include <gtk/gtktreestore.h>
#include <gtk/gtktreemodel.h>
#include <gtk/gtktreeview.h>
#include <gtk/gtktreeselection.h>
#include <gtk/gtkcellrenderer.h>
#include <gtk/gtkcellrenderertext.h>
#include <gtk/gtkcellrenderertoggle.h>
#include <gtk/gtkadjustment.h>
#include <gtk/gtkscrolledwindow.h>
#include <gtk/gtkimage.h>
#include <gtk/gtktextbuffer.h>
#include <gtk/gtktextview.h>
#include <gtk/gtkrange.h>
#include <gtk/gtkscale.h>
#include <gtk/gtkhscale.h>
#include <gtk/gtkvscale.h>
#include <gtk/gtk.h>
#include <glade/glade.h>

#include <gtksourceview/gtksourceview.h>
#include <gtksourceview/gtksourcebuffer.h>

#include "gtkrequest.h"
#include "gtkasyncrequests.h"

#include <PyrObject.h>
#include <PyrKernel.h>
#include <PyrPrimitive.h>
#include <PyrMessage.h>
#include <GC.h>
#include <VMGlobals.h>

extern pthread_mutex_t gLangMutex;

#include "gtkbindings.h"

static GQuark boxed_type_id = 0;
static GQuark gobject_sctype_id = 0;
static GQuark gobject_scwrapper_id = 0;

#undef CLOSURE_MARSHAL_DEBUG
#undef GBOXED_DEBUG
#undef DEBUG_DTORS

#define SC_CHECK_VALID_GENERIC_WRAPPER(slot) {        \
  if (slot->uo->slots[0].ui == 0) {                   \
    error("invalid generic wrapper (null pointer):\n"); \
    dumpObjectSlot(slot);                             \
    return errFailed;                                 \
  }                                                   \
}

#define SC_CHECK_VALID_GOBJECT_WRAPPER(slot) {        \
  if (!isKindOfSlot(slot, getsym("GObject")->u.classobj)) {                    \
    error("a GObject object was expected but got: \n");                       \
    dumpObjectSlot(slot);                                                    \
    return errWrongType;                                                     \
  }                                                   \
  if (slot->uo->slots[0].ui == 0) {                   \
    error("invalid gobject wrapper (null pointer):\n"); \
    dumpObjectSlot(slot);                             \
    return errFailed;                                 \
  }                                                   \
}

#define SC_CHECK_VALID_GBOXED_WRAPPER(slot) {                                \
  if (!isKindOfSlot(slot, getsym("GBoxed")->u.classobj)) {                    \
    error("a GBoxed object was expected but got: \n");                       \
    dumpObjectSlot(slot);                                                    \
    return errWrongType;                                                     \
  }                                                   \
  if (slot->uo->slots[0].ui == 0) {                   \
    error("invalid gboxed wrapper (null pointer):\n"); \
    dumpObjectSlot(slot);                             \
    return errFailed;                                 \
  }                                                   \
}

#define SC_CHECK_VALID_TYPED_GBOXED_WRAPPER(slot, className) {              \
  if (!isKindOfSlot(slot, getsym(className)->u.classobj)) {                  \
    error("an GBoxed object was expected but got: \n");                       \
    dumpObjectSlot(slot);                                                    \
    return errWrongType;                                                     \
  }                                                   \
  if (slot->uo->slots[0].ui == 0) {                   \
    error("invalid gboxed wrapper (null pointer):\n"); \
    dumpObjectSlot(slot);                             \
    return errFailed;                                 \
  }                                                   \
}

#define WRONG_TYPE_PARAMETER(name, expected, slot) {                         \
  error("Parameter '" name "' expected to be of type " expected ". Got:\n");   \
  dumpObjectSlot(slot);                                                      \
  return errWrongType;                                                       \
}

#define WRONG_TYPE_MEMBER(membername, name, expected, slot) {                         \
  error("Data member '" membername "' of parameter '" name "' expected to be of type " expected ". Got:\n");   \
  dumpObjectSlot(slot);                                                      \
  return errWrongType;                                                       \
}




#define UNWRAP_GBOXED(slot, argname, argtype, ptrtype, ptrname)        \
  if (!isKindOfSlot(slot, getsym(argtype)->u.classobj))                \
    WRONG_TYPE_PARAMETER(argname, argtype, slot);                      \
  if (slot->uo->slots[0].ui == 0) {                                    \
    error("Invalid (null pointer) gboxed wrapper provided for parameter " argname ":\n"); \
    dumpObjectSlot(slot);                                              \
    return errFailed;                                                  \
  }                                                                    \
  ptrtype *ptrname = (ptrtype *)(slot->uo->slots[0].ui);

#define UNWRAP_GOBJECT(slot, argname, argtype, ptrtype, ptrname, castmacro, checkmacro)           \
  SC_CHECK_VALID_GOBJECT_WRAPPER(slot);                                                        \
  if (!checkmacro(slot->uo->slots[0].uptr)) WRONG_TYPE_PARAMETER(argname, argtype, slot);      \
  ptrtype *ptrname = castmacro(slot->uo->slots[0].uptr);

#define WRONG_VALUE_ENUM(name, expected, slot) {                            \
  error("Wrong flag value for parameter " name " of type " expected ":\n"); \
  dumpObjectSlot(slot);                                                     \
  return errWrongType;                                                      \
}

#define UNWRAP_GENUM(slot, argname, enumctype, varname, enumtypemacro)                                       \
  enumctype varname;                                                                                         \
  if (slotEnumVal(enumtypemacro, slot, (int *)&varname)) WRONG_VALUE_ENUM(argname, #enumctype, slot);

#define UNWRAP_GDKEVENT(slot, argname, variable)                         \
  GdkEvent *variable;                                                    \
  if (!isKindOfSlot(slot, getsym("GdkEvent")->u.classobj))               \
  WRONG_TYPE_PARAMETER(argname, "GdkEvent", slot);                       \
  variable = (GdkEvent *)slot->uo->slots[0].uptr;                        \
  if (variable == 0) {                                                   \
    error("invalid GdkEvent wrapper (null pointer):\n");                 \
    dumpObjectSlot(slot);                                                \
    return errFailed;                                                    \
  }

#define UNWRAP_GDKCOLOR(slot, argname, variable)                           \
  GdkColor variable;                                                       \
  if (!isKindOfSlot(slot, getsym("GdkColor")->u.classobj))                 \
    WRONG_TYPE_PARAMETER(argname, "GdkColor", slot);                       \
  {                                                                        \
    int component;                                                         \
    if (slotIntVal(slot->uo->slots + 0, &component))                       \
      WRONG_TYPE_MEMBER("red", argname, "GdkColor", slot->uo->slots + 0);   \
    variable.red = component;                                              \
    if (slotIntVal(slot->uo->slots + 1, &component))                       \
      WRONG_TYPE_MEMBER("green", argname, "GdkColor", slot->uo->slots + 1); \
    variable.green = component;                                            \
    if (slotIntVal(slot->uo->slots + 2, &component))                       \
      WRONG_TYPE_MEMBER("blue", argname, "GdkColor", slot->uo->slots + 2);  \
    variable.blue = component;                                             \
  }

int slotStrPtrLen(PyrSlot *slot, const char **ptr, int *len, int maxlen = 0) {
  if (IsSym(slot)) {
    int n = strlen(slot->us->name);
    if (maxlen > 0 && n > maxlen - 1) return errWrongType; 
    *ptr = slot->us->name;
    *len = n;
    return errNone;
  } else if (isKindOfSlot(slot, class_string)) {
    if (maxlen > 0 && slot->uo->size > maxlen - 1) return errWrongType; 
    *ptr = slot->uos->s;
    *len = slot->uo->size;
    return errNone;
  }
  return errWrongType;
}

int slotBoolVal(PyrSlot *slot, bool &value) {
  if (!IsTrue(slot) && !IsFalse(slot)) return errWrongType;
  value = IsTrue(slot);
  return errNone;
}

int sc_gboxed_wrapper_finalizer(VMGlobals *g, PyrObject *object) {
  int type, ptr;

  if (slotIntVal(object->slots + 0, &ptr)) return errWrongType;  
  if (ptr == 0) return errWrongType;
  void *boxed = (void *)ptr;

  if (slotIntVal(object->slots + 1, &type)) return errWrongType;
  if (type == 0) return errNone;
  GType boxedtype = type;

  bool borrowed;
  if (slotBoolVal(object->slots + 2, borrowed)) return errWrongType;

  if (!borrowed) {
    GTK_REQUEST_QUEUE_SEND_OR_FAIL_WITHOUT_ANSWER(GBoxedFreeRequest(boxedtype, boxed));
#ifdef GBOXED_DEBUG
    std::cout << "GBoxedFreeRequest sent by sc_gboxed_wrapper_free" << std::endl;
#endif
  } else {
#ifdef GBOXED_DEBUG
    std::cout << "A GBoxed wrapper with a finalizer was put in borrowed mode. Skipping GBoxedFreeRequest." << std::endl;
#endif
  }

  SetInt(object->slots + 0, 0);
  SetInt(object->slots + 1, 0);
  SetNil(object->slots + 2);
  SetNil(object->slots + 3);

  return errNone;
}

bool createGBoxedWrapper(VMGlobals *g, PyrSlot *slot, const char *classname, void *ptr, GType type, bool borrowed) {
  PyrClass *classObject = getsym(classname)->u.classobj;
  if (classObject == 0) return true;
  return createGBoxedWrapper(g, slot, classObject, ptr, type, borrowed);
}

bool createGBoxedWrapper(VMGlobals *g, PyrSlot *slot, PyrClass *classObject, void *ptr, GType type, bool borrowed) {
#ifdef GBOXED_DEBUG
  std::cout << "creating a GBoxed wrapper of kind " << classObject->name.us->name << std::endl;
#endif
  if (classObject == 0) return true;
  PyrObject *wrapperObject = instantiateObject(g->gc, classObject, 0, true, true);
  
  SetInt(wrapperObject->slots + 0, (int)ptr);
  SetInt(wrapperObject->slots + 1, type);
  SetBool(wrapperObject->slots + 2, borrowed);

  if (borrowed) {
#ifdef GBOXED_DEBUG
    std::cout << "no finalizer installed for the GBoxed wrapper of a borrowed object" << std::endl;
#endif
    SetNil(wrapperObject->slots + 3);
  } else {
    InstallFinalizer(g, wrapperObject, 3, sc_gboxed_wrapper_finalizer);
#ifdef GBOXED_DEBUG
  std::cout << "finalizer installed for the GBoxed wrapper of a non-borrowed object" << std::endl;
#endif
  }

  SetObject(slot, wrapperObject);

  return false;
}

bool slot_to_GValue(PyrSlot *slot, GValue *value) {
  switch (G_TYPE_FUNDAMENTAL(G_VALUE_TYPE(value))) {
  case G_TYPE_BOOLEAN:
    g_value_set_boolean(value, IsTrue(slot));
    return false;
    break;
  case G_TYPE_INT:
    {
      int i;
      if (slotIntVal(slot, &i)) return true;
      g_value_set_int(value, i);
      return false;
    }
  case G_TYPE_UINT:
    {
      int i;
      if (slotIntVal(slot, &i)) return true;
      g_value_set_uint(value, i);
      return false;
    }
  case G_TYPE_LONG:
    {
      int i;
      if (slotIntVal(slot, &i)) return true;
      g_value_set_long(value, i);
      return false;
    }
  case G_TYPE_ULONG:
    {
      int i;
      if (slotIntVal(slot, &i)) return true;
      g_value_set_ulong(value, i);
      return false;
    }
  case G_TYPE_FLOAT:
    {
      float f;
      if (slotFloatVal(slot, &f)) return true;
      g_value_set_float(value, f);
      return false;
    } 
  case G_TYPE_DOUBLE:
    {
      double d;
      if (slotDoubleVal(slot, &d)) return true;
      g_value_set_double(value, d);
      return false;
    }
  case G_TYPE_STRING:
    {
      char buffer[1024]; // AHAHAHAHAHAH
      if (slotStrVal(slot, buffer, 1024)) return true;
      g_value_set_string(value, buffer);
      return false;
    }
  default:
    return true;
  }
}

PyrObject *wrap_gobject_as_scobject(GObject *object) {
  PyrObject *wrapper;

#if 0
  wrapper = (PyrObject *)g_object_get_qdata(object, gobject_scwrapper_id);
  if (wrapper != 0) {
    
    return wrapper;
  }
#endif

  GType type = G_OBJECT_TYPE(object);
  VMGlobals *g = gMainVMGlobals;

  const char *klassname = (const char *)g_type_get_qdata(type, gobject_sctype_id);

  if (!gobject_scwrapper_id) gobject_scwrapper_id = g_quark_from_static_string("SC::GObject::wrapper");
  
  PyrSymbol *klassym;
  
  if (klassname != 0) {
    klassym = findsym(klassname);
    //std::cout << "Wrapping GObject of type: " << klassname << std::endl;
  }
  
  if (klassname == 0 || klassym == 0 || (klassym->flags & sym_Class) == 0) {
    klassym = findsym("GObject");
    //std::cout << "Wrapping GObject of type '" << G_OBJECT_TYPE_NAME(object) << " as generic GObject" << std::endl;
  }
  
  PyrClass *classobj = klassym->u.classobj;
  if (classobj != 0) {
    //std::cout << "Creating new SC wrapper object" << std::endl;
    wrapper = instantiateObject(g->gc, classobj, 0, true, true);
    SetPtr(wrapper->slots + 0, object);
#if 0
    g_object_set_qdata(object, gobject_scwrapper_id, wrapper);
#endif
    return wrapper;
  } else {
    return 0;
  }  
}

static int SetWrapperOrNil(PyrSlot *slot, GObject *gobj) {
  if (gobj == 0) {
    SetNil(slot);
    return true;
  } else {
    PyrObject *object;
    object = wrap_gobject_as_scobject(gobj);
    if (object == 0) {
      SetNil(slot);
      return true;
    } else {
      SetObject(slot, object);
      return false;
    }
  }
}

bool GValue_to_slot(const GValue *value, PyrSlot *slot) {
  switch (G_TYPE_FUNDAMENTAL(G_VALUE_TYPE(value))) {
  case G_TYPE_CHAR:
    SetChar(slot, g_value_get_char(value));
    return false;
  case G_TYPE_UCHAR:
    SetChar(slot, g_value_get_uchar(value));
    return false;
  case G_TYPE_BOOLEAN:
    if (g_value_get_boolean(value))
      SetTrue(slot);
    else
      SetFalse(slot);
    return false;
  case G_TYPE_INT:
    SetInt(slot, g_value_get_int(value));
    return false;
  case G_TYPE_UINT:
    SetInt(slot, g_value_get_uint(value));
    return false;
  case G_TYPE_FLOAT:
    SetFloat(slot, g_value_get_float(value));
    return false;
  case G_TYPE_DOUBLE:
    SetFloat(slot, g_value_get_double(value));
    return false;
  case G_TYPE_ENUM:
    SetInt(slot, g_value_get_enum(value));
    return false;
  case G_TYPE_FLAGS:
    SetInt(slot, g_value_get_flags(value));
    return false;
  case G_TYPE_STRING:
    {
      VMGlobals *g = gMainVMGlobals;
      const char *s = g_value_get_string(value);
      if (s == 0) {
	//std::cout << "GValue of type string but null value" << std::endl;
	SetNil(slot);
      } else {
	int size = strlen(s);
	PyrString *strobj = newPyrStringN(g->gc, size, 0, true);
	memcpy(strobj->s, s, size);
	SetObject(slot, strobj);
	//SetSymbol(slot, getsym(s));
      }
    }
    return false;
  case G_TYPE_BOXED:
    {
      GType boxedtype = G_VALUE_TYPE(value);
      VMGlobals *g = gMainVMGlobals;
      const char *klassname = (const char *)g_type_get_qdata(boxedtype, boxed_type_id);
      
      PyrSymbol *klassym;
      
      if (klassname != 0) {
	klassym = findsym(klassname);
	//std::cout << "Wrapping GBoxed of type: " << klassname << std::endl;
      }

      if (klassname == 0 || klassym == 0 || (klassym->flags & sym_Class) == 0) {
	klassym = findsym("GBoxed");
	//std::cout << "Wrapping GBoxed of as generic GBoxed" << std::endl;
      }

      //std::cout << "Creating new GBoxed subclass object" << std::endl;
      if (createGBoxedWrapper(g, slot, klassname, g_value_get_boxed(value), boxedtype, true)) {      
	//std::cout << "No matching SC class. Using nil" << std::endl;
	SetNil(slot);
      }
    }
    return false;
  case G_TYPE_OBJECT:
    SetWrapperOrNil(slot, (GObject *)g_value_get_object(value));
    return false;
  default:
    SetNil(slot);
    return true;
  }  
}

bool GValue_to_slot2(const GValue *value, PyrSlot *tagslot, PyrSlot *slot) {
  switch (G_TYPE_FUNDAMENTAL(G_VALUE_TYPE(value))) {
  case G_TYPE_CHAR:
    SetChar(slot, g_value_get_char(value));
    SetNil(tagslot);
    return false;
  case G_TYPE_UCHAR:
    SetChar(slot, g_value_get_uchar(value));
    SetNil(tagslot);
    return false;
  case G_TYPE_BOOLEAN:
    if (g_value_get_boolean(value))
      SetTrue(slot);
    else
      SetFalse(slot);
    SetNil(tagslot);
    return false;
  case G_TYPE_INT:
    SetInt(slot, g_value_get_int(value));
    SetNil(tagslot);
    return false;
  case G_TYPE_UINT:
    SetInt(slot, g_value_get_uint(value));
    SetNil(tagslot);
    return false;
  case G_TYPE_FLOAT:
    SetFloat(slot, g_value_get_float(value));
    SetNil(tagslot);
    return false;
  case G_TYPE_DOUBLE:
    SetFloat(slot, g_value_get_double(value));
    SetNil(tagslot);
    return false;
  case G_TYPE_ENUM:
    SetInt(slot, g_value_get_enum(value));
    SetNil(tagslot);
    return false;
  case G_TYPE_FLAGS:
    SetInt(slot, g_value_get_flags(value));
    SetNil(tagslot);
    return false;
  case G_TYPE_STRING:
    {
      VMGlobals *g = gMainVMGlobals;
      const char *s = g_value_get_string(value);
      if (s == 0) {
	//std::cout << "GValue of type string but null value" << std::endl;
	SetNil(slot);
      } else {
	int size = strlen(s);
	PyrString *strobj = newPyrStringN(g->gc, size, 0, true);
	memcpy(strobj->s, s, size);
	SetObject(slot, strobj);
	//SetSymbol(slot, getsym(s));
      }
    }
    SetNil(tagslot);
    return false;
  case G_TYPE_BOXED:
    {
      GType boxedtype = G_VALUE_TYPE(value);
      VMGlobals *g = gMainVMGlobals;
      const char *klassname = (const char *)g_type_get_qdata(boxedtype, boxed_type_id);
      
      PyrSymbol *klassym;
      
      if (klassname != 0) {
	klassym = findsym(klassname);
	//std::cout << "Wrapping GBoxed of type: " << klassname << std::endl;
      }

      if (klassname == 0 || klassym == 0 || (klassym->flags & sym_Class) == 0) {
	klassym = findsym("GBoxed");
	//std::cout << "Wrapping GBoxed of as generic GBoxed" << std::endl;
      }

      //std::cout << "Creating new object" << std::endl;
      if (createGBoxedWrapper(g, slot, klassname, g_value_get_boxed(value), boxedtype, true)) {
	//std::cout << "No matching SC class. Using nil" << std::endl;
	SetNil(slot);
      }
    }
    SetNil(tagslot);
    return false;
  case G_TYPE_OBJECT:
    SetInt(slot, int(g_value_get_object(value)));
    SetInt(tagslot, 1);
    return false;
  default:
    SetNil(slot);
    return true;
  }  
}

void SetEnumSlot(PyrSlot *slot, GType enum_type, int value) {
  GEnumClass *eclass = NULL;
  GEnumValue *info;
  eclass = G_ENUM_CLASS(g_type_class_ref(enum_type));
  info = g_enum_get_value(eclass, value);
  g_type_class_unref(eclass);  
  if (info == 0) {
    SetSymbol(slot, getsym("badvalue"));
  } else {
    SetSymbol(slot, getsym(info->value_nick));
  }
}

static int slotEnumVal(GType enum_type, PyrSlot *slot, int &val) {
  GEnumClass *eclass = NULL;
  char name[128];
  int res = -1;
  
  if (slot == 0) {
    val = 0;
    res = 0;
  } else if (!slotIntVal(slot, &val)) {
    res = errNone;
  } else if (!slotStrVal(slot, name, 128)) {
    GEnumValue *info;

    if (enum_type != G_TYPE_NONE) {
      eclass = G_ENUM_CLASS(g_type_class_ref(enum_type));
      info = g_enum_get_value_by_name(eclass, name);
      g_type_class_unref(eclass);
      
      if (!info)
	info = g_enum_get_value_by_nick(eclass, name);

      if (info) {
	val = info->value;
	res = errNone;
      } else {
	res = errWrongType;
      }
    } else {
      res = errWrongType;
    }
    
  } else {
    res = errWrongType;
  }

  return res;
}

static int slotEnumVal(GType enum_type, PyrSlot *slot, int *val) {
  return slotEnumVal(enum_type, slot, val[0]);
}

static int slotFlagsVal(GType flags_type, PyrSlot *slot, int &mask) {
  if (slot->utag == tagObj && isKindOf(slot->uo, class_array)) {
    PyrObject *array = slot->uo;
    PyrSlot *slots = array->slots;
    
    mask = 0;
    for (int i = 0; i < array->size; ++i) {
      int value;
      if (slotFlagsVal(flags_type, slots + i, value)) return errWrongType;
      mask = mask | value;
    }
  } else {
    GFlagsClass *fclass = NULL;
    char name[128];
    int res = -1;
  
    if (slot == 0) {
      mask = 0;
      return errNone;
    } else if (!slotIntVal(slot, &mask)) {
      return errNone;
    } else if (!slotStrVal(slot, name, 128)) {
      GFlagsValue *info;
      
      if (flags_type != G_TYPE_NONE) {
	fclass = G_FLAGS_CLASS(g_type_class_ref(flags_type));
	info = g_flags_get_value_by_name(fclass, name);
	g_type_class_unref(fclass);
	
	if (!info)
	  info = g_flags_get_value_by_nick(fclass, name);
	
	if (info) {
	  mask = info->value;
	  return errNone;
	} else {
	  return errWrongType;
	}
      } else {
	return errWrongType;
      }
      
    } else {
      return errWrongType;
    }
    
  }
  return errNone;
}

static int slotFlagsVal(GType flags_type, PyrSlot *slot, int *mask) {
  return slotFlagsVal(flags_type, slot, mask[0]);
}

struct SC_GClosure {
  GClosure closure;
  int id1;
  PyrObject *wrapper;
};

static void
sc_gclosure_invalidate(gpointer data, GClosure *closure) {
  ((SC_GClosure *)closure)->id1 = 0;
  ((SC_GClosure *)closure)->wrapper = 0;
}

bool secondary_ticking_loop(VMGlobals *g, GtkRequestQueue::PollResult expected = GtkRequestQueue::INCOMING_RESULT) {
  const int MAX_COUNT = 25;
  int count = 0;
  bool result;
  GtkRequestQueue::PollResult presult;

#ifdef CLOSURE_MARSHAL_DEBUG
  static int ticking_level = 0;
  ticking_level++;
  std::cout << "entering the secondary ticking loop at level " << ticking_level << std::endl;
#endif

  // GUI Thread is now locked inside this call
  // we tick it here so that the interpreter can perform
  // callbacks
  while (1) {
    count += 1;

    presult = GtkRequestQueue::smartWait(50);
    if (presult == GtkRequestQueue::ABORT) {
      std::cout << "*** CALLBACK ABORT RECEIVED" << std::endl;
      result = true;
      break;
    }
    if (presult == expected) {
      if (presult == GtkRequestQueue::INCOMING_RESULT) {
	result = false;
	break;
      } else if (presult == GtkRequestQueue::INCOMING_VOID) {
	result = false;
	break;
      }
    } else {
      if (presult == GtkRequestQueue::INCOMING_RESULT) {
	std::cout << "Secondary Ticking Loop: ingnoring an unwanted NON-VOID result" << std::endl;
      }
      if (presult == GtkRequestQueue::INCOMING_VOID) {
	std::cout << "Secondary Ticking Loop: ingnoring an unwanted VOID result" << std::endl;
      }
    }

    pthread_mutex_lock(&gLangMutex);
    g->canCallOS = true;
    runLibrary(getsym("tick"));
    g->canCallOS = false;
    pthread_mutex_unlock(&gLangMutex);

    if (count >= MAX_COUNT) {
      std::cout << "*** ABORTING SECONDARY TICKING LOOP WITHOUT AN ANSWER" << std::endl;
      std::cout << "*** ANSWER STACK MAY GET CORRUPT IF THE ANSWER COMES LATER" << std::endl;
#ifdef GTKSERVER_FATAL_FAILURES
      abort();
#endif
      result = true;
      break;
    }
      
    //std::cout << "no signal answer pending after " << count << " secondary ticks" << std::endl;
  }

#ifdef CLOSURE_MARSHAL_DEBUG
  std::cout << "secondary ticking loop exited from level " << ticking_level << std::endl;
  ticking_level--;
#endif

  return result;
} 

static bool retrieve_sclang_callback_result(PyrSlot *resultArraySlot, PyrSlot &result) {
  if (!isKindOfSlot(resultArraySlot, class_array)) {
    std::cout << "*** callback result queue slot is not an array" << std::endl;
    return false;
  }
      
  PyrObject *array = resultArraySlot->uo;
    
  if (array->obj_format != obj_slot) {
    std::cout << "*** callback result queue is not array of slots." << std::endl;
    return false;
  }
      
  // This whole function must be executed inside the interpreter locked
  // because we use a temporary copy of a slot outside the
  // managed memory by the GC of sclang
  if (array->size > 0) {
    // Take the head of the array
    result.ucopy = array->slots[0].ucopy;
    for (int i = 0; i < array->size - 2; ++i)
      array->slots[i].ucopy = array->slots[i+1].ucopy;
    array->size -= 1;
    return true;
  } else {
    return false;
  }
}

static void dump_invocation_hint(gpointer invocation_hint) {
  if (invocation_hint != 0) {
    std::cout << "INVOCATION HINT:" << std::endl;
    GSignalInvocationHint *hint = (GSignalInvocationHint *)invocation_hint;
    const char *detail = g_quark_to_string(hint->detail);
    std::cout << "  NAME:" << ((char *)g_signal_name(hint->signal_id)) << std::endl;
    if (detail != 0) std::cout << "  DETAIL IS " << detail << std::endl;
  } else {
    std::cout << "NO INVOCATION HINT AVAILABLE" << std::endl;
  }

}

static void
sc_gclosure_marshal(GClosure *closure,
                    GValue *return_value,
                    guint n_param_values,
                    const GValue *param_values,
                    gpointer invocation_hint,
                    gpointer marshal_data) {

  SC_GClosure *c = (SC_GClosure *)closure;

#ifdef CLOSURE_MARSHAL_DEBUG
  std::cout << "MARSHALLING A SC-GCLOSURE" << std::cout;
  dump_invocation_hint(invocation_hint);
#endif

  if (closure == 0) {
    std::cout << "NULL closure (that's strange)" << std::endl;
    return;
  }

  VMGlobals *g = gMainVMGlobals;

  pthread_mutex_lock(&gLangMutex);

  int size = param_values ? n_param_values : 0;
  PyrObject *tagArray = newPyrArray(g->gc, size, 0, false);
  PyrSlot *tagslots = tagArray->slots;
  PyrObject *argArray = newPyrArray(g->gc, size, 0, false);
  PyrSlot *argslots = argArray->slots;

  if (param_values != 0) {
    //std::cout << "PARAMS: " << n_param_values << std::endl;
    for (int i = 0; i < size; ++i) {
      GValue_to_slot2(param_values + i, tagslots + i, argslots + i);
    }
  }
    
  tagArray->size = size;
  argArray->size = size;

  PyrObject *wrapper =  ((SC_GClosure*)closure)->wrapper;

  ++g->sp; SetObject(g->sp, wrapper);
  ++g->sp; SetInt(g->sp, ((SC_GClosure*)closure)->id1);
  ++g->sp; SetObject(g->sp, tagArray);
  ++g->sp; SetObject(g->sp, argArray);
  g->canCallOS = true;
  runInterpreter(g, getsym("prScheduleCallback"), 4);
  g->canCallOS = false;
  pthread_mutex_unlock(&gLangMutex);

  if (secondary_ticking_loop(g)) {
    dump_invocation_hint(invocation_hint);
    return;
  }

  pthread_mutex_lock(&gLangMutex);

  bool resultAvailable;
  PyrSlot result;
  PyrSlot *answerArraySlot = wrapper->slots + 1;

  resultAvailable = retrieve_sclang_callback_result(answerArraySlot, result);
      
  if (return_value) {
    if (resultAvailable) {
#ifdef CLOSURE_MARSHAL_DEBUG
      std::cout << "Passing return slot as callback result" << std::endl;
#endif
      slot_to_GValue(&result, return_value);
    } else {
      std::cout << "Signal needed a result but none is available" << std::endl;
    }
  }

  pthread_mutex_unlock(&gLangMutex);

#ifdef CLOSURE_MARSHAL_DEBUG
  std::cout << "EXITING CLOSURE MARSHALING CODE" << std::endl;
#endif
}

int prGTypeNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *object = g->sp - 1;

  char name[128];
  if (slotStrVal(g->sp, name, 128)) return errWrongType;

  GType type = g_type_from_name(name);
  
  // Assume the id is the first slot
  SetInt(object->uo->slots + 0, type);

  return errNone;
}

int prGObjectDump(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *object = g->sp;
  std::cout << ((void *)object->uo->slots[0].ui) << std::endl;

  return errNone;
}

int prGObjectGetTypeName(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!G_IS_OBJECT(objectSlot->uo->slots[0].uptr)) return errWrongType;
  GObject *object = (GObject *)objectSlot->uo->slots[0].uptr;

  const char *name = G_OBJECT_TYPE_NAME(object);

  if (name != 0) {  
    int size = strlen(name);
    PyrString *strobj = newPyrStringN(g->gc, size, 0, true);
    memcpy(strobj->s, name, size);
  
    SetObject(objectSlot, strobj);
  } else {
    SetNil(objectSlot);
  }

  return errNone;
}

int prGtkServerSendResultTrigger(struct VMGlobals *g, int numArgsPushed) {
  GtkRequestQueue::trigger('a');
  return errNone;
}
int prGtkServerSendVoidTrigger(struct VMGlobals *g, int numArgsPushed) {
  GtkRequestQueue::trigger('v');
  return errNone;
}
int prGtkServerSendAbortTrigger(struct VMGlobals *g, int numArgsPushed) {
  GtkRequestQueue::trigger('%');
  return errNone;
}

int prGObjectGetObjectTypeName(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 1;
  PyrSlot *ptrSlot = g->sp;

  int ptrAsInt;
  slotIntVal(ptrSlot, &ptrAsInt);

  GObject *object = (GObject *)ptrAsInt;
  if (object == 0) {
    SetNil(classSlot);
    return errNone;
  }
  
  if (!G_IS_OBJECT(object)) {
    SetNil(classSlot);
    return errNone;
  }

  const char *name = G_OBJECT_TYPE_NAME(object);

  if (name != 0) {  
    int size = strlen(name);
    PyrString *strobj = newPyrStringN(g->gc, size, 0, true);
    memcpy(strobj->s, name, size);  
    SetObject(classSlot, strobj);
    g->gc->GCWrite(g->process->interpreter.uo, strobj);
  } else {
    SetNil(classSlot);
  }

  return errNone;
}

int prGObjectGetClassName(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!G_IS_OBJECT(objectSlot->uo->slots[0].uptr)) return errWrongType;
  GObject *object = (GObject *)objectSlot->uo->slots[0].uptr;

  const char *name = G_OBJECT_CLASS_NAME(object);
  
  if (name != 0) {
    int size = strlen(name);
    PyrString *strobj = newPyrStringN(g->gc, size, 0, true);
    memcpy(strobj->s, name, size);
  
    SetObject(objectSlot, strobj);
    g->gc->GCWrite(g->process->interpreter.uo, strobj);
  } else {
    SetNil(objectSlot);
  }

  return errNone;
}

int prGObjectIsObject(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (G_IS_OBJECT(objectSlot->uo->slots[0].uptr))
    SetTrue(objectSlot);
  else
    SetFalse(objectSlot);

  return errNone;
}

int prGObjectRef(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *wrapperSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(wrapperSlot);
  PyrObject *wrapper = wrapperSlot->uo;
  GObject *object = (GObject *)(wrapper->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GObjectRefRequest(object));

  return errNone;
}

int prGObjectUnref(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *wrapperSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(wrapperSlot);
  PyrObject *wrapper = wrapperSlot->uo;
  GObject *object = (GObject *)(wrapper->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GObjectUnrefRequest(object));

  return errNone;
}

int prGtkObjectSink(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *wrapperSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(wrapperSlot);
  GtkObject *object = (GtkObject *)(wrapperSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkObjectSinkRequest(object));

  return errNone;
}

int prGObjectAsInteger(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  void *ptr = objectSlot->uo->slots[0].uptr;

  SetInt(objectSlot, int(ptr));

  return errNone;
}

int prGObjectConnect(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *wrapperSlot = g->sp - 3;
  PyrSlot *nameSlot = g->sp - 2;
  PyrSlot *id1Slot = g->sp - 1;
  PyrSlot *afterSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(wrapperSlot);

  PyrObject *wrapper = wrapperSlot->uo;
  GObject *object = (GObject *)(wrapper->slots[0].ui);
  
  char name[128];
  guint id2, sigid;
  GQuark detail = 0;

  if (slotStrVal(nameSlot, name, 128)) return errWrongType; 
  //std::cout << "A: " << name << std::endl;
  if (!g_signal_parse_name(name, G_OBJECT_TYPE(object), &sigid, &detail, TRUE)) {
    dumpObjectSlot(nameSlot);
    error("invalid signal name for object wrapped by\n");
    dumpObjectSlot(wrapperSlot);
    return errFailed;
  }

  int id1;
  if (slotIntVal(id1Slot, &id1)) WRONG_TYPE_PARAMETER("id1", "Integer", id1Slot);

  bool after;
  if (slotBoolVal(afterSlot, after)) WRONG_TYPE_PARAMETER("after", "Boolean", afterSlot);

  //std::cout << "D" << std::endl;
  GClosure *closure = g_closure_new_simple(sizeof(SC_GClosure), NULL);
  g_closure_add_invalidate_notifier(closure, NULL, sc_gclosure_invalidate);
  g_closure_set_marshal(closure, sc_gclosure_marshal);

  ((SC_GClosure *)closure)->id1 = id1;
  ((SC_GClosure *)closure)->wrapper = wrapper;

  id2 = g_signal_connect_closure_by_id(object, sigid, detail, closure, FALSE);

  SetInt(wrapperSlot, id2);
 
  return errNone;
}

int prGObjectGetProperty(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *nameSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  GObject *object = (GObject *)objectSlot->uo->slots[0].uptr;

  const char *namePtr;
  int nameSize;
  if (slotStrPtrLen(nameSlot, &namePtr, &nameSize, MEDIUM_STRING_SIZE))
	WRONG_TYPE_PARAMETER("name", "String", nameSlot);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GObjectGetPropertyRequest(object, namePtr, nameSize));
  
  return errNone;
}

int prGObjectSetProperty(struct VMGlobals *g, int numArgsPushed) {
  int result;

  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *nameSlot = g->sp - 1;
  PyrSlot *valueSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  GObject *object = (GObject *)objectSlot->uo->slots[0].uptr;

  const char *namePtr;
  int nameSize;
  if (slotStrPtrLen(nameSlot, &namePtr, &nameSize, MEDIUM_STRING_SIZE))
	WRONG_TYPE_PARAMETER("name", "String", nameSlot);

  GValue value = {0};

  GParamSpec *propspec;
  propspec = g_object_class_find_property(G_OBJECT_GET_CLASS(object), namePtr);
  if (propspec == 0) {
	error("unknown property name:\n");
    dumpObjectSlot(nameSlot);
    return errFailed;
  }

  g_value_init(&value, propspec->value_type);

  if (slot_to_GValue(valueSlot, &value)) {
    result = errWrongType;
  } else {
	GTK_REQUEST_QUEUE_SEND_OR_FAIL(GObjectSetPropertyRequest(object, namePtr, nameSize, &value));
    result = errNone;
  }

  g_value_unset(&value);
 
  return errNone;
}

int prGObjectGetData(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *wrapperSlot = g->sp - 1;
  PyrSlot *keySlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(wrapperSlot);
  PyrObject *wrapper = wrapperSlot->uo;
  GObject *object = (GObject *)(wrapper->slots[0].ui);
  
  char key[128];
  if (slotStrVal(keySlot, key, 128)) return errWrongType;

  void *data = g_object_get_data(object, key);

  SetInt(wrapperSlot, int(data));
 
  return errNone;
}

static void standard_destroy_handler(GtkObject *object, gpointer user_data) {
  PyrObject *wrapper = (PyrObject *)user_data;
  VMGlobals *g = gMainVMGlobals;
  bool canCallOS = g->canCallOS;

#ifdef DEBUG_DTORS
  std::cout << "standard_destroy_handler: " << ((void *)wrapper) << std::endl;
#endif

  pthread_mutex_lock(&gLangMutex);
  g->canCallOS = true;

  ++g->sp; SetObject(g->sp, wrapper);
  runInterpreter(g, getsym("prDestroySignalHandler"), 1);

  g->canCallOS = canCallOS;
  pthread_mutex_unlock(&gLangMutex);

  secondary_ticking_loop(g, GtkRequestQueue::INCOMING_VOID);

#ifdef DEBUG_DTORS
  std::cout << "standard_destroy_handler exited" << std::endl;
#endif
}

int prGtkObjectConnectStandardDestroySignalHandler(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  PyrObject *wrapper = objectSlot->uo;

  GObject *object = (GObject *)wrapper->slots[0].ui;

  if (object != 0)
    g_signal_connect(G_OBJECT(object), "destroy", G_CALLBACK(standard_destroy_handler), (gpointer)wrapper);
  
  return errNone;
}


void standard_weak_reference_notify(gpointer data, GObject *ptr) {
  PyrObject *wrapper = (PyrObject *)data;
  VMGlobals *g = gMainVMGlobals;
  bool canCallOS = g->canCallOS;
  
#ifdef DEBUG_DTORS
  std::cout << "standard_weak_reference_notify: " << ((void *)ptr) << std::endl;
#endif

  pthread_mutex_lock(&gLangMutex);
  g->canCallOS = true;

  ++g->sp; SetObject(g->sp, wrapper);

  runInterpreter(g, getsym("prWeakReferenceNotifyHandler"), 1);

  g->canCallOS = canCallOS;
  pthread_mutex_unlock(&gLangMutex);

  secondary_ticking_loop(g, GtkRequestQueue::INCOMING_VOID);
}

int prGObjectAddStandardWeakReference(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  PyrObject *wrapper = objectSlot->uo;
  GObject *ptr = (GObject *)wrapper->slots[0].ui;

  g_object_weak_ref(ptr, standard_weak_reference_notify, (gpointer)wrapper);

  return errNone;
}


int prGObjectDisconnect(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *idSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  GObject *object = (GObject *)objectSlot->uo->slots[0].ui;

  int id;
  if (slotIntVal(idSlot, &id)) return errWrongType;
  
  g_signal_handler_disconnect(object, (gulong)id);

  return errNone;
}

static gboolean timeout_c_handler(gpointer user_data) {
  int id1 = (int)user_data;
  VMGlobals *g = gMainVMGlobals;
  bool canCallOS = g->canCallOS;
  bool value;

  //std::cout << "timeout_c_handler: " << id1 << std::endl;

  pthread_mutex_lock(&gLangMutex);
  g->canCallOS = true;

  PyrClass *classobject = getsym("Gtk")->u.classobj;
  ++g->sp; SetObject(g->sp, classobject);
  ++g->sp; SetInt(g->sp, id1);

  runInterpreter(g, getsym("prTimeoutHandler"), 2);

  g->canCallOS = canCallOS;
  pthread_mutex_unlock(&gLangMutex);

  if (secondary_ticking_loop(g)) return false;

  pthread_mutex_lock(&gLangMutex);

  bool resultAvailable;
  PyrSlot result;
  int varindex = classVarOffset("Gtk", "callbackResults", &classobject);

  if (varindex >= 0) {
    PyrSlot *answerArraySlot = g->classvars[classobject->classIndex.ui].uo->slots + varindex;

    resultAvailable = retrieve_sclang_callback_result(answerArraySlot, result);
  } else {
    resultAvailable = false;
  }

  if (resultAvailable) {
    int integer;
#ifdef CLOSURE_MARSHAL_DEBUG
    std::cout << "Passing return slot as callback result" << std::endl;
#endif
    if (IsTrue(&result)) {
      value = true;
    } else if (IsFalse(&result)) {
      value = false;
    } else if (!slotIntVal(&result, &integer)) {
      value = bool(integer);
    } else {
      std::cout << "Timeout signal needs a bool result. got something else. returning false" << std::endl;
      value = false;
    }
  } else {
    std::cout << "Timeout signal needs a bool result but none is available" << std::endl;
  }

  pthread_mutex_unlock(&gLangMutex);
  
  return value;
}

int prGtkTimeoutAdd(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 2;
  PyrSlot *intervalSlot = g->sp - 1;
  PyrSlot *id1Slot = g->sp - 0;

  int interval, id1;
  guint id2;

  if (slotIntVal(intervalSlot, &interval)) WRONG_TYPE_PARAMETER("interval", "Integer", intervalSlot);
  if (slotIntVal(id1Slot, &id1)) WRONG_TYPE_PARAMETER("id1", "Integer", id1Slot);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTimeoutAddRequest(interval, id1, timeout_c_handler));
 
  return errNone;
}

int prGtkTimeoutRemove(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 1;
  PyrSlot *id2Slot = g->sp - 0;

  int id2;
  if (slotIntVal(id2Slot, &id2)) WRONG_TYPE_PARAMETER("id2", "Integer", id2Slot);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTimeoutRemoveRequest(id2));

  return errNone;
}

int prGBoxedWrap(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 3;
  PyrSlot *ptrSlot = g->sp - 2;
  PyrSlot *typeSlot = g->sp - 1;
  PyrSlot *borrowedSlot = g->sp - 0;

  PyrClass *classObject = (PyrClass *)classSlot->uo;
  
  int type, ptr;

  if (slotIntVal(ptrSlot, &ptr)) return errWrongType;
  if (ptr == 0) return errWrongType;
  void *boxed = (void *)ptr;

  if (slotIntVal(typeSlot, &type)) return errWrongType;
  if (type == 0) return errNone;
  GType boxedtype = type;

  bool borrowed;
  if (slotBoolVal(borrowedSlot, borrowed)) return errWrongType;

  if (createGBoxedWrapper(g, classSlot, classObject, boxed, boxedtype, borrowed)) {
    SetNil(classSlot);
  }

  return errNone;
}

int prGBoxedFree(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  PyrObject *object = g->sp->uo;

  int type, ptr;

  if (slotIntVal(objectSlot->uo->slots + 0, &ptr)) return errWrongType;  
  if (ptr == 0) return errWrongType;
  void *boxed = (void *)ptr;

  if (slotIntVal(objectSlot->uo->slots + 1, &type)) return errWrongType;
  if (type == 0) return errNone;
  GType boxedtype = type;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL_WITHOUT_ANSWER(GBoxedFreeRequest(boxedtype, boxed));
  //GTK_REQUEST_QUEUE_SEND_OR_FAIL(GBoxedFreeRequest(boxedtype, boxed));

  SetInt(object->slots + 0, 0);
  SetInt(object->slots + 1, 0);
  SetNil(object->slots + 2);
  SetNil(object->slots + 3);

  return errNone;
}

int prGBoxedCopy(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  int type, ptr;

  if (slotIntVal(objectSlot->uo->slots + 0, &ptr)) return errWrongType;  
  void *boxed = (void *)ptr;

  if (slotIntVal(objectSlot->uo->slots + 1, &type)) return errWrongType;
  GType boxedtype = type;

  if (ptr == 0) return errWrongType;
  if (type == 0) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GBoxedCopyRequest(objectSlot->uo->classptr, boxedtype, boxed));

  return errNone;
}

int prGBoxedRegisterWrapperClass(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 1;
  PyrSlot *nameSlot = g->sp - 0;

  if (classSlot->utag != tagObj || !isKindOf(classSlot->uo, class_class)) WRONG_TYPE_PARAMETER("class", "Class", classSlot);
  PyrClass *klass = (PyrClass *)classSlot->uo;

  char classname[1024];
  if (slotStrVal(&klass->name, classname, 1024)) return errWrongType;
  char boxedtypename[1024];
  if (slotStrVal(nameSlot, boxedtypename, 1024)) WRONG_TYPE_PARAMETER("name", "String", nameSlot);

  std::cout << "Registering " << classname << " as wrapper class for GBoxed of type " << boxedtypename << std::endl;

  if (!boxed_type_id) boxed_type_id = g_quark_from_static_string("SC::GBoxed::class");

  GType boxedtype = g_type_from_name(boxedtypename);

  if (boxedtype == 0) {
    error("Unknown boxed type named:");
    dumpObjectSlot(nameSlot);
    return errFailed;
  }

  g_type_set_qdata(boxedtype, boxed_type_id, strdup(classname));

  return errNone;
}

int prGTypeRegisterWrapperClass(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 1;
  PyrSlot *nameSlot = g->sp - 0;

  if (classSlot->utag != tagObj || !isKindOf(classSlot->uo, class_class)) return errWrongType;
  PyrClass *klass = (PyrClass *)classSlot->uo;

  char classname[1024];
  if (slotStrVal(&klass->name, classname, 1024)) WRONG_TYPE_PARAMETER("class", "Class", classSlot);
  char gclassname[1024];
  if (slotStrVal(nameSlot, gclassname, 1024)) WRONG_TYPE_PARAMETER("name", "String", nameSlot);

  //std::cout << "Registering " << classname << " as wrapper class for GType " << gclassname << std::endl;

  if (!gobject_sctype_id) gobject_sctype_id = g_quark_from_static_string("SC::GObject::class");

  GType type = g_type_from_name(gclassname);

  if (type == 0) return errWrongType;

  g_type_set_qdata(type, gobject_sctype_id, strdup(classname));

  return errNone;
}

int prGdkScreenHeight(struct VMGlobals *g, int numArgsPushed) {
  int n = gdk_screen_height();

  SetInt(g->sp, n);

  return errNone;
}

int prGdkScreenWidth(struct VMGlobals *g, int numArgsPushed) {
  int n = gdk_screen_width();

  SetInt(g->sp, n);

  return errNone;
}

static GdkDevice *get_gdk_device_by_name(const char *name) {
  GList *deviceList = gdk_devices_list(), *item;
  for (item = deviceList; item != 0; item = item->next) {
    GdkDevice *device = (GdkDevice *)(item->data);
    if (strcmp(device->name, name) == 0) {
      return device;
    }
  }
  return 0;
}

int prGdkSetDeviceModeByName(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 2;
  PyrSlot *deviceSlot = g->sp - 1;
  PyrSlot *modeSlot = g->sp - 0;

  char deviceName[128];
  if (slotStrVal(deviceSlot, deviceName, 128)) return errWrongType;

  GdkDevice *device = get_gdk_device_by_name(deviceName);
  if (device == 0) {
    error("Unknown device with the give name\n");
    dumpObjectSlot(deviceSlot);
    SetFalse(classSlot);
    return errNone;
    /*    return errFailed; */
  }

  int mode;
  if (slotEnumVal(GDK_TYPE_INPUT_MODE, modeSlot, mode))
    return errWrongType;
  
  gdk_device_set_mode(device, GdkInputMode(mode));

  SetTrue(classSlot);
  return errNone;
}

int prGdkGCNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 1;
  PyrSlot *drawableSlot = g->sp;

  if (!GDK_IS_DRAWABLE(drawableSlot->uo->slots[0].uptr)) return errWrongType;
  GdkDrawable *drawable = GDK_DRAWABLE(drawableSlot->uo->slots[0].uptr);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GdkGCNewRequest(drawable));

  return errNone;
}

int prGdkGCSetRgbFgColorByName(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *nameSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GDK_IS_GC(objectSlot->uo->slots[0].uptr)) return errWrongType;
  GdkGC *gc = GDK_GC(objectSlot->uo->slots[0].uptr);

  const char *namePtr;
  int nameSize;
  if (slotStrPtrLen(nameSlot, &namePtr, &nameSize, SMALL_STRING_SIZE)) return errWrongType;
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GdkGCSetRgbFgColorByNameRequest(gc, namePtr, nameSize));

  return errNone;
}

int prGdkGCSetRgbBgColorByName(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *nameSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GDK_IS_GC(objectSlot->uo->slots[0].uptr)) return errWrongType;
  GdkGC *gc = GDK_GC(objectSlot->uo->slots[0].uptr);

  const char *namePtr;
  int nameSize;
  if (slotStrPtrLen(nameSlot, &namePtr, &nameSize, SMALL_STRING_SIZE)) return errWrongType;
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GdkGCSetRgbBgColorByNameRequest(gc, namePtr, nameSize));

  return errNone;
}

int prGdkGCSetRgbFgColor(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *colorSlot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "gc", "GdkGC", GdkGC, gc, GDK_GC, GDK_IS_GC);
  UNWRAP_GDKCOLOR(colorSlot, "color", color);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GdkGCSetRgbFgColorRequest(gc, color));

  return errNone;
}

int prGdkGCSetRgbBgColor(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *colorSlot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "gc", "GdkGC", GdkGC, gc, GDK_GC, GDK_IS_GC);
  UNWRAP_GDKCOLOR(colorSlot, "color", color);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GdkGCSetRgbBgColorRequest(gc, color));

  return errNone;
}

int prGdkGCSetLineAttributes(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 4;
  PyrSlot *widthSlot = g->sp - 3;
  PyrSlot *styleSlot = g->sp - 2;
  PyrSlot *capSlot = g->sp - 1;
  PyrSlot *joinSlot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "gc", "GdkGC", GdkGC, gc, GDK_GC, GDK_IS_GC);

  int width;
  if (slotIntVal(widthSlot, &width)) WRONG_TYPE_PARAMETER("width", "Integer", widthSlot);

  UNWRAP_GENUM(styleSlot, "style", GdkLineStyle, style, GDK_TYPE_LINE_STYLE);
  UNWRAP_GENUM(capSlot, "cap", GdkCapStyle, cap, GDK_TYPE_CAP_STYLE);
  UNWRAP_GENUM(joinSlot, "join", GdkJoinStyle, join, GDK_TYPE_JOIN_STYLE);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GdkGCSetLineAttributesRequest(gc, width, style, cap, join));

  return errNone;
}

int prGdkWindowSetCursor(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *cursorSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GDK_IS_WINDOW(objectSlot->uo->slots[0].uptr)) return errWrongType;
  GdkWindow *window = GDK_WINDOW(objectSlot->uo->slots[0].uptr);

  int type;
  if (slotEnumVal(GDK_TYPE_CURSOR_TYPE, cursorSlot, type)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GdkWindowSetCursorRequest(window, GdkCursorType(type)));

  return errNone;
}

int prGdkDrawableDrawRectangle(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 6;
  PyrSlot *gcSlot = g->sp - 5;
  PyrSlot *filledSlot = g->sp - 4;
  PyrSlot *xSlot = g->sp - 3;
  PyrSlot *ySlot = g->sp - 2;
  PyrSlot *wSlot = g->sp - 1;
  PyrSlot *hSlot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "drawable", "GdkDrawable", GdkDrawable, drawable, GDK_DRAWABLE, GDK_IS_DRAWABLE);
  UNWRAP_GOBJECT(gcSlot, "gc", "GdkGC", GdkGC, gc, GDK_GC, GDK_IS_GC);

  bool filled;
  if (slotBoolVal(filledSlot, filled)) WRONG_TYPE_PARAMETER("filled", "Boolean", filledSlot); 
 
  int x, y, w, h;
  if (slotIntVal(xSlot, &x)) WRONG_TYPE_PARAMETER("x", "Integer", xSlot);
  if (slotIntVal(ySlot, &y)) WRONG_TYPE_PARAMETER("y", "Integer", ySlot);
  if (slotIntVal(wSlot, &w)) WRONG_TYPE_PARAMETER("w", "Integer", wSlot);
  if (slotIntVal(hSlot, &h)) WRONG_TYPE_PARAMETER("h", "Integer", hSlot);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GdkDrawableDrawRectangleRequest(drawable, gc, filled, x, y, w, h));

  return errNone;
}

int prGdkDrawableDrawArc(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 8;
  PyrSlot *gcSlot = g->sp - 7;
  PyrSlot *filledSlot = g->sp - 6;
  PyrSlot *xSlot = g->sp - 5;
  PyrSlot *ySlot = g->sp - 4;
  PyrSlot *wSlot = g->sp - 3;
  PyrSlot *hSlot = g->sp - 2;
  PyrSlot *angle1Slot = g->sp - 1;
  PyrSlot *angle2Slot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "drawable", "GdkDrawable", GdkDrawable, drawable, GDK_DRAWABLE, GDK_IS_DRAWABLE);
  UNWRAP_GOBJECT(gcSlot, "gc", "GdkGC", GdkGC, gc, GDK_GC, GDK_IS_GC);

  bool filled;
  if (slotBoolVal(filledSlot, filled)) WRONG_TYPE_PARAMETER("filled", "Boolean", filledSlot); 
 
  int x, y, w, h, angle1, angle2;
  if (slotIntVal(xSlot, &x)) WRONG_TYPE_PARAMETER("x", "Integer", xSlot);
  if (slotIntVal(ySlot, &y)) WRONG_TYPE_PARAMETER("y", "Integer", ySlot);
  if (slotIntVal(wSlot, &w)) WRONG_TYPE_PARAMETER("w", "Integer", wSlot);
  if (slotIntVal(hSlot, &h)) WRONG_TYPE_PARAMETER("h", "Integer", hSlot);
  if (slotIntVal(angle1Slot, &angle1)) WRONG_TYPE_PARAMETER("angle1", "Integer", wSlot);
  if (slotIntVal(angle2Slot, &angle2)) WRONG_TYPE_PARAMETER("angle2", "Integer", hSlot);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GdkDrawableDrawArcRequest(drawable, gc, filled, x, y, w, h, angle1, angle2));

  return errNone;
}

int prGdkDrawableDrawLine(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 5;
  PyrSlot *gcSlot = g->sp - 4;
  PyrSlot *x1Slot = g->sp - 3;
  PyrSlot *y1Slot = g->sp - 2;
  PyrSlot *x2Slot = g->sp - 1;
  PyrSlot *y2Slot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "drawable", "GdkDrawable", GdkDrawable, drawable, GDK_DRAWABLE, GDK_IS_DRAWABLE);
  UNWRAP_GOBJECT(gcSlot, "gc", "GdkGC", GdkGC, gc, GDK_GC, GDK_IS_GC);

  int x1, y1, x2, y2;
  if (slotIntVal(x1Slot, &x1)) WRONG_TYPE_PARAMETER("x1", "Integer", x1Slot);
  if (slotIntVal(y1Slot, &y1)) WRONG_TYPE_PARAMETER("y1", "Integer", y1Slot);
  if (slotIntVal(x2Slot, &x2)) WRONG_TYPE_PARAMETER("x2", "Integer", x2Slot);
  if (slotIntVal(y2Slot, &y2)) WRONG_TYPE_PARAMETER("y2", "Integer", y2Slot);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GdkDrawableDrawLineRequest(drawable, gc, x1, y1, x2, y2));

  return errNone;
}

int prGdkDrawableDrawPoint(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 3;
  PyrSlot *gcSlot = g->sp - 2;
  PyrSlot *xSlot = g->sp - 1;
  PyrSlot *ySlot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "drawable", "GdkDrawable", GdkDrawable, drawable, GDK_DRAWABLE, GDK_IS_DRAWABLE);
  UNWRAP_GOBJECT(gcSlot, "gc", "GdkGC", GdkGC, gc, GDK_GC, GDK_IS_GC);

  int x, y;
  if (slotIntVal(xSlot, &x)) WRONG_TYPE_PARAMETER("x", "Integer", xSlot);
  if (slotIntVal(ySlot, &y)) WRONG_TYPE_PARAMETER("y", "Integer", ySlot);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GdkDrawableDrawPointRequest(drawable, gc, x, y));

  return errNone;
}


int prGdkDrawableDrawLayout(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 4;
  PyrSlot *gcSlot = g->sp - 3;
  PyrSlot *xSlot = g->sp - 2;
  PyrSlot *ySlot = g->sp - 1;
  PyrSlot *layoutSlot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "drawable", "GdkDrawable", GdkDrawable, drawable, GDK_DRAWABLE, GDK_IS_DRAWABLE);
  UNWRAP_GOBJECT(gcSlot, "gc", "GdkGC", GdkGC, gc, GDK_GC, GDK_IS_GC);

  int x, y;
  if (slotIntVal(xSlot, &x)) WRONG_TYPE_PARAMETER("x", "Integer", xSlot);
  if (slotIntVal(ySlot, &y)) WRONG_TYPE_PARAMETER("y", "Integer", ySlot);

  UNWRAP_GOBJECT(layoutSlot, "layout", "PangoLayout", PangoLayout, layout, PANGO_LAYOUT, PANGO_IS_LAYOUT);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GdkDrawableDrawLayoutRequest(drawable, gc, x, y, layout));

  return errNone;
}

int prGdkDrawableGetSize(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GDK_IS_DRAWABLE(objectSlot->uo->slots[0].uptr)) return errWrongType;
  GdkDrawable *drawable = GDK_DRAWABLE(objectSlot->uo->slots[0].uptr);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GdkDrawableGetSizeRequest(drawable));
  
  return errNone;
}

int prGdkEventGetCoords(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GDKEVENT(objectSlot, "event", event);

  double x, y;
  if (gdk_event_get_coords(event, &x, &y)) {
    PyrObject *array = newPyrArray(g->gc, 2, 0, false);
    PyrSlot *slots = array->slots;
    
    SetFloat(slots + 0, x);
    SetFloat(slots + 1, y);
    
    array->size = 2;
    
    SetObject(objectSlot, array);
  } else {
    SetNil(objectSlot);
  }

  return errNone;
}

int prGdkEventGetRootCoords(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GDKEVENT(objectSlot, "event", event);

  double x, y;
  if (gdk_event_get_root_coords(event, &x, &y)) {
    PyrObject *array = newPyrArray(g->gc, 2, 0, false);
    PyrSlot *slots = array->slots;
    
    SetFloat(slots + 0, x);
    SetFloat(slots + 1, y);
    
    array->size = 2;
    
    SetObject(objectSlot, array);
  } else {
    SetNil(objectSlot);
  }

  return errNone;
}

int prGdkEventGetState(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GDKEVENT(objectSlot, "event", event);

  GdkModifierType state;
  
  if (gdk_event_get_state(event, &state)) {
    SetInt(objectSlot, (int)state);
  } else {
    SetNil(objectSlot);
  }

  return errNone;
}

int prGdkEventGetKeyval(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GDKEVENT(objectSlot, "event", event);

  if (event->type == GDK_KEY_PRESS || event->type == GDK_KEY_RELEASE) {
    GdkEventKey *eventkey = (GdkEventKey *)event;
    SetInt(objectSlot, eventkey->keyval);
  } else {
    SetNil(objectSlot);
  }

  return errNone;
}

int prGdkEventGetKeyname(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GDKEVENT(objectSlot, "event", event);

  if (event->type == GDK_KEY_PRESS || event->type == GDK_KEY_RELEASE) {
    GdkEventKey *eventkey = (GdkEventKey *)event;
    const char *keyname;
   
    keyname = gdk_keyval_name(eventkey->keyval);
    if (keyname == 0) {
      SetNil(objectSlot);
    } else {
      int size = strlen(keyname);
      PyrString *strobj = newPyrStringN(g->gc, size, 0, true);
      memcpy(strobj->s, keyname, size);
      SetObject(objectSlot, strobj);
    }
  } else {
    SetNil(objectSlot);
  }

  return errNone;
}

int prGdkEventGetTime(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GDKEVENT(objectSlot, "event", event);

  guint32 time = gdk_event_get_time(event);
  
  SetInt(objectSlot, (int)time);
  
  return errNone;
}

int prGdkEventGetAxis(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *nSlot = g->sp;

  UNWRAP_GDKEVENT(objectSlot, "event", event);

  int n;
  if (slotIntVal(nSlot, &n)) return errWrongType;

  gdouble value;

  if (gdk_event_get_axis(event, GdkAxisUse(n), &value))
    SetFloat(objectSlot, value);
  else
    SetNil(objectSlot);
  
  return errNone;
}

int prGdkEventGetButton(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GDKEVENT(objectSlot, "event", event);

  guint32 time;
  
  if (event->type == GDK_BUTTON_PRESS || event->type == GDK_BUTTON_PRESS ||
      event->type == GDK_2BUTTON_PRESS || event->type == GDK_3BUTTON_PRESS ||
      event->type == GDK_BUTTON_RELEASE) {
    GdkEventButton *buttonEvent = (GdkEventButton *)event;
    SetInt(objectSlot, buttonEvent->button);
  } else {
    SetNil(objectSlot);
  }

  return errNone;
}

int prGdkEventIsHint(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GDKEVENT(objectSlot, "event", event);

  if (event->type == GDK_MOTION_NOTIFY) {
    GdkEventMotion *motionEvent = (GdkEventMotion *)event;
    SetBool(objectSlot, motionEvent->is_hint);
  } else {
    SetNil(objectSlot);
  }

  return errNone;
}

int prGdkEventGetDeviceState(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GDKEVENT(objectSlot, "event", event);

  if (event->type == GDK_MOTION_NOTIFY) {
    GdkEventMotion *motionEvent = (GdkEventMotion *)event;
    GdkDevice *device;
    GdkModifierType state;
    gdk_device_get_state(motionEvent->device, motionEvent->window, NULL, &state);

    SetInt(objectSlot, int(state));
  } else {
    SetNil(objectSlot);
  }

  return errNone;
}

int prGdkEventGetType(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GDKEVENT(objectSlot, "event", event);

  SetEnumSlot(objectSlot, GDK_TYPE_EVENT_TYPE, event->type);

  return errNone;
}

int prGdkPixmapCreateFromXpm(struct VMGlobals *g, int numArgsPushed) {
  return errNone;
}

// ------ GtkWidget

int prGtkWidgetSetSizeRequest(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *wSlot = g->sp - 1;
  PyrSlot *hSlot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);

  int w, h;
  if (slotIntVal(wSlot, &w)) WRONG_TYPE_PARAMETER("width", "Integer", wSlot);
  if (slotIntVal(hSlot, &h)) WRONG_TYPE_PARAMETER("height", "Integer", hSlot);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetSetSizeRequestRequest(widget, w, h));

  return errNone;
}

int prGtkWidgetShow(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetShowRequest(widget));

  return errNone;
}

int prGtkWidgetShowAll(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetShowAllRequest(widget));

  return errNone;
}

int prGtkWidgetHide(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetHideRequest(widget));

  return errNone;
}

int prGtkWidgetHideAll(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetHideAllRequest(widget));

  return errNone;
}

int prGtkWidgetDestroy(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetDestroyRequest(widget));

  return errNone;
}

int prGtkWidgetQueueDraw(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetQueueDrawRequest(widget));

  return errNone;
}

int prGtkWidgetQueueResize(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetQueueResizeRequest(widget));

  return errNone;
}

int prGtkWidgetSetEvents(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *eventsSlot = g->sp;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);
  
  int eventmask;
  if (slotFlagsVal(GDK_TYPE_EVENT_MASK, eventsSlot, eventmask)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetSetEventsRequest(widget, GdkEventMask(eventmask)));

  return errNone;
}

int prGtkWidgetAddEvents(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *eventsSlot = g->sp;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);

  int eventmask;
  if (slotFlagsVal(GDK_TYPE_EVENT_MASK, eventsSlot, eventmask)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetAddEventsRequest(widget, GdkEventMask(eventmask)));

  return errNone;
}

int prGtkWidgetSetExtensionEvents(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *modeSlot = g->sp;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);

  char name[32];;
  if (slotStrVal(modeSlot, name, 32)) return errWrongType;

  int mode;
  if (slotEnumVal(GDK_TYPE_EXTENSION_MODE, modeSlot, mode))
    return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetSetExtensionEventsRequest(widget, GdkExtensionMode(mode)));
  
  return errNone;
}

int prGtkWidgetWindow(struct VMGlobals *g, int numArgsPushed) {
  if (!g->canCallOS) return errCantCallOS;
  
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);
  
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetGetWindowRequest(widget));
  
  return errNone;
}

int prGtkWidgetGetStyle(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetGetStyleRequest(widget));
  
  return errNone;
}

int prGtkWidgetGetPangoContext(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetGetPangoContextRequest(widget));

  return errNone;
}

int prGtkWidgetIsRealized(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetIsRealizedRequest(widget));

  return errNone;
}

int prGtkWidgetIsToplevel(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetIsToplevelRequest(widget));

  return errNone;
}

int prGtkWidgetIsMapped(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetIsMappedRequest(widget));

  return errNone;
}

int prGtkWidgetIsVisible(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetIsVisibleRequest(widget));

  return errNone;
}

int prGtkWidgetCanFocus(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetCanFocusRequest(widget));

  return errNone;
}

int prGtkWidgetGetDoubleBuffered(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetGetDoubleBufferedRequest(widget));

  return errNone;
}

int prGtkWidgetSetDoubleBuffered(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *valueSlot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);

  bool value;
  if (slotBoolVal(valueSlot, value)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetSetDoubleBufferedRequest(widget, value));

  return errNone;
}

int prGtkWidgetGetSensitive(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetGetSensitiveRequest(widget));

  return errNone;
}

int prGtkWidgetSetSensitive(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *valueSlot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);

  bool value;
  if (slotBoolVal(valueSlot, value)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetSetSensitiveRequest(widget, value));

  return errNone;
}


int prGtkWidgetModifyBg(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *stateSlot = g->sp - 1;
  PyrSlot *colorSlot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);
  UNWRAP_GENUM(stateSlot, "state", GtkStateType, state, GTK_TYPE_STATE_TYPE);
  UNWRAP_GDKCOLOR(colorSlot, "color", color);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetModifyBgRequest(widget, state, color));

  return errNone;
}

int prGtkWidgetGrabFocus(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetGrabFocusRequest(widget));

  return errNone;
}

int prGtkWidgetGetParent(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetGetParentRequest(widget));

  return errNone;
}

int prGtkWidgetSetParent(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *parentSlot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);
  UNWRAP_GOBJECT(parentSlot, "parent", "GtkWidget", GtkWidget, parent, GTK_WIDGET, GTK_IS_WIDGET);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetSetParentRequest(widget, parent));
  
  return errNone;
}

int prGtkWidgetUnparent(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetUnparentRequest(widget));
  
  return errNone;
}

int prGtkWidgetGetParentWindow(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetGetParentWindowRequest(widget));

  return errNone;
}

int prGtkWidgetGetSizeRequest(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetGetSizeRequestRequest(widget));
  
  return errNone;
}

int prGtkWidgetGetToplevel(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetGetToplevelRequest(widget));
  
  return errNone;
}

int prGtkWidgetSetState(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *stateSlot = g->sp;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);
  
  int state;
  if (slotEnumVal(GTK_TYPE_STATE_TYPE, stateSlot, state)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetSetStateRequest(widget, GtkStateType(state)));

  return errNone;
}

int prGtkWidgetReparent(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *newParentSlot = g->sp;

  UNWRAP_GOBJECT(objectSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);
  UNWRAP_GOBJECT(newParentSlot, "newParent", "GtkWidget", GtkWidget, newParent, GTK_WIDGET, GTK_IS_WIDGET);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWidgetReparentRequest(widget, newParent));

  return errNone;
}

int prGtkStyleGetWhiteGC(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_STYLE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkStyle *style = GTK_STYLE(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkStyleGetWhiteGCRequest(style));

  return errNone;
}
int prGtkStyleGetBlackGC(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_STYLE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkStyle *style = GTK_STYLE(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkStyleGetBlackGCRequest(style));

  return errNone;
}

int prGtkContainerAdd(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *childSlot = g->sp;

  UNWRAP_GOBJECT(objectSlot, "object", "GtkContainer", GtkContainer, container, GTK_CONTAINER, GTK_IS_CONTAINER);
  UNWRAP_GOBJECT(childSlot, "child", "GtkWidget", GtkWidget, child, GTK_WIDGET, GTK_IS_WIDGET);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkContainerAddRequest(container, child));

  return errNone;
}

int prGtkContainerRemove(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *childSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_CONTAINER(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkContainer *container = GTK_CONTAINER(objectSlot->uo->slots[0].ui);

  SC_CHECK_VALID_GOBJECT_WRAPPER(childSlot);
  if (!GTK_IS_WIDGET(childSlot->uo->slots[0].ui)) return errWrongType;
  GtkWidget *child = GTK_WIDGET(childSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkContainerRemoveRequest(container, child));

  return errNone;
}

int prGtkContainerChildren(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_CONTAINER(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkContainer *container = GTK_CONTAINER(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkContainerChildrenRequest(container));

  return errNone;
}

int prGtkContainerNChildren(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_CONTAINER(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkContainer *container = GTK_CONTAINER(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkContainerNChildrenRequest(container));

  return errNone;
}

int prGtkContainerSetBorderWidth(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_CONTAINER(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkContainer *container = GTK_CONTAINER(objectSlot->uo->slots[0].ui);

  int n;
  if (slotIntVal(g->sp, &n)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkContainerSetBorderWidthRequest(container, n));  

  return errNone;
}

int prGtkBinGetChild(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_BIN(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkBin *bin = GTK_BIN(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkBinGetChildRequest(bin));

  return errNone;
}

int prGtkBoxPackStart(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 4;
  PyrSlot *childSlot = g->sp - 3;
  PyrSlot *expandSlot = g->sp - 2;
  PyrSlot *fillSlot = g->sp - 1;
  PyrSlot *paddingSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_BOX(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkBox *box = GTK_BOX(objectSlot->uo->slots[0].ui);

  SC_CHECK_VALID_GOBJECT_WRAPPER(childSlot);
  if (!GTK_IS_WIDGET(childSlot->uo->slots[0].ui)) return errWrongType;
  GtkWidget *child = GTK_WIDGET(childSlot->uo->slots[0].ui);

  bool expand, fill;
  if (slotBoolVal(expandSlot, expand)) return errWrongType;
  if (slotBoolVal(fillSlot, fill)) return errWrongType;

  int padding;
  if (slotIntVal(paddingSlot, &padding)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkBoxPackStartRequest(box, child, expand, fill, padding));

  return errNone;
}

int prGtkBoxPackEnd(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 4;
  PyrSlot *childSlot = g->sp - 3;
  PyrSlot *expandSlot = g->sp - 2;
  PyrSlot *fillSlot = g->sp - 1;
  PyrSlot *paddingSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_BOX(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkBox *box = GTK_BOX(objectSlot->uo->slots[0].ui);

  SC_CHECK_VALID_GOBJECT_WRAPPER(childSlot);
  if (!GTK_IS_WIDGET(childSlot->uo->slots[0].ui)) return errWrongType;
  GtkWidget *child = GTK_WIDGET(childSlot->uo->slots[0].ui);

  bool expand, fill;
  if (slotBoolVal(expandSlot, expand)) return errWrongType;
  if (slotBoolVal(fillSlot, fill)) return errWrongType;

  int padding;
  if (slotIntVal(paddingSlot, &padding)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkBoxPackEndRequest(box, child, expand, fill, padding));

  return errNone;
}

int prGtkBoxReorderChild(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *childSlot = g->sp - 1;
  PyrSlot *nSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_BOX(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkBox *box = GTK_BOX(objectSlot->uo->slots[0].ui);

  SC_CHECK_VALID_GOBJECT_WRAPPER(childSlot);
  if (!GTK_IS_WIDGET(childSlot->uo->slots[0].ui)) return errWrongType;
  GtkWidget *child = GTK_WIDGET(childSlot->uo->slots[0].ui);

  int n;
  if (slotIntVal(nSlot, &n)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkBoxReorderChildRequest(box, child, n));

  return errNone;
}

int prGtkHBoxNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 2;
  PyrSlot *homoSlot = g->sp - 1;
  PyrSlot *spacingSlot = g->sp - 0;

  bool homogeneous;
  if (slotBoolVal(homoSlot, homogeneous)) return errWrongType;
  int spacing;
  if (slotIntVal(spacingSlot, &spacing)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkHBoxNewRequest(homogeneous, spacing));

  return errNone;
}

int prGtkVBoxNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 2;
  PyrSlot *homoSlot = g->sp - 1;
  PyrSlot *spacingSlot = g->sp - 0;

  bool homogeneous;
  if (slotBoolVal(homoSlot, homogeneous)) return errWrongType;
  int spacing;
  if (slotIntVal(spacingSlot, &spacing)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkVBoxNewRequest(homogeneous, spacing));

  return errNone;
}

int prGtkColorSelectionNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *resultSlot = g->sp - 0;

   GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkColorSelectionNewRequest());

  return errNone;
}

int prGtkColorSelectionSetCurrentColor(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *colorSlot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "selection", "GtkColorSelection", GtkColorSelection, selection, GTK_COLOR_SELECTION, GTK_IS_COLOR_SELECTION);
  UNWRAP_GDKCOLOR(colorSlot, "color", color);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkColorSelectionSetCurrentColorRequest(selection, color));

  return errNone;
}

int prGtkColorSelectionSetUpdatePolicy(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *policySlot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "selection", "GtkColorSelection", GtkColorSelection, selection, GTK_COLOR_SELECTION, GTK_IS_COLOR_SELECTION);
  UNWRAP_GENUM(policySlot, "policy", GtkUpdateType, policy, GTK_TYPE_UPDATE_TYPE);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkColorSelectionSetUpdatePolicyRequest(selection, policy));

  return errNone;
}

int prGtkColorSelectionGetCurrentColor(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_COLOR_SELECTION(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkColorSelection *selection = GTK_COLOR_SELECTION(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkColorSelectionGetCurrentColorRequest(selection));

  return errNone;
}

int prGtkColorSelectionSetHasOpacityControl(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *numericSlot = g->sp - 0;
 
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_COLOR_SELECTION(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkColorSelection *selection = GTK_COLOR_SELECTION(objectSlot->uo->slots[0].ui); 

  bool value;
  if (slotBoolVal(numericSlot, value)) return errWrongType;
 
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkColorSelectionSetHasOpacityControlRequest(selection, value));

  return errNone;
}

int prGtkColorSelectionGetHasOpacityControl(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_COLOR_SELECTION(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkColorSelection *selection = GTK_COLOR_SELECTION(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkColorSelectionGetHasOpacityControlRequest(selection));

  return errNone;
}

int prGtkFixedNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 0;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkFixedNewRequest());

  return errNone;
}

int prGtkFixedPut(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 3;
  PyrSlot *widgetSlot = g->sp - 2;
  PyrSlot *xSlot = g->sp - 1;
  PyrSlot *ySlot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "object", "GtkFixed", GtkFixed, fixed, GTK_FIXED, GTK_IS_FIXED);
  UNWRAP_GOBJECT(widgetSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);

  int x, y;
  if (slotIntVal(xSlot, &x)) WRONG_TYPE_PARAMETER("x", "Integer", xSlot);
  if (slotIntVal(ySlot, &y)) WRONG_TYPE_PARAMETER("y", "Integer", ySlot);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkFixedPutRequest(fixed, widget, x, y));
 
  return errNone;
}

int prGtkFixedMove(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 3;
  PyrSlot *widgetSlot = g->sp - 2;
  PyrSlot *xSlot = g->sp - 1;
  PyrSlot *ySlot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "object", "GtkFixed", GtkFixed, fixed, GTK_FIXED, GTK_IS_FIXED);
  UNWRAP_GOBJECT(widgetSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);

  int x, y;
  if (slotIntVal(xSlot, &x)) WRONG_TYPE_PARAMETER("x", "Integer", xSlot);
  if (slotIntVal(ySlot, &y)) WRONG_TYPE_PARAMETER("y", "Integer", ySlot);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkFixedMoveRequest(fixed, widget, x, y));
 
  return errNone;
}

int prGtkFixedSetHasWindow(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *valueSlot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "object", "GtkFixed", GtkFixed, fixed, GTK_FIXED, GTK_IS_FIXED);
  bool value;
  if (slotBoolVal(valueSlot, value)) WRONG_TYPE_PARAMETER("value", "Bool", valueSlot);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkFixedSetHasWindowRequest(fixed, value));
 
  return errNone;
}

int prGtkFixedGetHasWindow(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "object", "GtkFixed", GtkFixed, fixed, GTK_FIXED, GTK_IS_FIXED);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkFixedGetHasWindowRequest(fixed));
 
  return errNone;
}

int prGtkWindowNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *resultSlot = g->sp - 1;
  PyrSlot *typeSlot = g->sp;

  int type;
  if (slotEnumVal(GTK_TYPE_WINDOW_TYPE, typeSlot, type)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWindowNewRequest(GtkWindowType(type)));

  return errNone;
}

int prGtkWindowSetTitle(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *titleSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_WINDOW(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkWindow *window = GTK_WINDOW(objectSlot->uo->slots[0].ui);

  const char *titlePtr;
  int titleSize;
  if (slotStrPtrLen(titleSlot, &titlePtr, &titleSize, MEDIUM_STRING_SIZE)) return errWrongType;
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWindowSetTitleRequest(window, titlePtr, titleSize));

  return errNone;
}

int prGtkWindowGetTitle(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_WINDOW(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkWindow *window = GTK_WINDOW(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWindowGetTitleRequest(window));

  return errNone;
}

int prGtkWindowSetDefaultSize(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_WINDOW(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkWindow *window = GTK_WINDOW(objectSlot->uo->slots[0].ui);

  int w, h;
  if (slotIntVal(g->sp - 1, &w)) return errWrongType;
  if (slotIntVal(g->sp - 0, &h)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWindowSetDefaultSizeRequest(window, w, h));

  return errNone;
}

int prGtkWindowResize(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_WINDOW(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkWindow *window = GTK_WINDOW(objectSlot->uo->slots[0].ui);

  int w, h;
  if (slotIntVal(g->sp - 1, &w)) return errWrongType;
  if (slotIntVal(g->sp - 0, &h)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWindowResizeRequest(window, w, h));

  return errNone;
}

int prGtkWindowGetDefaultSize(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_WINDOW(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkWindow *window = GTK_WINDOW(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWindowGetDefaultSizeRequest(window));
  
  return errNone;
}

int prGtkWindowGetSize(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_WINDOW(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkWindow *window = GTK_WINDOW(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWindowGetSizeRequest(window));
  
  return errNone;
}

int prGtkWindowMove(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_WINDOW(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkWindow *window = GTK_WINDOW(objectSlot->uo->slots[0].ui);

  int x, y;
  if (slotIntVal(g->sp - 1, &x)) return errWrongType;
  if (slotIntVal(g->sp - 0, &y)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWindowMoveRequest(window, x, y));

  return errNone;
}

int prGtkWindowSetPosition(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *positionSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_WINDOW(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkWindow *window = GTK_WINDOW(objectSlot->uo->slots[0].ui);

  int position;
  if (slotEnumVal(GTK_TYPE_WINDOW_POSITION, positionSlot, position)) {
    return errWrongType;
  }

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWindowSetPositionRequest(window, GtkWindowPosition(position)));

  return errNone;
}

int prGtkWindowSetFocus(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *windowSlot = g->sp - 1;
  PyrSlot *widgetSlot = g->sp;

  UNWRAP_GOBJECT(windowSlot, "object", "GtkWindow", GtkWindow, window, GTK_WINDOW, GTK_IS_WINDOW);

  GtkWidget *widget;

  if (IsNil(widgetSlot)) {
	widget = 0;
  } else {
	UNWRAP_GOBJECT(widgetSlot, "widget", "GtkWidget", GtkWidget, widget0, GTK_WIDGET, GTK_IS_WIDGET);
	widget = widget0;
  }

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWindowSetFocusRequest(window, widget));

  return errNone;
}

int prGtkWindowGetFocus(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *windowSlot = g->sp;

  UNWRAP_GOBJECT(windowSlot, "object", "GtkWindow", GtkWindow, window, GTK_WINDOW, GTK_IS_WINDOW);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWindowGetFocusRequest(window));

  return errNone;
}

int prGtkWindowSetTransientFor(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *windowSlot = g->sp - 1;
  PyrSlot *parentSlot = g->sp;

  UNWRAP_GOBJECT(windowSlot, "object", "GtkWindow", GtkWindow, window, GTK_WINDOW, GTK_IS_WINDOW);
  UNWRAP_GOBJECT(parentSlot, "parent", "GtkWindow", GtkWindow, parent, GTK_WINDOW, GTK_IS_WINDOW);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWindowSetTransientForRequest(window, parent));

  return errNone;
}

int prGtkWindowSetDestroyWithParent(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *windowSlot = g->sp - 1;
  PyrSlot *valueSlot = g->sp;

  UNWRAP_GOBJECT(windowSlot, "object", "GtkWindow", GtkWindow, window, GTK_WINDOW, GTK_IS_WINDOW);

  bool value;
  if (slotBoolVal(valueSlot, value)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWindowSetDestroyWithParentRequest(window, value));

  return errNone;
}

int prGtkWindowSetHideOnDelete(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_WINDOW(objectSlot->uo->slots[0].ui)) return errWrongType;

  GtkWindow *window = GTK_WINDOW(objectSlot->uo->slots[0].ui);
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkWindowSetHideOnDeleteRequest(window));

  return errNone;
}

int prGtkFileSelectionNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 1;
  PyrSlot *titleSlot = g->sp - 0;

  const char *titlePtr;
  int titleSize;
  if (slotStrPtrLen(titleSlot, &titlePtr, &titleSize, MEDIUM_STRING_SIZE)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkFileSelectionNewRequest(titlePtr, titleSize));

  return errNone;
}

int prGtkFileSelectionSetFilename(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *filenameSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_FILE_SELECTION(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkFileSelection *fileselection = GTK_FILE_SELECTION(objectSlot->uo->slots[0].ui);

  const char *filenamePtr;
  int filenameSize;
  if (slotStrPtrLen(filenameSlot, &filenamePtr, &filenameSize, FILENAME_STRING_SIZE)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkFileSelectionSetFilenameRequest(fileselection, filenamePtr, filenameSize));  

  return errNone;
}

int prGtkFileSelectionGetFilename(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_FILE_SELECTION(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkFileSelection *fileselection = GTK_FILE_SELECTION(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkFileSelectionGetFilenameRequest(fileselection));  

  return errNone;
}

int prGtkFileSelectionShowFileopButtons(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_FILE_SELECTION(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkFileSelection *fileselection = GTK_FILE_SELECTION(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkFileSelectionShowFileopButtonsRequest(fileselection));

  return errNone;
}

int prGtkFileSelectionHideFileopButtons(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_FILE_SELECTION(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkFileSelection *fileselection = GTK_FILE_SELECTION(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkFileSelectionHideFileopButtonsRequest(fileselection));

  return errNone;
}

int prGtkFileSelectionGetSelections(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_FILE_SELECTION(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkFileSelection *fileselection = GTK_FILE_SELECTION(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkFileSelectionGetSelectionsRequest(fileselection));

  return errNone;
}

int prGtkFileSelectionSetSelectMultiple(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *valueSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_FILE_SELECTION(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkFileSelection *fileselection = GTK_FILE_SELECTION(objectSlot->uo->slots[0].ui);

  bool value;
  if (slotBoolVal(valueSlot, value)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkFileSelectionSetSelectMultipleRequest(fileselection, value));  

  return errNone;
}

int prGtkFileSelectionGetSelectMultiple(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_FILE_SELECTION(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkFileSelection *fileselection = GTK_FILE_SELECTION(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkFileSelectionGetSelectMultipleRequest(fileselection));  

  return errNone;
}

int prGtkFileSelectionComplete(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *patternSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_FILE_SELECTION(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkFileSelection *fileselection = GTK_FILE_SELECTION(objectSlot->uo->slots[0].ui);

  const char *patternPtr;
  int patternSize;
  if (slotStrPtrLen(patternSlot, &patternPtr, &patternSize, FILENAME_STRING_SIZE)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkFileSelectionCompleteRequest(fileselection, patternPtr, patternSize));  

  return errNone;
}

int prGtkFileSelectionDirList(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_FILE_SELECTION(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkFileSelection *fileselection = GTK_FILE_SELECTION(objectSlot->uo->slots[0].ui);

  SetInt(objectSlot, int(fileselection->dir_list));

  return errNone;
}

int prGtkFileSelectionFileList(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_FILE_SELECTION(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkFileSelection *fileselection = GTK_FILE_SELECTION(objectSlot->uo->slots[0].ui);

  SetInt(objectSlot, int(fileselection->file_list));

  return errNone;
}

int prGtkFileSelectionSelectionEntry(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_FILE_SELECTION(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkFileSelection *fileselection = GTK_FILE_SELECTION(objectSlot->uo->slots[0].ui);

  SetInt(objectSlot, int(fileselection->selection_entry));

  return errNone;
}

int prGtkFileSelectionSelectionText(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_FILE_SELECTION(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkFileSelection *fileselection = GTK_FILE_SELECTION(objectSlot->uo->slots[0].ui);

  SetInt(objectSlot, int(fileselection->selection_text));

  return errNone;
}

int prGtkFileSelectionMainVBox(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_FILE_SELECTION(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkFileSelection *fileselection = GTK_FILE_SELECTION(objectSlot->uo->slots[0].ui);

  SetInt(objectSlot, int(fileselection->main_vbox));

  return errNone;
}

int prGtkFileSelectionOkButton(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_FILE_SELECTION(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkFileSelection *fileselection = GTK_FILE_SELECTION(objectSlot->uo->slots[0].ui);

  SetInt(objectSlot, int(fileselection->ok_button));

  return errNone;
}

int prGtkFileSelectionCancelButton(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_FILE_SELECTION(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkFileSelection *fileselection = GTK_FILE_SELECTION(objectSlot->uo->slots[0].ui);

  SetInt(objectSlot, int(fileselection->cancel_button));

  return errNone;
}

int prGtkFileSelectionHelpButton(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_FILE_SELECTION(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkFileSelection *fileselection = GTK_FILE_SELECTION(objectSlot->uo->slots[0].ui);

  SetInt(objectSlot, int(fileselection->help_button));

  return errNone;
}

int prGtkFrameNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 1;
  PyrSlot *textSlot = g->sp - 0;

  GtkWidget *frame;

  if (IsNil(textSlot)) {
    GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkFrameNewRequest(0, 0));
  } else {
    const char *textPtr;
    int textSize;
    if (slotStrPtrLen(textSlot, &textPtr, &textSize, MEDIUM_STRING_SIZE)) return errWrongType;
    GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkFrameNewRequest(textPtr, textSize));
  }

  return errNone;
}

int prGtkFrameSetLabel(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *textSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_FRAME(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkFrame *frame = GTK_FRAME(objectSlot->uo->slots[0].ui);

  if (IsNil(textSlot)) {
    GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkFrameSetLabelRequest(frame, 0, 0));
  } else {
    const char *textPtr;
    int textSize;
    if (slotStrPtrLen(textSlot, &textPtr, &textSize, MEDIUM_STRING_SIZE)) return errWrongType;
    GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkFrameSetLabelRequest(frame, textPtr, textSize));
  }

  return errNone;
}

int prGtkFrameSetShadowType(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *shadowTypeSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_FRAME(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkFrame *frame = GTK_FRAME(objectSlot->uo->slots[0].ui);

  int type;
  if (slotEnumVal(GTK_TYPE_SHADOW_TYPE, shadowTypeSlot, type)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkFrameSetShadowTypeRequest(frame, GtkShadowType(type)));

  return errNone;
}


int prGtkPanedAdd1(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *childSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_PANED(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkPaned *paned = GTK_PANED(objectSlot->uo->slots[0].ui);
 
  SC_CHECK_VALID_GOBJECT_WRAPPER(childSlot);
  if (!GTK_IS_WIDGET(childSlot->uo->slots[0].ui)) return errWrongType;
  GtkWidget *widget = GTK_WIDGET(childSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkPanedAdd1Request(paned, widget));

  return errNone;
}

int prGtkPanedAdd2(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *childSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_PANED(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkPaned *paned = GTK_PANED(objectSlot->uo->slots[0].ui);
 
  SC_CHECK_VALID_GOBJECT_WRAPPER(childSlot);
  if (!GTK_IS_WIDGET(childSlot->uo->slots[0].ui)) return errWrongType;
  GtkWidget *widget = GTK_WIDGET(childSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkPanedAdd2Request(paned, widget));

  return errNone;
}

int prGtkPanedPack1(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 3;
  PyrSlot *childSlot = g->sp - 2;
  PyrSlot *resizeSlot = g->sp - 1;
  PyrSlot *shrinkSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_PANED(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkPaned *paned = GTK_PANED(objectSlot->uo->slots[0].ui);
 
  SC_CHECK_VALID_GOBJECT_WRAPPER(childSlot);
  if (!GTK_IS_WIDGET(childSlot->uo->slots[0].ui)) return errWrongType;
  GtkWidget *widget = GTK_WIDGET(childSlot->uo->slots[0].ui);
 
  bool resize, shrink;
  if (slotBoolVal(resizeSlot, resize)) return errWrongType;
  if (slotBoolVal(shrinkSlot, shrink)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkPanedPack1Request(paned, widget, resize, shrink));

  return errNone;
}

int prGtkPanedPack2(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 3;
  PyrSlot *childSlot = g->sp - 2;
  PyrSlot *resizeSlot = g->sp - 1;
  PyrSlot *shrinkSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_PANED(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkPaned *paned = GTK_PANED(objectSlot->uo->slots[0].ui);
 
  SC_CHECK_VALID_GOBJECT_WRAPPER(childSlot);
  if (!GTK_IS_WIDGET(childSlot->uo->slots[0].ui)) return errWrongType;
  GtkWidget *widget = GTK_WIDGET(childSlot->uo->slots[0].ui);
 
  bool resize, shrink;
  if (slotBoolVal(resizeSlot, resize)) return errWrongType;
  if (slotBoolVal(shrinkSlot, shrink)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkPanedPack2Request(paned, widget, resize, shrink));

  return errNone;
}

int prGtkPanedSetPosition(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_PANED(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkPaned *paned = GTK_PANED(objectSlot->uo->slots[0].ui);

  int n;
  if (slotIntVal(g->sp, &n)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkPanedSetPositionRequest(paned, n));

  return errNone;
}

int prGtkPanedGetPosition(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_PANED(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkPaned *paned = GTK_PANED(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkPanedGetPositionRequest(paned));

  return errNone;
}

int prGtkHPanedNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkHPanedNewRequest());  

  return errNone;
}

int prGtkVPanedNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkVPanedNewRequest());

  return errNone;
}

int prGtkButtonNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkButtonNewRequest());

  return errNone;
}

int prGtkButtonNewWithLabel(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 1;
  PyrSlot *textSlot = g->sp - 0;

  const char *textPtr;
  int textSize;
  if (slotStrPtrLen(textSlot, &textPtr, &textSize, MEDIUM_STRING_SIZE)) return errWrongType;
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkButtonNewWithLabelRequest(textPtr, textSize));

  return errNone;
}

int prGtkButtonNewFromStock(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 1;
  PyrSlot *nameSlot = g->sp - 0;

  const char *namePtr;
  int nameSize;
  if (slotStrPtrLen(nameSlot, &namePtr, &nameSize, SMALL_STRING_SIZE)) return errWrongType;
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkButtonNewFromStockRequest(namePtr, nameSize));

  return errNone;
}

int prGtkButtonSetLabel(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *textSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_BUTTON(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkButton *button = GTK_BUTTON(objectSlot->uo->slots[0].ui);

  const char *textPtr;
  int textSize;
  if (slotStrPtrLen(textSlot, &textPtr, &textSize, MEDIUM_STRING_SIZE)) return errWrongType;
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkButtonSetLabelRequest(button, textPtr, textSize));

  return errNone;
}

int prGtkButtonGetLabel(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_BUTTON(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkButton *button = GTK_BUTTON(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkButtonGetLabelRequest(button));

  return errNone;
}

int prGtkToggleButtonNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkToggleButtonNewRequest());

  return errNone;
}

int prGtkToggleButtonNewWithLabel(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 1;
  PyrSlot *textSlot = g->sp - 0;

  const char *textPtr;
  int textSize;
  if (slotStrPtrLen(textSlot, &textPtr, &textSize, MEDIUM_STRING_SIZE)) return errWrongType;
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkToggleButtonNewWithLabelRequest(textPtr, textSize));

  return errNone;
}

int prGtkToggleButtonSetMode(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *valueSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TOGGLE_BUTTON(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkToggleButton *button = GTK_TOGGLE_BUTTON(objectSlot->uo->slots[0].ui);

  bool value;
  if (slotBoolVal(valueSlot, value)) return errWrongType;
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkToggleButtonSetModeRequest(button, value));

  return errNone;
}

int prGtkToggleButtonGetMode(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TOGGLE_BUTTON(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkToggleButton *button = GTK_TOGGLE_BUTTON(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkToggleButtonGetModeRequest(button));

  return errNone;
}

int prGtkToggleButtonSetActive(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *valueSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TOGGLE_BUTTON(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkToggleButton *button = GTK_TOGGLE_BUTTON(objectSlot->uo->slots[0].ui);

  bool value;
  if (slotBoolVal(valueSlot, value)) return errWrongType;
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkToggleButtonSetActiveRequest(button, value));

  return errNone;
}

int prGtkToggleButtonGetActive(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TOGGLE_BUTTON(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkToggleButton *button = GTK_TOGGLE_BUTTON(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkToggleButtonGetActiveRequest(button));

  return errNone;
}

int prGtkToggleButtonSetInconsistent(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *valueSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TOGGLE_BUTTON(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkToggleButton *button = GTK_TOGGLE_BUTTON(objectSlot->uo->slots[0].ui);

  bool value;
  if (slotBoolVal(valueSlot, value)) return errWrongType;
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkToggleButtonSetInconsistentRequest(button, value));

  return errNone;
}

int prGtkToggleButtonGetInconsistent(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TOGGLE_BUTTON(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkToggleButton *button = GTK_TOGGLE_BUTTON(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkToggleButtonGetInconsistentRequest(button));

  return errNone;
}

int prGtkCheckButtonNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkCheckButtonNewRequest());

  return errNone;
}

int prGtkCheckButtonNewWithLabel(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 1;
  PyrSlot *textSlot = g->sp - 0;

  const char *textPtr;
  int textSize;
  if (slotStrPtrLen(textSlot, &textPtr, &textSize, MEDIUM_STRING_SIZE)) return errWrongType;
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkCheckButtonNewWithLabelRequest(textPtr, textSize));

  return errNone;
}

int prGtkRadioButtonNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 1;
  PyrSlot *groupSlot = g->sp;

  void *group;
  if (IsNil(groupSlot)) {
    group = 0;
  } else {
    if (!IsPtr(groupSlot)) return errWrongType;
    group = (void *)groupSlot->uo->slots[0].uptr;
  }

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkRadioButtonNewRequest(group));

  return errNone;
}

int prGtkRadioButtonNewWithLabel(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 2;
  PyrSlot *groupSlot = g->sp - 1;
  PyrSlot *textSlot = g->sp - 0;

  void *group;
  if (IsNil(groupSlot)) {
    group = 0;
  } else {
    if (!IsPtr(groupSlot)) return errWrongType;
    group = (void *)groupSlot->uo->slots[0].uptr;
  }

  const char *textPtr;
  int textSize;
  if (slotStrPtrLen(textSlot, &textPtr, &textSize, MEDIUM_STRING_SIZE)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkRadioButtonNewWithLabelRequest(group, textPtr, textSize));

  return errNone;
}

int prGtkRadioButtonNewFromWidget(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 1;
  PyrSlot *otherSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(otherSlot);
  if (!GTK_IS_RADIO_BUTTON(otherSlot->uo->slots[0].ui)) return errWrongType;
  GtkRadioButton *other = GTK_RADIO_BUTTON(otherSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkRadioButtonNewFromWidgetRequest(other));

  return errNone;
}

int prGtkRadioButtonNewWithLabelFromWidget(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 2;
  PyrSlot *otherSlot = g->sp - 1;
  PyrSlot *textSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(otherSlot);
  if (!GTK_IS_RADIO_BUTTON(otherSlot->uo->slots[0].ui)) return errWrongType;
  GtkRadioButton *other = GTK_RADIO_BUTTON(otherSlot->uo->slots[0].ui);

  const char *textPtr;
  int textSize;
  if (slotStrPtrLen(textSlot, &textPtr, &textSize, MEDIUM_STRING_SIZE)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkRadioButtonNewWithLabelFromWidgetRequest(other, textPtr, textSize));

  return errNone;
}

int prGtkRadioButtonGetGroup(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_RADIO_BUTTON(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkRadioButton *button = GTK_RADIO_BUTTON(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkRadioButtonGetGroupRequest(button));

  return errNone;
}

int prGtkRadioButtonSetGroup(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *groupSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_RADIO_BUTTON(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkRadioButton *button = GTK_RADIO_BUTTON(objectSlot->uo->slots[0].ui);

  void *group;
  if (IsNil(groupSlot)) {
    group = 0;
  } else {
    if (!IsPtr(groupSlot)) return errWrongType;
    group = (void *)groupSlot->uo->slots[0].uptr;
  }

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkRadioButtonSetGroupRequest(button, group));

  return errNone;
}

int prGtkMiscSetAlignment(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_MISC(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkMisc *misc = GTK_MISC(objectSlot->uo->slots[0].ui);

  double x, y;
  if (slotDoubleVal(g->sp - 1, &y)) return errWrongType;
  if (slotDoubleVal(g->sp - 0, &x)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkMiscSetAlignmentRequest(misc, x, y));

  return errNone;
}

int prGtkMiscSetPadding(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_MISC(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkMisc *misc = GTK_MISC(objectSlot->uo->slots[0].ui);

  int x, y;
  if (slotIntVal(g->sp - 1, &y)) return errWrongType;
  if (slotIntVal(g->sp - 0, &x)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkMiscSetPaddingRequest(misc, x, y));

  return errNone;
}

int prGtkSizeGroupNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 1;
  PyrSlot *modeSlot = g->sp - 0;

  UNWRAP_GENUM(modeSlot, "mode", GtkSizeGroupMode, mode, GTK_TYPE_SIZE_GROUP_MODE);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSizeGroupNewRequest(mode));
  
  return errNone;
}

int prGtkSizeGroupAddWidget(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *widgetSlot = g->sp;

  UNWRAP_GOBJECT(objectSlot, "object", "GtkSizeGroup", GtkSizeGroup, group, GTK_SIZE_GROUP, GTK_IS_SIZE_GROUP);
  UNWRAP_GOBJECT(widgetSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSizeGroupAddWidgetRequest(group, widget));

  return errNone;
}

int prGtkSizeGroupRemoveWidget(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *widgetSlot = g->sp;

  UNWRAP_GOBJECT(objectSlot, "object", "GtkSizeGroup", GtkSizeGroup, group, GTK_SIZE_GROUP, GTK_IS_SIZE_GROUP);
  UNWRAP_GOBJECT(widgetSlot, "widget", "GtkWidget", GtkWidget, widget, GTK_WIDGET, GTK_IS_WIDGET);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSizeGroupRemoveWidgetRequest(group, widget));

  return errNone;
}

int prGtkSizeGroupSetMode(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *modeSlot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "object", "GtkSizeGroup", GtkSizeGroup, group, GTK_SIZE_GROUP, GTK_IS_SIZE_GROUP);
  UNWRAP_GENUM(modeSlot, "mode", GtkSizeGroupMode, mode, GTK_TYPE_SIZE_GROUP_MODE);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSizeGroupSetModeRequest(group, mode));

  return errNone;
}

int prGtkLabelNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 1;
  PyrSlot *textSlot = g->sp - 0;

  GtkWidget *widget;

  if (IsNil(textSlot)) {
    GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkLabelNewRequest(0, 0));    
  } else {
    const char *textPtr;
    int textSize;
    if (slotStrPtrLen(textSlot, &textPtr, &textSize, MEDIUM_STRING_SIZE)) return errWrongType;
    GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkLabelNewRequest(textPtr, textSize));
  }

  return errNone;
}

int prGtkLabelSetText(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *textSlot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "label", "GtkLabel", GtkLabel, label, GTK_LABEL, GTK_IS_LABEL);

  if (IsNil(textSlot)) {
    GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkLabelSetTextRequest(label, 0, 0));    
  } else {
    const char *textPtr;
    int textSize;
    if (slotStrPtrLen(textSlot, &textPtr, &textSize, MEDIUM_STRING_SIZE))
      WRONG_TYPE_PARAMETER("text", "String", textSlot);
    GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkLabelSetTextRequest(label, textPtr, textSize));
  }

  return errNone;
}

int prGtkLabelSetMarkup(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *markupSlot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "label", "GtkLabel", GtkLabel, label, GTK_LABEL, GTK_IS_LABEL);

  if (IsNil(markupSlot)) {
    GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkLabelSetMarkupRequest(label, 0, 0));    
  } else {
    const char *markupPtr;
    int markupSize;
    if (slotStrPtrLen(markupSlot, &markupPtr, &markupSize, MEDIUM_STRING_SIZE))
      WRONG_TYPE_PARAMETER("markup", "String", markupSlot);
    GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkLabelSetMarkupRequest(label, markupPtr, markupSize));
  }

  return errNone;
}

int prGtkLabelGetText(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_LABEL(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkLabel *label = GTK_LABEL(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkLabelGetTextRequest(label));    

  return errNone;
}

int prGtkLabelSetJustify(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *justificationSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_LABEL(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkLabel *label = GTK_LABEL(objectSlot->uo->slots[0].ui);

  int justify;
  if (slotEnumVal(GTK_TYPE_JUSTIFICATION, justificationSlot, justify)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkLabelSetJustifyRequest(label, GtkJustification(justify)));

  return errNone;
}

int prGtkEntryNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classObject = g->sp;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkEntryNewRequest());

  return errNone;
}

int prGtkEntrySetText(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *textSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_ENTRY(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkEntry *entry = GTK_ENTRY(objectSlot->uo->slots[0].ui);

  const char *textPtr;
  int textSize;
  if (slotStrPtrLen(textSlot, &textPtr, &textSize, MEDIUM_STRING_SIZE)) return errWrongType;
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkEntrySetTextRequest(entry, textPtr, textSize));

  return errNone;
}

int prGtkEntryGetText(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_ENTRY(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkEntry *entry = GTK_ENTRY(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkEntryGetTextRequest(entry));

  return errNone;
}

int prGtkEntrySetEditable(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *windowSlot = g->sp - 1;
  PyrSlot *valueSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(windowSlot);
  if (!GTK_IS_ENTRY(windowSlot->uo->slots[0].ui)) return errWrongType;
  GtkEntry *entry = GTK_ENTRY(windowSlot->uo->slots[0].ui);

  bool value;
  if (slotBoolVal(valueSlot, value)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkEntrySetEditableRequest(entry, value));

  return errNone;
}

int prGtkDrawingAreaNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkDrawingAreaNewRequest());

  return errNone;
}

int prGtkTreeModelGetColumnType(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *columnSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_MODEL(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeModel *model = GTK_TREE_MODEL(objectSlot->uo->slots[0].ui);

  int column;
  if (slotIntVal(columnSlot, &column)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeModelGetColumnTypeRequest(model, column));  

  return errNone;
}

int prGtkTreeModelGetIter(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *pathSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_MODEL(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeModel *model = GTK_TREE_MODEL(objectSlot->uo->slots[0].ui);

  SC_CHECK_VALID_GBOXED_WRAPPER(pathSlot);
  GtkTreePath *path = (GtkTreePath *)(pathSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeModelGetIterRequest(model, path));  

  return errNone;
}

int prGtkTreeModelGetIterFirst(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_MODEL(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeModel *model = GTK_TREE_MODEL(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeModelGetIterFirstRequest(model));  

  return errNone;
}

int prGtkTreeModelGetPath(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *iterSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_MODEL(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeModel *model = GTK_TREE_MODEL(objectSlot->uo->slots[0].ui);

  SC_CHECK_VALID_GBOXED_WRAPPER(iterSlot);
  GtkTreeIter *iter = (GtkTreeIter *)(iterSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeModelGetPathRequest(model, iter));

  return errNone;
}

int prGtkTreeModelGetPathString(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *iterSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_MODEL(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeModel *model = GTK_TREE_MODEL(objectSlot->uo->slots[0].ui);

  SC_CHECK_VALID_GBOXED_WRAPPER(iterSlot);
  GtkTreeIter *iter = (GtkTreeIter *)(iterSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeModelGetPathStringRequest(model, iter));

  return errNone;
}

int prGtkTreeModelIterParent(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *iterSlot = g->sp - 1;
  PyrSlot *parentSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_MODEL(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeModel *model = GTK_TREE_MODEL(objectSlot->uo->slots[0].ui);

  SC_CHECK_VALID_GBOXED_WRAPPER(iterSlot);
  GtkTreeIter *iter = (GtkTreeIter *)(iterSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeModelIterParentRequest(model, iter));

  return errNone;
}

int prGtkTreeModelIterNext(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *iterSlot = g->sp - 1;
  PyrSlot *parentSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_MODEL(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeModel *model = GTK_TREE_MODEL(objectSlot->uo->slots[0].ui);

  SC_CHECK_VALID_GBOXED_WRAPPER(iterSlot);
  GtkTreeIter *iter = (GtkTreeIter *)(iterSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeModelIterNextRequest(model, iter));

  return errNone;
}

int prGtkTreeModelIterChildren(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *iterSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_MODEL(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeModel *model = GTK_TREE_MODEL(objectSlot->uo->slots[0].ui);

  SC_CHECK_VALID_GBOXED_WRAPPER(iterSlot);
  GtkTreeIter *iter = (GtkTreeIter *)(iterSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeModelIterChildrenRequest(model, iter));

  return errNone;
}

int prGtkTreeModelIterHasChild(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *iterSlot = g->sp - 1;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_MODEL(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeModel *model = GTK_TREE_MODEL(objectSlot->uo->slots[0].ui);

  SC_CHECK_VALID_GBOXED_WRAPPER(iterSlot);
  GtkTreeIter *iter = (GtkTreeIter *)(iterSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeModelIterHasChildRequest(model, iter));

  return errNone;
}

int prGtkTreeModelIterNChildren(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *iterSlot = g->sp - 1;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_MODEL(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeModel *model = GTK_TREE_MODEL(objectSlot->uo->slots[0].ui);

  SC_CHECK_VALID_GBOXED_WRAPPER(iterSlot);
  GtkTreeIter *iter = (GtkTreeIter *)(iterSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeModelIterNChildrenRequest(model, iter));

  return errNone;
}

int prGtkTreeModelIterNthChild(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *iterSlot = g->sp - 1;
  PyrSlot *nSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_MODEL(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeModel *model = GTK_TREE_MODEL(objectSlot->uo->slots[0].ui);

  GtkTreeIter *iter;
  if (IsNil(iterSlot)) {
    iter = 0;
  } else {
    SC_CHECK_VALID_GBOXED_WRAPPER(iterSlot);
    iter = (GtkTreeIter *)(iterSlot->uo->slots[0].ui);
  }

  int n;
  if (slotIntVal(nSlot, &n)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeModelIterNthChildRequest(model, iter, n));

  return errNone;
}

int prGtkTreeModelGetValue(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *iterSlot = g->sp - 1;
  PyrSlot *columnSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_MODEL(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeModel *model = GTK_TREE_MODEL(objectSlot->uo->slots[0].ui);

  SC_CHECK_VALID_GBOXED_WRAPPER(iterSlot);
  GtkTreeIter *iter = (GtkTreeIter *)(iterSlot->uo->slots[0].ui);

  int column;
  if (slotIntVal(columnSlot, &column)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeModelGetValueRequest(model, iter, column));

  return errNone;
}

int prGtkTreeRowReferenceNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *modelSlot = g->sp - 1;
  PyrSlot *pathSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(modelSlot);
  if (!GTK_IS_TREE_MODEL(modelSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeModel *model = GTK_TREE_MODEL(modelSlot->uo->slots[0].ui);

  SC_CHECK_VALID_GBOXED_WRAPPER(pathSlot);
  GtkTreePath *path = (GtkTreePath *)(pathSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeRowReferenceNewRequest(model, path));

  return errNone;
}

int prGtkTreeRowReferenceGetPath(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GENERIC_WRAPPER(objectSlot);
  GtkTreeRowReference *reference = (GtkTreeRowReference *)(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeRowReferenceGetPathRequest(reference));

  return errNone;
}

int prGtkTreeRowReferenceValid(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GENERIC_WRAPPER(objectSlot);
  GtkTreeRowReference *reference = (GtkTreeRowReference *)(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeRowReferenceValidRequest(reference));

  return errNone;
}

int prGtkTreeRowReferenceFree(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GENERIC_WRAPPER(objectSlot);
  GtkTreeRowReference *reference = (GtkTreeRowReference *)(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeRowReferenceFreeRequest(reference));

  return errNone;
}

int prGtkTreeRowReferenceCopy(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GENERIC_WRAPPER(objectSlot);
  GtkTreeRowReference *reference = (GtkTreeRowReference *)(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeRowReferenceCopyRequest(reference));

  return errNone;
}

int prGtkListStoreNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 1;
  PyrSlot *typeArraySlot = g->sp - 0;

  const int N = 32;
  GType types[N];

  if (typeArraySlot->utag != tagObj || !isKindOf(typeArraySlot->uo, class_array)) return errWrongType;
  PyrObject *array = typeArraySlot->uo;

  int n = (array->size < N) ? array->size : N;
  for (int i = 0; i < n; ++i) {
    char typeString[32];
    if (slotStrVal(array->slots + i, typeString, 32)) return errWrongType;
    types[i] = g_type_from_name(typeString);
  }

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkListStoreNewRequest(n, types));

  return errNone;
}

int prGtkListStoreAppend(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_LIST_STORE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkListStore *store = GTK_LIST_STORE(objectSlot->uo->slots[0].ui);
 
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkListStoreAppendRequest(store));

  return errNone;
}


int prGtkListStoreClear(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_LIST_STORE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkListStore *store = GTK_LIST_STORE(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkListStoreClearRequest(store));

  return errNone;
}

int prGtkListStoreSetValue(struct VMGlobals *g, int numArgsPushed) {
  int result;

  PyrSlot *objectSlot = g->sp - 4;
  PyrSlot *iterSlot = g->sp - 3;
  PyrSlot *columnSlot = g->sp - 2;
  PyrSlot *typeSlot = g->sp - 1;
  PyrSlot *valueSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_LIST_STORE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkListStore *store = GTK_LIST_STORE(objectSlot->uo->slots[0].ui);

  SC_CHECK_VALID_GBOXED_WRAPPER(iterSlot);
  GtkTreeIter *iter = (GtkTreeIter *)(iterSlot->uo->slots[0].ui);

  int column;
  if (slotIntVal(columnSlot, &column)) return errWrongType;
    
  int typeAsInt;
  if (slotIntVal(typeSlot, &typeAsInt)) return errWrongType;

  GValue value = {0};

  g_value_init(&value, GType(typeAsInt));

  if (slot_to_GValue(valueSlot, &value)) {
    result = errWrongType;
  } else {
    GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkListStoreSetValueRequest(store, iter, column, &value));    
    result = errNone;
  }

  g_value_unset(&value);
 
  return result;
}

int prGtkListStoreRemove(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *iterSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_LIST_STORE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkListStore *store = GTK_LIST_STORE(objectSlot->uo->slots[0].ui);

  SC_CHECK_VALID_GBOXED_WRAPPER(iterSlot);
  GtkTreeIter *iter = (GtkTreeIter *)(iterSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkListStoreRemoveRequest(store, iter));

  return errNone;
}

int prGtkTreeStoreNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 1;
  PyrSlot *typeArraySlot = g->sp - 0;

  const int N = 32;
  GType types[N];

  if (!isKindOfSlot(typeArraySlot, class_array))
    WRONG_TYPE_PARAMETER("typeArray", "Array", typeArraySlot);
  PyrObject *array = typeArraySlot->uo;

  int n = (array->size < N) ? array->size : N;
  for (int i = 0; i < n; ++i) {
    char typeString[32];
    if (slotStrVal(array->slots + i, typeString, 32)) WRONG_TYPE_PARAMETER("type", "String", array->slots + i);
    types[i] = g_type_from_name(typeString);
  }

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeStoreNewRequest(n, types));

  return errNone;
}

int prGtkTreeStoreAppend(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *parentSlot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "object", "GtkTreeStore", GtkTreeStore, store, GTK_TREE_STORE, GTK_IS_TREE_STORE);

  GtkTreeIter *parent;
  if (IsNil(parentSlot)) {
    parent = 0;
  } else {
    SC_CHECK_VALID_GBOXED_WRAPPER(parentSlot);
    parent = (GtkTreeIter *)(parentSlot->uo->slots[0].ui);
  }

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeStoreAppendRequest(store, parent));

  return errNone;
}

int prGtkTreeStoreInsertBefore(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *parentSlot = g->sp - 1;
  PyrSlot *siblingSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_STORE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeStore *store = GTK_TREE_STORE(objectSlot->uo->slots[0].ui);

  GtkTreeIter *parent, *sibling;
  if (IsNil(parentSlot)) {
    parent = 0;
  } else {
    SC_CHECK_VALID_GBOXED_WRAPPER(parentSlot);
    parent = (GtkTreeIter *)(parentSlot->uo->slots[0].ui);
  }

  if (IsNil(siblingSlot)) {
    sibling = 0;
  } else {
    SC_CHECK_VALID_GBOXED_WRAPPER(siblingSlot);
    sibling = (GtkTreeIter *)(siblingSlot->uo->slots[0].ui);
  }

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeStoreInsertBeforeRequest(store, parent, sibling));

  return errNone;
}

int prGtkTreeStoreInsertAfter(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *parentSlot = g->sp - 1;
  PyrSlot *siblingSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_STORE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeStore *store = GTK_TREE_STORE(objectSlot->uo->slots[0].ui);

  GtkTreeIter *parent, *sibling;
  if (IsNil(parentSlot)) {
    parent = 0;
  } else {
    SC_CHECK_VALID_GBOXED_WRAPPER(parentSlot);
    parent = (GtkTreeIter *)(parentSlot->uo->slots[0].ui);
  }

  if (IsNil(siblingSlot)) {
    sibling = 0;
  } else {
    SC_CHECK_VALID_GBOXED_WRAPPER(siblingSlot);
    sibling = (GtkTreeIter *)(siblingSlot->uo->slots[0].ui);
  }

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeStoreInsertAfterRequest(store, parent, sibling));

  return errNone;
}

int prGtkTreeStoreSetValue(struct VMGlobals *g, int numArgsPushed) {
  int result;

  PyrSlot *objectSlot = g->sp - 4;
  PyrSlot *iterSlot = g->sp - 3;
  PyrSlot *columnSlot = g->sp - 2;
  PyrSlot *typeSlot = g->sp - 1;
  PyrSlot *valueSlot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "object", "GtkTreeStore", GtkTreeStore, store, GTK_TREE_STORE, GTK_IS_TREE_STORE);
  UNWRAP_GBOXED(iterSlot, "iter", "GtkTreeIter", GtkTreeIter, iter);
  int column, typeAsInt;
  if (slotIntVal(columnSlot, &column)) WRONG_TYPE_PARAMETER("column", "Integer", columnSlot);
  if (slotIntVal(typeSlot, &typeAsInt)) WRONG_TYPE_PARAMETER("type", "Integer", typeSlot);

  GValue value = {0};

  g_value_init(&value, GType(typeAsInt));

  if (slot_to_GValue(valueSlot, &value)) {
    error("value has wrong type");
    dumpObjectSlot(valueSlot);
    result = errWrongType;
  } else {
    GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeStoreSetValueRequest(store, iter, column, &value));    
    result = errNone;
  }

  g_value_unset(&value);
 
  return result;
}

int prGtkTreeStoreClear(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_STORE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeStore *store = GTK_TREE_STORE(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeStoreClearRequest(store));

  return errNone;
}

int prGtkTreeStoreRemove(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *iterSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_STORE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeStore *store = GTK_TREE_STORE(objectSlot->uo->slots[0].ui);
 
  SC_CHECK_VALID_GBOXED_WRAPPER(iterSlot);
  PyrObject *iterObject = iterSlot->uo;
  GtkTreeIter *iter = (GtkTreeIter *)(iterObject->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeStoreRemoveRequest(store, iter));

  return errNone;
}

int prGtkTreeViewNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeViewNewRequest());

  return errNone;
}

int prGtkTreeViewNewWithModel(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 1;
  PyrSlot *modelSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(modelSlot);
  if (!GTK_IS_TREE_MODEL(modelSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeModel *model = GTK_TREE_MODEL(modelSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeViewNewWithModelRequest(model));

  return errNone;
}

int prGtkTreeViewAppendColumn(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *columnSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_VIEW(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeView *view = GTK_TREE_VIEW(objectSlot->uo->slots[0].ui);

  SC_CHECK_VALID_GOBJECT_WRAPPER(columnSlot);
  if (!GTK_IS_TREE_VIEW_COLUMN(columnSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeViewColumn *column = GTK_TREE_VIEW_COLUMN(columnSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeViewAppendColumnRequest(view, column));

  return errNone;
}

int prGtkTreeViewScrollToCell(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 4;
  PyrSlot *pathSlot = g->sp - 3;
  PyrSlot *columnSlot = g->sp - 2;
  PyrSlot *rowAlignSlot = g->sp - 1;
  PyrSlot *colAlignSlot = g->sp - 1;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_VIEW(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeView *view = GTK_TREE_VIEW(objectSlot->uo->slots[0].ui);

  GtkTreePath *path;
  if (IsNil(pathSlot)) {
    path = 0;
  } else {
    SC_CHECK_VALID_GBOXED_WRAPPER(pathSlot);
    path = (GtkTreePath *)(pathSlot->uo->slots[0].ui);
  }

  GtkTreeViewColumn *column;
  if (IsNil(columnSlot)) {
    column = 0;
  } else {
    SC_CHECK_VALID_GOBJECT_WRAPPER(columnSlot);
    if (!GTK_IS_TREE_VIEW_COLUMN(columnSlot->uo->slots[0].ui)) return errWrongType;
    column = GTK_TREE_VIEW_COLUMN(columnSlot->uo->slots[0].ui);
  }

  bool useAlign = true;
  double rowAlign, colAlign;

  if (IsNil(rowAlignSlot)) {
    useAlign = false;
    rowAlign = 0.5;
  } else {
    if (slotDoubleVal(rowAlignSlot, &rowAlign)) return errWrongType;
  }
  if (IsNil(colAlignSlot)) {
    useAlign = false;
    colAlign = 0.5;
  } else {
    if (slotDoubleVal(colAlignSlot, &colAlign)) return errWrongType;
  }

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeViewScrollToCellRequest(view, path, column, useAlign, rowAlign, colAlign));

  return errNone;
}

int prGtkTreeViewGetSelection(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_VIEW(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeView *view = GTK_TREE_VIEW(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeViewGetSelectionRequest(view));

  return errNone;
}

int prGtkTreeViewGetModel(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_VIEW(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeView *view = GTK_TREE_VIEW(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeViewGetModelRequest(view));

  return errNone;
}

int prGtkTreeViewEnableModelDragSource(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 3;
  PyrSlot *modMaskSlot = g->sp - 2;
  PyrSlot *targetArraySlot = g->sp - 1;
  PyrSlot *actionSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_VIEW(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeView *view = GTK_TREE_VIEW(objectSlot->uo->slots[0].ui);

  int modmask;
  if (slotFlagsVal(GDK_TYPE_MODIFIER_TYPE, modMaskSlot, modmask)) return errWrongType;

  if (targetArraySlot->utag != tagObj || !isKindOf(targetArraySlot->uo, class_array)) return errWrongType;
  PyrObject *array = targetArraySlot->uo;

  int nTargets = array->size;
  if (nTargets > DRAG_TARGETS_MAX) {
    nTargets = DRAG_TARGETS_MAX;
    std::cout << "dropping drag targets" << std::endl;
  }

  DragTargetNameArray nameArray;
  DragTargetFlagArray flagArray;
  DragTargetInfoArray infoArray;

  for (int i; i < nTargets; ++i) {
    PyrObject *object = (array->slots + i)->uo;
    const char *str;
    int size;
    if (slotStrPtrLen(object->slots + 0, &str, &size, MEDIUM_STRING_SIZE)) return errWrongType;
    safe_string_copy(nameArray[i], str, size, SMALL_STRING_SIZE);
    if (slotEnumVal(GTK_TYPE_TARGET_FLAGS, object->slots + 1, flagArray[i])) return errWrongType;
    if (slotIntVal(object->slots + 2, &infoArray[i])) return errWrongType;
  }

  int actions;
  if (slotFlagsVal(GDK_TYPE_DRAG_ACTION, actionSlot, actions)) return errWrongType;
  
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeViewEnableModelDragSourceRequest(view, GdkModifierType(modmask), nTargets,
									 nameArray, flagArray, infoArray,
									 GdkDragAction(actions)));

  return errNone;
}

int prGtkTreeViewEnableModelDragDest(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *targetArraySlot = g->sp - 1;
  PyrSlot *actionSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_VIEW(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeView *view = GTK_TREE_VIEW(objectSlot->uo->slots[0].ui);

  if (targetArraySlot->utag != tagObj || !isKindOf(targetArraySlot->uo, class_array)) return errWrongType;
  PyrObject *array = targetArraySlot->uo;

  int nTargets = array->size;
  if (nTargets > DRAG_TARGETS_MAX) {
    nTargets = DRAG_TARGETS_MAX;
    std::cout << "dropping drag targets" << std::endl;
  }

  DragTargetNameArray nameArray;
  DragTargetFlagArray flagArray;
  DragTargetInfoArray infoArray;

  for (int i; i < nTargets; ++i) {
    PyrObject *object = (array->slots + i)->uo;
    const char *str;
    int size;
    if (slotStrPtrLen(object->slots + 0, &str, &size, MEDIUM_STRING_SIZE)) return errWrongType;
    safe_string_copy(nameArray[i], str, size, SMALL_STRING_SIZE);
    if (slotEnumVal(GTK_TYPE_TARGET_FLAGS, object->slots + 1, flagArray[i])) return errWrongType;
    if (slotIntVal(object->slots + 2, &infoArray[i])) return errWrongType;
  }

  int actions;
  if (slotFlagsVal(GDK_TYPE_DRAG_ACTION, actionSlot, actions)) return errWrongType;
  
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeViewEnableModelDragDestRequest(view, nTargets,
								       nameArray, flagArray, infoArray,
								       GdkDragAction(actions)));

  return errNone;
}


int prGtkTreeViewGetDestRowAtPos(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *xSlot = g->sp - 1;
  PyrSlot *ySlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_VIEW(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeView *view = GTK_TREE_VIEW(objectSlot->uo->slots[0].ui);

  gint x, y;
  if (slotIntVal(xSlot, &x)) return errWrongType;
  if (slotIntVal(ySlot, &y)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeViewGetDestRowAtPosRequest(view, x, y));

  return errNone;
}

int prGtkTreeViewExpandRow(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *pathSlot = g->sp - 1;
  PyrSlot *openAllSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_VIEW(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeView *view = GTK_TREE_VIEW(objectSlot->uo->slots[0].ui);

  SC_CHECK_VALID_GBOXED_WRAPPER(pathSlot);
  GtkTreePath *path = (GtkTreePath *)(pathSlot->uo->slots[0].ui);

  bool openAll;
  if (slotBoolVal(openAllSlot, openAll)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeViewExpandRowRequest(view, path, openAll));

  return errNone;
}

int prGtkTreeViewExpandAll(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_VIEW(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeView *view = GTK_TREE_VIEW(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeViewExpandAllRequest(view));

  return errNone;
}

int prGtkTreeViewColumnNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeViewColumnNewRequest());
  
  return errNone;
}

int prGtkTreeViewColumnSetTitle(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *textSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_VIEW_COLUMN(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeViewColumn *column = GTK_TREE_VIEW_COLUMN(objectSlot->uo->slots[0].ui);

  const char *textPtr;
  int textSize;
  if (slotStrPtrLen(textSlot, &textPtr, &textSize, MEDIUM_STRING_SIZE)) return errWrongType;
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeViewColumnSetTitleRequest(column, textPtr, textSize));

  return errNone;
}

int prGtkTreeViewColumnNewWithAttributes(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 3;
  PyrSlot *titleSlot = g->sp - 2;
  PyrSlot *rendererSlot = g->sp - 1;
  PyrSlot *attributeArraySlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(rendererSlot);
  if (!GTK_IS_CELL_RENDERER(rendererSlot->uo->slots[0].ui)) return errWrongType;
  GtkCellRenderer *renderer = GTK_CELL_RENDERER(rendererSlot->uo->slots[0].ui);

  if (attributeArraySlot->utag != tagObj || !isKindOf(attributeArraySlot->uo, class_array)) return errWrongType;
  PyrObject *array = attributeArraySlot->uo;
  if (array->size % 2 != 0) errWrongType; // Should be something else
  int nAttributes = array->size / 2;

  if (nAttributes > 16) {
    nAttributes = 16;
    std::cout << "Dropping attributes" << std::endl;
  }

  const char *titlePtr;
  int titleSize;
  if (slotStrPtrLen(titleSlot, &titlePtr, &titleSize, MEDIUM_STRING_SIZE)) return errWrongType;
  
  AttributeNameArray anames;
  AttributeColumnArray acols;  

  // XXX: we can avoid this copy
  int i, j;
  for (i = 0, j = 0; j < nAttributes; i += 2, ++j) {
    int column;
    if (slotStrVal(array->slots + i, &(anames[j][0]), SMALL_STRING_SIZE)) return errWrongType;
    if (slotIntVal(array->slots + i + 1, &column)) return errWrongType;
    acols[j] = column;
  }

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeViewColumnNewWithAttributesRequest(titlePtr, titleSize, renderer, nAttributes, anames, acols));

  return errNone;
}

int prGtkTreeViewColumnSetClickable(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *valueSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_VIEW_COLUMN(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeViewColumn *column = GTK_TREE_VIEW_COLUMN(objectSlot->uo->slots[0].ui);

  if (!IsTrue(valueSlot) && !IsFalse(valueSlot))  return errWrongType;
  bool value = IsTrue(valueSlot);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeViewColumnSetClickableRequest(column, value));

  return errNone;
}

int prGtkTreeViewColumnSetSizing(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *valueSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_VIEW_COLUMN(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeViewColumn *column = GTK_TREE_VIEW_COLUMN(objectSlot->uo->slots[0].ui);

  int sizing;
  if (slotEnumVal(GTK_TYPE_TREE_VIEW_COLUMN_SIZING, valueSlot, sizing)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeViewColumnSetSizingRequest(column, GtkTreeViewColumnSizing(sizing)));

  return errNone;
}

int prGtkTreeViewColumnSetFixedWidth(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *valueSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_VIEW_COLUMN(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeViewColumn *column = GTK_TREE_VIEW_COLUMN(objectSlot->uo->slots[0].ui);

  int value;
  if (slotIntVal(valueSlot, &value)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeViewColumnSetFixedWidthRequest(column, value));

  return errNone;
}

int prGtkTreeViewSetRulesHint(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *valueSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_VIEW(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeView *view = GTK_TREE_VIEW(objectSlot->uo->slots[0].ui);

  if (!IsTrue(valueSlot) && !IsFalse(valueSlot))  return errWrongType;
  bool value = IsTrue(valueSlot);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeViewSetRulesHintRequest(view,  value));

  return errNone;
}

int prGtkTreeViewSetSearchColumn(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *valueSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_VIEW(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeView *view = GTK_TREE_VIEW(objectSlot->uo->slots[0].ui);

  int value;
  if (slotIntVal(valueSlot, &value)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeViewSetSearchColumnRequest(view, value));

  return errNone;
}

int prGtkTreeViewSetReorderable(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *valueSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_VIEW(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeView *view = GTK_TREE_VIEW(objectSlot->uo->slots[0].ui);

  if (!IsTrue(valueSlot) && !IsFalse(valueSlot))  return errWrongType;
  bool value = IsTrue(valueSlot);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeViewSetReorderableRequest(view,  value));

  return errNone;
}

int prGtkTreeViewGetReorderable(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_VIEW(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeView *view = GTK_TREE_VIEW(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeViewGetReorderableRequest(view));

  return errNone;
}

int prGtkCellRendererTextNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkCellRendererTextNewRequest());

  return errNone;
}

int prGtkCellRendererToggleNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkCellRendererToggleNewRequest());

  return errNone;
}

int prGtkTreeSelectionGetSelected(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_SELECTION(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeSelection *selection = GTK_TREE_SELECTION(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeSelectionGetSelectedRequest(selection));  

  return errNone;
}

int prGtkTreeSelectionGetSelectedRows(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_SELECTION(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeSelection *selection = GTK_TREE_SELECTION(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeSelectionGetSelectedRowsRequest(selection));  

  return errNone;
}

int prGtkTreeSelectionSelectIter(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *iterSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_SELECTION(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeSelection *selection = GTK_TREE_SELECTION(objectSlot->uo->slots[0].ui);

  SC_CHECK_VALID_GBOXED_WRAPPER(iterSlot);
  GtkTreeIter *iter = (GtkTreeIter *)(iterSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeSelectionSelectIterRequest(selection, iter));  

  return errNone;
}

int prGtkTreeSelectionSelectAll(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_SELECTION(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeSelection *selection = GTK_TREE_SELECTION(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeSelectionSelectAllRequest(selection));  

  return errNone;
}

int prGtkTreeSelectionUnselectAll(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_SELECTION(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeSelection *selection = GTK_TREE_SELECTION(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeSelectionUnselectAllRequest(selection));  

  return errNone;
}

int prGtkTreeSelectionSetMode(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *modeSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TREE_SELECTION(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeSelection *selection = GTK_TREE_SELECTION(objectSlot->uo->slots[0].ui);

  int mode;
  if (slotEnumVal(GTK_TYPE_SELECTION_MODE, modeSlot, mode)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreeSelectionSetModeRequest(selection, GtkSelectionMode(mode)));

  return errNone;
}

int prGtkAdjustmentNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 6;

  float value;
  if (slotFloatVal(g->sp + 5, &value)) return errWrongType;
  float lower;
  if (slotFloatVal(g->sp + 4, &lower)) return errWrongType;
  float upper;
  if (slotFloatVal(g->sp + 3, &upper)) return errWrongType;
  float step_increment;
  if (slotFloatVal(g->sp + 3, &step_increment)) return errWrongType;
  float page_increment;
  if (slotFloatVal(g->sp + 3, &page_increment)) return errWrongType;
  float page_size;
  if (slotFloatVal(g->sp + 3, &page_size)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkAdjustmentNewRequest(value, lower, upper, step_increment, page_increment, page_size));

  return errNone;
}

int prGtkAdjustmentGetValue(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_ADJUSTMENT(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkAdjustment *adjustment = GTK_ADJUSTMENT(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkAdjustmentGetValueRequest(adjustment));

  return errNone;
}

int prGtkAdjustmentSetValue(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *valueSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_ADJUSTMENT(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkAdjustment *adjustment = GTK_ADJUSTMENT(objectSlot->uo->slots[0].ui);

  double value;
  if (slotDoubleVal(valueSlot, &value)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkAdjustmentSetValueRequest(adjustment, value));

  return errNone;
}

int prGtkAdjustmentClampPage(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *lowerSlot = g->sp - 1;
  PyrSlot *upperSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_ADJUSTMENT(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkAdjustment *adjustment = GTK_ADJUSTMENT(objectSlot->uo->slots[0].ui);

  double lower, upper;
  if (slotDoubleVal(lowerSlot, &lower)) return errWrongType;
  if (slotDoubleVal(upperSlot, &upper)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkAdjustmentClampPageRequest(adjustment, lower, upper));

  return errNone;
}

int prGtkAdjustmentChanged(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_ADJUSTMENT(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkAdjustment *adjustment = GTK_ADJUSTMENT(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkAdjustmentChangedRequest(adjustment));

  return errNone;
}

int prGtkAdjustmentValueChanged(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_ADJUSTMENT(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkAdjustment *adjustment = GTK_ADJUSTMENT(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkAdjustmentValueChangedRequest(adjustment));

  return errNone;
}

int prGtkScrolledWindowNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 2;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkScrolledWindowNewRequest());

  return errNone;
}

int prGtkScrolledWindowSetPolicy(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_SCROLLED_WINDOW(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkScrolledWindow *swindow = GTK_SCROLLED_WINDOW(objectSlot->uo->slots[0].ui);

  int htype, vtype;
  if (slotEnumVal(GTK_TYPE_POLICY_TYPE, g->sp - 1, htype)) return errWrongType;
  if (slotEnumVal(GTK_TYPE_POLICY_TYPE, g->sp - 0, vtype)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkScrolledWindowSetPolicyRequest(swindow, GtkPolicyType(htype), GtkPolicyType(vtype)));

  return errNone;
}

int prGtkScrolledWindowSetShadowType(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *shadowTypeSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_SCROLLED_WINDOW(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkScrolledWindow *swindow = GTK_SCROLLED_WINDOW(objectSlot->uo->slots[0].ui);

  int type;
  if (slotEnumVal(GTK_TYPE_SHADOW_TYPE, shadowTypeSlot, type)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkScrolledWindowSetShadowTypeRequest(swindow, GtkShadowType(type)));

  return errNone;
}

int prGtkScrolledWindowAddWithViewport(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *childSlot = g->sp;

  UNWRAP_GOBJECT(objectSlot, "object", "GtkScrolledWindow", GtkScrolledWindow, scrolled, GTK_SCROLLED_WINDOW, GTK_IS_SCROLLED_WINDOW);
  UNWRAP_GOBJECT(childSlot, "child", "GtkWidget", GtkWidget, child, GTK_WIDGET, GTK_IS_WIDGET);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkScrolledWindowAddWithViewportRequest(scrolled, child));

  return errNone;
}

int prPangoContextNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 0;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(PangoContextNewRequest());

  return errNone;
}

int prPangoLayoutNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 1;
  PyrSlot *contextSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(contextSlot);
  if (!PANGO_IS_CONTEXT(contextSlot->uo->slots[0].ui)) return errWrongType;
  PangoContext *context = PANGO_CONTEXT(contextSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(PangoLayoutNewRequest(context));

  return errNone;
}

int prPangoLayoutSetText(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *textSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!PANGO_IS_LAYOUT(objectSlot->uo->slots[0].ui)) return errWrongType;
  PangoLayout *layout = PANGO_LAYOUT(objectSlot->uo->slots[0].ui);

  const char *textPtr;
  int textSize;
  if (slotStrPtrLen(textSlot, &textPtr, &textSize, LARGE_STRING_SIZE)) return errWrongType;
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(PangoLayoutSetTextRequest(layout, textPtr, textSize));

  return errNone;
}

int prPangoLayoutSetMarkup(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *textSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!PANGO_IS_LAYOUT(objectSlot->uo->slots[0].ui)) return errWrongType;
  PangoLayout *layout = PANGO_LAYOUT(objectSlot->uo->slots[0].ui);

  const char *textPtr;
  int textSize;
  if (slotStrPtrLen(textSlot, &textPtr, &textSize, LARGE_STRING_SIZE)) return errWrongType;
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(PangoLayoutSetMarkupRequest(layout, textPtr, textSize));

  return errNone;
}

int prPangoLayoutGetPixelSize(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!PANGO_IS_LAYOUT(objectSlot->uo->slots[0].ui)) return errWrongType;
  PangoLayout *layout = PANGO_LAYOUT(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(PangoLayoutGetPixelSizeRequest(layout));  

  return errNone;
}

void closeAllGUIScreens(void) {
#if 0
  extern long compiledOK;
  VMGlobals *g = gMainVMGlobals;
  if (compiledOK) {
    bool canCallOS = g->canCallOS;

    PyrClass *classobj = getsym("GtkWindow")->u.classobj;
    if (classobj == 0) return;
    
    pthread_mutex_lock(&gLangMutex);

    g->canCallOS = true;
    
    ++g->sp; SetObject(g->sp, classobj);

    runInterpreter(g, getsym("destroyAll"), 1);

    g->canCallOS = canCallOS;

    pthread_mutex_unlock(&gLangMutex);
  }
#endif
}

int prGtkTableNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 3;
  PyrSlot *nSlot = g->sp - 2;
  PyrSlot *mSlot = g->sp - 1;
  PyrSlot *homoSlot = g->sp - 0;

  int n, m;
  if (slotIntVal(nSlot, &n)) return errWrongType;
  if (slotIntVal(mSlot, &m)) return errWrongType;
  
  bool homogeneous;
  if (slotBoolVal(homoSlot, homogeneous)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTableNewRequest(n, m, homogeneous));

  return errNone;
}

int prGtkTableResize(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TABLE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTable *table = GTK_TABLE(objectSlot->uo->slots[0].ui);

  int n, m;
  if (slotIntVal(g->sp - 1, &n)) return errWrongType;
  if (slotIntVal(g->sp - 0, &m)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTableResizeRequest(table, n, m));

  return errNone;
}

int prGtkTableAttach(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 9;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TABLE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTable *table = GTK_TABLE(objectSlot->uo->slots[0].ui);

  PyrSlot *childSlot = g->sp - 8;
  SC_CHECK_VALID_GOBJECT_WRAPPER(childSlot);
  if (!GTK_IS_WIDGET(childSlot->uo->slots[0].ui)) return errWrongType;
  GtkWidget *child = GTK_WIDGET(childSlot->uo->slots[0].ui);

  int leftAttach, rightAttach, topAttach, bottomAttach;
  if (slotIntVal(g->sp - 7, &leftAttach)) return errWrongType;
  if (slotIntVal(g->sp - 6, &rightAttach)) return errWrongType;
  if (slotIntVal(g->sp - 5, &topAttach)) return errWrongType;
  if (slotIntVal(g->sp - 4, &bottomAttach)) return errWrongType;

  int xOptions, yOptions;
  if (slotFlagsVal(GTK_TYPE_ATTACH_OPTIONS, g->sp - 3, xOptions)) return errWrongType;
  if (slotFlagsVal(GTK_TYPE_ATTACH_OPTIONS, g->sp - 2, yOptions)) return errWrongType;

  int xPadding, yPadding;
  if (slotIntVal(g->sp - 1, &xPadding)) return errWrongType;
  if (slotIntVal(g->sp - 0, &yPadding)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTableAttachRequest(table, child, leftAttach, rightAttach, topAttach, bottomAttach,
						       GtkAttachOptions(xOptions), GtkAttachOptions(yOptions),
						       xPadding, yPadding));

  return errNone;
}

int prGtkTableSetColSpacings(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *nSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TABLE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTable *table = GTK_TABLE(objectSlot->uo->slots[0].ui);

  int n;
  if (slotIntVal(nSlot, &n)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTableSetColSpacingsRequest(table, n));

  return errNone;
}

int prGtkTableSetRowSpacings(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *nSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TABLE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTable *table = GTK_TABLE(objectSlot->uo->slots[0].ui);

  int n;
  if (slotIntVal(nSlot, &n)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTableSetRowSpacingsRequest(table, n));

  return errNone;
}

int prGtkImageNewFromFile(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 1;
  PyrSlot *filenameSlot = g->sp - 0;

  const char *filenamePtr;
  int filenameSize;
  if (slotStrPtrLen(filenameSlot, &filenamePtr, &filenameSize, FILENAME_STRING_SIZE)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkImageNewFromFileRequest(filenamePtr, filenameSize));  

  return errNone;
}

int prGtkImageNewFromStock(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 2;
  PyrSlot *nameSlot = g->sp - 1;
  PyrSlot *sizeSlot = g->sp - 0;

  int size;
  if (slotEnumVal(GTK_TYPE_ICON_SIZE, sizeSlot, size)) return errWrongType;

  const char *namePtr;
  int nameSize;
  if (slotStrPtrLen(nameSlot, &namePtr, &nameSize, SMALL_STRING_SIZE)) return errWrongType;
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkImageNewFromStockRequest(namePtr, nameSize, GtkIconSize(size)));  

  return errNone;
}

int prGtkMenuShellAppend(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *childSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_MENU_SHELL(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkMenuShell *menushell = GTK_MENU_SHELL(objectSlot->uo->slots[0].ui);

  SC_CHECK_VALID_GOBJECT_WRAPPER(childSlot);
  if (!GTK_IS_WIDGET(childSlot->uo->slots[0].ui)) return errWrongType;
  GtkWidget *child = GTK_WIDGET(childSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkMenuShellAppendRequest(menushell, child));

  return errNone;
}

int prGtkMenuShellPrepend(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *childSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_MENU_SHELL(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkMenuShell *menushell = GTK_MENU_SHELL(objectSlot->uo->slots[0].ui);

  SC_CHECK_VALID_GOBJECT_WRAPPER(childSlot);
  if (!GTK_IS_WIDGET(childSlot->uo->slots[0].ui)) return errWrongType;
  GtkWidget *child = GTK_WIDGET(childSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkMenuShellPrependRequest(menushell, child));

  return errNone;
}

int prGtkMenuShellInsert(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *childSlot = g->sp - 1;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_MENU_SHELL(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkMenuShell *menushell = GTK_MENU_SHELL(objectSlot->uo->slots[0].ui);

  SC_CHECK_VALID_GOBJECT_WRAPPER(childSlot);
  if (!GTK_IS_WIDGET(childSlot->uo->slots[0].ui)) return errWrongType;
  GtkWidget *child = GTK_WIDGET(childSlot->uo->slots[0].ui);

  int position;
  if (slotIntVal(g->sp, &position)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkMenuShellInsertRequest(menushell, child, position));

  return errNone;
}

int prGtkMenuNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkMenuNewRequest());

  return errNone;
}

int prGtkMenuPopup(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_MENU(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkMenu *menu = GTK_MENU(objectSlot->uo->slots[0].ui);

  int button, time;
  if (slotIntVal(g->sp - 1, &button)) return errWrongType;
  if (slotIntVal(g->sp - 0, &time)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkMenuPopupRequest(menu, button, time));

  return errNone;
}

int prGtkMenuItemNewWithLabel(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 1;
  PyrSlot *textSlot = g->sp - 0;

  const char *textPtr;
  int textSize;
  if (slotStrPtrLen(textSlot, &textPtr, &textSize, MEDIUM_STRING_SIZE)) return errWrongType;
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkMenuItemNewWithLabelRequest(textPtr, textSize));

  return errNone;
}

int prGtkSeparatorMenuItemNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 0;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSeparatorMenuItemNewRequest());

  return errNone;
}

int prGtkMenuItemSetSubmenu(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *widgetSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_MENU_ITEM(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkMenuItem *menuitem = GTK_MENU_ITEM(objectSlot->uo->slots[0].ui);

  SC_CHECK_VALID_GOBJECT_WRAPPER(widgetSlot);
  if (!GTK_IS_WIDGET(widgetSlot->uo->slots[0].ui)) return errWrongType;
  GtkWidget *widget = GTK_WIDGET(widgetSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkMenuItemSetSubmenuRequest(menuitem, widget));

  return errNone;
}

int prGtkTextIterIsStart(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  UNWRAP_GBOXED(objectSlot, "object", "GtkTextIter", GtkTextIter, iter);
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextIterIsStartRequest(iter));
  return errNone;
}

int prGtkTextIterIsEnd(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  UNWRAP_GBOXED(objectSlot, "object", "GtkTextIter", GtkTextIter, iter);
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextIterIsEndRequest(iter));
  return errNone;
}

int prGtkTextIterGetLine(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  UNWRAP_GBOXED(objectSlot, "object", "GtkTextIter", GtkTextIter, iter);
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextIterGetLineRequest(iter));
  return errNone;
}
int prGtkTextIterGetOffset(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  UNWRAP_GBOXED(objectSlot, "object", "GtkTextIter", GtkTextIter, iter);
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextIterGetOffsetRequest(iter));
  return errNone;
}
int prGtkTextIterGetLineOffset(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  UNWRAP_GBOXED(objectSlot, "object", "GtkTextIter", GtkTextIter, iter);
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextIterGetLineOffsetRequest(iter));
  return errNone;
}
int prGtkTextIterGetLineIndex(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  UNWRAP_GBOXED(objectSlot, "object", "GtkTextIter", GtkTextIter, iter);
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextIterGetLineIndexRequest(iter));
  return errNone;
}

int prGtkTextIterSetLineOffset(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *offsetSlot = g->sp - 1;
  UNWRAP_GBOXED(objectSlot, "object", "GtkTextIter", GtkTextIter, iter);
  int offset;
  if (slotIntVal(offsetSlot, &offset)) WRONG_TYPE_PARAMETER("offset", "Integer", offsetSlot);
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextIterSetLineOffsetRequest(iter, offset));
  return errNone;
}
int prGtkTextIterForwardToLineEnd(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  UNWRAP_GBOXED(objectSlot, "object", "GtkTextIter", GtkTextIter, iter);
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextIterForwardToLineEndRequest(iter));
  return errNone;
}
int prGtkTextIterForwardLine(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  UNWRAP_GBOXED(objectSlot, "object", "GtkTextIter", GtkTextIter, iter);
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextIterForwardLineRequest(iter));
  return errNone;
}


int prGtkTextBufferNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextBufferNewRequest());  

  return errNone;
}

int prGtkTextBufferSetText(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *textSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TEXT_BUFFER(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTextBuffer *buffer = GTK_TEXT_BUFFER(objectSlot->uo->slots[0].ui);

  const char *textPtr;
  int textSize;
  if (slotStrPtrLen(textSlot, &textPtr, &textSize, LARGE_STRING_SIZE)) return errWrongType;
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextBufferSetTextRequest(buffer, textPtr, textSize));

  return errNone;
}

int prGtkTextBufferGetText(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 3;
  PyrSlot *beginSlot = g->sp - 2;
  PyrSlot *endSlot = g->sp - 1;
  PyrSlot *hiddenSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TEXT_BUFFER(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTextBuffer *buffer = GTK_TEXT_BUFFER(objectSlot->uo->slots[0].ui);

  if (!IsTrue(hiddenSlot) && !IsFalse(hiddenSlot)) return errWrongType;
  bool hidden = IsTrue(hiddenSlot);
  
  GtkTextIter *beginIter, *endIter;

  if (IsNil(beginSlot)) {
    beginIter = 0;
  } else {
    beginIter = (GtkTextIter *)(beginSlot->uo->slots[0].ui);
  }

  if (IsNil(endSlot)) {
    endIter = 0;
  } else {
    endIter = (GtkTextIter *)(endSlot->uo->slots[0].ui);
  }

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextBufferGetTextRequest(buffer, beginIter, endIter, hidden));

  return errNone;
}

int prGtkTextBufferGetIterAtMark(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *markSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TEXT_BUFFER(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTextBuffer *buffer = GTK_TEXT_BUFFER(objectSlot->uo->slots[0].ui);

  SC_CHECK_VALID_GOBJECT_WRAPPER(markSlot);
  if (!GTK_IS_TEXT_MARK(markSlot->uo->slots[0].ui)) return errWrongType;
  GtkTextMark *mark = GTK_TEXT_MARK(markSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextBufferGetIterAtMarkRequest(buffer, mark));

  return errNone;
}

int prGtkTextBufferGetIterAtLineOffset(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *lineSlot = g->sp - 1;
  PyrSlot *offsetSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TEXT_BUFFER(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTextBuffer *buffer = GTK_TEXT_BUFFER(objectSlot->uo->slots[0].ui);

  int line, offset;
  if (slotIntVal(lineSlot, &line)) return errWrongType;
  if (slotIntVal(offsetSlot, &offset)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextBufferGetIterAtLineOffsetRequest(buffer, line, offset));

  return errNone;
}

int prGtkTextBufferGetIterAtLineIndex(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *lineSlot = g->sp - 1;
  PyrSlot *indexSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TEXT_BUFFER(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTextBuffer *buffer = GTK_TEXT_BUFFER(objectSlot->uo->slots[0].ui);

  int line, index;
  if (slotIntVal(lineSlot, &line)) return errWrongType;
  if (slotIntVal(indexSlot, &index)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextBufferGetIterAtLineIndexRequest(buffer, line, index));

  return errNone;
}

int prGtkTextBufferGetIterAtLine(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *lineSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TEXT_BUFFER(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTextBuffer *buffer = GTK_TEXT_BUFFER(objectSlot->uo->slots[0].ui);

  int line;
  if (slotIntVal(lineSlot, &line)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextBufferGetIterAtLineRequest(buffer, line));

  return errNone;
}

int prGtkTextBufferGetIterAtOffset(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *offsetSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TEXT_BUFFER(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTextBuffer *buffer = GTK_TEXT_BUFFER(objectSlot->uo->slots[0].ui);

  int offset;
  if (slotIntVal(offsetSlot, &offset)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextBufferGetIterAtOffsetRequest(buffer, offset));

  return errNone;
}

int prGtkTextBufferGetStartIter(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TEXT_BUFFER(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTextBuffer *buffer = GTK_TEXT_BUFFER(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextBufferGetStartIterRequest(buffer));

  return errNone;
}

int prGtkTextBufferGetEndIter(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TEXT_BUFFER(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTextBuffer *buffer = GTK_TEXT_BUFFER(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextBufferGetEndIterRequest(buffer));

  return errNone;
}

int prGtkTextBufferGetBounds(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TEXT_BUFFER(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTextBuffer *buffer = GTK_TEXT_BUFFER(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextBufferGetBoundsRequest(buffer));

  return errNone;
}

int prGtkTextBufferInsert(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *iterSlot = g->sp - 1;
  PyrSlot *textSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TEXT_BUFFER(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTextBuffer *buffer = GTK_TEXT_BUFFER(objectSlot->uo->slots[0].ui);

  SC_CHECK_VALID_GBOXED_WRAPPER(iterSlot);
  GtkTextIter *iter = (GtkTextIter *)(iterSlot->uo->slots[0].ui);
  
  const char *textPtr;
  int textSize;
  if (slotStrPtrLen(textSlot, &textPtr, &textSize, LARGE_STRING_SIZE)) return errWrongType;
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextBufferInsertRequest(buffer, iter, textPtr, textSize));

  return errNone;
}

int prGtkTextBufferInsertAtCursor(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *textSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TEXT_BUFFER(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTextBuffer *buffer = GTK_TEXT_BUFFER(objectSlot->uo->slots[0].ui);

  const char *textPtr;
  int textSize;
  if (slotStrPtrLen(textSlot, &textPtr, &textSize, LARGE_STRING_SIZE)) return errWrongType;
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextBufferInsertAtCursorRequest(buffer, textPtr, textSize));

  return errNone;
}

int prGtkTextBufferInsertWithTag(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 3;
  PyrSlot *iterSlot = g->sp - 2;
  PyrSlot *textSlot = g->sp - 1;
  PyrSlot *tagSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TEXT_BUFFER(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTextBuffer *buffer = GTK_TEXT_BUFFER(objectSlot->uo->slots[0].ui);

  SC_CHECK_VALID_GBOXED_WRAPPER(iterSlot);
  GtkTextIter *iter = (GtkTextIter *)(iterSlot->uo->slots[0].ui);
  
  char tagName[128];
  if (slotStrVal(tagSlot, tagName, 128)) return errWrongType;

  const char *textPtr;
  int textSize;
  if (slotStrPtrLen(textSlot, &textPtr, &textSize, LARGE_STRING_SIZE)) return errWrongType;
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextBufferInsertWithTagRequest(buffer, iter, textPtr, textSize, tagName));

  return errNone;
}

int prGtkTextBufferGetMark(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *nameSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TEXT_BUFFER(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTextBuffer *buffer = GTK_TEXT_BUFFER(objectSlot->uo->slots[0].ui);

  const char *namePtr;
  int nameSize;
  if (slotStrPtrLen(nameSlot, &namePtr, &nameSize, MEDIUM_STRING_SIZE))
    WRONG_TYPE_PARAMETER("name", "String", nameSlot);
  
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextBufferGetMarkRequest(buffer, namePtr, nameSize));

  return errNone;
}

int prGtkTextBufferGetInsert(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TEXT_BUFFER(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTextBuffer *buffer = GTK_TEXT_BUFFER(objectSlot->uo->slots[0].ui);
  
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextBufferGetInsertRequest(buffer));

  return errNone;
}

int prGtkTextBufferGetSelectionBound(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TEXT_BUFFER(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTextBuffer *buffer = GTK_TEXT_BUFFER(objectSlot->uo->slots[0].ui);
  
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextBufferGetSelectionBoundRequest(buffer));

  return errNone;
}

int prGtkTextBufferGetSelectionBounds(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TEXT_BUFFER(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTextBuffer *buffer = GTK_TEXT_BUFFER(objectSlot->uo->slots[0].ui);
  
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextBufferGetSelectionBoundsRequest(buffer));

  return errNone;
}

int prGtkTextBufferCreateMark(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 3;
  PyrSlot *nameSlot = g->sp - 2;
  PyrSlot *iterSlot = g->sp - 1;
  PyrSlot *gravitySlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TEXT_BUFFER(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTextBuffer *buffer = GTK_TEXT_BUFFER(objectSlot->uo->slots[0].ui);
  
  SC_CHECK_VALID_GBOXED_WRAPPER(iterSlot);
  GtkTextIter *iter = (GtkTextIter *)iterSlot->uo->slots[0].ui;

  char name[128];
  char *markName;

  if (IsNil(nameSlot)) {
    markName = 0;
  } else {
    if (slotStrVal(nameSlot, name, 128)) return errWrongType;
    markName = name;
  }

  bool leftGravity;
  if (slotBoolVal(gravitySlot, leftGravity)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextBufferCreateMarkRequest(buffer, markName, iter, leftGravity));

  return errNone;
}

int prGtkTextBufferDeleteMark(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *markSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TEXT_BUFFER(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTextBuffer *buffer = GTK_TEXT_BUFFER(objectSlot->uo->slots[0].ui);

  SC_CHECK_VALID_GOBJECT_WRAPPER(markSlot);
  if (!GTK_IS_TEXT_MARK(markSlot->uo->slots[0].ui)) return errWrongType;
  GtkTextMark *mark = GTK_TEXT_MARK(markSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextBufferDeleteMarkRequest(buffer, mark));

  return errNone;
}

int prGtkTextViewNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextViewNewRequest());  

  return errNone;
}

int prGtkTextViewNewWithBuffer(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 1;
  PyrSlot *bufferSlot = g->sp - 0;

  UNWRAP_GOBJECT(bufferSlot, "buffer", "GtkTextBuffer", GtkTextBuffer, buffer, GTK_TEXT_BUFFER, GTK_IS_TEXT_BUFFER);    

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextViewNewWithBufferRequest(buffer));

  return errNone;
}

int prGtkTextViewGetBuffer(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TEXT_VIEW(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTextView *textview = GTK_TEXT_VIEW(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextViewGetBufferRequest(textview));

  return errNone;
}

int prGtkTextViewSetBuffer(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *bufferSlot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "object", "GtkTextView", GtkTextView, textview, GTK_TEXT_VIEW, GTK_IS_TEXT_VIEW);    
  UNWRAP_GOBJECT(bufferSlot, "buffer", "GtkTextBuffer", GtkTextBuffer, buffer, GTK_TEXT_BUFFER, GTK_IS_TEXT_BUFFER);    

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextViewSetBufferRequest(textview, buffer));

  return errNone;
}

int prGtkTextViewSetEditable(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *valueSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TEXT_VIEW(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTextView *textview = GTK_TEXT_VIEW(objectSlot->uo->slots[0].ui);

  bool value;
  if (slotBoolVal(valueSlot, value)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextViewSetEditableRequest(textview, value));

  return errNone;
}

int prGtkTextViewGetEditable(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TEXT_VIEW(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTextView *textview = GTK_TEXT_VIEW(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextViewGetEditableRequest(textview));

  return errNone;
}

int prGtkTextViewSetCursorVisible(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *valueSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TEXT_VIEW(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTextView *textview = GTK_TEXT_VIEW(objectSlot->uo->slots[0].ui);

  bool value;
  if (slotBoolVal(valueSlot, value)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextViewSetCursorVisibleRequest(textview, value));

  return errNone;
}

int prGtkTextViewScrollToMark(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 5;
  PyrSlot *markSlot = g->sp - 4;
  PyrSlot *marginSlot = g->sp - 3;
  PyrSlot *useAlignSlot = g->sp - 2;
  PyrSlot *xAlignSlot = g->sp - 1;
  PyrSlot *yAlignSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_TEXT_VIEW(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkTextView *textview = GTK_TEXT_VIEW(objectSlot->uo->slots[0].ui);

  SC_CHECK_VALID_GOBJECT_WRAPPER(markSlot);
  if (!GTK_IS_TEXT_MARK(markSlot->uo->slots[0].ui)) return errWrongType;
  GtkTextMark *mark = GTK_TEXT_MARK(markSlot->uo->slots[0].ui);
  
  double margin;
  if (slotDoubleVal(marginSlot, &margin)) return errWrongType;

  bool useAlign;
  double xAlign, yAlign;

  if (IsTrue(useAlignSlot)) {
    useAlign = true;
    if (slotDoubleVal(xAlignSlot, &xAlign)) return errWrongType;
    if (slotDoubleVal(yAlignSlot, &yAlign)) return errWrongType;
  } else  {
    useAlign = 0;
    xAlign = yAlign = 0;
  }

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextViewScrollToMarkRequest(textview, mark, margin, useAlign, xAlign, yAlign));

  return errNone;
}

int prGtkTextViewScrollToIter(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 5;
  PyrSlot *iterSlot = g->sp - 4;
  PyrSlot *marginSlot = g->sp - 3;
  PyrSlot *useAlignSlot = g->sp - 2;
  PyrSlot *xAlignSlot = g->sp - 1;
  PyrSlot *yAlignSlot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "object", "GtkTextView", GtkTextView, textview, GTK_TEXT_VIEW, GTK_IS_TEXT_VIEW);
  UNWRAP_GBOXED(iterSlot, "iter", "GtkTextIter", GtkTextIter, iter);
  
  double margin;
  if (slotDoubleVal(marginSlot, &margin)) WRONG_TYPE_PARAMETER("margin", "Float", marginSlot);

  bool useAlign;
  double xAlign, yAlign;

  if (IsTrue(useAlignSlot)) {
    useAlign = true;
    if (slotDoubleVal(xAlignSlot, &xAlign)) WRONG_TYPE_PARAMETER("xAlign", "Float", xAlignSlot);
    if (slotDoubleVal(yAlignSlot, &yAlign)) WRONG_TYPE_PARAMETER("yAlign", "Float", yAlignSlot);
  } else  {
    useAlign = 0;
    xAlign = yAlign = 0;
  }

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTextViewScrollToIterRequest(textview, iter, margin, useAlign, xAlign, yAlign));

  return errNone;
}

int prGtkNotebookNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkNotebookNewRequest());

  return errNone;
}

int prGtkNotebookAppendPage(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *notebookSlot = g->sp - 2;
  PyrSlot *childSlot = g->sp - 1;
  PyrSlot *labelSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(notebookSlot);
  SC_CHECK_VALID_GOBJECT_WRAPPER(childSlot);
  SC_CHECK_VALID_GOBJECT_WRAPPER(labelSlot);

  if (!GTK_IS_NOTEBOOK(notebookSlot->uo->slots[0].ui)) return errWrongType;
  GtkNotebook *notebook = GTK_NOTEBOOK(notebookSlot->uo->slots[0].ui);
  if (!GTK_IS_WIDGET(childSlot->uo->slots[0].ui)) return errWrongType;
  GtkWidget *child = GTK_WIDGET(childSlot->uo->slots[0].ui);
  if (!GTK_IS_WIDGET(labelSlot->uo->slots[0].ui)) return errWrongType;
  GtkWidget *label = GTK_WIDGET(labelSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkNotebookAppendPageRequest(notebook, child, label));

  return errNone;
}

int prGtkNotebookPrependPage(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *notebookSlot = g->sp - 2;
  PyrSlot *childSlot = g->sp - 1;
  PyrSlot *labelSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(notebookSlot);
  SC_CHECK_VALID_GOBJECT_WRAPPER(childSlot);
  SC_CHECK_VALID_GOBJECT_WRAPPER(labelSlot);

  if (!GTK_IS_NOTEBOOK(notebookSlot->uo->slots[0].ui)) return errWrongType;
  GtkNotebook *notebook = GTK_NOTEBOOK(notebookSlot->uo->slots[0].ui);
  if (!GTK_IS_WIDGET(childSlot->uo->slots[0].ui)) return errWrongType;
  GtkWidget *child = GTK_WIDGET(childSlot->uo->slots[0].ui);
  if (!GTK_IS_WIDGET(labelSlot->uo->slots[0].ui)) return errWrongType;
  GtkWidget *label = GTK_WIDGET(labelSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkNotebookPrependPageRequest(notebook, child, label));

  return errNone;
}

int prGtkNotebookInsertPage(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *notebookSlot = g->sp - 3;
  PyrSlot *childSlot = g->sp - 2;
  PyrSlot *labelSlot = g->sp - 1;
  PyrSlot *positionSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(notebookSlot);
  SC_CHECK_VALID_GOBJECT_WRAPPER(childSlot);
  SC_CHECK_VALID_GOBJECT_WRAPPER(labelSlot);
 
  if (!GTK_IS_NOTEBOOK(notebookSlot->uo->slots[0].ui)) return errWrongType;
  GtkNotebook *notebook = GTK_NOTEBOOK(notebookSlot->uo->slots[0].ui);
  if (!GTK_IS_WIDGET(childSlot->uo->slots[0].ui)) return errWrongType;
  GtkWidget *child = GTK_WIDGET(childSlot->uo->slots[0].ui);
  if (!GTK_IS_WIDGET(labelSlot->uo->slots[0].ui)) return errWrongType;
  GtkWidget *label = GTK_WIDGET(labelSlot->uo->slots[0].ui);

  int position;
  if (slotIntVal(positionSlot, &position)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkNotebookInsertPageRequest(notebook, child, label, position));

  return errNone;
}

int prGtkNotebookRemovePage(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *notebookSlot = g->sp - 1;
  PyrSlot *positionSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(notebookSlot);

  if (!GTK_IS_NOTEBOOK(notebookSlot->uo->slots[0].ui)) return errWrongType;
  GtkNotebook *notebook = GTK_NOTEBOOK(notebookSlot->uo->slots[0].ui);

  int position;
  if (slotIntVal(positionSlot, &position)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkNotebookRemovePageRequest(notebook, position));

  return errNone;
}

int prGtkNotebookGetCurrentPage(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *notebookSlot = g->sp - 0;
  SC_CHECK_VALID_GOBJECT_WRAPPER(notebookSlot);
  if (!GTK_IS_NOTEBOOK(notebookSlot->uo->slots[0].ui)) return errWrongType;
  GtkNotebook *notebook = GTK_NOTEBOOK(notebookSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkNotebookGetCurrentPageRequest(notebook));

  return errNone;
}

int prGtkNotebookGetNthPage(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *notebookSlot = g->sp - 1;
  PyrSlot *nSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(notebookSlot);

  if (!GTK_IS_NOTEBOOK(notebookSlot->uo->slots[0].ui)) return errWrongType;
  GtkNotebook *notebook = GTK_NOTEBOOK(notebookSlot->uo->slots[0].ui);

  int n;
  if (slotIntVal(nSlot, &n)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkNotebookGetNthPageRequest(notebook, n));

  return errNone;
}

int prGtkNotebookGetNPages(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *notebookSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(notebookSlot);

  if (!GTK_IS_NOTEBOOK(notebookSlot->uo->slots[0].ui)) return errWrongType;
  GtkNotebook *notebook = GTK_NOTEBOOK(notebookSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkNotebookGetNPagesRequest(notebook));

  return errNone;
}

int prGtkNotebookPageNum(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *notebookSlot = g->sp - 1;
  PyrSlot *childSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(notebookSlot);
  SC_CHECK_VALID_GOBJECT_WRAPPER(childSlot);

  if (!GTK_IS_NOTEBOOK(notebookSlot->uo->slots[0].ui)) return errWrongType;
  GtkNotebook *notebook = GTK_NOTEBOOK(notebookSlot->uo->slots[0].ui);
  if (!GTK_IS_WIDGET(childSlot->uo->slots[0].ui)) return errWrongType;
  GtkWidget *child = GTK_WIDGET(childSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkNotebookPageNumRequest(notebook, child));

  return errNone;
}

int prGtkNotebookSetCurrentPage(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *notebookSlot = g->sp - 1;
  PyrSlot *nSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(notebookSlot);
  if (!GTK_IS_NOTEBOOK(notebookSlot->uo->slots[0].ui)) return errWrongType;
  GtkNotebook *notebook = GTK_NOTEBOOK(notebookSlot->uo->slots[0].ui);

  int n;
  if (slotIntVal(nSlot, &n)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkNotebookSetCurrentPageRequest(notebook, n));

  return errNone;
}

int prGtkNotebookPrevPage(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *notebookSlot = g->sp - 0;
  SC_CHECK_VALID_GOBJECT_WRAPPER(notebookSlot);
  if (!GTK_IS_NOTEBOOK(notebookSlot->uo->slots[0].ui)) return errWrongType;
  GtkNotebook *notebook = GTK_NOTEBOOK(notebookSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkNotebookPrevPageRequest(notebook));

  return errNone;
}

int prGtkNotebookNextPage(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *notebookSlot = g->sp - 0;
  SC_CHECK_VALID_GOBJECT_WRAPPER(notebookSlot);
  if (!GTK_IS_NOTEBOOK(notebookSlot->uo->slots[0].ui)) return errWrongType;
  GtkNotebook *notebook = GTK_NOTEBOOK(notebookSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkNotebookNextPageRequest(notebook));

  return errNone;
}

int prGtkNotebookSetScrollable(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *notebookSlot = g->sp - 1;
  PyrSlot *valueSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(notebookSlot);
  if (!GTK_IS_NOTEBOOK(notebookSlot->uo->slots[0].ui)) return errWrongType;
  GtkNotebook *notebook = GTK_NOTEBOOK(notebookSlot->uo->slots[0].ui);

  bool value;
  if (slotBoolVal(valueSlot, value)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkNotebookSetScrollableRequest(notebook, value));

  return errNone;
}

int prGtkNotebookGetScrollable(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *notebookSlot = g->sp;
  SC_CHECK_VALID_GOBJECT_WRAPPER(notebookSlot);
  if (!GTK_IS_NOTEBOOK(notebookSlot->uo->slots[0].ui)) return errWrongType;
  GtkNotebook *notebook = GTK_NOTEBOOK(notebookSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkNotebookGetScrollableRequest(notebook));

  return errNone;
}

int prGtkNotebookPopupDisable(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *notebookSlot = g->sp;
  SC_CHECK_VALID_GOBJECT_WRAPPER(notebookSlot);
  if (!GTK_IS_NOTEBOOK(notebookSlot->uo->slots[0].ui)) return errWrongType;
  GtkNotebook *notebook = GTK_NOTEBOOK(notebookSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkNotebookPopupDisableRequest(notebook));

  return errNone;
}

int prGtkNotebookPopupEnable(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *notebookSlot = g->sp;
  SC_CHECK_VALID_GOBJECT_WRAPPER(notebookSlot);
  if (!GTK_IS_NOTEBOOK(notebookSlot->uo->slots[0].ui)) return errWrongType;
  GtkNotebook *notebook = GTK_NOTEBOOK(notebookSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkNotebookPopupEnableRequest(notebook));

  return errNone;
}

int prGtkEventBoxNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkEventBoxNewRequest());

  return errNone;
}

int prGtkHScaleNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 1;
  PyrSlot *adjustmentSlot = g->sp - 0;

  GtkAdjustment *adjustment;
  if (IsNil(adjustmentSlot)) {
    adjustment = 0;
  } else {
    SC_CHECK_VALID_GOBJECT_WRAPPER(adjustmentSlot);
    if (!GTK_IS_ADJUSTMENT(adjustmentSlot->uo->slots[0].ui)) return errWrongType;
    adjustment = GTK_ADJUSTMENT(adjustmentSlot->uo->slots[0].ui);
  }

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkHScaleNewRequest(adjustment));

  return errNone;
}

int prGtkHScaleNewWithRange(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 3;
  PyrSlot *minSlot = g->sp - 2;
  PyrSlot *maxSlot = g->sp - 1;
  PyrSlot *stepSlot = g->sp - 0;

  double min, max, step;

  if (slotDoubleVal(minSlot, &min)) return errWrongType;
  if (slotDoubleVal(maxSlot, &max)) return errWrongType;
  if (slotDoubleVal(stepSlot, &step)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkHScaleNewWithRangeRequest(min, max, step));

  return errNone;
}

int prGtkVScaleNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 1;
  PyrSlot *adjustmentSlot = g->sp - 0;

  GtkAdjustment *adjustment;
  if (IsNil(adjustmentSlot)) {
    adjustment = 0;
  } else {
    SC_CHECK_VALID_GOBJECT_WRAPPER(adjustmentSlot);
    if (!GTK_IS_ADJUSTMENT(adjustmentSlot->uo->slots[0].ui)) return errWrongType;
    adjustment = GTK_ADJUSTMENT(adjustmentSlot->uo->slots[0].ui);
  }

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkVScaleNewRequest(adjustment));

  return errNone;
}

int prGtkVScaleNewWithRange(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 3;
  PyrSlot *minSlot = g->sp - 2;
  PyrSlot *maxSlot = g->sp - 1;
  PyrSlot *stepSlot = g->sp - 0;

  double min, max, step;

  if (slotDoubleVal(minSlot, &min)) return errWrongType;
  if (slotDoubleVal(maxSlot, &max)) return errWrongType;
  if (slotDoubleVal(stepSlot, &step)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkVScaleNewWithRangeRequest(min, max, step));

  return errNone;
}

int prGtkScaleSetDigits(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *nSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_SCALE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkScale *scale = GTK_SCALE(objectSlot->uo->slots[0].ui);

  int n;
  if (slotIntVal(nSlot, &n)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkScaleSetDigitsRequest(scale, n));

  return errNone;
}

int prGtkScaleGetDigits(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_SCALE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkScale *scale = GTK_SCALE(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkScaleGetDigitsRequest(scale));

  return errNone;
}

int prGtkScaleSetDrawValue(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *nSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_SCALE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkScale *scale = GTK_SCALE(objectSlot->uo->slots[0].ui);

  bool value;
  if (slotBoolVal(nSlot, value)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkScaleSetDrawValueRequest(scale, value));

  return errNone;
}

int prGtkScaleGetDrawValue(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_SCALE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkScale *scale = GTK_SCALE(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkScaleGetDrawValueRequest(scale));

  return errNone;
}

int prGtkScaleSetValuePosition(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *positionSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_SCALE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkScale *scale = GTK_SCALE(objectSlot->uo->slots[0].ui);

  int position;
  if (slotEnumVal(GTK_TYPE_POSITION_TYPE, positionSlot, position))
    return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkScaleSetValuePositionRequest(scale, GtkPositionType(position)));

  return errNone;
}

int prGtkScaleGetValuePosition(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_SCALE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkScale *scale = GTK_SCALE(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkScaleGetValuePositionRequest(scale));

  return errNone;
}

int prGtkRangeSetUpdatePolicy(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *policySlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_RANGE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkRange *range = GTK_RANGE(objectSlot->uo->slots[0].ui);

  int policy;
  if (slotEnumVal(GTK_TYPE_UPDATE_TYPE, policySlot, policy))
    return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkRangeSetUpdatePolicyRequest(range, GtkUpdateType(policy)));

  return errNone;
}

int prGtkRangeGetUpdatePolicy(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_RANGE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkRange *range = GTK_RANGE(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkRangeGetUpdatePolicyRequest(range));

  return errNone;
}

int prGtkRangeSetValue(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *valueSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_RANGE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkRange *range = GTK_RANGE(objectSlot->uo->slots[0].ui);

  double value;
  if (slotDoubleVal(valueSlot, &value)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkRangeSetValueRequest(range, value));

  return errNone;
}

int prGtkRangeGetValue(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_RANGE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkRange *range = GTK_RANGE(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkRangeGetValueRequest(range));

  return errNone;
}

int prGtkRangeSetAdjustment(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *adjustmentSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_RANGE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkRange *range = GTK_RANGE(objectSlot->uo->slots[0].ui);

  SC_CHECK_VALID_GOBJECT_WRAPPER(adjustmentSlot);
  if (!GTK_IS_ADJUSTMENT(adjustmentSlot->uo->slots[0].ui)) return errWrongType;
  GtkAdjustment *adjustment = GTK_ADJUSTMENT(adjustmentSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkRangeSetAdjustmentRequest(range, adjustment));

  return errNone;
}

int prGtkRangeGetAdjustment(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_RANGE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkRange *range = GTK_RANGE(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkRangeGetAdjustmentRequest(range));

  return errNone;
}

int prGtkRangeSetInverted(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *invertedSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_RANGE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkRange *range = GTK_RANGE(objectSlot->uo->slots[0].ui);

  bool value;
  if (slotBoolVal(invertedSlot, value)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkRangeSetInvertedRequest(range, value));

  return errNone;
}

int prGtkRangeGetInverted(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_RANGE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkRange *range = GTK_RANGE(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkRangeGetInvertedRequest(range));

  return errNone;
}

int prGtkRangeSetIncrements(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *stepSlot = g->sp - 1;
  PyrSlot *pageSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_RANGE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkRange *range = GTK_RANGE(objectSlot->uo->slots[0].ui);

  double step;
  if (slotDoubleVal(stepSlot, &step)) return errWrongType;

  double page;
  if (slotDoubleVal(pageSlot, &page)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkRangeSetIncrementsRequest(range, step, page));

  return errNone;
}

int prGtkRangeSetRange(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *minSlot = g->sp - 1;
  PyrSlot *maxSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_RANGE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkRange *range = GTK_RANGE(objectSlot->uo->slots[0].ui);

  double min;
  if (slotDoubleVal(minSlot, &min)) return errWrongType;

  double max;
  if (slotDoubleVal(maxSlot, &max)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkRangeSetRangeRequest(range, min, max));

  return errNone;
}

int prGtkEditableSetEditable(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *editableSlot = g->sp - 0;
  
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_EDITABLE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkEditable *editable = GTK_EDITABLE(objectSlot->uo->slots[0].ui);

  bool value;
  if (slotBoolVal(editableSlot, value)) return errWrongType;
 
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkEditableSetEditableRequest(editable, value));

  return errNone;
}

int prGtkEditableGetEditable(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;
  
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_EDITABLE(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkEditable *editable = GTK_EDITABLE(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkEditableGetEditableRequest(editable));

  return errNone;
}

int prGtkSpinButtonNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 3;
  PyrSlot *adjustmentSlot = g->sp - 2;
  PyrSlot *climbRateSlot = g->sp - 1;
  PyrSlot *digitsSlot = g->sp - 0;

  GtkAdjustment *adjustment;
  SC_CHECK_VALID_GOBJECT_WRAPPER(adjustmentSlot);
  if (!GTK_IS_ADJUSTMENT(adjustmentSlot->uo->slots[0].ui)) return errWrongType;
  adjustment = GTK_ADJUSTMENT(adjustmentSlot->uo->slots[0].ui);
  
  double climbRate;
  if(slotDoubleVal(climbRateSlot, &climbRate)) return errWrongType;

  int digits;
  if(slotIntVal(climbRateSlot, &digits)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSpinButtonNewRequest(adjustment, climbRate, digits));

  return errNone;
}

int prGtkSpinButtonNewWithRange(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 3;
  PyrSlot *minSlot = g->sp - 2;
  PyrSlot *maxSlot = g->sp - 1;
  PyrSlot *stepSlot = g->sp - 0;
 
  double min, max, step;
  if(slotDoubleVal(minSlot, &min)) return errWrongType;
  if(slotDoubleVal(maxSlot, &max)) return errWrongType;
  if(slotDoubleVal(stepSlot, &step)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSpinButtonNewWithRangeRequest(min, max, step));

  return errNone;
}

int prGtkSpinButtonSetAdjustment(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *adjustmentSlot = g->sp - 0;
  
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_SPIN_BUTTON(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkSpinButton *spinbutton = GTK_SPIN_BUTTON(objectSlot->uo->slots[0].ui);

  SC_CHECK_VALID_GOBJECT_WRAPPER(adjustmentSlot);
  if (!GTK_IS_ADJUSTMENT(adjustmentSlot->uo->slots[0].ui)) return errWrongType;
  GtkAdjustment *adjustment = GTK_ADJUSTMENT(adjustmentSlot->uo->slots[0].ui);
  
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSpinButtonSetAdjustmentRequest(spinbutton, adjustment));

  return errNone;
}

int prGtkSpinButtonGetAdjustment(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;
  
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_SPIN_BUTTON(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkSpinButton *spinbutton = GTK_SPIN_BUTTON(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSpinButtonGetAdjustmentRequest(spinbutton));

  return errNone;
}

int prGtkSpinButtonSetDigits(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *digitsSlot = g->sp - 0;
  
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_SPIN_BUTTON(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkSpinButton *spinbutton = GTK_SPIN_BUTTON(objectSlot->uo->slots[0].ui);

  int digits;
  if(slotIntVal(digitsSlot, &digits)) return errWrongType;
  
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSpinButtonSetDigitsRequest(spinbutton, digits));

  return errNone;
}

int prGtkSpinButtonGetDigits(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;
  
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_SPIN_BUTTON(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkSpinButton *spinbutton = GTK_SPIN_BUTTON(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSpinButtonGetDigitsRequest(spinbutton));

  return errNone;
}

int prGtkSpinButtonSetIncrements(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *stepSlot = g->sp - 1;
  PyrSlot *pageSlot = g->sp - 0;
  
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_SPIN_BUTTON(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkSpinButton *spinbutton = GTK_SPIN_BUTTON(objectSlot->uo->slots[0].ui);

  double step, page;
  if(slotDoubleVal(stepSlot, &step)) return errWrongType;
  if(slotDoubleVal(pageSlot, &page)) return errWrongType;
 
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSpinButtonSetIncrementsRequest(spinbutton, step, page));

  return errNone;
}

int prGtkSpinButtonGetIncrements(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;
  
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_SPIN_BUTTON(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkSpinButton *spinbutton = GTK_SPIN_BUTTON(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSpinButtonGetIncrementsRequest(spinbutton));

  return errNone;
}

int prGtkSpinButtonSetRange(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *minSlot = g->sp - 1;
  PyrSlot *maxSlot = g->sp - 0;
  
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_SPIN_BUTTON(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkSpinButton *spinbutton = GTK_SPIN_BUTTON(objectSlot->uo->slots[0].ui);

  double min, max;
  if(slotDoubleVal(minSlot, &min)) return errWrongType;
  if(slotDoubleVal(maxSlot, &max)) return errWrongType;
 
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSpinButtonSetRangeRequest(spinbutton, min, max));

  return errNone;
}

int prGtkSpinButtonGetRange(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;
  
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_SPIN_BUTTON(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkSpinButton *spinbutton = GTK_SPIN_BUTTON(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSpinButtonGetRangeRequest(spinbutton));

  return errNone;
}

int prGtkSpinButtonSetValue(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *valueSlot = g->sp - 0;
  
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_SPIN_BUTTON(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkSpinButton *spinbutton = GTK_SPIN_BUTTON(objectSlot->uo->slots[0].ui);

  double value;
  if(slotDoubleVal(valueSlot, &value)) return errWrongType;
 
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSpinButtonSetValueRequest(spinbutton, value));

  return errNone;
}

int prGtkSpinButtonGetValue(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;
  
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_SPIN_BUTTON(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkSpinButton *spinbutton = GTK_SPIN_BUTTON(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSpinButtonGetValueRequest(spinbutton));

  return errNone;
}

int prGtkSpinButtonGetValueAsInteger(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;
  
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_SPIN_BUTTON(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkSpinButton *spinbutton = GTK_SPIN_BUTTON(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSpinButtonGetValueAsIntegerRequest(spinbutton));

  return errNone;
}

int prGtkSpinButtonSetUpdatePolicy(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *policySlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_SPIN_BUTTON(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkSpinButton *spinbutton = GTK_SPIN_BUTTON(objectSlot->uo->slots[0].ui);

  int policy;
  if (slotEnumVal(GTK_TYPE_SPIN_BUTTON_UPDATE_POLICY, policySlot, policy))
    return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSpinButtonSetUpdatePolicyRequest(spinbutton, GtkSpinButtonUpdatePolicy(policy)));

  return errNone;
}

int prGtkSpinButtonGetUpdatePolicy(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;
  
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_SPIN_BUTTON(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkSpinButton *spinbutton = GTK_SPIN_BUTTON(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSpinButtonGetUpdatePolicyRequest(spinbutton));

  return errNone;
}

int prGtkSpinButtonSetNumeric(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *numericSlot = g->sp - 0;
  
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_SPIN_BUTTON(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkSpinButton *spinbutton = GTK_SPIN_BUTTON(objectSlot->uo->slots[0].ui);

  bool value;
  if (slotBoolVal(numericSlot, value)) return errWrongType;
 
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSpinButtonSetNumericRequest(spinbutton, value));

  return errNone;
}

int prGtkSpinButtonGetNumeric(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;
  
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_SPIN_BUTTON(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkSpinButton *spinbutton = GTK_SPIN_BUTTON(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSpinButtonGetNumericRequest(spinbutton));

  return errNone;
}

int prGtkSpinButtonSetWrap(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *wrapSlot = g->sp - 0;
  
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_SPIN_BUTTON(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkSpinButton *spinbutton = GTK_SPIN_BUTTON(objectSlot->uo->slots[0].ui);

  bool value;
  if (slotBoolVal(wrapSlot, value)) return errWrongType;
 
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSpinButtonSetWrapRequest(spinbutton, value));

  return errNone;
}

int prGtkSpinButtonGetWrap(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;
  
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_SPIN_BUTTON(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkSpinButton *spinbutton = GTK_SPIN_BUTTON(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSpinButtonGetWrapRequest(spinbutton));

  return errNone;
}

int prGtkSpinButtonSpin(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *typeSlot = g->sp - 1;
  PyrSlot *amountSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_SPIN_BUTTON(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkSpinButton *spinbutton = GTK_SPIN_BUTTON(objectSlot->uo->slots[0].ui);

  int type;
  if (slotEnumVal(GTK_TYPE_SPIN_TYPE, typeSlot, type))
    return errWrongType;

  double amount;
  if(slotDoubleVal(amountSlot, &amount)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSpinButtonSpinRequest(spinbutton, GtkSpinType(type), amount));

  return errNone;
}

int prGtkSpinButtonSetSnapToTicks(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *snapToTicksSlot = g->sp - 0;
  
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_SPIN_BUTTON(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkSpinButton *spinbutton = GTK_SPIN_BUTTON(objectSlot->uo->slots[0].ui);

  bool value;
  if (slotBoolVal(snapToTicksSlot, value)) return errWrongType;
 
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSpinButtonSetSnapToTicksRequest(spinbutton, value));

  return errNone;
}

int prGtkSpinButtonGetSnapToTicks(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;
  
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_SPIN_BUTTON(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkSpinButton *spinbutton = GTK_SPIN_BUTTON(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSpinButtonGetSnapToTicksRequest(spinbutton));

  return errNone;
}

int prGtkSpinButtonUpdate(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;
  
  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_SPIN_BUTTON(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkSpinButton *spinbutton = GTK_SPIN_BUTTON(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSpinButtonUpdateRequest(spinbutton));

  return errNone;
}

int prGtkComboNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *resultSlot = g->sp - 0;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkComboNewRequest());

  return errNone;
}

int prGtkComboSetPopdownStrings(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *stringArraySlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_COMBO(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkCombo *combo = GTK_COMBO(objectSlot->uo->slots[0].ui);

  if (stringArraySlot->utag != tagObj || !isKindOf(stringArraySlot->uo, class_array)) return errWrongType;
  PyrObject *array = stringArraySlot->uo;
  
  int n = array->size;
  if (n > MAX_COMBO_STRINGS) {
    n = MAX_COMBO_STRINGS;
    std::cout << "Dropping combo items" << std::endl;
  }

  ComboStringArray strings;

  // XXX: we can avoid this copy
  //      Since the calling methods won't terminate before this request
  //      has been performed and the result has been delivered to the language
  //      we are sure that the sclang strings won't be garbage collected
  int i;
  for (i = 0; i < n; ++i) {
    if (slotStrVal(array->slots + i, &(strings[i][0]), MEDIUM_STRING_SIZE)) return errWrongType;
  }

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkComboSetPopdownStringsRequest(combo, n, strings));

  return errNone;
}

int prGtkComboSetValueInList(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *valueSlot = g->sp - 1;
  PyrSlot *okIfEmptySlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_COMBO(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkCombo *combo = GTK_COMBO(objectSlot->uo->slots[0].ui);

  bool value;
  if (slotBoolVal(valueSlot, value)) return errWrongType;
  bool okIfEmpty;
  if (slotBoolVal(okIfEmptySlot, okIfEmpty)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkComboSetValueInListRequest(combo, value, okIfEmpty));

  return errNone;
}

int prGtkComboSetUseArrows(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *valueSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_COMBO(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkCombo *combo = GTK_COMBO(objectSlot->uo->slots[0].ui);

  bool value;
  if (slotBoolVal(valueSlot, value)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkComboSetUseArrowsRequest(combo, value));

  return errNone;
}

int prGtkComboSetCaseSensitive(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *valueSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_COMBO(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkCombo *combo = GTK_COMBO(objectSlot->uo->slots[0].ui);

  bool value;
  if (slotBoolVal(valueSlot, value)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkComboSetCaseSensitiveRequest(combo, value));

  return errNone;
}

int prGtkComboSetItemString(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *itemSlot = g->sp - 1;
  PyrSlot *stringSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_COMBO(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkCombo *combo = GTK_COMBO(objectSlot->uo->slots[0].ui);

  SC_CHECK_VALID_GOBJECT_WRAPPER(itemSlot);
  if (!GTK_IS_ITEM(itemSlot->uo->slots[0].ui)) return errWrongType;
  GtkItem *item = GTK_ITEM(itemSlot->uo->slots[0].ui);

  const char *stringPtr;
  int stringSize;
  if (slotStrPtrLen(stringSlot, &stringPtr, &stringSize, MEDIUM_STRING_SIZE)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkComboSetItemStringRequest(combo, item, stringPtr, stringSize));

  return errNone;
}

int prGtkComboDisableActivate(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_COMBO(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkCombo *combo = GTK_COMBO(objectSlot->uo->slots[0].ui);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkComboDisableActivateRequest(combo));

  return errNone;
}

int prGtkComboEntry(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_COMBO(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkCombo *combo = GTK_COMBO(objectSlot->uo->slots[0].ui);

  SetInt(objectSlot, int(combo->entry));

  return errNone;
}

int prGtkComboList(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GOBJECT_WRAPPER(objectSlot);
  if (!GTK_IS_COMBO(objectSlot->uo->slots[0].ui)) return errWrongType;
  GtkCombo *combo = GTK_COMBO(objectSlot->uo->slots[0].ui);

  SetInt(objectSlot, int(combo->list));

  return errNone;
}

int prGtkTreePathNewFromString(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 2;
  PyrSlot *modelSlot = g->sp - 1;
  PyrSlot *pathSlot = g->sp;

  SC_CHECK_VALID_GOBJECT_WRAPPER(modelSlot);
  if (!GTK_IS_TREE_MODEL(modelSlot->uo->slots[0].ui)) return errWrongType;
  GtkTreeModel *model = GTK_TREE_MODEL(modelSlot->uo->slots[0].ui);

  const char *pathPtr;
  int pathSize;
  if (slotStrPtrLen(pathSlot, &pathPtr, &pathSize, SMALL_STRING_SIZE)) return errWrongType;
  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreePathNewFromStringRequest(model, pathPtr, pathSize));  

  return errNone;
}

int prGtkTreePathAsString(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  GtkTreePath *path = (GtkTreePath *)objectSlot->uo->slots[0].ui;

  if (path == 0) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreePathAsStringRequest(path));

  return errNone;
}

int prGtkTreePathGetIndices(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  GtkTreePath *path = (GtkTreePath *)objectSlot->uo->slots[0].ui;

  if (path == 0) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkTreePathGetIndicesRequest(path));

  return errNone;
}

int prGladeXmlNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 2;
  PyrSlot *filenameSlot = g->sp - 2;
  PyrSlot *rootSlot = g->sp - 1;
  PyrSlot *lamespaceSlot = g->sp - 0;

  const char *filename;
  int filenamesize;
  if (slotStrPtrLen(filenameSlot, &filename, &filenamesize, FILENAME_STRING_SIZE)) return errWrongType;

  const char *root;
  int rootsize;

  if (IsNil(rootSlot)) {
    root = 0;
  } else {
    if (slotStrPtrLen(rootSlot, &root, &rootsize, SMALL_STRING_SIZE)) return errWrongType;
  }

  const char *lamespace;
  int lamespacesize;

  if (IsNil(lamespaceSlot)) {
    lamespace = 0;
  } else {
    if (slotStrPtrLen(lamespaceSlot, &lamespace, &lamespacesize, SMALL_STRING_SIZE)) return errWrongType;
  }

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GladeXMLNewRequest(filename, filenamesize, root, rootsize, lamespace, lamespacesize));

  return errNone;
}

int prGladeXmlSignalAutoconnect(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "object", "GladeXML", GladeXML, object, GLADE_XML, GLADE_IS_XML);    

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GladeXMLSignalAutoconnectRequest(object, objectSlot->uo));

  return errNone;
}

int prGtkSelectionDataSetToPointer(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *ptrSlot = g->sp - 0;

  SC_CHECK_VALID_GBOXED_WRAPPER(objectSlot);
  GtkSelectionData *data = (GtkSelectionData *)objectSlot->uo->slots[0].ui;

  int ptr;
  if (slotIntVal(ptrSlot, &ptr)) return errWrongType;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSelectionDataSetToPointerRequest(data, ptr));

  return errNone;
}

int prGtkSelectionDataGetAsPointer(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  SC_CHECK_VALID_GBOXED_WRAPPER(objectSlot);
  GtkSelectionData *data = (GtkSelectionData *)objectSlot->uo->slots[0].ui;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSelectionDataGetAsPointerRequest(data));

  return errNone;
}

int prGtkHSeparatorNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkHSeparatorNewRequest());

  return errNone;
}

int prGtkVSeparatorNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkVSeparatorNewRequest());

  return errNone;
}

int prGnomeCanvasNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *antialiasedSlot = g->sp - 0;
  
  bool antialiased;
  if (slotBoolVal(antialiasedSlot, antialiased)) WRONG_TYPE_PARAMETER("antialised", "Boolean", antialiasedSlot);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GnomeCanvasNewRequest(antialiased));

  return errNone;
}

int prGnomeCanvasSetScrollRegion(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 4;
  PyrSlot *xSlot = g->sp - 3;
  PyrSlot *ySlot = g->sp - 2;
  PyrSlot *wSlot = g->sp - 1;
  PyrSlot *hSlot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "object", "GnomeCanvas", GnomeCanvas, object, GNOME_CANVAS, GNOME_IS_CANVAS);  

  double x, y, w, h;
  if (slotDoubleVal(xSlot, &x)) WRONG_TYPE_PARAMETER("x", "Double", xSlot);
  if (slotDoubleVal(ySlot, &y)) WRONG_TYPE_PARAMETER("y", "Double", ySlot);
  if (slotDoubleVal(wSlot, &w)) WRONG_TYPE_PARAMETER("w", "Double", wSlot);
  if (slotDoubleVal(hSlot, &h)) WRONG_TYPE_PARAMETER("h", "Double", hSlot);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GnomeCanvasSetScrollRegionRequest(object, x, y, w, h));

  return errNone;
}

int prGnomeCanvasRoot(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "object", "GnomeCanvas", GnomeCanvas, object, GNOME_CANVAS, GNOME_IS_CANVAS);  

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GnomeCanvasRootRequest(object));

  return errNone;
}

bool slotCanvasItemType(PyrSlot *slot, GType *type) {
  char itemtypename[SMALL_STRING_SIZE];
  if (slotStrVal(slot, itemtypename, SMALL_STRING_SIZE))
    return errWrongType;
  *type = g_type_from_name(itemtypename);
  if (*type == G_TYPE_INVALID)
    return errFailed;
  else
    return errNone;
}

int prGnomeCanvasGroupAdd(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *typeSlot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "object", "GnomeCanvasGroup", GnomeCanvasGroup, object, GNOME_CANVAS_GROUP, GNOME_IS_CANVAS_GROUP);  

  GType type;
  if (slotCanvasItemType(typeSlot, &type)) WRONG_TYPE_PARAMETER("type", "String", typeSlot);

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GnomeCanvasGroupAddRequest(object, type));

  return errNone;
}

int prGnomeCanvasItemSet(struct VMGlobals *g, int numArgsPushed) {
  int result;

  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *propertyNameSlot = g->sp - 1;
  PyrSlot *valueSlot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "item", "GnomeCanvasItem", GnomeCanvasItem, item, GNOME_CANVAS_ITEM, GNOME_IS_CANVAS_ITEM);

  char propertyName[SMALL_STRING_SIZE];
  if (slotStrVal(propertyNameSlot, propertyName, SMALL_STRING_SIZE))
    WRONG_TYPE_PARAMETER("propertyName", "String", propertyNameSlot);
    
  GObjectClass *gclass = G_OBJECT_GET_CLASS(item);

  GParamSpec *gparam = g_object_class_find_property(gclass, propertyName);
  if (gparam == 0) {
    error("Unknown property name\n");
    return errFailed;
  }

  GValue value = {0};
  
  g_value_init(&value, gparam->value_type);
  
  if (slot_to_GValue(valueSlot, &value)) {
    result = errWrongType;
  } else {
    GTK_REQUEST_QUEUE_SEND_OR_FAIL(GnomeCanvasItemSetRequest(item, propertyName, &value));    
    result = errNone;
  }
  
  g_value_unset(&value);
  
  return result;
}

int prGnomeCanvasItemSetPoints(struct VMGlobals *g, int numArgsPushed) {
  int result;

  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *coordinatesSlot = g->sp - 0;

  UNWRAP_GOBJECT(objectSlot, "item", "GnomeCanvasItem", GnomeCanvasItem, item, GNOME_CANVAS_ITEM, GNOME_IS_CANVAS_ITEM);

  if (!isKindOfSlot(coordinatesSlot, class_array)) {
    WRONG_TYPE_PARAMETER("coordinates", "Array", coordinatesSlot);
  }

  PyrObject *array = coordinatesSlot->uo;

  if ((array->size % 2) != 0) {
    error("Parameter 'coordinates' must be an array with even size\n");
    return errWrongType;
  }

  PyrSlot *slots = array->slots;
  int n;

  if (n <= MAX_GNOME_ITEMS_POINTS * 2) {
    n = array->size / 2;
  } else {
    n = MAX_GNOME_ITEMS_POINTS;
    std::cout << "dropping gnome canvas item points" << std::endl; 
  }
  
  GnomeItemPointsArray points;

  for (int i = 0; i < n *2; ++i) {
    double value;
    if (slotDoubleVal(slots + i, &value)) {
      error("All elements of coordinates mus be of type Float. Got:\n");
      dumpObjectSlot(coordinatesSlot);
      return errWrongType;
    }
    points[i] = value;
  }

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GnomeCanvasItemSetPointsRequest(item, points, n));    
    
  return errNone;
}

// Gtk

int prGtkMainQuit(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;
  gtk_main_quit();
  return errNone;
}


// GtkSourceView

int prGtkSourceBufferNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp;

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSourceBufferNewRequest());  

  return errNone;
}

int prGtkSourceViewNewWithBuffer(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *classSlot = g->sp - 1;
  PyrSlot *bufferSlot = g->sp - 0;

  UNWRAP_GOBJECT(bufferSlot, "buffer", "GtkSourceBuffer", GtkSourceBuffer, buffer, GTK_SOURCE_BUFFER, GTK_IS_SOURCE_BUFFER);    

  GTK_REQUEST_QUEUE_SEND_OR_FAIL(GtkSourceViewNewWithBufferRequest(buffer));

  return errNone;
}

// ------ AsyncRequest support

static int prRoutineWaitAsyncAnswer(struct VMGlobals *g, int numArgsPushed)
{
#if 0
	PyrSlot value;
	
	//postfl("->prRoutineYield %d %08X\n", g->level, g->thread);
	//assert(g->gc->SanityCheck());
	//CallStackSanity(g);
	//postfl("->numArgsPushed %d\n", numArgsPushed);
	if (!isKindOf((PyrObject*)g->thread, class_routine)) {
		//debugf ("yield was called outside of a Routine.\n");
		error ("yield was called outside of a Routine.\n");
		return errFailed;
	}
	/*if (!g->thread->parent.uot) {
		error ("yield was called from a thread with no parent.\n");
		return errFailed;
	}*/
	value.ucopy = g->sp->ucopy;

	//debugf("yield from thread %08X to parent %08X\n", g->thread, g->thread->parent.uot);
	switchToThread(g, g->thread->parent.uot, tYieldToParent, &numArgsPushed);
	// on the other side of the looking glass, put the yielded value on the stack as the result..
	(g->sp - numArgsPushed + 1)->ucopy = value.ucopy;
	//postfl("<-numArgsPushed %d\n", numArgsPushed);
	//postfl("<-prRoutineYield %d\n", g->level);
	//assert(g->gc->SanityCheck());
	//CallStackSanity(g);
	return errNone;
#endif
}

// -------

int prGetGtkTextBufferForConsoleOutput(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 0;
  SetInt(objectSlot, int(SclangInterpreter::get_instance()->consoleBuffer()));
  return errNone;
}


void initialize_gtk_primitives(void) {
  int base = nextPrimitiveIndex();
  int index = 0;

  definePrimitive(base, index++, "_prGTypeNew", prGTypeNew, 2, 0);
  definePrimitive(base, index++, "_prGTypeRegisterWrapperClass", prGTypeRegisterWrapperClass, 2, 0);

  definePrimitive(base, index++, "_prGtkTimeoutAdd", prGtkTimeoutAdd, 3, 0);
  definePrimitive(base, index++, "_prGtkTimeoutRemove", prGtkTimeoutRemove, 2, 0);

  definePrimitive(base, index++, "_prGtkServerSendResultTrigger", prGtkServerSendResultTrigger, 1, 0);
  definePrimitive(base, index++, "_prGtkServerSendVoidTrigger", prGtkServerSendVoidTrigger, 1, 0);
  definePrimitive(base, index++, "_prGtkServerSendAbortTrigger", prGtkServerSendAbortTrigger, 1, 0);

  definePrimitive(base, index++, "_prGObjectDump", prGObjectDump, 1, 0);
  definePrimitive(base, index++, "_prGObjectGetTypeName", prGObjectGetTypeName, 1, 0);
  definePrimitive(base, index++, "_prGObjectGetObjectTypeName", prGObjectGetObjectTypeName, 2, 0);
  definePrimitive(base, index++, "_prGObjectGetProperty", prGObjectGetProperty, 2, 0);
  definePrimitive(base, index++, "_prGObjectSetProperty", prGObjectSetProperty, 3, 0);
  definePrimitive(base, index++, "_prGObjectGetClassName", prGObjectGetClassName, 1, 0);
  definePrimitive(base, index++, "_prGObjectIsObject", prGObjectIsObject, 1, 0);
  definePrimitive(base, index++, "_prGObjectRef", prGObjectRef, 1, 0);
  definePrimitive(base, index++, "_prGObjectUnref", prGObjectUnref, 1, 0);
  definePrimitive(base, index++, "_prGObjectConnect", prGObjectConnect, 4, 0);
  definePrimitive(base, index++, "_prGObjectGetData", prGObjectGetData, 2, 0);
  definePrimitive(base, index++, "_prGObjectDisconnect", prGObjectDisconnect, 2, 0);
  definePrimitive(base, index++, "_prGObjectAsInteger", prGObjectAsInteger, 1, 0);
  definePrimitive(base, index++, "_prGObjectAddStandardWeakReference", prGObjectAddStandardWeakReference, 1, 0);
  definePrimitive(base, index++, "_prGtkObjectConnectStandardDestroySignalHandler",
		  prGtkObjectConnectStandardDestroySignalHandler, 1, 0);

  definePrimitive(base, index++, "_prGBoxedWrap", prGBoxedWrap, 4, 0); 
  definePrimitive(base, index++, "_prGBoxedFree", prGBoxedFree, 1, 0);
  definePrimitive(base, index++, "_prGBoxedCopy", prGBoxedCopy, 1, 0);
  definePrimitive(base, index++, "_prGBoxedRegisterWrapperClass", prGBoxedRegisterWrapperClass, 2, 0);

  definePrimitive(base, index++, "_prGdkScreenWidth", prGdkScreenWidth, 1, 0);
  definePrimitive(base, index++, "_prGdkScreenHeight", prGdkScreenHeight, 1, 0);
  definePrimitive(base, index++, "_prGdkSetDeviceModeByName", prGdkSetDeviceModeByName, 3, 0);

  definePrimitive(base, index++, "_prGdkGCNew", prGdkGCNew, 2, 0);
  definePrimitive(base, index++, "_prGdkGCSetRgbFgColorByName", prGdkGCSetRgbFgColorByName, 2, 0);
  definePrimitive(base, index++, "_prGdkGCSetRgbBgColorByName", prGdkGCSetRgbBgColorByName, 2, 0);
  definePrimitive(base, index++, "_prGdkGCSetLineAttributes", prGdkGCSetLineAttributes, 5, 0);
  definePrimitive(base, index++, "_prGdkGCSetRgbFgColor", prGdkGCSetRgbFgColor, 2, 0);
  definePrimitive(base, index++, "_prGdkGCSetRgbBgColor", prGdkGCSetRgbBgColor, 2, 0);

  definePrimitive(base, index++, "_prGdkDrawableDrawRectangle", prGdkDrawableDrawRectangle, 7, 0);
  definePrimitive(base, index++, "_prGdkDrawableDrawLine", prGdkDrawableDrawLine, 6, 0);
  definePrimitive(base, index++, "_prGdkDrawableDrawPoint", prGdkDrawableDrawPoint, 4, 0);
  definePrimitive(base, index++, "_prGdkDrawableDrawArc", prGdkDrawableDrawArc, 9, 0);
  definePrimitive(base, index++, "_prGdkDrawableDrawLayout", prGdkDrawableDrawLayout, 5, 0);
  definePrimitive(base, index++, "_prGdkDrawableGetSize", prGdkDrawableGetSize, 1, 0);

  definePrimitive(base, index++, "_prGdkEventGetCoords", prGdkEventGetCoords, 1, 0);
  definePrimitive(base, index++, "_prGdkEventGetState", prGdkEventGetState, 1, 0);
  definePrimitive(base, index++, "_prGdkEventGetRootCoords", prGdkEventGetRootCoords, 1, 0);
  definePrimitive(base, index++, "_prGdkEventGetKeyval", prGdkEventGetKeyval, 1, 0);
  definePrimitive(base, index++, "_prGdkEventGetKeyname", prGdkEventGetKeyname, 1, 0);
  definePrimitive(base, index++, "_prGdkEventGetButton", prGdkEventGetButton, 1, 0);
  definePrimitive(base, index++, "_prGdkEventGetTime", prGdkEventGetTime, 1, 0);
  definePrimitive(base, index++, "_prGdkEventGetType", prGdkEventGetType, 1, 0);
  definePrimitive(base, index++, "_prGdkEventIsHint", prGdkEventIsHint, 1, 0);
  definePrimitive(base, index++, "_prGdkEventGetAxis", prGdkEventGetAxis, 2, 0);
  definePrimitive(base, index++, "_prGdkEventGetDeviceState", prGdkEventGetDeviceState, 1, 0);

  definePrimitive(base, index++, "_prGdkColorNew", prGdkPixmapCreateFromXpm, 3, 0);
  definePrimitive(base, index++, "_prGdkPixmapCreateFromXpm", prGdkPixmapCreateFromXpm, 3, 0);
  //definePrimitive(base, index++, "_prGdkCursorNew", prGdkCursorNew, 2, 0);
  definePrimitive(base, index++, "_prGdkWindowSetCursor", prGdkWindowSetCursor, 2, 0);

  definePrimitive(base, index++, "_prGtkObjectSink", prGtkObjectSink, 1, 0);
  definePrimitive(base, index++, "_prGtkMainQuit", prGtkMainQuit, 1, 0);

  definePrimitive(base, index++, "_prGtkWidgetSetSizeRequest", prGtkWidgetSetSizeRequest, 3, 0);
  definePrimitive(base, index++, "_prGtkWidgetShowAll", prGtkWidgetShowAll, 1, 0);
  definePrimitive(base, index++, "_prGtkWidgetShow", prGtkWidgetShow, 1, 0);
  definePrimitive(base, index++, "_prGtkWidgetShowAll", prGtkWidgetShowAll, 1, 0);
  definePrimitive(base, index++, "_prGtkWidgetHide", prGtkWidgetHide, 1, 0);
  definePrimitive(base, index++, "_prGtkWidgetHideAll", prGtkWidgetHideAll, 1, 0);
  definePrimitive(base, index++, "_prGtkWidgetDestroy", prGtkWidgetDestroy, 1, 0);
  definePrimitive(base, index++, "_prGtkWidgetQueueDraw", prGtkWidgetQueueDraw, 1, 0);
  definePrimitive(base, index++, "_prGtkWidgetQueueResize", prGtkWidgetQueueResize, 1, 0);
  definePrimitive(base, index++, "_prGtkWidgetSetEvents", prGtkWidgetSetEvents, 2, 0);
  definePrimitive(base, index++, "_prGtkWidgetSetExtensionEvents", prGtkWidgetSetExtensionEvents, 2, 0);
  definePrimitive(base, index++, "_prGtkWidgetAddEvents", prGtkWidgetAddEvents, 2, 0);
  definePrimitive(base, index++, "_prGtkWidgetWindow", prGtkWidgetWindow, 1, 0);
  definePrimitive(base, index++, "_prGtkWidgetGetStyle", prGtkWidgetGetStyle, 1, 0);
  definePrimitive(base, index++, "_prGtkWidgetGetPangoContext", prGtkWidgetGetPangoContext, 1, 0);
  definePrimitive(base, index++, "_prGtkWidgetIsRealized", prGtkWidgetIsRealized, 1, 0);
  definePrimitive(base, index++, "_prGtkWidgetIsToplevel", prGtkWidgetIsToplevel, 1, 0);
  definePrimitive(base, index++, "_prGtkWidgetIsMapped", prGtkWidgetIsMapped, 1, 0);
  definePrimitive(base, index++, "_prGtkWidgetIsVisible", prGtkWidgetIsVisible, 1, 0);
  definePrimitive(base, index++, "_prGtkWidgetCanFocus", prGtkWidgetCanFocus, 1, 0);
  definePrimitive(base, index++, "_prGtkWidgetGetSizeRequest", prGtkWidgetGetSizeRequest, 1, 0);
  definePrimitive(base, index++, "_prGtkWidgetGetParent", prGtkWidgetGetParent, 1, 0);
  definePrimitive(base, index++, "_prGtkWidgetGetParentWindow", prGtkWidgetGetParentWindow, 1, 0);
  definePrimitive(base, index++, "_prGtkWidgetGetDoubleBuffered", prGtkWidgetGetDoubleBuffered, 1, 0);
  definePrimitive(base, index++, "_prGtkWidgetSetDoubleBuffered", prGtkWidgetSetDoubleBuffered, 2, 0);
  definePrimitive(base, index++, "_prGtkWidgetGrabFocus", prGtkWidgetGrabFocus, 1, 0);
  definePrimitive(base, index++, "_prGtkWidgetSetState", prGtkWidgetSetState, 2, 0);
  definePrimitive(base, index++, "_prGtkWidgetGetToplevel", prGtkWidgetGetToplevel, 1, 0);
  definePrimitive(base, index++, "_prGtkWidgetGetSensitive", prGtkWidgetGetSensitive, 1, 0);
  definePrimitive(base, index++, "_prGtkWidgetSetSensitive", prGtkWidgetSetSensitive, 2, 0);
  definePrimitive(base, index++, "_prGtkWidgetModifyBg", prGtkWidgetModifyBg, 3, 0);
  definePrimitive(base, index++, "_prGtkWidgetReparent", prGtkWidgetReparent, 2, 0);

  definePrimitive(base, index++, "_prGtkStyleGetWhiteGC", prGtkStyleGetWhiteGC, 1, 0);
  definePrimitive(base, index++, "_prGtkStyleGetBlackGC", prGtkStyleGetBlackGC, 1, 0);

  definePrimitive(base, index++, "_prGtkSizeGroupNew", prGtkSizeGroupNew, 2, 0);
  definePrimitive(base, index++, "_prGtkSizeGroupAddWidget", prGtkSizeGroupAddWidget, 2, 0);
  definePrimitive(base, index++, "_prGtkSizeGroupRemoveWidget", prGtkSizeGroupRemoveWidget, 2, 0);

  definePrimitive(base, index++, "_prGtkContainerAdd", prGtkContainerAdd, 2, 0);
  definePrimitive(base, index++, "_prGtkContainerSetBorderWidth", prGtkContainerSetBorderWidth, 2, 0);
  definePrimitive(base, index++, "_prGtkContainerRemove", prGtkContainerRemove, 2, 0);
  definePrimitive(base, index++, "_prGtkContainerChildren", prGtkContainerChildren, 1, 0);
  definePrimitive(base, index++, "_prGtkContainerNChildren", prGtkContainerNChildren, 1, 0);

  definePrimitive(base, index++, "_prGtkBinGetChild", prGtkBinGetChild, 1, 0);

  definePrimitive(base, index++, "_prGtkBoxPackStart", prGtkBoxPackStart, 5, 0);
  definePrimitive(base, index++, "_prGtkBoxPackEnd", prGtkBoxPackEnd, 5, 0);
  definePrimitive(base, index++, "_prGtkBoxReorderChild", prGtkBoxReorderChild, 3, 0);

  definePrimitive(base, index++, "_prGtkHBoxNew", prGtkHBoxNew, 3, 0);
  definePrimitive(base, index++, "_prGtkVBoxNew", prGtkVBoxNew, 3, 0);

  definePrimitive(base, index++, "_prGtkFrameNew", prGtkFrameNew, 2, 0);
  definePrimitive(base, index++, "_prGtkFrameSetLabel", prGtkFrameSetLabel, 2, 0);
  definePrimitive(base, index++, "_prGtkFrameSetShadowType", prGtkFrameSetShadowType, 2, 0);

  definePrimitive(base, index++, "_prGtkPanedAdd1", prGtkPanedAdd1, 2, 0);
  definePrimitive(base, index++, "_prGtkPanedAdd2", prGtkPanedAdd2, 2, 0);
  definePrimitive(base, index++, "_prGtkPanedPack1", prGtkPanedPack1, 4, 0);
  definePrimitive(base, index++, "_prGtkPanedPack2", prGtkPanedPack2, 4, 0);
  definePrimitive(base, index++, "_prGtkPanedSetPosition", prGtkPanedSetPosition, 2, 0);
  definePrimitive(base, index++, "_prGtkPanedGetPosition", prGtkPanedGetPosition, 1, 0);
  definePrimitive(base, index++, "_prGtkHPanedNew", prGtkHPanedNew, 1, 0);
  definePrimitive(base, index++, "_prGtkVPanedNew", prGtkVPanedNew, 1, 0);

  definePrimitive(base, index++, "_prGtkFixedNew", prGtkFixedNew, 1, 0);
  definePrimitive(base, index++, "_prGtkFixedPut", prGtkFixedPut, 4, 0);
  definePrimitive(base, index++, "_prGtkFixedMove", prGtkFixedMove, 4, 0);
  definePrimitive(base, index++, "_prGtkFixedSetHasWindow", prGtkFixedSetHasWindow, 2, 0);
  definePrimitive(base, index++, "_prGtkFixedGetHasWindow", prGtkFixedGetHasWindow, 1, 0);

  definePrimitive(base, index++, "_prGtkWindowNew", prGtkWindowNew, 2, 0);
  definePrimitive(base, index++, "_prGtkWindowSetTitle", prGtkWindowSetTitle, 2, 0);
  definePrimitive(base, index++, "_prGtkWindowGetTitle", prGtkWindowGetTitle, 1, 0);
  definePrimitive(base, index++, "_prGtkWindowSetPosition", prGtkWindowSetPosition, 2, 0);
  definePrimitive(base, index++, "_prGtkWindowSetDefaultSize", prGtkWindowSetDefaultSize, 3, 0);
  definePrimitive(base, index++, "_prGtkWindowGetDefaultSize", prGtkWindowGetDefaultSize, 1, 0);
  definePrimitive(base, index++, "_prGtkWindowSetFocus", prGtkWindowSetFocus, 2, 0);
  definePrimitive(base, index++, "_prGtkWindowGetFocus", prGtkWindowGetFocus, 1, 0);
  definePrimitive(base, index++, "_prGtkWindowMove", prGtkWindowMove, 3, 0);
  definePrimitive(base, index++, "_prGtkWindowSetTransientFor", prGtkWindowSetTransientFor, 2, 0);
  definePrimitive(base, index++, "_prGtkWindowSetDestroyWithParent", prGtkWindowSetDestroyWithParent, 2, 0);
  definePrimitive(base, index++, "_prGtkWindowSetHideOnDelete", prGtkWindowSetHideOnDelete, 1, 0);
  definePrimitive(base, index++, "_prGtkWindowResize", prGtkWindowResize, 3, 0);
  definePrimitive(base, index++, "_prGtkWindowGetSize", prGtkWindowGetSize, 1, 0);

  definePrimitive(base, index++, "_prGtkFileSelectionNew", prGtkFileSelectionNew, 2, 0);
  definePrimitive(base, index++, "_prGtkFileSelectionSetFilename", prGtkFileSelectionSetFilename, 2, 0);
  definePrimitive(base, index++, "_prGtkFileSelectionGetFilename", prGtkFileSelectionGetFilename, 1, 0);
  definePrimitive(base, index++, "_prGtkFileSelectionShowFileopButtons", prGtkFileSelectionShowFileopButtons, 1, 0);
  definePrimitive(base, index++, "_prGtkFileSelectionHideFileopButtons", prGtkFileSelectionHideFileopButtons, 1, 0);
  definePrimitive(base, index++, "_prGtkFileSelectionGetSelections", prGtkFileSelectionGetSelections, 1, 0);
  definePrimitive(base, index++, "_prGtkFileSelectionSetSelectMultiple", prGtkFileSelectionSetSelectMultiple, 2, 0);
  definePrimitive(base, index++, "_prGtkFileSelectionGetSelectMultiple", prGtkFileSelectionGetSelectMultiple, 1, 0);
  definePrimitive(base, index++, "_prGtkFileSelectionComplete", prGtkFileSelectionComplete, 2, 0);
  definePrimitive(base, index++, "_prGtkFileSelectionDirList", prGtkFileSelectionDirList, 1, 0);
  definePrimitive(base, index++, "_prGtkFileSelectionFileList", prGtkFileSelectionFileList, 1, 0);
  definePrimitive(base, index++, "_prGtkFileSelectionSelectionEntry", prGtkFileSelectionSelectionEntry, 1, 0);
  definePrimitive(base, index++, "_prGtkFileSelectionSelectionText", prGtkFileSelectionSelectionText, 1, 0);
  definePrimitive(base, index++, "_prGtkFileSelectionMainVBox", prGtkFileSelectionMainVBox, 1, 0);
  definePrimitive(base, index++, "_prGtkFileSelectionOkButton", prGtkFileSelectionOkButton, 1, 0);
  definePrimitive(base, index++, "_prGtkFileSelectionCancelButton", prGtkFileSelectionCancelButton, 1, 0);
  definePrimitive(base, index++, "_prGtkFileSelectionHelpButton", prGtkFileSelectionHelpButton, 1, 0);

  definePrimitive(base, index++, "_prGtkButtonNew", prGtkButtonNew, 1, 0);
  definePrimitive(base, index++, "_prGtkButtonNewWithLabel", prGtkButtonNewWithLabel, 2, 0);
  definePrimitive(base, index++, "_prGtkButtonNewFromStock", prGtkButtonNewFromStock, 2, 0);
  definePrimitive(base, index++, "_prGtkButtonSetLabel", prGtkButtonSetLabel, 2, 0);
  definePrimitive(base, index++, "_prGtkButtonGetLabel", prGtkButtonGetLabel, 1, 0);

  definePrimitive(base, index++, "_prGtkToggleButtonNew", prGtkToggleButtonNew, 1, 0);
  definePrimitive(base, index++, "_prGtkToggleButtonNewWithLabel", prGtkToggleButtonNewWithLabel, 2, 0);
  definePrimitive(base, index++, "_prGtkToggleButtonSetMode", prGtkToggleButtonSetMode, 2, 0);
  definePrimitive(base, index++, "_prGtkToggleButtonGetMode", prGtkToggleButtonGetMode, 1, 0);
  definePrimitive(base, index++, "_prGtkToggleButtonSetActive", prGtkToggleButtonSetActive, 2, 0);
  definePrimitive(base, index++, "_prGtkToggleButtonGetActive", prGtkToggleButtonGetActive, 1, 0);
  definePrimitive(base, index++, "_prGtkToggleButtonSetInconsistent", prGtkToggleButtonSetInconsistent, 2, 0);
  definePrimitive(base, index++, "_prGtkToggleButtonGetInconsistent", prGtkToggleButtonGetInconsistent, 1, 0);

  definePrimitive(base, index++, "_prGtkCheckButtonNew", prGtkCheckButtonNew, 1, 0);
  definePrimitive(base, index++, "_prGtkCheckButtonNewWithLabel", prGtkCheckButtonNewWithLabel, 2, 0);

  definePrimitive(base, index++, "_prGtkRadioButtonNew", prGtkRadioButtonNew, 2, 0);
  definePrimitive(base, index++, "_prGtkRadioButtonNewWithLabel", prGtkRadioButtonNewWithLabel, 3, 0);
  definePrimitive(base, index++, "_prGtkRadioButtonNewFromWidget", prGtkRadioButtonNewFromWidget, 2, 0);
  definePrimitive(base, index++, "_prGtkRadioButtonNewWithLabelFromWidget", prGtkRadioButtonNewWithLabelFromWidget, 3, 0);
  definePrimitive(base, index++, "_prGtkRadioButtonSetGroup", prGtkRadioButtonSetGroup, 2, 0);
  definePrimitive(base, index++, "_prGtkRadioButtonGetGroup", prGtkRadioButtonGetGroup, 1, 0);

  definePrimitive(base, index++, "_prGtkMiscSetAlignment", prGtkMiscSetAlignment, 3, 0);
  definePrimitive(base, index++, "_prGtkMiscSetPadding", prGtkMiscSetPadding, 3, 0);
 
  definePrimitive(base, index++, "_prGtkLabelNew", prGtkLabelNew, 2, 0);
  definePrimitive(base, index++, "_prGtkLabelSetText", prGtkLabelSetText, 2, 0);
  definePrimitive(base, index++, "_prGtkLabelSetMarkup", prGtkLabelSetMarkup, 2, 0);
  definePrimitive(base, index++, "_prGtkLabelGetText", prGtkLabelGetText, 1, 0);
  definePrimitive(base, index++, "_prGtkLabelSetJustify", prGtkLabelSetJustify, 2, 0);

  definePrimitive(base, index++, "_prGtkEntryNew", prGtkEntryNew, 1, 0);
  definePrimitive(base, index++, "_prGtkEntrySetText", prGtkEntrySetText, 2, 0);
  definePrimitive(base, index++, "_prGtkEntryGetText", prGtkEntryGetText, 1, 0);
  definePrimitive(base, index++, "_prGtkEntrySetEditable", prGtkEntrySetEditable, 2, 0);

  definePrimitive(base, index++, "_prGtkDrawingAreaNew", prGtkDrawingAreaNew, 1, 0);

  definePrimitive(base, index++, "_prGtkTreePathNewFromString", prGtkTreePathNewFromString, 3, 0);
  definePrimitive(base, index++, "_prGtkTreePathAsString", prGtkTreePathAsString, 1, 0);
  definePrimitive(base, index++, "_prGtkTreePathGetIndices", prGtkTreePathGetIndices, 1, 0);

  definePrimitive(base, index++, "_prGtkTreeModelGetColumnType", prGtkTreeModelGetColumnType, 2, 0);
  definePrimitive(base, index++, "_prGtkTreeModelGetIter", prGtkTreeModelGetIter, 2, 0);
  definePrimitive(base, index++, "_prGtkTreeModelGetIterFirst", prGtkTreeModelGetIterFirst, 1, 0);
  definePrimitive(base, index++, "_prGtkTreeModelGetPath", prGtkTreeModelGetPath, 2, 0);
  definePrimitive(base, index++, "_prGtkTreeModelGetPathString", prGtkTreeModelGetPathString, 2, 0);
  definePrimitive(base, index++, "_prGtkTreeModelIterParent", prGtkTreeModelIterParent, 3, 0);
  definePrimitive(base, index++, "_prGtkTreeModelIterNext", prGtkTreeModelIterNext, 3, 0);
  definePrimitive(base, index++, "_prGtkTreeModelIterHasChild", prGtkTreeModelIterHasChild, 2, 0);
  definePrimitive(base, index++, "_prGtkTreeModelIterNChildren", prGtkTreeModelIterNChildren, 2, 0);
  definePrimitive(base, index++, "_prGtkTreeModelIterChildren", prGtkTreeModelIterChildren, 2, 0);
  definePrimitive(base, index++, "_prGtkTreeModelIterChildren", prGtkTreeModelIterNthChild, 3, 0);
  definePrimitive(base, index++, "_prGtkTreeModelGetValue", prGtkTreeModelGetValue, 3, 0);

  definePrimitive(base, index++, "_prGtkTreeRowReferenceNew", prGtkTreeRowReferenceNew, 3, 0);
  definePrimitive(base, index++, "_prGtkTreeRowReferenceGetPath", prGtkTreeRowReferenceGetPath, 1, 0);
  definePrimitive(base, index++, "_prGtkTreeRowReferenceValid", prGtkTreeRowReferenceValid, 1, 0);
  definePrimitive(base, index++, "_prGtkTreeRowReferenceFree", prGtkTreeRowReferenceFree, 1, 0);
  definePrimitive(base, index++, "_prGtkTreeRowReferenceCopy", prGtkTreeRowReferenceCopy, 1, 0);

  definePrimitive(base, index++, "_prGtkListStoreNew", prGtkListStoreNew, 2, 0);
  definePrimitive(base, index++, "_prGtkListStoreAppend", prGtkListStoreAppend, 1, 0);
  definePrimitive(base, index++, "_prGtkListStoreSetValue", prGtkListStoreSetValue, 5, 0);
  definePrimitive(base, index++, "_prGtkListStoreClear", prGtkListStoreClear, 1, 0);
  definePrimitive(base, index++, "_prGtkListStoreRemove", prGtkListStoreRemove, 2, 0);

  definePrimitive(base, index++, "_prGtkTreeStoreNew", prGtkTreeStoreNew, 2, 0);
  definePrimitive(base, index++, "_prGtkTreeStoreAppend", prGtkTreeStoreAppend, 2, 0);
  definePrimitive(base, index++, "_prGtkTreeStoreInsertBefore", prGtkTreeStoreInsertBefore, 3, 0);
  definePrimitive(base, index++, "_prGtkTreeStoreInsertAfter", prGtkTreeStoreInsertAfter, 3, 0);
  definePrimitive(base, index++, "_prGtkTreeStoreSetValue", prGtkTreeStoreSetValue, 5, 0);
  definePrimitive(base, index++, "_prGtkTreeStoreClear", prGtkTreeStoreClear, 1, 0);
  definePrimitive(base, index++, "_prGtkTreeStoreRemove", prGtkTreeStoreRemove, 2, 0);

  definePrimitive(base, index++, "_prGtkTreeViewNew", prGtkTreeViewNew, 1, 0);
  definePrimitive(base, index++, "_prGtkTreeViewNewWithModel", prGtkTreeViewNewWithModel, 2, 0);
  definePrimitive(base, index++, "_prGtkTreeViewSetRulesHint", prGtkTreeViewSetRulesHint, 2, 0);
  definePrimitive(base, index++, "_prGtkTreeViewSetSearchColumn", prGtkTreeViewSetSearchColumn, 2, 0);
  definePrimitive(base, index++, "_prGtkTreeViewAppendColumn", prGtkTreeViewAppendColumn, 2, 0);
  definePrimitive(base, index++, "_prGtkTreeViewScrollToCell", prGtkTreeViewScrollToCell, 5, 0);
  definePrimitive(base, index++, "_prGtkTreeViewGetSelection", prGtkTreeViewGetSelection, 1, 0);
  definePrimitive(base, index++, "_prGtkTreeViewGetModel", prGtkTreeViewGetModel, 1, 0);
  definePrimitive(base, index++, "_prGtkTreeViewSetReorderable", prGtkTreeViewSetReorderable, 2, 0);
  definePrimitive(base, index++, "_prGtkTreeViewGetReorderable", prGtkTreeViewGetReorderable, 1, 0);
  definePrimitive(base, index++, "_prGtkTreeViewEnableModelDragSource", prGtkTreeViewEnableModelDragSource, 4, 0);
  definePrimitive(base, index++, "_prGtkTreeViewEnableModelDragDest", prGtkTreeViewEnableModelDragDest, 3, 0);
  definePrimitive(base, index++, "_prGtkTreeViewGetDestRowAtPos", prGtkTreeViewGetDestRowAtPos, 3, 0);
  definePrimitive(base, index++, "_prGtkTreeViewExpandRow", prGtkTreeViewExpandRow, 3, 0);
  definePrimitive(base, index++, "_prGtkTreeViewExpandAll", prGtkTreeViewExpandAll, 1, 0);

  definePrimitive(base, index++, "_prGtkTreeViewColumnNew", prGtkTreeViewColumnNew, 1, 0);
  definePrimitive(base, index++, "_prGtkTreeViewColumnNewWithAttributes", prGtkTreeViewColumnNewWithAttributes, 4, 0);
  definePrimitive(base, index++, "_prGtkTreeViewColumnSetClickable", prGtkTreeViewColumnSetClickable, 2, 0);
  definePrimitive(base, index++, "_prGtkTreeViewColumnSetSizing", prGtkTreeViewColumnSetSizing, 2, 0);
  definePrimitive(base, index++, "_prGtkTreeViewColumnSetFixedWidth", prGtkTreeViewColumnSetFixedWidth, 2, 0);
  definePrimitive(base, index++, "_prGtkTreeViewColumnSetTitle", prGtkTreeViewColumnSetTitle, 2, 0);

  definePrimitive(base, index++, "_prGtkTreeSelectionGetSelected", prGtkTreeSelectionGetSelected, 1, 0);
  definePrimitive(base, index++, "_prGtkTreeSelectionGetSelectedRows", prGtkTreeSelectionGetSelectedRows, 1, 0);
  definePrimitive(base, index++, "_prGtkTreeSelectionSelectIter", prGtkTreeSelectionSelectIter, 2, 0);
  definePrimitive(base, index++, "_prGtkTreeSelectionSetMode", prGtkTreeSelectionSetMode, 2, 0);
  definePrimitive(base, index++, "_prGtkTreeSelectionSelectAll", prGtkTreeSelectionSelectAll, 1, 0);
  definePrimitive(base, index++, "_prGtkTreeSelectionUnselectAll", prGtkTreeSelectionUnselectAll, 1, 0);

  definePrimitive(base, index++, "_prGtkCellRendererTextNew", prGtkCellRendererTextNew, 1, 0);

  definePrimitive(base, index++, "_prGtkCellRendererToggleNew", prGtkCellRendererToggleNew, 1, 0);

  definePrimitive(base, index++, "_prGtkAdjustmentNew", prGtkAdjustmentNew, 7, 0);
  definePrimitive(base, index++, "_prGtkAdjustmentGetValue", prGtkAdjustmentGetValue, 1, 0);
  definePrimitive(base, index++, "_prGtkAdjustmentSetValue", prGtkAdjustmentSetValue, 2, 0);
  definePrimitive(base, index++, "_prGtkAdjustmentClampPage", prGtkAdjustmentClampPage, 3, 0);
  definePrimitive(base, index++, "_prGtkAdjustmentChanged", prGtkAdjustmentChanged, 1, 0);
  definePrimitive(base, index++, "_prGtkAdjustmentValueChanged", prGtkAdjustmentValueChanged, 1, 0);

  definePrimitive(base, index++, "_prGtkScrolledWindowNew", prGtkScrolledWindowNew, 3, 0);
  definePrimitive(base, index++, "_prGtkScrolledWindowSetPolicy", prGtkScrolledWindowSetPolicy, 3, 0);
  definePrimitive(base, index++, "_prGtkScrolledWindowSetShadowType", prGtkScrolledWindowSetShadowType, 2, 0);
  definePrimitive(base, index++, "_prGtkScrolledWindowAddWithViewport", prGtkScrolledWindowAddWithViewport, 2, 0);

  definePrimitive(base, index++, "_prPangoLayoutNew", prPangoLayoutNew, 2, 0);
  definePrimitive(base, index++, "_prPangoLayoutSetText", prPangoLayoutSetText, 2, 0);
  definePrimitive(base, index++, "_prPangoLayoutSetMarkup", prPangoLayoutSetMarkup, 2, 0);
  definePrimitive(base, index++, "_prPangoLayoutGetPixelSize", prPangoLayoutGetPixelSize, 1, 0);

  definePrimitive(base, index++, "_prGtkTableNew", prGtkTableNew, 4, 0);
  definePrimitive(base, index++, "_prGtkTableResize", prGtkTableResize, 3, 0);
  definePrimitive(base, index++, "_prGtkTableAttach", prGtkTableAttach, 10, 0);
  definePrimitive(base, index++, "_prGtkTableSetColSpacings", prGtkTableSetColSpacings, 2, 0);
  definePrimitive(base, index++, "_prGtkTableSetRowSpacings", prGtkTableSetRowSpacings, 2, 0);

  definePrimitive(base, index++, "_prGtkImageNewFromFile", prGtkImageNewFromFile, 2, 0);
  definePrimitive(base, index++, "_prGtkImageNewFromStock", prGtkImageNewFromStock, 3, 0);

  definePrimitive(base, index++, "_prGtkMenuShellAppend", prGtkMenuShellAppend, 2, 0);
  definePrimitive(base, index++, "_prGtkMenuShellPrepend", prGtkMenuShellPrepend, 2, 0);
  definePrimitive(base, index++, "_prGtkMenuShellInsert", prGtkMenuShellInsert, 3, 0);

  definePrimitive(base, index++, "_prGtkMenuNew", prGtkMenuNew, 1, 0);
  definePrimitive(base, index++, "_prGtkMenuPopup", prGtkMenuPopup, 3, 0);

  definePrimitive(base, index++, "_prGtkMenuItemNewWithLabel", prGtkMenuItemNewWithLabel, 2, 0);
  definePrimitive(base, index++, "_prGtkMenuItemSetSubmenu", prGtkMenuItemSetSubmenu, 2, 0);

  definePrimitive(base, index++, "_prGtkSeparatorMenuItemNew", prGtkSeparatorMenuItemNew, 1, 0);

  definePrimitive(base, index++, "_prGtkTextIterIsStart", prGtkTextIterIsStart, 1, 0);
  definePrimitive(base, index++, "_prGtkTextIterIsEnd", prGtkTextIterIsEnd, 1, 0);
  definePrimitive(base, index++, "_prGtkTextIterGetLine", prGtkTextIterGetLine, 1, 0);
  definePrimitive(base, index++, "_prGtkTextIterGetOffset", prGtkTextIterGetOffset, 1, 0);
  definePrimitive(base, index++, "_prGtkTextIterGetLineOffset", prGtkTextIterGetLineOffset, 1, 0);
  definePrimitive(base, index++, "_prGtkTextIterGetLineIndex", prGtkTextIterGetLineIndex, 1, 0);
  definePrimitive(base, index++, "_prGtkTextIterSetLineOffset", prGtkTextIterSetLineOffset, 2, 0);
  definePrimitive(base, index++, "_prGtkTextIterForwardToLineEnd", prGtkTextIterForwardToLineEnd, 1, 0);
  definePrimitive(base, index++, "_prGtkTextIterForwardLine", prGtkTextIterForwardLine, 1, 0);

  definePrimitive(base, index++, "_prGtkTextBufferNew", prGtkTextBufferNew, 1, 0);
  definePrimitive(base, index++, "_prGtkTextBufferSetText", prGtkTextBufferSetText, 2, 0);
  definePrimitive(base, index++, "_prGtkTextBufferGetText", prGtkTextBufferGetText, 4, 0);
  definePrimitive(base, index++, "_prGtkTextBufferGetIterAtMark", prGtkTextBufferGetIterAtMark, 2, 0);
  definePrimitive(base, index++, "_prGtkTextBufferGetIterAtLineOffset", prGtkTextBufferGetIterAtLineOffset, 3, 0);
  definePrimitive(base, index++, "_prGtkTextBufferGetIterAtLineIndex", prGtkTextBufferGetIterAtLineIndex, 3, 0);
  definePrimitive(base, index++, "_prGtkTextBufferGetIterAtLine", prGtkTextBufferGetIterAtLine, 2, 0);
  definePrimitive(base, index++, "_prGtkTextBufferGetIterAtOffset", prGtkTextBufferGetIterAtOffset, 2, 0);
  definePrimitive(base, index++, "_prGtkTextBufferGetStartIter", prGtkTextBufferGetStartIter, 1, 0);
  definePrimitive(base, index++, "_prGtkTextBufferGetEndIter", prGtkTextBufferGetEndIter, 1, 0);
  definePrimitive(base, index++, "_prGtkTextBufferGetBounds", prGtkTextBufferGetBounds, 1, 0);
  definePrimitive(base, index++, "_prGtkTextBufferInsert", prGtkTextBufferInsert, 3, 0);
  definePrimitive(base, index++, "_prGtkTextBufferInsertAtCursor", prGtkTextBufferInsertAtCursor, 2, 0);
  definePrimitive(base, index++, "_prGtkTextBufferInsertWithTag", prGtkTextBufferInsertWithTag, 4, 0);
  definePrimitive(base, index++, "_prGtkTextBufferGetMark", prGtkTextBufferGetMark, 2, 0);
  definePrimitive(base, index++, "_prGtkTextBufferGetInsert", prGtkTextBufferGetInsert, 1, 0);
  definePrimitive(base, index++, "_prGtkTextBufferGetSelectionBound", prGtkTextBufferGetSelectionBound, 1, 0);
  definePrimitive(base, index++, "_prGtkTextBufferCreateMark", prGtkTextBufferCreateMark, 4, 0);
  definePrimitive(base, index++, "_prGtkTextBufferDeleteMark", prGtkTextBufferDeleteMark, 2, 0);
  definePrimitive(base, index++, "_prGtkTextBufferGetSelectionBounds", prGtkTextBufferGetSelectionBounds, 1, 0);
 
  definePrimitive(base, index++, "_prGtkTextViewNew", prGtkTextViewNew, 1, 0);
  definePrimitive(base, index++, "_prGtkTextViewNewWithBuffer", prGtkTextViewNewWithBuffer, 2, 0);
  definePrimitive(base, index++, "_prGtkTextViewGetBuffer", prGtkTextViewGetBuffer, 1, 0);
  definePrimitive(base, index++, "_prGtkTextViewSetBuffer", prGtkTextViewSetBuffer, 2, 0);
  definePrimitive(base, index++, "_prGtkTextViewScrollToMark", prGtkTextViewScrollToMark, 6, 0);
  definePrimitive(base, index++, "_prGtkTextViewScrollToIter", prGtkTextViewScrollToIter, 6, 0);
  definePrimitive(base, index++, "_prGtkTextViewGetEditable", prGtkTextViewGetEditable, 1, 0);
  definePrimitive(base, index++, "_prGtkTextViewSetEditable", prGtkTextViewSetEditable, 2, 0);
  definePrimitive(base, index++, "_prGtkTextViewSetCursorVisible", prGtkTextViewSetCursorVisible, 2, 0);

  definePrimitive(base, index++, "_prGtkNotebookNew", prGtkNotebookNew, 1, 0);
  definePrimitive(base, index++, "_prGtkNotebookAppendPage", prGtkNotebookAppendPage, 3, 0);
  definePrimitive(base, index++, "_prGtkNotebookPrependPage", prGtkNotebookPrependPage, 3, 0);
  definePrimitive(base, index++, "_prGtkNotebookInsertPage", prGtkNotebookInsertPage, 4, 0);
  definePrimitive(base, index++, "_prGtkNotebookRemovePage", prGtkNotebookRemovePage, 2, 0);
  definePrimitive(base, index++, "_prGtkNotebookGetCurrentPage", prGtkNotebookGetCurrentPage, 1, 0);
  definePrimitive(base, index++, "_prGtkNotebookGetNthPage", prGtkNotebookGetNthPage, 2, 0);
  definePrimitive(base, index++, "_prGtkNotebookGetNPages", prGtkNotebookGetNPages, 1, 0);
  definePrimitive(base, index++, "_prGtkNotebookPageNum", prGtkNotebookPageNum, 2, 0);
  definePrimitive(base, index++, "_prGtkNotebookSetCurrentPage", prGtkNotebookSetCurrentPage, 2, 0);
  definePrimitive(base, index++, "_prGtkNotebookNextPage", prGtkNotebookNextPage, 1, 0);
  definePrimitive(base, index++, "_prGtkNotebookPrevPage", prGtkNotebookPrevPage, 1, 0);
  definePrimitive(base, index++, "_prGtkNotebookSetScrollable", prGtkNotebookSetScrollable, 2, 0);
  definePrimitive(base, index++, "_prGtkNotebookGetScrollable", prGtkNotebookGetScrollable, 1, 0);
  definePrimitive(base, index++, "_prGtkNotebookPopupDisable", prGtkNotebookPopupDisable, 1, 0);
  definePrimitive(base, index++, "_prGtkNotebookPopupEnable", prGtkNotebookPopupEnable, 1, 0);

  definePrimitive(base, index++, "_prGtkEventBoxNew", prGtkEventBoxNew, 1, 0);

  definePrimitive(base, index++, "_prGtkRangeSetUpdatePolicy", prGtkRangeSetUpdatePolicy, 2, 0);
  definePrimitive(base, index++, "_prGtkRangeGetUpdatePolicy", prGtkRangeGetUpdatePolicy, 1, 0);
  definePrimitive(base, index++, "_prGtkRangeSetValue", prGtkRangeSetValue, 2, 0);
  definePrimitive(base, index++, "_prGtkRangeGetValue", prGtkRangeGetValue, 1, 0);
  definePrimitive(base, index++, "_prGtkRangeSetAdjustment", prGtkRangeSetAdjustment, 2, 0);
  definePrimitive(base, index++, "_prGtkRangeGetAdjustment", prGtkRangeGetAdjustment, 1, 0);
  definePrimitive(base, index++, "_prGtkRangeSetInverted", prGtkRangeSetInverted, 2, 0);
  definePrimitive(base, index++, "_prGtkRangeGetInverted", prGtkRangeGetInverted, 1, 0);
  definePrimitive(base, index++, "_prGtkRangeSetIncrements", prGtkRangeSetIncrements, 3, 0);
  definePrimitive(base, index++, "_prGtkRangeSetRange", prGtkRangeSetRange, 3, 0);

  definePrimitive(base, index++, "_prGtkScaleSetDigits", prGtkScaleSetDigits, 2, 0);
  definePrimitive(base, index++, "_prGtkScaleGetDigits", prGtkScaleGetDigits, 1, 0);
  definePrimitive(base, index++, "_prGtkScaleSetDrawValue", prGtkScaleSetDrawValue, 2, 0);
  definePrimitive(base, index++, "_prGtkScaleGetDrawValue", prGtkScaleGetDrawValue, 1, 0);
  definePrimitive(base, index++, "_prGtkScaleSetValuePosition", prGtkScaleSetValuePosition, 2, 0);
  definePrimitive(base, index++, "_prGtkScaleGetValuePosition", prGtkScaleGetValuePosition, 1, 0);

  definePrimitive(base, index++, "_prGtkHScaleNew", prGtkHScaleNew, 2, 0);
  definePrimitive(base, index++, "_prGtkHScaleNewWithRange", prGtkHScaleNewWithRange, 4, 0);
  definePrimitive(base, index++, "_prGtkVScaleNew", prGtkVScaleNew, 2, 0);
  definePrimitive(base, index++, "_prGtkVScaleNewWithRange", prGtkVScaleNewWithRange, 4, 0);

  definePrimitive(base, index++, "_prGtkSpinButtonNew", prGtkSpinButtonNew, 3, 0);
  definePrimitive(base, index++, "_prGtkSpinButtonNewWithRange", prGtkSpinButtonNewWithRange, 4, 0);
  definePrimitive(base, index++, "_prGtkSpinButtonSetAdjustment", prGtkSpinButtonSetAdjustment, 2, 0);
  definePrimitive(base, index++, "_prGtkSpinButtonGetAdjustment", prGtkSpinButtonGetAdjustment, 1, 0);
  definePrimitive(base, index++, "_prGtkSpinButtonSetDigits", prGtkSpinButtonSetDigits, 2, 0);
  definePrimitive(base, index++, "_prGtkSpinButtonGetDigits", prGtkSpinButtonGetDigits, 1, 0);
  definePrimitive(base, index++, "_prGtkSpinButtonSetIncrements", prGtkSpinButtonSetIncrements, 3, 0);
  definePrimitive(base, index++, "_prGtkSpinButtonGetIncrements", prGtkSpinButtonGetIncrements, 1, 0);
  definePrimitive(base, index++, "_prGtkSpinButtonSetRange", prGtkSpinButtonSetRange, 3, 0);
  definePrimitive(base, index++, "_prGtkSpinButtonGetRange", prGtkSpinButtonGetRange, 1, 0);
  definePrimitive(base, index++, "_prGtkSpinButtonSetValue", prGtkSpinButtonSetValue, 2, 0);
  definePrimitive(base, index++, "_prGtkSpinButtonGetValue", prGtkSpinButtonGetValue, 1, 0);
  definePrimitive(base, index++, "_prGtkSpinButtonGetValueAsInteger", prGtkSpinButtonGetValueAsInteger, 1, 0);
  definePrimitive(base, index++, "_prGtkSpinButtonSetUpdatePolicy", prGtkSpinButtonSetUpdatePolicy, 2, 0);
  definePrimitive(base, index++, "_prGtkSpinButtonGetUpdatePolicy", prGtkSpinButtonGetUpdatePolicy, 1, 0);
  definePrimitive(base, index++, "_prGtkSpinButtonSetNumeric", prGtkSpinButtonSetNumeric, 2, 0);
  definePrimitive(base, index++, "_prGtkSpinButtonGetNumeric", prGtkSpinButtonGetNumeric, 1, 0);
  definePrimitive(base, index++, "_prGtkSpinButtonSetWrap", prGtkSpinButtonSetWrap, 2, 0);
  definePrimitive(base, index++, "_prGtkSpinButtonGetWrap", prGtkSpinButtonGetWrap, 1, 0);
  definePrimitive(base, index++, "_prGtkSpinButtonSpin", prGtkSpinButtonSpin, 3, 0);
  definePrimitive(base, index++, "_prGtkSpinButtonSetSnapToTicks", prGtkSpinButtonSetSnapToTicks, 2, 0);
  definePrimitive(base, index++, "_prGtkSpinButtonGetSnapToTicks", prGtkSpinButtonGetSnapToTicks, 1, 0);
  definePrimitive(base, index++, "_prGtkSpinButtonUpdate", prGtkSpinButtonUpdate, 1, 0);

  definePrimitive(base, index++, "_prGtkComboNew", prGtkComboNew, 1, 0);
  definePrimitive(base, index++, "_prGtkComboSetPopdownStrings", prGtkComboSetPopdownStrings, 2, 0);
  definePrimitive(base, index++, "_prGtkComboSetValueInList", prGtkComboSetValueInList, 3, 0);
  definePrimitive(base, index++, "_prGtkComboSetUseArrows", prGtkComboSetUseArrows, 2, 0);
  definePrimitive(base, index++, "_prGtkComboSetCaseSensitive", prGtkComboSetCaseSensitive, 2, 0);
  definePrimitive(base, index++, "_prGtkComboSetItemString", prGtkComboSetItemString, 3, 0);
  definePrimitive(base, index++, "_prGtkComboDisableActivate", prGtkComboDisableActivate, 1, 0);
  definePrimitive(base, index++, "_prGtkComboEntry", prGtkComboEntry, 1, 0);
  definePrimitive(base, index++, "_prGtkComboList", prGtkComboList, 1, 0);

  definePrimitive(base, index++, "_prGtkSelectionDataSetToPointer", prGtkSelectionDataSetToPointer, 2, 0);
  definePrimitive(base, index++, "_prGtkSelectionDataGetAsPointer", prGtkSelectionDataGetAsPointer, 1, 0);

  definePrimitive(base, index++, "_prGtkColorSelectionNew", prGtkColorSelectionNew, 1, 0);
  definePrimitive(base, index++, "_prGtkColorSelectionSetCurrentColor", prGtkColorSelectionSetCurrentColor, 2, 0);
  definePrimitive(base, index++, "_prGtkColorSelectionGetCurrentColor", prGtkColorSelectionGetCurrentColor, 1, 0);
  definePrimitive(base, index++, "_prGtkColorSelectionSetHasOpacityControl", prGtkColorSelectionSetHasOpacityControl, 2, 0);
  definePrimitive(base, index++, "_prGtkColorSelectionGetHasOpacityControl", prGtkColorSelectionGetHasOpacityControl, 1, 0);
  definePrimitive(base, index++, "_prGtkColorSelectionSetUpdatePolicy", prGtkColorSelectionSetUpdatePolicy, 2, 0);

  definePrimitive(base, index++, "_prGtkHSeparatorNew", prGtkHSeparatorNew, 1, 0);
  definePrimitive(base, index++, "_prGtkVSeparatorNew", prGtkVSeparatorNew, 1, 0);

  // ------ Gnome

  definePrimitive(base, index++, "_prGnomeCanvasNew", prGnomeCanvasNew, 2, 0);
  definePrimitive(base, index++, "_prGnomeCanvasSetScrollRegion", prGnomeCanvasSetScrollRegion, 5, 0);
  definePrimitive(base, index++, "_prGnomeCanvasRoot", prGnomeCanvasRoot, 1, 0);

  definePrimitive(base, index++, "_prGnomeCanvasGroupAdd", prGnomeCanvasGroupAdd, 2, 0);

  definePrimitive(base, index++, "_prGnomeCanvasItemSet", prGnomeCanvasItemSet, 3, 0);
  definePrimitive(base, index++, "_prGnomeCanvasItemSetPoints", prGnomeCanvasItemSetPoints, 2, 0);

  // ------ GLADE

  definePrimitive(base, index++, "_prGladeXmlNew", prGladeXmlNew, 4, 0);
  definePrimitive(base, index++, "_prGladeXmlSignalAutoconnect", prGladeXmlSignalAutoconnect, 1, 0);

  // ------ SourceView
  definePrimitive(base, index++, "_prGtkSourceBufferNew", prGtkSourceBufferNew, 1, 0);
  definePrimitive(base, index++, "_prGtkSourceViewNewWithBuffer", prGtkSourceViewNewWithBuffer, 2, 0);

  // ------- HOST APPLICATION

  definePrimitive(base, index++, "_prGetGtkTextBufferForConsoleOutput", prGetGtkTextBufferForConsoleOutput, 1, 0);
  //definePrimitive(base, index++, "_prHostReturnCmdLineResult", prHostReturnCmdLineResult, 2, 0);
}
